#include "schpin_lib/rviz.h"
#include <algorithm>
#include <vector>

#pragma once

namespace schpin
{

template < typename Link, typename Pin >
struct ChainView
{
        Pin&  source_pin;
        Link& link;
        Pin&  target_pin;

        ChainView( Pin& source_pin, Link& link, Pin& target_pin )
          : source_pin( source_pin )
          , link( link )
          , target_pin( target_pin )
        {
        }

        // todo: limit this to same types of link/pin
        template < typename OtherLink, typename OtherPin >
        ChainView( const ChainView< OtherLink, OtherPin >& other )
          : source_pin( other.source_pin )
          , link( other.link )
          , target_pin( other.target_pin )
        {
        }

        // This overload required for ChainIterator::operator->() to work.
        ChainView* operator->()
        {
                return this;
        }

        const ChainView* operator->() const
        {
                return this;
        }
};

template < typename Link >
using ChainLinkContainer = std::vector< Link >;

template < typename Pin >
using ChainPinContainer = std::vector< Pin >;

template < typename LinkT, typename PinT, bool is_const = false >
class ChainIterator
{
        using LinkIterator = std::conditional_t< is_const,
                                                 typename ChainLinkContainer< LinkT >::const_iterator,
                                                 typename ChainLinkContainer< LinkT >::iterator >;
        using PinIterator  = std::conditional_t< is_const,
                                                typename ChainPinContainer< PinT >::const_iterator,
                                                typename ChainPinContainer< PinT >::iterator >;

      public:
        using Link              = std::conditional_t< is_const, const LinkT, LinkT >;
        using Pin               = std::conditional_t< is_const, const PinT, PinT >;
        using value_type        = std::conditional_t< is_const, const ChainView< Link, Pin >, ChainView< Link, Pin > >;
        using difference_type   = std::ptrdiff_t;
        using pointer           = value_type*;
        using reference         = value_type&;
        using iterator_category = std::bidirectional_iterator_tag;

      private:
        LinkIterator link_iter_;
        PinIterator  pin_iter_;

        friend ChainIterator< LinkT, PinT, !is_const >;

      public:
        ChainIterator( LinkIterator link_iter, PinIterator pin_iter )
          : link_iter_( link_iter )
          , pin_iter_( pin_iter )
        {
        }

        template < bool is_actual_const = is_const >
        ChainIterator( const ChainIterator< LinkT, PinT, false >& other,
                       typename std::enable_if_t< is_actual_const >* /*unused*/ = nullptr )
          : link_iter_( other.link_iter_ )
          , pin_iter_( other.pin_iter_ )
        {
        }

        ChainIterator()                           = default;
        ChainIterator( const ChainIterator& )     = default;
        ChainIterator( ChainIterator&& ) noexcept = default;
        ChainIterator& operator=( const ChainIterator& ) = default;
        ChainIterator& operator=( ChainIterator&& ) noexcept = default;

        ChainView< Link, Pin > operator*()
        {
                return {*pin_iter_, *link_iter_, *std::next( pin_iter_ )};
        }
        ChainView< Link, Pin > operator->()
        {
                return **this;
        }
        ChainView< const Link, const Pin > operator*() const
        {
                return {*pin_iter_, *link_iter_, *std::next( pin_iter_ )};
        }
        ChainView< const Link, const Pin > operator->() const
        {
                return **this;
        }
        ChainIterator operator+( unsigned off )
        {
                link_iter_ += off;
                pin_iter_ += off;
                return *this;
        }
        ChainIterator& operator++()
        {
                ++link_iter_;
                ++pin_iter_;
                return *this;
        }
        const ChainIterator operator++( int )
        {
                ChainIterator copy( *this );
                ++( *this );
                return copy;
        }
        ChainIterator& operator--()
        {
                --link_iter_;
                --pin_iter_;
                return *this;
        }
        const ChainIterator operator--( int )
        {
                ChainIterator copy( *this );
                --( *this );
                return copy;
        }

        template < bool other_is_const >
        bool operator==( const ChainIterator< LinkT, PinT, other_is_const >& other ) const
        {
                return getLinkIter() == other.getLinkIter();
        }

        const LinkIterator& getLinkIter() const
        {
                return link_iter_;
        }

        const PinIterator& getPinIter() const
        {
                return pin_iter_;
        }

        ~ChainIterator() = default;
};

template < typename Link, typename Pin, bool is_const_lh, bool is_const_rh >
bool operator!=( const ChainIterator< Link, Pin, is_const_lh >& lh, const ChainIterator< Link, Pin, is_const_rh >& rh )
{
        return !( lh == rh );
}

template < typename LinkT, typename PinT >
class Chain
{
        ChainLinkContainer< LinkT > links_;
        ChainPinContainer< PinT >   pins_;

      public:
        using Link           = LinkT;
        using Pin            = PinT;
        using ConstView      = ChainView< const Link, const Pin >;
        using const_iterator = ChainIterator< Link, Pin, true >;
        using iterator       = ChainIterator< Link, Pin >;
        using value_type     = ChainView< Link, Pin >;

        std::reverse_iterator< ChainIterator< Link, Pin, true > > rbegin() const
        {
                return std::make_reverse_iterator( end() );
        }
        std::reverse_iterator< ChainIterator< Link, Pin, true > > rend() const
        {
                return std::make_reverse_iterator( begin() );
        }

        bool empty()
        {
                return pins_.empty();
        }

        ChainView< Link, Pin > back()
        {
                auto iter = end();
                --iter;
                return *iter;
        }

        ChainIterator< Link, Pin, true > begin() const
        {
                if ( links_.empty() ) {
                        return end();
                }
                return {links_.begin(), pins_.begin()};
        }

        ChainIterator< Link, Pin, true > end() const
        {
                return {links_.end(), std::prev( pins_.end() )};
        }

        ChainIterator< Link, Pin > begin()
        {
                if ( links_.empty() ) {
                        return end();
                }
                return {links_.begin(), pins_.begin()};
        }

        ChainIterator< Link, Pin > end()
        {
                return {links_.end(), pins_.end()};
        }

        ChainView< Link, Pin > operator[]( std::size_t i )
        {
                return {pins_[i], links_[i], pins_[i + 1]};
        }

        ChainView< const Link, const Pin > operator[]( std::size_t i ) const
        {
                return {pins_[i], links_[i], pins_[i + 1]};
        }

        void reverse()
        {
                std::reverse( pins_.begin(), pins_.end() );
                std::reverse( links_.begin(), links_.end() );
        }

        void pushBack( const Pin& pin )
        {
                if ( !pins_.empty() ) {
                        links_.emplace_back();
                }
                pins_.push_back( pin );
        }

        template < typename Container >
        iterator insert( const const_iterator& iter, Container pins )
        {
                std::size_t count = pins.size();
                if ( pins_.empty() && count > 0 ) {
                        count -= 1;
                }
                auto link_iter = links_.insert( iter.getLinkIter(), count, Link() );
                auto pin_iter  = pins_.insert( iter.getPinIter(), pins.begin(), pins.end() );
                return {link_iter, pin_iter};
        }

        iterator insert( const const_iterator& iter, const Pin& pin )
        {
                if ( !pins_.empty() ) {
                        auto link_iter = links_.insert( iter.getLinkIter(), Link() );
                        auto pin_iter  = pins_.insert( iter.getPinIter(), pin );
                        return {link_iter, pin_iter};
                }

                pins_.insert( iter.getPinIter(), pin );
                return end();
        }

        template < typename Container >
        ChainIterator< Link, Pin > erase( const Container& cont )
        {
                auto link_iter = links_.erase( cont.begin().getLinkIter(), cont.end().getLinkIter() );
                auto pin_iter  = pins_.erase( cont.begin().getPinIter(), cont.end().getPinIter() );

                if ( link_iter == links_.end() ) {
                        return end();
                }
                return {link_iter, pin_iter};
        }

        ChainIterator< Link, Pin > erase( const ChainIterator< Link, Pin >& iter )
        {
                auto link_iter = links_.erase( iter.getLinkIter() );
                auto pin_iter  = pins_.erase( iter.getPinIter() );

                if ( link_iter == links_.end() ) {
                        return end();
                }
                return {link_iter, pin_iter};
        }

        std::size_t size() const
        {
                return links_.size();
        }

        void clear()
        {
                links_.clear();
                pins_.clear();
        }

        auto getPins() const
        {
                return pins_;
        }
};

inline visualization_msgs::Marker createChainLinkMarker( unsigned id, std::string frame )
{
        visualization_msgs::Marker res = createMarker( "chainlink" + frame, id, frame, 1.f );

        res.type    = visualization_msgs::Marker::LINE_LIST;
        res.scale.x = 0.003;
        res.scale.y = 0.003;
        res.scale.z = 0.003;

        res.color.r = 0.f;
        res.color.g = 1.f;
        res.color.b = 0.f;

        return res;
}
inline visualization_msgs::Marker createChainPinMarker( unsigned id, std::string frame )
{
        visualization_msgs::Marker res = createMarker( "chainlink" + frame, id, frame, 1.f );

        res.type    = visualization_msgs::Marker::SPHERE_LIST;
        res.scale.x = 0.004;
        res.scale.y = 0.004;
        res.scale.z = 0.004;

        res.color.r = 0.f;
        res.color.g = 1.f;
        res.color.b = 0.f;

        return res;
}

template < typename ChainT, typename PinToPointFunction >
visualization_msgs::MarkerArray createChainVisualization( const ChainT&      ch,
                                                          std::string        frame,
                                                          PinToPointFunction pin_to_point )
{
        visualization_msgs::Marker link_msg = createChainLinkMarker( 0, frame );
        visualization_msgs::Marker pins_msg = createChainPinMarker( 1, frame );

        auto f = [&]( const auto& pin ) -> geometry_msgs::Point { return pointToMsg( pin_to_point( pin ) ); };

        if ( ch.begin() != ch.end() ) {
                pins_msg.points.push_back( f( ch.begin()->source_pin ) );
        }
        forEach( ch, [&]( const auto& view ) {
                link_msg.points.push_back( f( view.source_pin ) );
                link_msg.points.push_back( f( view.target_pin ) );
                pins_msg.points.push_back( f( view.target_pin ) );
        } );

        visualization_msgs::MarkerArray res;
        res.markers.push_back( link_msg );
        res.markers.push_back( pins_msg );
        return res;
}

}  // namespace schpin
