#include "schpin_lib/either.h"
#include "schpin_lib/error.h"
#include "schpin_lib/util.h"
#include "termios.h"
#include <boost/core/demangle.hpp>
#include <readline/history.h>
#include <readline/readline.h>
#include <tuple>
#include <variant>

#pragma once

namespace schpin
{
namespace console
{
        enum class Color
        {
                RESET   = 0,
                RED     = 31,
                BLUE    = 34,
                YELLOW  = 33,
                CYAN    = 36,
                MAGENTA = 35
        };

        std::ostream& operator<<( std::ostream& os, Color c )
        {
                os << "\033[0;" << int( c ) << "m";
                return os;
        }

        template < typename T = uint8_t >
        struct MatchResult
        {
                std::optional< T >         item;
                std::vector< std::string > return_path;
        };

        using NoEvent = void;

        struct ExitEvent
        {
        };

        struct CompletionContext
        {
                const std::string_view                        key;
                const std::vector< const CompletionContext* > children;

                CompletionContext( std::string_view key, std::vector< const CompletionContext* > children )
                  : key( key )
                  , children( children )
                {
                }
        };

        template < typename EventTypeT >
        struct ConsoleCmdContext
        {
                const std::vector< std::string > keywords;
                const Color                      color;
                const std::string                merged_keywords;
                const CompletionContext          completion;
                static constexpr bool            has_event = !std::is_same_v< EventTypeT, NoEvent >;

                using EventType = EventTypeT;

                ConsoleCmdContext( std::vector< std::string >              keywords,
                                   Color                                   color,
                                   std::vector< const CompletionContext* > compl_child = {} )
                  : keywords( keywords )
                  , color( color )
                  , merged_keywords( joined( keywords, std::string{", "} ) )
                  , completion( std::string_view{keywords[0]}, compl_child )
                {
                }
        };

        std::vector< std::string > prompt( const std::vector< std::string >& p, const CompletionContext* compl_node );

        struct HelpCommand
        {
                ConsoleCmdContext< NoEvent >     context{{"help", "h"}, Color::CYAN};
                const std::vector< std::string > description{"prints help of this level"};

                template < typename UserInput, typename ParentLevel, typename Iterator >
                MatchResult<> match( const UserInput& input, const Iterator&, const ParentLevel& parent ) const
                {
                        std::cout << std::endl;  // hmmmm ?
                        std::vector< std::string > keywords;

                        std::size_t indent = max(
                            parent.items, [&]( const auto& item ) { return item.context.merged_keywords.size(); } );
                        std::cout << std::left;

                        forEach( parent.items, [&]( const auto& item ) {
                                std::cout << "    " << item.context.color << std::setw( indent )
                                          << item.context.merged_keywords << Color::RESET << " - "
                                          << item.description.front() << std::endl;

                                forEach( tail( item.description ), [&]( const std::string& line ) {
                                        std::cout << "    " << std::setw( indent ) << " "
                                                  << "   " << line << std::endl;
                                } );
                        } );

                        std::cout << std::endl;
                        return {init( input )};
                }
        };

        struct ParentCommand
        {
                ConsoleCmdContext< NoEvent >     context{{".."}, Color::CYAN};
                const std::vector< std::string > description{"goes up in the control level"};

                template < typename UserInput, typename ParentLevel, typename Iterator >
                MatchResult<> match( const UserInput& input, const Iterator&, const ParentLevel& ) const
                {
                        if ( input.size() == 1 ) {
                                return {init( input )};
                        }
                        return {init( init( input ) )};
                }
        };

        struct ExitCommand
        {
                ConsoleCmdContext< ExitEvent >   context{{"quit", "q", "exit"}, Color::CYAN};
                const std::vector< std::string > description{"exits the entire app"};

                template < typename UserInput, typename ParentLevel >
                MatchResult< ExitEvent > match( const UserInput& input, const ParentLevel& ) const
                {
                        return {ExitEvent{}};
                }
        };

        struct LevelEvent
        {
        };

        struct LevelCommand
        {
                ConsoleCmdContext< LevelEvent > context{{""}, Color::CYAN};
        };

        template < typename... >
        struct FilterResultType;

        template <>
        struct FilterResultType<>
        {
                using type = std::tuple<>;
        };

        template < typename T, typename... Ts >
        struct FilterResultType< T, Ts... >
        {
                template < typename, typename >
                struct Concat;

                template < typename U, typename... Args >
                struct Concat< U, std::tuple< Args... > >
                {
                        using type = std::tuple< U, Args... >;
                };

                template < typename U, typename W >
                using ConcatT = typename Concat< U, W >::type;

                template < typename... Args >
                using FilterResultTypeT = typename FilterResultType< Args... >::type;

                static constexpr bool has_event = decltype( std::declval< T >().context )::has_event;

                using type = std::conditional_t< has_event,  //
                                                 ConcatT< T, FilterResultTypeT< Ts... > >,
                                                 FilterResultTypeT< Ts... > >;
        };

        template < typename T >
        using EventType = typename decltype( std::declval< T >().context )::EventType;

        template < typename... >
        struct LevelVariant
        {
        };

        template < typename... Ts >
        struct LevelVariant< std::tuple< Ts... > >
        {
                using type = std::variant< EventType< Ts >... >;
        };

        template < typename... Ts >
        struct LevelResultType
        {
                using Filtered = typename FilterResultType< Ts... >::type;

                using type = typename LevelVariant< Filtered >::type;
        };

        template < typename... Ts >
        using LevelResultTypeT = typename LevelResultType< Ts... >::type;

        template < typename >
        struct NoLevelResultType
        {
        };

        template < template < typename > class ResultTemplate, typename... Ts >
        struct Level
        {
                using EventType =
                    std::conditional_t< std::is_same_v< NoLevelResultType< uint8_t >, ResultTemplate< uint8_t > >,  //
                                        LevelResultTypeT< LevelCommand, Ts... >,                                    //
                                        ResultTemplate< LevelResultTypeT< LevelCommand, Ts... > > >;

                std::string                      name;
                const std::vector< std::string > description;
                std::tuple< Ts... >              items;
                ConsoleCmdContext< EventType >   context;

                Level( std::string name, std::string description_, Ts... items_ )
                  : name( name )
                  , description( {description_} )
                  , items( items_... )
                  , context( {name}, Color::MAGENTA, mapFtoV( items, []( const auto& item ) {
                                     return &item.context.completion;
                             } ) )
                {
                }

                template < typename UserInput, typename ParentLevel, typename Iterator >
                MatchResult< EventType > match( const UserInput& input,
                                                const Iterator&  input_cmd_iter,
                                                const ParentLevel& )
                {
                        return match( input, input_cmd_iter );
                }
                template < typename UserInput, typename ParentLevel, typename Iterator >
                MatchResult< EventType > match( const UserInput& input, const Iterator& input_cmd_iter )
                {

                        MatchResult< EventType > res{input};
                        if ( std::distance( input, input_cmd_iter ) == 1 ) {
                                return {input, EventType{LevelEvent{}}};
                        }
                        std::optional< MatchResult< EventType > > opt_child_res =
                            firstOptional( items, [&]( auto& item ) -> std::optional< EventType > {
                                    auto iter = findIf( item.context.keywords, [&]( const std::string& key ) {  //
                                            return key == *input_cmd_iter;
                                    } );
                                    if ( iter == item.context.keywords.end() ) {
                                            return {};
                                    }

                                    auto sub_res = item.match( input, std::next( input_cmd_iter ), *this );

                                    if constexpr ( decltype( item.context )::has_event ) {
                                            return {MatchResult{
                                                sub_res.return_path,
                                                EventType{std::move( *sub_res.item )},
                                            }};
                                    }

                                    return {MatchResult{sub_res.return_path}};
                            } );

                        if ( opt_child_res ) {
                                return *opt_child_res;
                        }
                        return res;
                }
        };  // namespace schpin

        template < typename... Ts >
        class Console
        {
                using RootLevel = Level< NoLevelResultType, Ts..., HelpCommand, ExitCommand >;
                using EventType = typename RootLevel::EventType;

                RootLevel root_;

              public:
                Console( Ts... items )
                  : root_( "", "", items..., HelpCommand{}, ExitCommand{} )
                {
                }

                MatchResult< EventType > promptMatch( const MatchResult< EventType >& prev ) const
                {
                        auto user_input = prompt( prev.return_path, &root_.context.completion );
                        return match(user_input);
                }

                template < typename UserInput >
                MatchResult< EventType > match( const UserInput& input ) const
                {
                        return root_.match( input, input.begin() );
                }
        };

        template < template < typename > class ResultType, typename... Ts >
        Level< ResultType, Ts..., HelpCommand, ParentCommand, ExitCommand > level( std::string name,
                                                                                   std::string description,
                                                                                   Ts... items )
        {
                return {name, description, items..., HelpCommand{}, ParentCommand{}, ExitCommand{}};
        }

        template < class T, typename UserInput, class Tuple, size_t... Is >
        Either< T, ParseError > constructFromArgs( const UserInput& input, Tuple&& tuple, std::index_sequence< Is... > )
        {
                auto access = [&]( std::size_t i ) {
                        if ( i < input.size() ) {
                                return input[i];
                        }
                        return std::string{};
                };
                return assembleLeftCollectRight(
                           std::get< Is >( std::forward< Tuple >( tuple ) ).parse( access( Is ) )... )
                    .convertLeft( [&]( const auto& t ) {  //
                            return T{std::get< Is >( t )...};
                    } )
                    .convertRight( [&]( const std::vector< ParseError >& errors ) {  //
                            return E( ParseError ).attach( errors ) << "Failed to construct return type";
                    } );
        }

        template < class T, typename UserInput, class Tuple >
        Either< T, ParseError > constructFromArgs( const UserInput& input, Tuple&& tuple )
        {
                if constexpr ( std::tuple_size_v< std::decay_t< Tuple > > != 0 ) {
                        return constructFromArgs< T >(
                            input,
                            std::forward< Tuple >( tuple ),
                            std::make_index_sequence< std::tuple_size< std::decay_t< Tuple > >::value >{} );
                } else {
                        return T{};
                }
        }

        template < typename EventTypeT, typename... Args >
        struct Command
        {
                using EventType = EventTypeT;
                ConsoleCmdContext< EventType > context;
                std::string                    name;
                std::vector< std::string >     description;
                std::tuple< Args... >          args;

                Command( std::string name, std::string description_line, Args... args_ )
                  : name( name )
                  , context( {name}, Color::BLUE )
                  , description()
                  , args( args_... )
                {
                        description.push_back( description_line );

                        if ( std::tuple_size< std::tuple< Args... > >::value != 0 ) {
                                description.push_back( "Arguments: " );
                        }

                        forEach( args, [&]( const auto& arg ) {  //
                                std::stringstream ss;
                                ss << "  " << Color::CYAN << arg.name << Color::RESET << " - " + arg.description << "("
                                   << arg.getTypeName() << ")";
                                description.push_back( ss.str() );
                        } );
                }

                template < typename UserInput, typename ParentLevel, typename Iterator >
                MatchResult< EventType > match( const UserInput& input, const Iterator& iter, const ParentLevel& ) const
                {
                        return constructFromArgs< EventType >( range( std::next( iter ), input.end() ), args )
                            .convertLeft( [&]( EventType e ) {
                                    return MatchResult< EventType >{init( input ), e};
                            } )
                            .convertRight( [&]( ParseError e ) {
                                    std::cout << e;
                                    return MatchResult< EventType >{init( input )};
                            } )
                            .join();
                }
        };

        template < typename EventType, typename... Args >
        Command< EventType, Args... > command( std::string name, std::string description, Args... args )
        {
                return {name, description, args...};
        }

        template < typename T >
        struct ArgParser
        {
                static Either< T, ParseError > parse( const std::string& word )
                {
                        if ( word.empty() ) {
                                return {E( ParseError ) << "Word is empty"};
                        }
                        T                 tmp;
                        std::stringstream ss{word};
                        ss >> tmp;
                        return {tmp};
                }
        };

        template < typename T >
        struct ArgParser< std::optional< T > >
        {
                static Either< std::optional< T >, ParseError > parse( const std::string& word )
                {
                        if ( word.empty() ) {
                                return {std::optional< T >{}};
                        }

                        return ArgParser< T >::parse( word ).convertLeft( [&]( T t ) {  //
                                return std::make_optional< T >( t );
                        } );
                }
        };

        template < typename T >
        struct Arg
        {
                std::string name;
                std::string description;

                using Type = T;

                Arg( std::string name, std::string description )
                  : name( name )
                  , description( description )
                {
                }

                std::string getTypeName() const
                {
                        char const* type_name = typeid( T ).name();
                        return boost::core::demangle( type_name );
                }

                Either< T, ParseError > parse( const std::string& word ) const
                {
                        return ArgParser< T >::parse( word );
                }
        };
}  // namespace console
using console::Arg;
using console::command;
using console::Console;
using console::ExitEvent;
using console::level;
using console::MatchResult;
using console::prompt;
}  // namespace schpin
