#include <ostream>
#include <type_traits>
#include <utility>

#pragma once

namespace schpin
{
template < typename LH, typename RH >
class Either
{
        union
        {
                LH left;
                RH right;
        };

        enum
        {
                LEFT_ITEM,
                RIGHT_ITEM
        } id;

      public:
        using LeftItem  = LH;
        using RightItem = RH;

        Either( LH&& lh ) noexcept
          : id( LEFT_ITEM )
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &left ) LH( std::move( lh ) );
        }

        Either( const LH& lh ) noexcept
          : id( LEFT_ITEM )
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &left ) LH( lh );
        }

        template < typename U = RH, typename = std::enable_if_t< !std::is_same_v< LH, U > > >
        Either( RH&& rh ) noexcept
          : id( RIGHT_ITEM )
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &right ) RH( std::move( rh ) );
        }

        template < typename U = RH, typename = std::enable_if_t< !std::is_same_v< LH, U > > >
        Either( const RH& rh ) noexcept
          : id( RIGHT_ITEM )
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &right ) RH( rh );
        }

        Either( const Either& other ) noexcept
          : id( other.id )
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        new ( &left ) LH( other.left );
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        new ( &right ) RH( other.right );
                }
        }

        Either( Either&& other ) noexcept
          : id( other.id )
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        new ( &left ) LH( std::move( other.left ) );
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        new ( &right ) RH( std::move( other.right ) );
                }
        }

        Either& operator=( const Either& other )
        {
                if ( other.id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        *this = other.left;
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        *this = other.right;
                }
                return *this;
        }

        Either& operator=( Either&& other ) noexcept
        {
                if ( other.id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        *this = std::move( other.left );
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        *this = std::move( other.right );
                }
                return *this;
        }

        Either& operator=( const LH& other )
        {
                destruct();
                id = LEFT_ITEM;
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &left ) LH( other );
                return *this;
        }
        Either& operator=( LH&& other )
        {
                destruct();
                id = LEFT_ITEM;
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &left ) LH( std::move( other ) );
                return *this;
        }

        template < typename U = RH, typename = std::enable_if_t< !std::is_same_v< LH, U > > >
        Either& operator=( const RH& other )
        {
                destruct();
                id = RIGHT_ITEM;
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &right ) RH( other );
                return *this;
        }

        template < typename U = RH, typename = std::enable_if_t< !std::is_same_v< LH, U > > >
        Either& operator=( RH&& other )
        {
                destruct();
                id = RIGHT_ITEM;
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                new ( &right ) RH( std::move( other ) );
                return *this;
        }

        bool isLeft() const
        {
                return id == LEFT_ITEM;
        }

        template < typename LeftFunction >
        auto convertLeft( LeftFunction left_f ) &
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                using ReturnEither = Either< decltype( left_f( left ) ), RightItem >;

                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return ReturnEither{left_f( left )};
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return ReturnEither{right};
        }

        template < typename LeftFunction >
        auto convertLeft( LeftFunction left_f ) &&
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                using ReturnEither = Either< decltype( left_f( std::move( left ) ) ), RightItem >;

                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return ReturnEither{left_f( std::move( left ) )};
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return ReturnEither{std::move( right )};
        }

        template < typename RightFunction >
        auto convertRight( RightFunction right_f ) &
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                using ReturnEither = Either< LeftItem, decltype( right_f( right ) ) >;

                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return ReturnEither{left};
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return ReturnEither{right_f( right )};
        }

        template < typename RightFunction >
        auto convertRight( RightFunction right_f ) &&
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                using ReturnEither = Either< LeftItem, decltype( right_f( std::move( right ) ) ) >;

                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return ReturnEither{std::move( left )};
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return ReturnEither{right_f( std::move( right ) )};
        }

        template < typename LeftFunction, typename RightFunction >
        void match( LeftFunction left_f, RightFunction right_f ) &
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        left_f( left );
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        right_f( right );
                }
        }

        template < typename LeftFunction, typename RightFunction >
        void match( LeftFunction left_f, RightFunction right_f ) const&
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        left_f( left );
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        right_f( right );
                }
        }

        template < typename LeftFunction, typename RightFunction >
        void match( LeftFunction left_f, RightFunction right_f ) &&
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        left_f( std::move( left ) );
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        right_f( std::move( right ) );
                }
        }

        template < typename U = LH, typename K = RH >
        std::enable_if_t< std::is_same_v< U, K >, LH > join() &&
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return std::move( left );
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return std::move( right );
        }
        template < typename U = LH, typename K = RH >
        std::enable_if_t< std::is_same_v< U, K >, LH > join() &
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return left;
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return right;
        }

        template < typename UnaryFunction >
        auto transformLeft( UnaryFunction left_f ) &
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                using ReturnEither = decltype( left_f( left ) );

                static_assert( std::is_same_v< typename ReturnEither::RightItem, RH >,
                               "In transform left, the RightTypes has to be same!" );

                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return left_f( left );
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return ReturnEither{right};
        }
        template < typename UnaryFunction >
        auto transformLeft( UnaryFunction left_f ) &&
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                using ReturnEither = decltype( left_f( std::move( left ) ) );

                static_assert( std::is_same_v< typename ReturnEither::RightItem, RH >,
                               "In transform left, the RightTypes has to be same!" );

                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        return left_f( std::move( left ) );
                }

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                return ReturnEither{std::move( right )};
        }

        ~Either()
        {
                destruct();
        }

      private:
        void destruct()
        {
                if ( id == LEFT_ITEM ) {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        left.~LH();
                } else {
                        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
                        right.~RH();
                }
        }
};

template < typename LH, typename RH >
std::ostream& operator<<( std::ostream& os, const Either< LH, RH >& env )
{
        env.match( [&]( const LH& lh ) { os << lh; }, [&]( const RH& rh ) { os << rh; } );
        return os;
}

}  // namespace schpin
