#include "schpin_lib/either.h"
#include <optional>
#include <schpin_lib/util.h>
#include <sstream>
#include <string>
#include <type_traits>
#include <utility>
#include <vector>

#pragma once

namespace schpin
{

struct ErrorTagLess
{
};

#define E( error_t ) error_t( __PRETTY_FUNCTION__ )

template < typename Tag >
class Error
{
        std::vector< Error< ErrorTagLess > > collection_;

        std::string msg_;
        std::string function_name_;

      public:
        explicit Error( std::string function_name )
          : function_name_( std::move( function_name ) )
        {
        }

        Error( Error&& other ) noexcept = default;
        Error( const Error& other )     = default;
        Error& operator=( Error&& other ) noexcept = default;
        Error& operator=( const Error& other ) = default;

        const std::vector< Error< ErrorTagLess > >& getCollection() const
        {
                return collection_;
        }

        const std::string& getMessage() const
        {
                return msg_;
        }

        const std::string& getFunctionName() const
        {
                return function_name_;
        }

        template < typename T >
        Error& operator<<( const T& item )
        {
                std::stringstream ss;
                ss << item;
                msg_ += ss.str();
                return *this;
        }

        template < typename OtherTag >
        Error& attach( const Error< OtherTag >& other )
        {
                collection_.emplace_back( other.getFunctionName() );
                collection_.back().attach( other.getCollection() ) << other.getMessage();
                return *this;
        }

        template < typename OtherTag >
        Error& attach( const std::vector< Error< OtherTag > >& other )
        {
                forEach( other, [&]( const Error< OtherTag >& other_error ) { this->attach( other_error ); } );
                return *this;
        }

        void printRec( std::ostream& os, std::string prefix, bool is_last ) const
        {
                os << prefix << "+-\033[1;31m " << function_name_ << "\033[0m \n";

                prefix += is_last ? "  " : "| ";

                if ( !msg_.empty() ) {
                        os << prefix << msg_ << "\n";
                }
                forEach( range( collection_.size() ),
                         [&]( std::size_t i ) { collection_[i].printRec( os, prefix, i == collection_.size() - 1 ); } );
        };

        ~Error() noexcept;
};

template < typename T >
Error< T >::~Error() noexcept = default;

template < typename Tag >
std::ostream& operator<<( std::ostream& os, const Error< Tag >& error )
{
        error.printRec( os, std::string{""}, true );
        return os;
}

template < typename Error >
class OptError
{
        std::optional< Error > error_;

      public:
        OptError( Error err )
          : error_( err )
        {
        }

        OptError()
          : error_()
        {
        }

        template < typename UnaryFunction >
        auto convert( UnaryFunction f ) -> OptError< decltype( f( *error_ ) ) >
        {
                if ( error_ ) {
                        return {f( *error_ )};
                }
                return {};
        }

        template < typename UnaryFunction >
        void optionally( UnaryFunction f )
        {
                if ( error_ ) {
                        f( *error_ );
                }
        }

        template < typename T >
        Either< T, Error > operator||( const T& item )
        {
                if ( error_ ) {
                        return {*error_};
                }

                return {item};
        }
        template < typename T >
        Either< T, Error > operator||( T&& item )
        {
                if ( error_ ) {
                        return {*error_};
                }

                return {std::forward< T >( item )};
        }
};

template < typename Error >
std::vector< Error > collectOptErrors( std::initializer_list< OptError< Error > > errors )
{
        std::vector< Error > res;
        forEach( errors, [&]( OptError< Error > opt_error ) {
                opt_error.optionally( [&]( Error err ) { res.push_back( err ); } );
        } );
        return res;
}

struct ParseErrorTag
{
};
using ParseError = Error< ParseErrorTag >;

}  // namespace schpin
