/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 *
 * This file is part of Schpin.
 *
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */

#include "schpin_lib/util.h"
#include <boost/integer.hpp>
#include <nlohmann/json.hpp>

#pragma once

namespace schpin
{
/** Class designed to create a identificator data-type for data containers.
 *
 *
 * The class derived than stores the unsigned value passed during the creation and allows for modification by suffix
 * and preffix operators --/++. There is also a full set of operators overloads to compare identificators of same
 * type. The actual value can either by accessed with the operator*() method or the object can explicitly be converted
 * into std::size_t. This is considered enough for identificator of container.
 *
 * Various parts of the class have asserts for integer overflow or underflow.
 *
 * Usage:
 *
 *  MyIdentificator : Identificator< MyIdentificator, 32 > {}
 *
 */
template < typename Tag, unsigned bits >
class Identificator
{

public:
  typedef typename boost::uint_t< bits >::least    ValueType;
  typedef typename boost::int_t< bits + 1 >::least DifferenceType;

private:
  ValueType value_;

public:
  /** Constructors, the default should always be 0 */
  Identificator()
    : value_( 0 )
  {
  }

  /** Primary constructor for usage, this is able to use maximal size of unsigned integer and use an assert to make
   * sure that it is within range of actually used data type. */
  Identificator( uintmax_t val )
    : value_( ValueType( val ) )
  {
    // WARN: as long as ValueType is unsigned, this is enough
    SCHPIN_ASSERT( val < std::numeric_limits< ValueType >::max() );
  }

  /** Manual conversion of the object into std::size_t */
  explicit operator std::size_t() const
  {
    return std::size_t( value_ );
  }

  /** Accessor for the internal value, only const version is available */
  const ValueType& operator*() const
  {
    return value_;
  }

  /** increment/decrement operators */
  // -----------------------------------------------------------------
  Identificator& operator++()
  {
    SCHPIN_ASSERT( value_ != std::numeric_limits< ValueType >::max() );
    value_ += 1;
    return *this;
  }
  Identificator& operator--()
  {
    SCHPIN_ASSERT( value_ != std::numeric_limits< ValueType >::min() );
    value_ -= 1;
    return *this;
  }

  Identificator operator++( int )
  {
    Identificator tmp( *this );

    operator++();

    return tmp;
  }

  Identificator operator--( int )
  {
    Identificator tmp( *this );

    operator--();

    return tmp;
  }
  //-------------------------------------------------------------------
};

template < typename Tag, unsigned bits >
void to_json( nlohmann::json& j, const Identificator< Tag, bits >& id )
{
  j = *id;
}

template < typename Tag, unsigned bits >
void from_json( const nlohmann::json& j, Identificator< Tag, bits >& id )
{
  id = Identificator< Tag, bits >{j.get< unsigned long >()};
}

// Definition of operators for the identificaotrs
// ------------------------------------------------------------------
// First set defines basic operators operators

template < typename Tag, unsigned bits >
std::ostream& operator<<( std::ostream& os, const Identificator< Tag, bits >& id )
{
  os << *id;
  return os;
}

template < typename Tag, unsigned bits >
typename Identificator< Tag, bits >::DifferenceType operator-( const Identificator< Tag, bits >& lhs,
                                                               const Identificator< Tag, bits >& rhs )
{
  typedef typename Identificator< Tag, bits >::DifferenceType DifferenceType;
  return DifferenceType( *lhs ) - DifferenceType( *rhs );
}

template < typename Tag, unsigned bits >
inline bool operator<( const Identificator< Tag, bits >& lhs, const Identificator< Tag, bits >& rhs )
{
  return *lhs < *rhs;
}

template < typename Tag, unsigned bits >
inline bool operator==( const Identificator< Tag, bits >& lhs, const Identificator< Tag, bits >& rhs )
{
  return *lhs == *rhs;
}

// Second set is defined solely based on the firt set
template < typename Tag, unsigned bits >
inline bool operator!=( const Identificator< Tag, bits >& lhs, const Identificator< Tag, bits >& rhs )
{
  return !operator==( lhs, rhs );
}

template < typename Tag, unsigned bits >
inline bool operator>( const Identificator< Tag, bits >& lhs, const Identificator< Tag, bits >& rhs )
{
  return operator<( rhs, lhs );
}

template < typename Tag, unsigned bits >
inline bool operator<=( const Identificator< Tag, bits >& lhs, const Identificator< Tag, bits >& rhs )
{
  return !operator>( lhs, rhs );
}

template < typename Tag, unsigned bits >
inline bool operator>=( const Identificator< Tag, bits >& lhs, const Identificator< Tag, bits >& rhs )
{
  return !operator<( lhs, rhs );
}
}  // namespace schpin
