#include <list>
#include <nlohmann/json.hpp>
#include <string>

#pragma once

namespace schpin
{

template < typename Item, typename Identificator, template < typename... > class Storage >
class Container
{
      public:
        using StorageType    = Storage< Item >;
        using Iterator       = typename StorageType::iterator;
        using iterator       = Iterator;
        using ConstIterator  = typename StorageType::const_iterator;
        using const_iterator = ConstIterator;
        using ValueType      = Item;
        using value_type     = ValueType;
        using Reference      = ValueType&;
        using ConstReference = const ValueType&;
        using SizeType       = Identificator;

        using ReverseIterator      = typename StorageType::reverse_iterator;
        using ConstReverseIterator = typename StorageType::const_reverse_iterator;

      private:
        StorageType storage_;

      public:
        Container()
          : storage_()
        {
        }
        Container( const StorageType& storage )
          : storage_( storage )
        {
        }
        Container( StorageType&& storage )
          : storage_( std::move( storage ) )
        {
        }

        const StorageType& operator*() const
        {
                return storage_;
        }

        const StorageType* operator->() const
        {
                return storage_;
        }

        ConstReference operator[]( const Identificator& id ) const
        {
                return storage_[*id];
        }
        Reference operator[]( const Identificator& id )
        {
                return storage_[*id];
        }

        Iterator begin()
        {
                return storage_.begin();
        }

        Iterator end()
        {
                return storage_.end();
        }

        ConstIterator begin() const
        {
                return storage_.cbegin();
        }

        ConstIterator end() const
        {
                return storage_.cend();
        }

        ReverseIterator rbegin()
        {
                return storage_.rbegin();
        }

        ReverseIterator rend()
        {
                return storage_.rend();
        }

        ConstReverseIterator rbegin() const
        {
                return storage_.rbegin();
        }

        ConstReverseIterator rend() const
        {
                return storage_.rend();
        }

        bool empty() const
        {
                return storage_.empty();
        }

        SizeType size() const
        {
                return storage_.size();
        }

        SizeType capacity() const
        {
                return storage_.capacity();
        }

        bool hasID( const Identificator& id )
        {
                return *id >= 0 && *id < storage_.size();
        }

        void resize( const Identificator& id )
        {
                storage_.resize( *id );
        }

        void reserve( const Identificator& id )
        {
                storage_.reserve( *id );
        }

        void shrinkToFit()
        {
                storage_.shrink_to_fit();
        }

        template < typename Container >
        void insert( ConstIterator iter, Container& cont )
        {
                storage_.insert( iter, cont.begin(), cont.end() );
        }

        std::back_insert_iterator< Container > makeBackInserter()
        {
                return std::back_inserter( storage_ );
        }

        template < typename... Args >
        ConstIterator emplace( ConstIterator iter, Args... args )
        {
                return storage_.emplace( iter, args... );
        }

        template < typename... Args >
        Identificator emplaceBack( Args... args )
        {
                Identificator res( size() );

                storage_.emplace_back( args... );

                return res;
        }

        void popBack()
        {
                storage_.pop_back();
        }

        template < typename... Args >
        Identificator emplace_back( Args... args )
        {
                return emplaceBack( args... );
        }

        ValueType& front()
        {
                return storage_.front();
        }

        ValueType& back()
        {
                return storage_.back();
        }

        ValueType& at( const Identificator& id )
        {
                return storage_.at( *id );
        }

        const ValueType& at( const Identificator& id ) const
        {
                return storage_.at( *id );
        }

        void erase( const ConstIterator& iter )
        {
                storage_.erase( iter );
        }
};

template < typename Item, typename Identificator, template < typename... > class Storage >
void to_json( nlohmann::json& json, const Container< Item, Identificator, Storage >& container )
{
        json = nlohmann::json{*container};
}

template < typename Item, typename Identificator, template < typename... > class Storage >
void from_json( const nlohmann::json& json, Container< Item, Identificator, Storage >& container )
{
        container = Container< Item, Identificator, Storage >( json.get< Storage< Item > >() );
}

}  // namespace schpin
