/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 *
 * This file is part of Schpin.
 *
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */

#include "schpin_lib/quantity.h"

#pragma once

/** Credits should go to https://github.com/joboccara/NamedType as I inspired by project by this blogger!
     @todo Make the mechanism recognize 'strings' for each tag, 's', 'm', 'g' and so on  */

namespace schpin
{

template < int l, int mass, int t, int current, int temp, int mol, int li, int angle, int byte >
struct PhysicalQuantityTag : QuantityTag
{
  static std::string getUnit()
  {
    auto seg = []( std::string unit, int i ) -> std::string {
      if ( i == 0 )
      {
        return std::string();
      }
      if ( i == 1 )
      {
        return unit;
      }
      return unit + "^" + std::to_string( i );
    };
    return seg( "m", l ) + seg( "g", mass ) + seg( "s", t ) + seg( "A", current ) + seg( "K", temp ) +
           seg( "mol", mol ) + seg( "cd", li ) + seg( "rad", angle ) + seg( "B", byte );
  }
};

using Unitless          = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 0, 0, 0, 0, 0, 0, 0 >, float >;
using Length            = Quantity< NeutralRatio, PhysicalQuantityTag< 1, 0, 0, 0, 0, 0, 0, 0, 0 >, float >;
using Mass              = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 1, 0, 0, 0, 0, 0, 0, 0 >, float >;
using Time              = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 1, 0, 0, 0, 0, 0, 0 >, float >;
using Current           = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 0, 1, 0, 0, 0, 0, 0 >, float >;
using Temp              = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 0, 0, 1, 0, 0, 0, 0 >, float >;
using AmountOfSubstance = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 0, 0, 0, 1, 0, 0, 0 >, float >;
using LuminousIntensity = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 0, 0, 0, 0, 1, 0, 0 >, float >;
using Angle             = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 0, 0, 0, 0, 0, 1, 0 >, float >;
using Byte              = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, 0, 0, 0, 0, 0, 0, 1 >, float >;

using AngularVelocity = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, -1, 0, 0, 0, 0, 1, 0 >, float >;
using Area            = Quantity< NeutralRatio, PhysicalQuantityTag< 2, 0, 0, 0, 0, 0, 0, 0, 0 >, float >;
using Volume          = Quantity< NeutralRatio, PhysicalQuantityTag< 3, 0, 0, 0, 0, 0, 0, 0, 0 >, float >;
using Velocity        = Quantity< NeutralRatio, PhysicalQuantityTag< 1, 0, -1, 0, 0, 0, 0, 0, 0 >, float >;
using Frequency       = Quantity< NeutralRatio, PhysicalQuantityTag< 0, 0, -1, 0, 0, 0, 0, 0, 0 >, float >;
using Force           = Quantity< NeutralRatio, PhysicalQuantityTag< 1, 1, -2, 0, 0, 0, 0, 0, 0 >, float >;
using Power           = Quantity< NeutralRatio, PhysicalQuantityTag< 2, 1, -3, 0, 0, 0, 0, 0, 0 >, float >;
using Voltage         = Quantity< NeutralRatio, PhysicalQuantityTag< 2, 1, -3, -1, 0, 0, 0, 0, 0 >, float >;
using Resistance      = Quantity< NeutralRatio, PhysicalQuantityTag< 2, 1, -3, -2, 0, 0, 0, 0, 0 >, float >;

// note: the *,/ operators are not handling the Ratio correctly
//
template < typename Ratio0,
           int l0,
           int mass0,
           int t0,
           int curr0,
           int temp0,
           int mol0,
           int li0,
           int angle0,
           int byte0,
           typename Ratio1,
           int l1,
           int mass1,
           int t1,
           int curr1,
           int temp1,
           int mol1,
           int li1,
           int angle1,
           int byte1,
           typename ValueType >
auto operator*(
    const Quantity< Ratio0, PhysicalQuantityTag< l0, mass0, t0, curr0, temp0, mol0, li0, angle0, byte0 >, ValueType >&
        lh,
    const Quantity< Ratio1, PhysicalQuantityTag< l1, mass1, t1, curr1, temp1, mol1, li1, angle1, byte1 >, ValueType >&
        rh )
{
  return Quantity< std::ratio_multiply< Ratio0, Ratio1 >,
                   PhysicalQuantityTag< l0 + l1,
                                        mass0 + mass1,
                                        t0 + t1,
                                        curr0 + curr1,
                                        temp0 + temp1,
                                        mol0 + mol1,
                                        li0 + li1,
                                        angle0 + angle1,
                                        byte0 + byte1 >,
                   ValueType >{( *lh ) * ( *rh )};
}
template < typename Ratio0,
           int l0,
           int mass0,
           int t0,
           int curr0,
           int temp0,
           int mol0,
           int li0,
           int angle0,
           int byte0,
           typename Ratio1,
           int l1,
           int mass1,
           int t1,
           int curr1,
           int temp1,
           int mol1,
           int li1,
           int angle1,
           int byte1,
           typename ValueType >
auto operator/(
    const Quantity< Ratio0, PhysicalQuantityTag< l0, mass0, t0, curr0, temp0, mol0, li0, angle0, byte0 >, ValueType >&
        lh,
    const Quantity< Ratio1, PhysicalQuantityTag< l1, mass1, t1, curr1, temp1, mol1, li1, angle1, byte1 >, ValueType >&
        rh )
{
  return Quantity< std::ratio_divide< Ratio0, Ratio1 >,
                   PhysicalQuantityTag< l0 - l1,
                                        mass0 - mass1,
                                        t0 - t1,
                                        curr0 - curr1,
                                        temp0 - temp1,
                                        mol0 - mol1,
                                        li0 - li1,
                                        angle0 - angle1,
                                        byte0 - byte1 >,
                   ValueType >{( *lh ) / ( *rh )};
}

}  // namespace schpin
