#include <cmath>
#include <limits>
#include <nlohmann/json.hpp>
#include <ostream>
#include <ratio>
#include <string_view>
#include <type_traits>

#pragma once

namespace schpin
{

/** Base structure for other tags of quantity.
 *
 * Quantity tag is datatype that marks the semantical meaning of specific physical quantity.
 * That datatype is preferably struct, that inherits 'QuantityTag'. This is verified inside the Quantity
 * class.
 *
 * For a physical quantity that should represent 'distance', this means that the tag should be 'struct DistanceTag :
 * QuantityTag'
 *
 */
struct QuantityTag
{
};

template < typename T >
constexpr std::string_view ratio_prefix = "";

using Atto  = std::atto;
using Centi = std::centi;
template <>
constexpr std::string_view ratio_prefix< Centi > = "c";

using Deca = std::deca;
using Deci = std::deci;
template <>
constexpr std::string_view ratio_prefix< Deci > = "d";

using Exa   = std::exa;
using Femto = std::femto;
using Giga  = std::giga;
template <>
constexpr std::string_view ratio_prefix< Giga > = "g";

using Hecto = std::hecto;
using Kilo  = std::kilo;
template <>
constexpr std::string_view ratio_prefix< Kilo > = "k";

using Mega = std::mega;
template <>
constexpr std::string_view ratio_prefix< Mega > = "m";

using Micro = std::micro;
template <>
constexpr std::string_view ratio_prefix< Micro > = "u";

using Milli = std::milli;
template <>
constexpr std::string_view ratio_prefix< Milli > = "m";

using Nano = std::nano;
using Peta = std::peta;
using Pico = std::pico;
using Tera = std::tera;
template <>
constexpr std::string_view ratio_prefix< Tera > = "t";

using NeutralRatio = std::ratio< 1, 1 >;

/** Structure used to detect if type is std::ratio
 *
 * The is_ratio<T>::value is true if the type T is std::ratio<N,D> with some parameters N,D.
 * The is_ratio<T>::value is false otherwise.
 *
 * This is used internally by Quantity.
 *
 * @param T is type which is analyzed.
 */
template < typename T >
struct is_ratio
{
  template < typename U >
  struct test : std::false_type
  {
  };

  template < std::intmax_t Num, std::intmax_t Denom >
  struct test< std::ratio< Num, Denom > > : std::true_type
  {
  };

  static constexpr bool value = test< T >::value;
};

/** Class representing generic quantity.
 *
 * This class is used to generate a type for specific quantity.
 * Idea is that most of the quantites behave same, but we are missing strong typedefs in C++.
 * Hence the best solution to have strong types representing the quantities, is to create a template class that can be
 * used to represent the specific quantities.
 *
 * The 'Tag' datatype is what distinguishes quantities between each other on datatype level.
 * The standart set of operators is defined in a way that allows only quantities with same Tag to interact with
 * each other. The type itself is however not used internally.
 *
 * Idea is that you create your custom tag, such as 'struct DistanceTag : QuantityTag {}' and create a distance
 * datatype: 'using Distance = Quantity< NeutralRatio, DistanceTag>;'. Multiple types representing physical
 * quantities can be created and as long as they have different tags, they won't interact with each other by default.
 *
 * Set of operations that this class allows should be standart set required, so this 'feels' like using standart
 * numerical datatype.
 *
 * The 'Ratio' typename is used to represent scaled. The scale itself is stored with a usage of std::ratio<N,D>
 * template. This template also provided ratios for several known metric prefixes, which are provided with this class.
 *
 * Datatypes with different ratios, but same tags, are convertible between each other. Either automatically through copy
 * constructor or with the 'toRatio<R>()' template method. However operators are not made to interact between different
 * ratios, as the conversion can be costly in case two different ratios interact extensively.
 *
 * @param Ratio template param that represents scale of actual physical quantity.
 * @param Tag template param that specifies the semantical meaning of the physical quantity.
 *
 * Both parameters for this class are verified with static assert. It is expected that 'Ratio' parameter is std::ratio<>
 * template and that 'Tag' parameter is base of QuantityTag.
 *
 */
template < typename Ratio, typename Tag, typename ValueType = double >
class Quantity final
{
  static_assert( is_ratio< Ratio >::value, "The Ratio provided to Quantity should be std::ratio template" );
  static_assert( std::is_base_of< QuantityTag, Tag >::value,
                 "The Tag provided to Quantity should inherit from the QuantityTag" );
  ValueType value_;

public:
  Quantity() noexcept
    : value_( 0 )
  {
  }

  /* Default constructor used to create a physical quantity from value */
  explicit Quantity( ValueType val ) noexcept
    : value_( val )
  {
  }

  /* Constructo that allows conversion between physical quantities with same tag, but different ratio. The internall
   * value is correctly recalculated */
  template < typename OtherRatio >
  Quantity( const Quantity< OtherRatio, Tag, ValueType >& other )
    : value_( 0 )
  {
    ValueType scale = ValueType( OtherRatio::num * Ratio::den ) / ValueType( OtherRatio::den * Ratio::num );
    value_          = *other * scale;
  }

  /* Method to manually convert from actual ratio to specified ratio. */
  template < typename TargetRatio >
  Quantity< TargetRatio, Tag, ValueType > toRatio()
  {
    return Quantity< TargetRatio, Tag, ValueType >( *this );
  }

  /* Const reference to the internal value */
  const ValueType operator*() const
  {
    return value_;
  }

  Quantity& operator+=( const Quantity& other )
  {
    value_ += *other;
    return *this;
  }
  Quantity& operator-=( const Quantity& other )
  {
    value_ -= *other;
    return *this;
  }
  Quantity& operator/=( const ValueType val )
  {
    value_ /= val;
    return *this;
  }
  Quantity& operator*=( const ValueType val )
  {
    value_ *= val;
    return *this;
  }

  explicit operator float() const
  {
    return float( value_ );
  }
};

template < typename Ratio, typename Tag, typename ValueType >
inline Quantity< Ratio, Tag, ValueType > operator+( const Quantity< Ratio, Tag, ValueType >& lhs,
                                                    const Quantity< Ratio, Tag, ValueType >& rhs )
{
  Quantity< Ratio, Tag, ValueType > res( *lhs );
  res += rhs;
  return res;
}

template < typename Ratio, typename Tag, typename ValueType >
inline Quantity< Ratio, Tag, ValueType > operator-( const Quantity< Ratio, Tag, ValueType >& lhs,
                                                    const Quantity< Ratio, Tag, ValueType >& rhs )
{
  Quantity< Ratio, Tag, ValueType > res( *lhs );
  res -= rhs;
  return res;
}

template < typename Ratio, typename Tag, typename ValueType >
inline Quantity< Ratio, Tag, ValueType > operator-( const Quantity< Ratio, Tag, ValueType >& val )
{
  return Quantity< Ratio, Tag, ValueType >{-*val};
}

template < typename Ratio, typename Tag, typename ValueType >
inline bool operator==( const Quantity< Ratio, Tag, ValueType >& lhs, const Quantity< Ratio, Tag, ValueType >& rhs )
{
  return *lhs == *rhs;
}

template < typename Ratio, typename Tag, typename ValueType >
inline bool operator<( const Quantity< Ratio, Tag, ValueType >& lhs, const Quantity< Ratio, Tag, ValueType >& rhs )
{
  return *lhs < *rhs;
}

template < typename Ratio, typename Tag, typename ValueType >
Quantity< Ratio, Tag, ValueType > operator*( const Quantity< Ratio, Tag, ValueType > pq, const ValueType val )
{
  Quantity< Ratio, Tag, ValueType > res( *pq );
  res *= val;
  return res;
}

template < typename Ratio, typename Tag, typename ValueType >
Quantity< Ratio, Tag, ValueType > operator/( const Quantity< Ratio, Tag, ValueType > pq, const ValueType val )
{
  Quantity< Ratio, Tag, ValueType > res( *pq );
  res /= val;
  return res;
}

template < typename Ratio, typename Tag, typename ValueType >
Quantity< Ratio, Tag, ValueType > abs( Quantity< Ratio, Tag, ValueType > a )
{
  return Quantity< Ratio, Tag, ValueType >( std::abs( *a ) );
}

template < typename Ratio, typename Tag, typename ValueType >
double cos( const Quantity< Ratio, Tag, ValueType > u )
{
  return std::cos( *u );
}

template < typename Ratio, typename Tag, typename ValueType >
double sin( const Quantity< Ratio, Tag, ValueType > u )
{
  return std::sin( *u );
}

template < typename Ratio, typename Tag, typename ValueType >
Quantity< Ratio, Tag, ValueType > max( const Quantity< Ratio, Tag, ValueType >& lh,
                                       const Quantity< Ratio, Tag, ValueType >& rh )
{
  return Quantity< Ratio, Tag, ValueType >( std::max( *lh, *rh ) );
}

template < typename Ratio, typename Tag, typename ValueType >
Quantity< Ratio, Tag, ValueType > min( const Quantity< Ratio, Tag, ValueType >& lh,
                                       const Quantity< Ratio, Tag, ValueType >& rh )
{
  return Quantity< Ratio, Tag, ValueType >( std::min( *lh, *rh ) );
}

/* @todo print also actual ratio and tag */
template < typename Ratio, typename Tag, typename ValueType >
std::ostream& operator<<( std::ostream& os, const Quantity< Ratio, Tag, ValueType >& id )
{
  os << *id << ratio_prefix< Ratio > << Tag::getUnit();
  return os;
}

template < typename Ratio, typename Tag, typename ValueType >
void to_json( nlohmann::json& j, const Quantity< Ratio, Tag, ValueType >& id )
{
  j = *id;
}

template < typename Ratio, typename Tag, typename ValueType >
void from_json( nlohmann::json& j, Quantity< Ratio, Tag, ValueType >& id )
{
  id = Quantity< Ratio, Tag, ValueType >( j.get< ValueType >() );
}

//---------------------------------------------------------------------------

template < typename Ratio, typename Tag, typename ValueType >
inline bool operator!=( const Quantity< Ratio, Tag, ValueType >& lhs, const Quantity< Ratio, Tag, ValueType >& rhs )
{
  return !operator==( lhs, rhs );
}
template < typename Ratio, typename Tag, typename ValueType >
inline bool operator>( const Quantity< Ratio, Tag, ValueType >& lhs, const Quantity< Ratio, Tag, ValueType >& rhs )
{
  return operator<( rhs, lhs );
}
template < typename Ratio, typename Tag, typename ValueType >
inline bool operator<=( const Quantity< Ratio, Tag, ValueType >& lhs, const Quantity< Ratio, Tag, ValueType >& rhs )
{
  return !operator>( lhs, rhs );
}
template < typename Ratio, typename Tag, typename ValueType >
inline bool operator>=( const Quantity< Ratio, Tag, ValueType >& lhs, const Quantity< Ratio, Tag, ValueType >& rhs )
{
  return !operator<( lhs, rhs );
}
//---------------------------------------------------------------------------

template < typename Ratio, typename Tag, typename ValueType >
Quantity< Ratio, Tag, ValueType > operator*( const ValueType val, const Quantity< Ratio, Tag, ValueType > pq )
{
  return pq * val;
}
template < typename Ratio, typename Tag, typename ValueType >
inline ValueType operator/( const ValueType val, const Quantity< Ratio, Tag, ValueType >& pq )
{
  return val / *pq;
}

}  // namespace schpin

namespace std
{
template < typename Ratio, typename Tag, typename ValueType >
class numeric_limits< schpin::Quantity< Ratio, Tag, ValueType > >
{
public:
  static schpin::Quantity< Ratio, Tag, ValueType > min()
  {
    return schpin::Quantity< Ratio, Tag, ValueType >{std::numeric_limits< ValueType >::min()};
  }
  static schpin::Quantity< Ratio, Tag, ValueType > max()
  {
    return schpin::Quantity< Ratio, Tag, ValueType >{std::numeric_limits< ValueType >::max()};
  }
};
}  // namespace std
namespace nlohmann
{
template < typename Ratio, typename Tag, typename ValueType >
struct adl_serializer< schpin::Quantity< Ratio, Tag, ValueType > >
{
  static void to_json( nlohmann::json& json, const schpin::Quantity< Ratio, Tag, ValueType >& quantity )
  {
    if ( Tag::getUnit() == "" )
    {
      json = *quantity;
    }
    else
    {
      json = nlohmann::json{{Tag::getUnit(), *quantity}};
    }
  }
  static schpin::Quantity< Ratio, Tag, ValueType > from_json( const nlohmann::json& j )
  {
    if ( Tag::getUnit() == "" )
    {
      return schpin::Quantity< Ratio, Tag, ValueType >{j.get< ValueType >()};
    }
    else
    {
      return schpin::Quantity< Ratio, Tag, ValueType >{j[Tag::getUnit()].template get< ValueType >()};
    }
  }
};
}  // namespace nlohmann
