#include <functional>

#pragma once

namespace schpin
{

// dummy token for construction purpose
struct SegmentizedIteratorEndToken
{
};

template < typename Iterator, typename SubIter >
class SegmentizedIterator
{
      public:
        using Segment        = decltype( *std::declval< Iterator >() );
        using AccessFunction = std::function< Range< SubIter >( const Segment& seg ) >;

      private:
        Range< Iterator > segment_range_;
        Iterator          segment_iter_;
        Range< SubIter >  sub_range_;
        SubIter           sub_range_iter_;
        AccessFunction    access_f_;

      public:
        using ValueType       = decltype( *sub_range_iter_ );
        using difference_type = decltype( std::declval< SubIter >() - std::declval< SubIter >() );

        SegmentizedIterator( Range< Iterator > segment_range, Iterator segment_iter, AccessFunction f )
          : segment_range_( segment_range )
          , segment_iter_( segment_iter )
          , sub_range_( f( *segment_iter_ ) )
          , sub_range_iter_( sub_range_.begin() )
          , access_f_( f )
        {
        }

        SegmentizedIterator( SegmentizedIteratorEndToken, Range< Iterator > segment_range, AccessFunction f )
          : segment_range_( segment_range )
          , segment_iter_( segment_range_.end() )
          , sub_range_()
          , sub_range_iter_()
          , access_f_( f )
        {
        }

        // TODO(squirrel): this needs testing

        SegmentizedIterator( const SegmentizedIterator& other )     = default;
        SegmentizedIterator( SegmentizedIterator&& other ) noexcept = default;
        SegmentizedIterator& operator=( const SegmentizedIterator& other ) = default;
        SegmentizedIterator& operator=( SegmentizedIterator&& other ) noexcept = default;

        SegmentizedIterator& operator++()
        {
                ++sub_range_iter_;

                if ( sub_range_iter_ != sub_range_.end() ) {
                        return *this;
                }

                ++segment_iter_;
                if ( segment_iter_ != segment_range_.end() ) {
                        sub_range_      = access_f_( *segment_iter_ );
                        sub_range_iter_ = sub_range_.begin();
                }
                return *this;
        }

        SegmentizedIterator& operator--()
        {
                --sub_range_iter_;
                if ( sub_range_iter_ != std::prev( sub_range_.begin() ) ) {
                        return *this;
                }
                --segment_iter_;
                if ( segment_iter_ != std::prev( segment_range_.begin() ) ) {
                        sub_range_      = access_f_( *segment_iter_ );
                        sub_range_iter_ = sub_range_.end();
                        --sub_range_iter_;
                }
                return *this;
        }

        const ValueType& operator*()
        {
                return *sub_range_iter_;
        }

        // TODO(squirrel): this may not be done this way??
        bool operator==( const SegmentizedIterator& other ) const
        {
                if ( segment_iter_ != other.segment_iter_ ) {
                        return false;
                }

                // both at end
                if ( segment_iter_ == segment_range_.end() && other.segment_iter_ == other.segment_range_.end() ) {
                        return true;
                }

                // this at end
                if ( segment_iter_ == segment_range_.end() ) {
                        return other.sub_range_iter_ == other.sub_range_.begin();
                }

                // other at end
                return sub_range_iter_ == sub_range_.begin();
        }
        ~SegmentizedIterator() = default;
};

template < typename Iterator, typename SubIter >
bool operator!=( const SegmentizedIterator< Iterator, SubIter >& lh,
                 const SegmentizedIterator< Iterator, SubIter >& rh )
{
        return !( lh == rh );
}

template < typename Container, typename AccessFunction >
auto makeSegmentizedRange( Container& cont, AccessFunction f )
{

        using Iterator = decltype( cont.begin() );
        using SubIter  = decltype( f( *cont.begin() ).begin() );
        using SegIter  = SegmentizedIterator< Iterator, SubIter >;
        using Value    = typename Iterator::value_type;

        return Range< SegIter >{SegIter( range( cont ), cont.begin(), f ),
                                SegIter( SegmentizedIteratorEndToken{}, range( cont ), f )};
}
template < typename Container >
auto makeSegmentizedRange( Container& cont )
{
        using SubVal  = decltype( *cont.begin() );
        using SubIter = decltype( range( *cont.begin() ).begin() );

        std::function< Range< SubIter >( const SubVal& ) > f = []( const SubVal& item ) -> Range< SubIter > {  //
                return range( item );
        };
        return makeSegmentizedRange( cont, f );
}

}  // namespace schpin

namespace std
{
template < typename Iterator, typename SubIter >
struct iterator_traits< schpin::SegmentizedIterator< Iterator, SubIter > > : public iterator_traits< SubIter >
{
};
}  // namespace std
