#include <array>
#include <boost/integer.hpp>
#include <nlohmann/json.hpp>

#pragma once

namespace schpin
{

template < typename Item, unsigned N >
class StaticVector
{
        union Placeholder
        {
                Item item;
                Placeholder(){}; // NOLINT(modernize-use-equals-default)
                ~Placeholder(){}; // NOLINT(modernize-use-equals-default)
        };

      public:
        using SizeType      = typename boost::uint_value_t< N >::least;
        using Container     = std::array< Placeholder, N >;
        using Iterator      = Item*;
        using ConstIterator = const Item*;

      private:
        Container container_;
        SizeType  size_;

      public:
        StaticVector()
          : container_()
          , size_()
        {
        }

        StaticVector( const StaticVector& other )
          : container_()
          , size_()
        {
                forEach( other, [&]( const Item& item ) { pushBack( item ); } );
        }

        StaticVector( StaticVector&& other ) noexcept
          : container_()
          , size_( other.size_ )
        {
                forRange( size_, [&]( std::size_t i ) { container_[i].item = std::move( other.container_[i].item ); } );
        }

        StaticVector& operator=( const StaticVector& other )
        {
                StaticVector copy( other );
                using std::swap;

                forRange( std::min( size_, copy.size_ ),
                          [&]( std::size_t i ) { swap( container_[i].item, copy.container_[i].item ); } );

                if ( size_ > copy.size_ ) {
                        forRange( size_ - copy.size_, [&]( std::size_t offset ) {
                                std::size_t i = copy.size_ + offset;
                                delete ( &container_[i].item );  // NOLINT(cppcoreguidelines-owning-memory)
                        } );
                        size_ = copy.size_;
                } else if ( copy.size_ > size_ ) {
                        forRange( copy.size_ - size_, [&]( std::size_t offset ) {
                                std::size_t i = size_ + offset;
                                new ( &container_[i].item ) Item( std::move( copy.container_[i].item ) );
                        } );
                        swap( copy.size_, size_ );
                }

                return *this;
        }

        StaticVector& operator=( StaticVector&& other ) noexcept
        {
                forRange( std::min( size_, other.size_ ),
                          [&]( std::size_t i ) { container_[i].item = std::move( other.container_[i].item ); } );
                if ( size_ >= other.size_ ) {
                        forRange( size_ - other.size_, [&]( std::size_t offset ) {
                                std::size_t i = other.size_ + offset;
                                delete ( &container_[i].item );  // NOLINT(cppcoreguidelines-owning-memory)
                        } );
                } else {
                        forRange( other.size_ - size_, [&]( std::size_t offset ) {
                                std::size_t i = size_ + offset;
                                new ( &container_[i].item ) Item( std::move( other.container_[i].item ) );
                        } );
                }
                size_ = other.size_;
                return *this;
        }

        Iterator begin()
        {
                return &container_[0].item;
        };

        Iterator end()
        {
                return &container_[size_].item;
        };

        ConstIterator begin() const
        {
                return &container_[0].item;
        };

        ConstIterator end() const
        {
                return &container_[size_].item;
        };

        std::array< Item, N > moveToArray() &&
        {
                return std::move( *this ).arrayHelper( std::make_index_sequence< N >() );
        }

        const Item& operator[]( std::size_t i ) const
        {
                return container_[i];
        }

        void pushBack( const Item& item )
        {
                SCHPIN_ASSERT( size_ != N );
                new ( &container_[size_].item ) Item( item );
        }
        void pushBack( Item&& item )
        {
                SCHPIN_ASSERT( size_ != N );
                new ( &container_[size_].item ) Item( std::move( item ) );
        }

        void push_back( const Item& item )
        {
                pushBack( item );
        }

        void push_back( Item&& item )
        {
                pushBack( std::move( item ) );
        }

        template < typename... Args >
        void emplaceBack( Args... args )
        {
                SCHPIN_ASSERT( size_ != N );
                new ( &container_[size_].item ) Item( args... );
        }

        ~StaticVector()
        {
                forRange( size_, [&]( std::size_t i ) {
                        delete &container_[i].item;  // NOLINT(cppcoreguidelines-owning-memory)
                } );
        }

      private:
        template < std::size_t... Is >
        std::array< Item, N > arrayHelper( std::index_sequence< Is... > /*unused*/ ) &&
        {
                return std::array< Item, N >{std::move( container_[Is].item )...};
        }
};

template < typename Item, unsigned N >
void from_json( const nlohmann::json& j, StaticVector< Item, N >& sv )
{
        SCHPIN_ASSERT( j.size() <= N );
        // TODO(squirrel): should raise exception
        sv = mapF< StaticVector< Item, N > >( j, [&]( const nlohmann::json& sub ) { return sub.get< Item >(); } );
}

template < typename Item, unsigned N >
void to_json( nlohmann::json& j, const StaticVector< Item, N >& sv )
{
        SCHPIN_ASSERT( j.size() <= N );
        // TODO(squirrel): should raise exception
        j = mapF< nlohmann::json >( sv, [&]( const Item& item ) { return nlohmann::json{item}; } );
}

}  // namespace schpin
