/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 *
 * This file is part of Schpin.
 *
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */

#include "schpin_lib/either.h"
#include "schpin_lib/util/algorithm.h"
#include "schpin_lib/util/logging.h"
#include "schpin_lib/util/range.h"
#include "schpin_lib/util/types.h"
#include <algorithm>
#include <fstream>
#include <iterator>
#include <nlohmann/json.hpp>
#include <numeric>
#include <regex>
#include <ros/package.h>
#include <type_traits>

#pragma once

namespace schpin
{

inline std::string getResolvedPath( std::string filename )
{
        std::regex  package_regex( "package://([a-zA-Z0-9_]+)/([a-zA-Z0-9_/.]+)", std::regex_constants::ECMAScript );
        std::smatch match;
        if ( std::regex_match( filename, match, package_regex ) ) {
                SCHPIN_DEBUG_LOG( "Converting to full path: " << filename );
                SCHPIN_ASSERT( match.size() == 3 );
                return ros::package::getPath( match[1].str() ) + "/" + match[2].str();
        }
        return filename;
}
template < typename... >
struct are_same;

template < typename T, typename... Ts >
struct are_same< T, Ts... >
{
        static constexpr bool value = std::conjunction< std::is_same< T, Ts >... >::value;
};

template <>
struct are_same<>
{
        static constexpr bool value = true;
};

struct EmptyAssemblyTag
{
};

template < typename... Ts >
Either< std::tuple< Ts... >, EmptyAssemblyTag > assembleOptionals( std::optional< Ts >&&... opt )
{
        if ( ( ... && opt ) ) {
                return std::tuple( std::forward< Ts >( *opt )... );
        }

        return EmptyAssemblyTag{};
}

template < typename FirstE, typename... Eithers >
auto assembleLeftCollectRight( FirstE&& first, Eithers&&... others )
{
        static_assert( are_same< typename Eithers::RightItem... >::value,
                       "Right items of eithers have to be same for collection!" );

        using RighType = typename FirstE::RightItem;

        std::vector< RighType > collection;

        auto convert = [&]( auto&& either ) {
                using LeftType = typename std::remove_reference_t< decltype( either ) >::LeftItem;

                return std::forward< decltype( either ) >( either )
                    .convertLeft( [&]( LeftType&& item ) { return std::make_optional( std::move( item ) ); } )
                    .convertRight( [&]( RighType&& right ) {
                            collection.push_back( std::move( right ) );
                            return std::optional< LeftType >();
                    } )
                    .join();
        };

        return assembleOptionals( convert( std::forward< FirstE >( first ) ), convert( std::move( others ) )... )
            .convertRight( [&]( EmptyAssemblyTag ) { return collection; } );
}

}  // namespace schpin
