
#include "schpin_lib/util/logging.h"
#include "schpin_lib/util/range.h"
#include "schpin_lib/util/types.h"

#pragma once

namespace schpin
{

template < typename Container, typename Iterator = decltype( std::declval< Container >().begin() ) >
Range< Iterator > tail( Container&& cont, int step = 1 )
{
        return Range< Iterator >( cont.begin() + step, cont.end() );
}

template < typename Container, typename Iterator = decltype( std::declval< Container >().begin() ) >
Range< Iterator > init( Container&& cont, int step = 1 )
{
        return Range< Iterator >( cont.begin(), cont.end() - step );
}

template < typename Container, typename UnaryFunction >
auto findIf( Container&& cont, UnaryFunction f )
{
        auto beg = cont.begin();
        for ( ; beg != cont.end(); ++beg ) {
                if ( f( *beg ) ) {
                        return beg;
                }
        }
        return cont.end();
}
template < typename LhContainer, typename RhContainer, typename BinaryFunction >
auto findIf( const LhContainer& lh_cont, const RhContainer& rh_cont, BinaryFunction f )
{

        auto lh_begin = lh_cont.begin();
        auto rh_begin = rh_cont.begin();

        for ( ; lh_begin != lh_cont.end() && rh_begin != rh_cont.end(); ++lh_begin, ++rh_begin ) {
                if ( f( *lh_begin, *rh_begin ) ) {
                        return std::make_pair( lh_begin, rh_begin );
                }
        }

        return std::make_pair( lh_cont.end(), rh_cont.end() );
}

template < typename Container, typename T = decltype( *std::declval< Container >().begin() ) >
auto find( const Container& cont, const T& item )
{
        auto beg = cont.begin();

        for ( ; beg != cont.end(); ++beg ) {
                if ( *beg == item ) {
                        return beg;
                }
        }
        return cont.end();
}

template < typename... Args, typename Func >
void forEach( std::tuple< Args... >& t, Func&& f )
{
        std::apply(
            [&]( auto&... items ) {  //
                    ( f( items ), ... );
            },
            t );
}
template < typename... Args, typename Func, std::size_t... Idx >
void forEach( const std::tuple< Args... >& t, Func&& f, std::index_sequence< Idx... > )
{
        ( f( std::get< Idx >( t ) ), ... );
}

template < typename... Args, typename Func >
void forEach( const std::tuple< Args... >& t, Func&& f )
{
        return forEach( t, f, std::index_sequence_for< Args... >{} );
}

template < typename Container, typename UnaryFunction >
void forEach( Container&& cont, UnaryFunction f )
{
        auto beg = cont.begin();
        for ( ; beg != cont.end(); ++beg ) {
                f( *beg );
        }
}

template < typename LhContainer, typename RhContainer, typename BinaryFunction >
void forEach( const LhContainer& lh_cont, const RhContainer& rh_cont, BinaryFunction f )
{
        SCHPIN_ASSERT( lh_cont.size() == rh_cont.size() );

        auto lh_begin = lh_cont.begin();
        auto rh_begin = rh_cont.begin();

        for ( ; lh_begin != lh_cont.end(); ++lh_begin, ++rh_begin ) {
                f( *lh_begin, *rh_begin );
        }
}
template < typename T >
struct MinMax
{
        T min;
        T max;
};

template < typename T >
std::ostream& operator<<( std::ostream& os, const MinMax< T >& mm )
{
        os << "Min: " << mm.min << " Max: " << mm.max;
        return os;
}

template < typename T >
MinMax< T > startMinMax()
{
        return MinMax< T >{std::numeric_limits< T >::min(), std::numeric_limits< T >::max()};
}

template < typename Container >
auto minmax( const Container& cont ) -> MinMax< std::remove_const_t< typename Container::value_type > >
{
        using ValueType = typename Container::value_type;
        return minmax( cont, [&]( const ValueType& val ) { return val; } );
}

template < typename Container, typename UnaryFunction >
auto minmax( const Container& cont, UnaryFunction f ) -> MinMax< extract_map_value_type_t< Container, UnaryFunction > >
{
        using T = extract_map_value_type_t< Container, UnaryFunction >;

        MinMax< T > res = startMinMax< T >();
        using std::max;
        using std::min;

        if ( cont.begin() == cont.end() ) {
                return res;
        }
        auto beg = cont.begin();
        res.min  = f( *beg );
        res.max  = res.min;
        ++beg;
        for ( ; beg != cont.end(); ++beg ) {
                T val   = f( *beg );
                res.min = min( res.min, val );
                res.max = max( res.max, val );
        }
        return res;
}

template < typename Container, typename UnaryFunction >
extract_map_value_type_t< Container, UnaryFunction > max( const Container& cont, UnaryFunction f )
{
        using T = extract_map_value_type_t< Container, UnaryFunction >;
        if ( std::distance( cont.begin(), cont.end() ) == 0 ) {
                return T{};
        }
        auto beg = cont.begin();
        T    res = f( *beg );
        ++beg;
        for ( ; beg != cont.end(); ++beg ) {
                res = std::max( res, f( *beg ) );
        }
        return res;
}

template < typename Container, typename UnaryFunction >
auto min( const Container& cont, UnaryFunction f ) -> decltype( f( *cont.begin() ) )
{
        using T = decltype( f( *cont.begin() ) );
        if ( cont.empty() ) {
                return T{};
        }

        T actual = f( *cont.begin() );

        for ( const auto& item : tail( cont ) ) {
                actual = std::min( f( item ), actual );
        }

        return actual;
}

template < typename Container, typename UnaryFunction >
auto minIterator( const Container& cont, UnaryFunction f ) -> decltype( cont.begin() )
{
        using T = decltype( f( *cont.begin() ) );
        if ( cont.empty() ) {
                return cont.end();
        }

        auto actual_iter = cont.begin();
        T    actual      = f( *actual_iter );

        auto iter = ++actual_iter;
        for ( ; iter != cont.end(); ++iter ) {
                if ( actual > f( *iter ) ) {
                        actual      = f( *iter );
                        actual_iter = iter;
                }
        }

        return iter;
}

template < typename Container, typename UnaryFunction >
std::size_t count( const Container& cont, UnaryFunction f )
{
        auto        begin = cont.begin();
        auto        end   = cont.end();
        std::size_t res   = 0;
        for ( ; begin != end; ++begin ) {
                if ( f( *begin ) ) {
                        res += 1;
                }
        }
        return 0;
}

template < typename Container, typename T >
T accumulate( const Container& cont, T init )
{
        auto begin = cont.begin();
        auto end   = cont.end();

        for ( ; begin != end; ++begin ) {
                init += *begin;
        }
        return init;
}

template < typename Container, typename T, typename BinaryFunction >
T accumulate( const Container& cont, T init, BinaryFunction f )
{
        auto begin = cont.begin();
        auto end   = cont.end();

        for ( ; begin != end; ++begin ) {
                init += f( *begin );
        }
        return init;
}

template < typename LhCont, typename RhCont, typename T, typename BinaryFunction >
T accumulate( const LhCont& lh, const RhCont& rh, T init, BinaryFunction f )
{
        auto lh_begin = lh.begin();
        auto lh_end   = lh.end();
        auto rh_begin = rh.begin();
        auto rh_end   = rh.end();

        SCHPIN_ASSERT( std::distance( lh_begin, lh_end ) == std::distance( rh_begin, rh_end ) );

        for ( ; lh_begin != lh_end; ++lh_begin, ++rh_begin ) {
                init += f( *lh_begin, *rh_begin );
        }

        return init;
}

template < typename LhContainer, typename RhContainer, typename BinaryFunction >
void forCrossJoint( const LhContainer& lh_cont, const RhContainer& rh_cont, BinaryFunction f )
{

        auto lh_begin = lh_cont.begin();
        auto lh_end   = lh_cont.end();
        auto rh_begin = rh_cont.begin();
        auto rh_end   = rh_cont.end();

        for ( ; lh_begin != lh_end; ++lh_begin ) {
                for ( ; rh_begin != rh_end; ++rh_begin ) {
                        f( *lh_begin, *rh_begin );
                }
        }
}

template < typename LhContainer, typename RhContainer, typename BinaryFunction >
bool allOf( const LhContainer& lh_cont, const RhContainer& rh_cont, BinaryFunction f )
{
        if ( std::distance( lh_cont.begin(), lh_cont.end() ) != std::distance( rh_cont.begin(), rh_cont.end() ) ) {
                return false;
        }

        auto lh_begin = lh_cont.begin();
        auto rh_begin = rh_cont.begin();
        for ( ; lh_begin != lh_cont.end(); ++lh_begin, ++rh_begin ) {
                if ( !f( *lh_begin, *rh_begin ) ) {
                        return false;
                }
        }

        return true;
}

template < typename LhContainer, typename RhContainer, typename BinaryFunction >
bool anyOf( const LhContainer& lh_cont, const RhContainer& rh_cont, BinaryFunction f )
{
        if ( std::distance( lh_cont.begin(), lh_cont.end() ) != std::distance( rh_cont.begin(), rh_cont.end() ) ) {
                return false;
        }

        auto lh_begin = lh_cont.begin();
        auto rh_begin = rh_cont.begin();
        for ( ; lh_begin != lh_cont.end(); ++lh_begin, ++rh_begin ) {
                if ( f( *lh_begin, *rh_begin ) ) {
                        return true;
                }
        }

        return false;
}

template < typename Container, typename UnaryFunction >
bool noneOf( const Container& cont, UnaryFunction f )
{
        return std::none_of( cont.begin(), cont.end(), f );
}

template < typename Container, typename UnaryFunction >
bool allOf( const Container& cont, UnaryFunction f )
{
        auto begin = cont.begin();
        for ( ; begin != cont.end(); ++begin ) {
                if ( !f( *begin ) ) {
                        return false;
                }
        }
        return true;
}

template < typename... Args, typename Func, std::size_t... Idx >
bool anyOf( const std::tuple< Args... >& t, Func&& f, std::index_sequence< Idx... > )
{
        return ( f( std::get< Idx >( t ) ) && ... );
}

template < typename... Args, typename Func >
bool anyOf( std::tuple< Args... >& t, Func&& f )
{
        return anyOf( t, f, std::index_sequence_for< Args... >{} );
}

template < typename Container, typename UnaryFunction >
bool anyOf( const Container& cont, UnaryFunction f )
{
        auto begin = cont.begin();
        for ( ; begin != cont.end(); ++begin ) {
                if ( f( *begin ) ) {
                        return true;
                }
        }
        return false;
}

template < typename Container >
bool equal( const Container& lh, const Container& rh )
{
        typedef typename Container::value_type LhValueType;
        typedef typename Container::value_type RhValueType;

        return allOf( lh, rh, []( const LhValueType& lh_v, const RhValueType& rh_v ) { return lh_v == rh_v; } );
}

template < typename Container, typename T >
bool binarySearch( const Container& cont, const T& value )
{
        return std::binary_search( cont.begin(), cont.end(), value );
}

template < typename ResultContainer,
           typename Container,
           typename UnaryFunction,
           typename = decltype( std::back_inserter( *reinterpret_cast< ResultContainer* >( 0 ) ) ) >
ResultContainer mapF( const Container& cont, UnaryFunction f )
{
        ResultContainer res{};
        for ( const auto& item : cont ) {
                res.push_back( std::move( f( item ) ) );
        }
        return res;
}

template < typename T, unsigned long size, typename UnaryFunction >
auto mapF( const std::array< T, size >& cont, UnaryFunction f ) -> std::array< decltype( f( *cont.begin() ) ), size >
{
        std::array< decltype( f( *cont.begin() ) ), size > res;
        std::transform( cont.begin(), cont.end(), res.begin(), f );
        return res;
}

template < typename T, unsigned long size, typename UnaryFunction >
auto mapF( std::array< T, size >& cont, UnaryFunction f ) -> std::array< decltype( f( *cont.begin() ) ), size >
{
        std::array< decltype( f( *cont.begin() ) ), size > res;
        std::transform( cont.begin(), cont.end(), res.begin(), f );
        return res;
}

template < unsigned N, typename Container, typename UnaryFunction >
std::array< extract_map_value_type_t< Container, UnaryFunction >, N > mapFtoA( const Container& cont, UnaryFunction f )
{
        std::array< decltype( f( *cont.begin() ) ), N > res;
        std::transform( cont.begin(), cont.end(), res.begin(), f );
        return res;
}

template < unsigned N, typename ContainerLh, typename ContainerRh, typename BinaryFunction >
auto mapFtoA( const ContainerLh& lh, const ContainerRh& rh, BinaryFunction f )
{
        std::array< decltype( f( *lh.begin(), *rh.begin() ) ), N > res;
        std::transform( lh.begin(), lh.end(), rh.begin(), res.begin(), f );
        return res;
}

template < typename Container, typename UnaryFunction >
using FResVector = std::vector< typename std::decay< extract_map_value_type_t< Container, UnaryFunction > >::type >;

template < typename Container, typename UnaryFunction >
FResVector< Container, UnaryFunction > mapFtoV( const Container& cont, UnaryFunction f )
{
        FResVector< Container, UnaryFunction > res;
        for ( const auto& item : cont ) {
                res.push_back( std::move( f( item ) ) );
        }
        return res;
}

template < typename Container, typename T >
T joined( const Container& cont, T val )
{
        if ( cont.empty() ) {
                return T{};
        }
        T res = *cont.begin();
        forEach( tail( cont ), [&]( const T& r ) { res += val + r; } );
        return res;
}

// TODO: note that copied from https://www.bfilipek.com/2018/07/string-view-perf-followup.html
template < typename T >
std::vector< std::string_view > split( T line, std::string_view delims = " " )
{
        std::string_view                strv{line};
        std::vector< std::string_view > output;
        size_t                          first = 0;

        while ( first < strv.size() ) {
                const auto second = strv.find_first_of( delims, first );

                if ( first != second ) {
                        output.emplace_back( strv.substr( first, second - first ) );
                }

                if ( second == std::string_view::npos ) {

                        break;
                }

                first = second + 1;
        }

        return output;
}

template < typename T, typename UnaryFunction >
void cartesianProduct( const std::vector< std::vector< T > >& data, UnaryFunction f )
{
        using Data         = std::vector< std::vector< T > >;
        using DataIterator = typename Data::const_iterator;
        std::vector< T > item_stack;

        std::function< void( const Range< DataIterator >& ) > combine_f = [&]( const Range< DataIterator >& actual ) {
                if ( actual.empty() ) {
                        f( item_stack );
                        return;
                }
                auto res = tail( actual );
                forEach( actual.front(), [&]( const T& item ) {
                        item_stack.push_back( item );
                        combine_f( res );
                        item_stack.pop_back();
                } );
        };

        combine_f( range( data ) );
}

template < typename T >
T clamp( T min_val, T val, T max_val )
{
        using namespace std;
        if ( min_val > val ) {
                return min_val;
        }
        if ( max_val < val ) {
                return max_val;
        }
        return val;
}

template < typename LhContainer, typename RhContainer, typename T = typename LhContainer::value_type >
std::vector< T > merge( const LhContainer& lh, const RhContainer& rh )
{
        std::vector< T > res;
        res.reserve( lh.size() + rh.size() );
        res.insert( res.end(), lh.begin(), lh.end() );
        res.insert( res.end(), rh.begin(), rh.end() );
        return res;
}

}  // namespace schpin
