#include "schpin_lib/util/iterator.h"
#include <nlohmann/json.hpp>
#pragma once

namespace schpin
{

template < typename Iterator >
class Range
{
        Iterator begin_;
        Iterator end_;

      public:
        using ValueType       = typename std::remove_reference_t< decltype( *std::declval< Iterator >() ) >;
        using value_type      = ValueType;
        using ReverseIterator = std::reverse_iterator< Iterator >;
        using iterator        = Iterator;

        Range() = default;

        Range( Iterator begin, Iterator end )
          : begin_( begin )
          , end_( end )
        {
        }

        template < typename T >
        Range( const Range< T >& other )
          : Range( other.begin(), other.end() )
        {
        }

        Iterator begin() const
        {
                return begin_;
        }

        Iterator end() const
        {
                return end_;
        }

        const ValueType& operator[]( std::size_t i ) const
        {
                return begin_[i];
        }

        ReverseIterator rbegin() const
        {
                return ReverseIterator{begin_};
        }

        ReverseIterator rend() const
        {
                return ReverseIterator{end_};
        }

        std::size_t size() const
        {
                return std::distance( begin(), end() );
        }

        bool empty() const
        {
                return begin() == end();
        }

        const ValueType& front() const
        {
                return *begin_;
        }

        const ValueType& back() const
        {
                return *std::prev( end_ );
        }

        template < typename T >
        T as() const
        {
                return T{begin(), end()};
        }
};

template < typename Iter, typename = std::enable_if_t< !std::is_arithmetic_v< Iter > > >
Range< Iter > range( Iter begin, Iter end )
{
        return {begin, end};
}

template < typename Container >
auto range( Container& cont ) -> Range< std::remove_reference_t< decltype( cont.begin() ) > >
{
        return {cont.begin(), cont.end()};
}

template < typename Iter, typename Count, typename = std::enable_if_t< std::is_arithmetic_v< Count > > >
Range< Iter > range( Iter begin, Count n )
{
        return {begin, begin + n};
}

template < typename Numeric, typename = std::enable_if_t< std::is_arithmetic_v< Numeric > > >
Range< NumericIterator< Numeric > > range( Numeric from, Numeric to )
{
        return {NumericIterator< Numeric >{from}, NumericIterator< Numeric >{to}};
}

template < typename Numeric, typename = std::enable_if_t< std::is_arithmetic_v< Numeric > > >
Range< NumericIterator< Numeric > > range( Numeric to )
{
        return range< Numeric >( 0, to );
}

template < typename Container >
auto iterRange( Container& cont ) -> Range< IteratorIterator< decltype( cont.begin() ) > >
{
        return {IteratorIterator( cont.begin() ), IteratorIterator( cont.end() )};
}

template < typename Container >
auto reversed( Container&& container )
{
        return Range{container.rbegin(), container.rend()};
}

template < typename Iter >
void to_json( nlohmann::json& j, const Range< Iter >& iter )
{
        forEach( iter, [&]( const auto& item ) { j.push_back( item ); } );
}

template < typename Iterator >
std::ostream& operator<<( std::ostream& os, const Range< Iterator >& output )
{
        using Item = decltype( *std::declval< Iterator >() );
        forEach( output, [&]( const Item& item ) { os << item << ","; } );
        return os;
}

}  // namespace schpin
