#!/usr/bin/env python
import rospy
from schpin_mech.algo import (calculate_angles, MaxRangeException,
                              MinRangeException)

import schpin_msgs.msg as schpin_msg

__author__ = 'squirrel'

ALFA = 'alfa'
BETA = 'beta'
GAMA = 'gama'


class SeqGenerator:
    def __init__(self, msg_type):
        self.msg_type = msg_type

        seq_type = None
        for i in range(len(msg_type.__slots__)):
            if msg_type.__slots__[i] == 'seq':
                seq_type = msg_type._slot_types[i]

        assert seq_type is not None, ("seq type could not be deducet from msq,"
                                      "'seq' attribute is missing")
        assert seq_type.startswith('uint'), "can work only with uint type"

        # we assume types as uint8,  uint32...
        self.max_size = pow(2, int(seq_type[len('uint'):]))

        self.range = range(self.max_size)

    def get(self):

        seq = next(self.range, default=None)

        if seq is None:
            self.range = range(self.max_size)
            seq = next(self.range)

        return seq


class Leg:
    def __init__(self):

        # ok, learn leg how to track joint position and calculate actual leg position based on that
        # with respect to covariance

        # don't forget that his should ten publish something like "PointWIthCovarianceStampeD"

        # we should laso take body_pos and target leg pos from tf?

        self.joints = dict()

        for key in [ALFA, BETA, GAMA]:
            self.joints[key] = dict(
                pub=rospy.Publisher("~/joints/" + key + "/move",
                                    schpin_msg.JointMove),
                last=None)

        self.joints_control_pub = rospy.Publisher(
            '~/joints/control', schpin_msg.JointControl, queue_size=1)
        self.joints_report_sub = rospy.Subscriber(
            '~/joints/report',
            schpin_msg.JointReport,
            callback=self.joints_report_cb)

        self.move_sub = rospy.Subscriber(
            '~/move', schpin_msg.LegMove, callback=self.move_cb)
        self.control_sub = rospy.Subscriber(
            'control', schpin_msg.LegControl, callback=self.control_cb)
        self.body_sub = rospy.Subscriber(
            'body', schpin_msg.BodyPosition, callback=self.body_cb)

        self.report_pub = rospy.Publisher(
            'report', schpin_msg.LegReport, queue_size=1)

        self.state = dict(disabled=True, blocked=False, body=None)

        self.gen = SeqGenerator(schpin_msg.JointMove)

        self.config = dict(
            base=rospy.get_param('base'),
            coxa=rospy.get_param('coxa'),
            tibia=rospy.get_param('tibia'),
            femur=rospy.get_param('femur'))

        self.moves = dict()

        self.body = dict(position=[0, 0, 0], orientation=[1, 0, 0, 0])

    def enable(self):
        self.state['disabled'] = False
        msg = schpin_msg.JointControl()
        msg.state = msg.ENABLE
        msg.servo = '*'
        self.joints_control_pub.publish(msg)
        self.moves = dict()

    def disable(self):
        self.state['disabled'] = True
        msg = schpin_msg.JointControl()
        msg.state = msg.DISABLE
        msg.servo = '*'
        self.joints_control_pub.publish(msg)

    def is_disabled(self):
        return self.state['disabled']

    def joints_report_cb(self, report):

        if self.is_disabled():
            return

        if report.seq not in self.moves:
            rospy.logwarn("got seq for unknown move {}".format(report))
            return

        d = self.moves[report.seq]
        if report.node not in d['states']:
            rospy.logwarn("unrelated communication {}".format(report))
            return

        if d['states'][report.node] is not None:
            rospy.logwarn("multiple messages for same seq {}".format(report))
            return

        d['states'][report.node] = report.status

        if report.status == report.OUT_OF_RANGE:

            self.cancel_seq(seq) # servos that didn't reported yet are sent message to cancel
            # beware of race conditions

        got_all = all([x is not None for x in d['states'].values()])

        if not got_all:
            return

        # with 3 servos and 4 available message types from servo
        # 4^3 makes it 64 options
        #
        # A. 2^3 - 8 - for: OK V DISABLED
        # B. 2^3-1 - 7 - for: OK V UNFINISHED ( OK,OK,OK is in previous )
        if not self.moves[report.seq]['responded']:

            orig_msg = self.moves[report.seq]['orig_msg']
            values = d['states'].values()

            if all([x in [report.OK, report.DISABLED] for x in values]):
                self.report(orig_msg, self.report_consts.OK)
            elif all([x in [report.OK, report.UNFISNIHED] for x in values]):
                self.report(orig_msg, self.report_consts.UNFISNIHED)

        del self.moves[report.seq]

    def body_cb(self, msg):
        self.body = dict(position=msg.position, orientation=msg.orientation)

    def move_cb(self, move_msg):

        if self.is_disabled():
            self.report(move_msg, self.report_consts.DISABLED)
            return

        try:
            angles = calculate_angles(
                body=self.body,
                base=self.config['base'],
                coxa=self.config['coxa'],
                tibia=self.config['tibia'],
                femur=self.config['femur'],
                target=move_msg.position)
        except MaxRangeException:
            self.report(move_msg, self.report_consts.MAX_RANGE_ERROR)
            return
        except MinRangeException:
            self.report(move_msg, self.report_consts.MIN_RANGE_ERROR)
            return

        keys = [ALFA, BETA, GAMA]

        seq = self.gen.get()

        self.moves[seq] = dict(
            responded=False,
            states=dict(),
            time=rospy.get_rostime(),
            orig_msg=move_msg)

        for i in range(3):
            key = keys[i]
            if self.move_joint(key, angles[i], seq):
                self.moves[seq]['states'][key] = None

        if len(self.moves[seq]['states']) == 0:
            del self.moves[seq]

    def control_cb(self, msg):

        if msg.leg not in [rospy.get_name(), '*']:
            return

        if msg.state == msg.DISABLE:
            self.disable()
        elif msg.state == msg.ENABLE:
            self.enable()

    def move_joint(self, key, position, seq):
        if self.joints[key]['last'] == position:
            return False
        self.joints[key]['last'] = position

        msg = schpin_msg.JointMove()
        msg.position = position
        msg.seq = seq
        self.joints[key]['pub'].publish(msg)

        return True


if __name__ == '__main__':
    rospy.init_node('leg')

    Leg()

    rospy.spin()
