#pragma once

#include "schpin_lib/error.h"
#include <fstream>
#include <iomanip>
#include <nlohmann/json.hpp>

namespace schpin
{

using JsonStorageError = Error< struct JsonStorageErrorTag >;

template < typename T, typename UnaryFunction >
inline Either< T, JsonStorageError > safeLoadJson( std::istream& input, UnaryFunction f )
{

        if ( !input ) {
                return {E( JsonStorageError ) << "The input json file is missing"};
        }
        nlohmann::json j;
        try {
                input >> j;
        }
        catch ( const std::ios_base::failure& e ) {
                return {E( JsonStorageError ) << " Failed to read the file: " << e.what()};
        }

        try {
                return {f( j )};
        }
        catch ( const nlohmann::json::exception& e ) {
                return {E( JsonStorageError ) << "Failed to parse the JSON: " << e.what()};
        }
}

template < typename T >
inline Either< T, JsonStorageError > safeLoadJson( std::istream& input )
{
        return safeLoadJson< T >( input, []( const nlohmann::json& j ) { return j.get< T >(); } );
}

template < typename Function, typename = std::enable_if_t< std::is_invocable_v< Function > > >
inline OptError< JsonStorageError > safeStoreJson( std::ostream& output, Function f )
{

        nlohmann::json res;
        try {
                res = f();
        }
        catch ( const nlohmann::json::exception& e ) {
                return {E( JsonStorageError ) << "JSON Exception raised when tried to store json: " << e.what()};
        }
        try {
                output << res << std::endl;
        }
        catch ( const std::ios_base::failure& e ) {
                return {E( JsonStorageError ) << "Failed to write to graph file: " << e.what()};
        }

        return {};
}

template < typename T, typename = std::enable_if_t< !std::is_invocable_v< T > > >
inline OptError< JsonStorageError > safeStoreJson( std::ostream& output, const T& item )
{
        return safeStoreJson( output, [&] { return nlohmann::json( item ); } );
}

}  // namespace schpin
