#include "base/math.h"
#include "base/physics.h"
#include "collission_interfaces.h"
#include <nlohmann/json.hpp>
#include <tf/tf.h>

#pragma once

namespace schpin
{

template < unsigned N >
class Point
{
      public:
        using Container = std::array< float, N >;

      private:
        // todo: patch for using aligned storage
        Container dims_{};

      public:
        using ConstIterator = typename Container::const_iterator;

        Point() = default;

        explicit Point( Container&& cont ) noexcept
          : dims_( std::move( cont ) )
        {
        }

        Point( const Point< N >& other ) noexcept = default;
        Point( Point< N >&& other ) noexcept      = default;

        template < typename... T >
        explicit Point( T... t )
          : dims_( {float( t )...} )
        {
                static_assert( sizeof...( T ) == N, "Number of parameters for point has to be N" );
        }

        const Container& operator*() const
        {
                return dims_;
        }

        Point& operator=( const Point& other ) = default;
        Point& operator=( Point&& other ) noexcept = default;

        ConstIterator begin() const
        {
                return dims_.begin();
        }

        ConstIterator end() const
        {
                return dims_.end();
        }

        auto length2() const
        {
                auto val =
                    float( schpin::accumulate( dims_, float( 0.0 ), []( float val ) { return pow( val, 2 ); } ) );
                return Length{val} * Length{1};  // just trick to make the unit squared too
        }
        Length length() const
        {
                return Length( sqrt( *length2() ) );
        }

        Point normalized() const
        {
                return Point( *this / length() );
        }

        Distance distance( const Point& p ) const
        {
                auto tmp = float( schpin::accumulate(
                    dims_, p.dims_, float( 0.0 ), []( float lh, float rh ) { return pow( lh - rh, 2 ); } ) );
                return Distance( sqrt( tmp ) );
        }

        double dot( const Point& p ) const
        {
                return schpin::accumulate(
                    dims_, p.dims_, double( 0.0 ), []( const float lh, const float rh ) { return lh * rh; } );
        }

        float operator[]( std::size_t i ) const
        {
                return dims_[i];
        }

        float& operator[]( std::size_t i )
        {
                return dims_[i];
        }

        Point< N > operator-() const
        {
                return *this * -1;
        }

        Point< N >& operator+=( const Point< N >& other )
        {
            // TODO: ugly
                *this = *this + other;
                return *this;
        }

        constexpr std::size_t size() const
        {
                return N;
        }

        ~Point() = default;
};

extern template class Point< 3 >;

template < unsigned N >
inline bool almostEqual( const Point< N >& lh, const Point< N >& rh, float eps = DEFAULT_EPSILON )
{
        return allOf( lh, rh, [&]( const float& lh_f, const float& rh_f ) { return almostEqual( lh_f, rh_f, eps ); } );
}

inline tf::Vector3 toTfVector( const Point< 3 >& p )
{
        return tf::Vector3( double( p[0] ), double( p[1] ), double( p[2] ) );
}

inline Point< 3 > cross( const Point< 3 >& lh, const Point< 3 >& rh )
{
        return Point< 3 >( lh[1] * rh[2] - lh[2] * rh[1],  //
                           lh[2] * rh[0] - lh[0] * rh[2],
                           lh[0] * rh[1] - lh[1] * rh[0] );
}

template < unsigned N >
inline bool operator<( const Point< N >& lh, const Point< N >& rh )
{
        std::size_t index = *findIf( range( N ), [&]( std::size_t i ) {  //
                return lh[i] != rh[i];
        } );
        if ( index == N ) {
                return false;
        }
        return lh[index] < rh[index];
}

template < unsigned N >
inline bool operator==( const Point< N >& lh, const Point< N >& rh )
{
        return allOf( lh, rh, []( const float lh, const float rh ) { return lh == rh; } );
}

template < unsigned N >
Point< N > operator*( const Point< N >& lh, const Point< N >& rh )
{
        Point< N > res( lh );
        forEach( range( N ), [&]( std::size_t i ) { res[i] *= rh[i]; } );
        return res;
}

template < unsigned N >
inline Point< N > operator*( const Point< N >& p, const float& s )
{
        Point< N > res( p );
        forEach( range( N ), [&]( std::size_t i ) { res[i] *= s; } );
        return res;
}

template < unsigned N >
inline Point< N > operator*( const float& s, const Point< N >& p )
{
        return p * s;
}

template < unsigned N >
inline Point< N > operator/( const Point< N >& p, const float& s )
{
        Point< N > res( p );
        forEach( range( N ), [&]( std::size_t i ) { res[i] /= s; } );
        return res;
}

template < unsigned N >
inline Point< N > operator/( const Point< N >& p, const Length& s )
{
        return p / float( *s );
}

template < unsigned N >
inline Point< N > operator-( const Point< N >& lh, const Point< N >& rh )
{
        Point< N > res( lh );
        forEach( range( N ), [&]( std::size_t i ) { res[i] -= rh[i]; } );
        return res;
}

template < unsigned N >
inline Point< N > operator+( const Point< N >& lh, const Point< N >& rh )
{
        Point< N > res( lh );
        forEach( range( N ), [&]( std::size_t i ) { res[i] += rh[i]; } );
        return res;
}

template < unsigned N >
inline Point< N > max( const Point< N >& lh, const Point< N >& rh )
{
        Point< N > res( lh );
        forEach( range( N ), [&]( std::size_t i ) { res[i] = std::max( res[i], rh[i] ); } );
        return res;
}

template < unsigned N >
inline Point< N > min( const Point< N >& lh, const Point< N >& rh )
{
        Point< N > res( lh );
        forEach( range( N ), [&]( std::size_t i ) { res[i] = std::min( res[i], rh[i] ); } );
        return res;
}

template < unsigned N >
inline std::ostream& operator<<( std::ostream& os, const Point< N >& p )
{
        os << "[";
        forEach( p, [&]( float val ) { os << val << ","; } );
        return os << "]";
}

inline Point< 2 > getNormal( const Point< 2 >& p )
{
        return Point< 2 >( p[1], -p[0] );
}

/* Function to calculate distance of projection of 'point'. That point is projected on axis defined only by it's
 * direction - 'axisDirection'. The distance of that projection from the [0,0,0] coordinate is returned.
 *
 * O(1)
 */
template < unsigned N >
inline Distance axisProjectionDistance( const Point< N >& point, const Point< N >& axisDirection )
{
        return Distance( float( point.dot( axisDirection ) ) );
}

inline geometry_msgs::Point pointToMsg( const Point< 3 >& p )
{
        geometry_msgs::Point msg;
        msg.x = double( p[0] );
        msg.y = double( p[1] );
        msg.z = double( p[2] );
        return msg;
}
inline Point< 3 > msgToPoint( const geometry_msgs::Point& msg )
{
        return Point< 3 >{{float( msg.x ), float( msg.y ), float( msg.z )}};
}

template < unsigned N >
class PointCollisionQuery : ICollisionItem< const Point< N >*, const Point< N >* >
{
        std::array< Point< N >, 1 > point_;

      public:
        explicit PointCollisionQuery( const Point< N >& other )
          : point_( {other} )
        {
        }

        using ItemType               = Point< N >;
        using PointIterator          = const Point< N >*;
        using SeparationAxisIterator = const Point< N >*;

        Range< PointIterator > getPoints() const override
        {
                return range( point_ );
        }

        Range< SeparationAxisIterator > getSeparationAxes() const override
        {
                return {nullptr, nullptr};
        }

        Range< SeparationAxisIterator > getIntersectionAxes() const override
        {
                return {nullptr, nullptr};
        }
};

template < unsigned N >
class FastPointCollisionQuery : IFastCollisionItem< PointCollisionQuery< N >, N >
{
        Point< N > point_;

      public:
        explicit FastPointCollisionQuery( const Point< N >& p )
          : point_( p )
        {
        }

        Point< N > getMinSpherePosition() const
        {
                return point_;
        }

        Radius getMinSphereRadius() const
        {
                return Radius{0.f};
        }
        PointCollisionQuery< N > getFull() const
        {
                return PointCollisionQuery{point_};
        }
};

static const Point< 3 > X_AXIS{1, 0, 0};
static const Point< 3 > Y_AXIS{0, 1, 0};
static const Point< 3 > Z_AXIS{0, 0, 1};

template < unsigned N >
void from_json( const nlohmann::json& j, Point< N >& p )
{
        using Container = typename Point< N >::Container;

        p = Point< N >( j.get< Container >() );
}

template < unsigned N >
void to_json( nlohmann::json& j, const Point< N >& p )
{
        j = nlohmann::json( *p );
}

}  // namespace schpin

namespace std
{
template < unsigned N >
class numeric_limits< schpin::Point< N > >
{
      public:
        static schpin::Point< N > min()
        {
                float                  min_val = std::numeric_limits< float >::min();
                std::array< float, N > data{};
                schpin::forEach( data, [=]( float& val ) { val = min_val; } );
                return schpin::Point< N >{std::move( data )};
        }
        static schpin::Point< N > max()
        {
                float                  max_val = std::numeric_limits< float >::max();
                std::array< float, N > data{};
                schpin::forEach( data, [=]( float& val ) { val = max_val; } );
                return schpin::Point< N >{std::move( data )};
        }
};
}  // namespace std
