#include "base/point.h"
#include "base/quaternion.h"
#include <utility>

#pragma once

namespace schpin
{

struct PoseDistance
{
        Distance dist;
        Angle    angle;
};

inline std::size_t steps( const PoseDistance& dist, Distance dist_step, Angle angle_step )
{
        std::size_t d_steps = std::size_t( *dist.dist / *dist_step );
        std::size_t a_steps = std::size_t( *dist.angle / *angle_step );
        return std::max( d_steps, a_steps );
}

class Pose
{
      public:
        Pose()
          : position( 0, 0, 0 )
          , orientation( NEUTRAL_QUAT )
        {
        }

        Pose( Point< 3 > position, Quaternion orientation )
          : position( std::move( position ) )
          , orientation( orientation )
        {
        }

        explicit Pose( Point< 3 > position )
          : position( std::move( position ) )
          , orientation( NEUTRAL_QUAT )
        {
        }

        explicit Pose( Quaternion orientation )
          : position( 0, 0, 0 )
          , orientation( orientation )
        {
        }

        PoseDistance distance( const Pose& other ) const
        {
                return {position.distance( other.position ), orientation.angleShortestPath( other.orientation )};
        }

        Point< 3 > position;
        Quaternion orientation;
};

inline void to_json( nlohmann::json& j, const Pose& p )
{
        j = nlohmann::json{{"position", p.position}, {"orientation", p.orientation}};
}

inline void from_json( const nlohmann::json& j, Pose& p )
{
        p = Pose{j["position"].get< Point< 3 > >(), j["orientation"].get< Quaternion >()};
}

inline bool almostEqual( const Pose& lh, const Pose& rh, float eps = DEFAULT_EPSILON )
{
        return almostEqual( lh.orientation, rh.orientation, eps ) && almostEqual( lh.position, rh.position, eps );
}

inline bool operator==( const Pose& lh, const Pose& rh )
{
        return lh.position == rh.position && lh.orientation == rh.orientation;
}

inline Pose interp( const Pose& base, const Pose& goal, float factor )
{
        return Pose{base.position + ( goal.position - base.position ) * factor,
                    base.orientation.slerp( goal.orientation, factor )};
}

inline Point< 3 > transform( const Point< 3 >& p, const Pose& transformation )
{
        return rotate( p, transformation.orientation ) + transformation.position;
}

inline Pose transform( const Pose& p, const Pose& transformation )
{
        return Pose{transform( p.position, transformation ), transformation.orientation * p.orientation};
}

inline Pose rotate( const Pose& p, const Quaternion& quat )
{
        return Pose{rotate( p.position, quat ), quat * p.orientation};
}

inline Pose operator+( const Pose& lh, const Pose& rh )
{
        return transform( rh, lh );
}

inline std::ostream& operator<<( std::ostream& os, const Pose& pose )
{
        os << "P: " << pose.position << " O: " << pose.orientation;
        return os;
}

inline Pose msgToPose( const geometry_msgs::Pose& msg )
{
        return Pose{msgToPoint( msg.position ), msgToQuaternion( msg.orientation )};
}

}  // namespace schpin
