#include "base/point.h"
#include <tf/tf.h>

#pragma once

namespace schpin
{

class Quaternion : private tf::Quaternion
{
      public:
        Quaternion() noexcept
        {
        }

        explicit Quaternion( const tf::Quaternion& other ) noexcept
          : tf::Quaternion( other.x(), other.y(), other.z(), other.w() )
        {
        }

        Quaternion( tfScalar x, tfScalar y, tfScalar z, tfScalar w ) noexcept
          : tf::Quaternion( x, y, z, w )
        {
        }

        Quaternion( const Point< 3 >& p, const Angle& a ) noexcept
          : tf::Quaternion( toTfVector( p ), double( *a ) )
        {
        }

        Quaternion slerp( const Quaternion& other, float dist ) const
        {
                return Quaternion{tf::Quaternion::slerp( other, double( dist ) )};
        }

        Quaternion operator*( const Quaternion& other ) const
        {
                tf::Quaternion copy( *this );
                copy *= other;
                return Quaternion( copy );
        }

        tfScalar x() const
        {
                return tf::Quaternion::x();
        }
        tfScalar y() const
        {
                return tf::Quaternion::y();
        }
        tfScalar z() const
        {
                return tf::Quaternion::z();
        }
        tfScalar w() const
        {
                return tf::Quaternion::w();
        }

        Angle angleShortestPath( const Quaternion& other ) const
        {
                return Angle( float( tf::Quaternion::angleShortestPath( other ) ) );
        }

      private:
        friend tf::Quaternion toTfQuaternion( const Quaternion& /*quat*/ );
};

inline void to_json( nlohmann::json& j, const Quaternion& quat )
{
        j = nlohmann::json{quat.x(), quat.y(), quat.z(), quat.w()};
}

inline void from_json( const nlohmann::json& j, Quaternion& quaternion )
{
        quaternion = Quaternion{//
                                j[0].get< tfScalar >(),
                                j[1].get< tfScalar >(),
                                j[2].get< tfScalar >(),
                                j[3].get< tfScalar >()};
}

inline bool almostEqual( const Quaternion& lh, const Quaternion& rh, float eps = DEFAULT_EPSILON )
{
        return almostEqual( lh.x(), rh.x(), eps ) && almostEqual( lh.y(), rh.y(), eps ) &&
               almostEqual( lh.z(), rh.z(), eps ) && almostEqual( lh.w(), rh.w(), eps );
}

inline bool operator==( const Quaternion& lh, const Quaternion& rh )
{
        return lh.x() == rh.x() && lh.y() == rh.y() && lh.z() == rh.z() && lh.w() == rh.w();
}

inline tf::Quaternion toTfQuaternion( const Quaternion& quat )
{
        return quat;
}

const Quaternion NEUTRAL_QUAT{Z_AXIS, Angle( 0 )};
const Quaternion X_QUAT{X_AXIS, Angle( 0 )};
const Quaternion Y_QUAT{Y_AXIS, Angle( 0 )};
const Quaternion Z_QUAT{Z_AXIS, Angle( 0 )};

inline Quaternion shortestArcQuat( const Point< 3 >& lh, const Point< 3 >& rh )
{
        return Quaternion{tf::shortestArcQuat( toTfVector( lh ), toTfVector( rh ) )};
}

inline Point< 3 > rotate( const Point< 3 >& p, const Quaternion& quat )
{
        tf::Vector3 res = tf::quatRotate( toTfQuaternion( quat ), toTfVector( p ) );

        return Point< 3 >{float( res.x() ), float( res.y() ), float( res.z() )};
}

inline std::ostream& operator<<( std::ostream& os, const Quaternion& q )
{
        os << "[" << q.x() << "," << q.y() << "," << q.z() << "," << q.w() << "]";
        return os;
}

inline geometry_msgs::Quaternion quaternionToMsg( const Quaternion& q )
{
        geometry_msgs::Quaternion res;
        res.x = q.x();
        res.y = q.y();
        res.z = q.z();
        res.w = q.w();
        return res;
}

inline Quaternion msgToQuaternion( const geometry_msgs::Quaternion& msg )
{
        return Quaternion( msg.x, msg.y, msg.z, msg.w );
}

}  // namespace schpin
