#include "base/json.h"
#include "base/math.h"
#include "base/point.h"
#include "schpin_lib/util.h"
#include <algorithm>
#include <array>
#include <vector>

#pragma once

namespace schpin
{

template < typename Item, unsigned N >
class Simplex
{
      public:
        using Container = std::array< Item, N + 1 >;

      private:
        Container points_;

      public:
        using Iterator      = typename Container::iterator;
        using ConstIterator = typename Container::const_iterator;

        explicit Simplex() = default;

        explicit Simplex( Item item )
          : points_( {item} )
        {
        }

        Simplex( const Simplex< Item, N - 1 >& sub, Item item )
          : points_()
        {
                std::copy( sub.begin(), sub.end(), points_.begin() );
                *points_.rbegin() = item;
                std::sort( points_.begin(), points_.end() );
        }

        explicit Simplex( std::array< Item, N + 1 > data )
          : points_( std::move( data ) )
        {
                std::sort( points_.begin(), points_.end() );
        }

        template < typename... Ts >
        explicit Simplex( Ts... ts )
          : points_( {ts...} )
        {
                static_assert( sizeof...( Ts ) == N + 1, "Number of parameters for simplex has to be N+1" );
        }

        const Container& operator*() const
        {
                return points_;
        }

        ConstIterator begin() const
        {
                return points_.begin();
        }

        ConstIterator end() const
        {
                return points_.end();
        }

        const Item& operator[]( std::size_t index ) const
        {
                return points_[index];
        }

        std::size_t size() const
        {
                return points_.size();
        }
};

template < unsigned N, unsigned U >
Point< N > center( const Simplex< Point< N >, U >& s )
{
        return accumulate( s, Point< N >{}, [&]( const Point< N >& p ) {  //
                return p / s.size();
        } );
}

template < typename Item, unsigned N >
void to_json( nlohmann::json& j, const Simplex< Item, N >& simplex )
{
        j = nlohmann::json( *simplex );
}

template < typename Item, unsigned N >
void from_json( const nlohmann::json& j, Simplex< Item, N >& simplex )
{
        simplex = Simplex< Item, N >{j.get< std::array< Item, N + 1 > >()};
}

template < unsigned N >
float getVolume( const Simplex< Point< N >, N >& simplex )
{
        std::array< float, N * N > matrix{};
        forEach( range( simplex.size() - 1 ), [&]( std::size_t i ) {
                auto diff = simplex[i + 1] - simplex[0];
                std::copy( diff.begin(), diff.end(), &matrix[N * i] );
        } );

        return std::abs( ( 1 / fac( N ) ) * determinant< N >( matrix ) );
}

template < typename Item, unsigned N >
std::ostream& operator<<( std::ostream& os, const Simplex< Item, N >& s )
{
        os << N << "-Simplex[";
        forEach( s, [&os]( const Item& it ) { os << it << ","; } );
        os << "]";
        return os;
}

template < typename Item, unsigned N >
bool operator<( const Simplex< Item, N >& lh, const Simplex< Item, N >& rh )
{
        // there are actually N+1 items in N simplex;
        std::size_t i = *findIf( range( N ), [&]( std::size_t i ) { return lh[i] != rh[i]; } );
        return lh[i] < rh[i];
}

template < typename Item, unsigned N >
bool operator==( const Simplex< Item, N >& lh, const Simplex< Item, N >& rh )
{
        return allOf( lh, rh, []( const Item& l, const Item& r ) { return l == r; } );
}

template < typename Item, unsigned N >
bool operator!=( const Simplex< Item, N >& lh, const Simplex< Item, N >& rh )
{
        return !( lh == rh );
}

template < typename Item, unsigned N >
using SimplexContainer = std::vector< Simplex< Item, N > >;

template < typename Item, unsigned N >
using SimplexIterator = typename SimplexContainer< Item, N >::iterator;

template < typename Item, unsigned N >
using SimplexConstIterator = typename SimplexContainer< Item, N >::const_iterator;
// TODO: this is pointless
template < typename Item, unsigned N >
class SimplexStorage
{
        static_assert( N > 0, "The simplex container is only valid ofr N > 0" );

        SimplexContainer< Item, N > simplexes_;

      public:
        explicit SimplexStorage()
          : simplexes_()
        {
        }
        template < typename Container >
        explicit SimplexStorage( const Container& cont )
          : simplexes_( cont.begin(), cont.end() )
        {
        }

        const SimplexContainer< Item, N >& operator*() const
        {
                return simplexes_;
        }

        SimplexIterator< Item, N > begin()
        {
                return simplexes_.begin();
        }

        SimplexIterator< Item, N > end()
        {
                return simplexes_.end();
        }

        SimplexConstIterator< Item, N > begin() const
        {
                return simplexes_.begin();
        }

        SimplexConstIterator< Item, N > end() const
        {
                return simplexes_.end();
        }

        std::size_t size() const
        {
                return simplexes_.size();
        }

        const Simplex< Item, N >& operator[]( std::size_t i ) const
        {
                return simplexes_[i];
        }
};

template < typename Item, unsigned N >
void logSimplexStorageStats( const SimplexStorage< Item, N >& stor )
{
        SCHPIN_DEBUG_LOG( "Simplex<" << N << "> count: " << std::distance( stor.begin(), stor.end() ) );
}

}  // namespace schpin

namespace nlohmann
{
template < typename Item, unsigned N >
struct adl_serializer< schpin::SimplexStorage< Item, N > >
{
        static void to_json( nlohmann::json& j, const schpin::SimplexStorage< Item, N >& storage )
        {
                j["simplexes"] = *storage;
        }

        static schpin::SimplexStorage< Item, N > from_json( const nlohmann::json& j )
        {
                return schpin::SimplexStorage< Item, N >{
                    j["simplexes"].get< std::vector< schpin::Simplex< Item, N > > >()};
        }
};

}  // namespace nlohmann
