#include "base/physics.h"
#include "base/point.h"
#include "base/pose.h"
#include "base/simplex.h"
#include "collission_interfaces.h"

#pragma once

namespace schpin
{

template < unsigned N >
using Triangle = Simplex< Point< N >, 2 >;

extern template class Simplex< Point< 3 >, 2 >;

// TODO: propably rename to normal()
inline Point< 3 > getNormal( const Triangle< 3 >& triangle )
{
        return cross( triangle[0] - triangle[1], triangle[0] - triangle[2] );
}

template < unsigned N >
inline std::ostream& operator<<( std::ostream& os, const Triangle< N >& t )
{
        return os << t[0] << t[1] << t[2];
}

template < unsigned N >
class TriangleCollisionQuery : ICollisionItem< const Point< N >*, const Point< N >* >
{
        std::array< Point< N >, 3 > intersection_axis_;
        std::array< Point< N >, 1 > separation_axis_;
        const Triangle< N >&        triangle_;

        static_assert( N == 3, "TriangleCollisionQuery is working only for N = 3" );

      public:
        using ItemType               = Triangle< N >;
        using PointIterator          = const Point< N >*;
        using SeparationAxisIterator = const Point< N >*;

        explicit TriangleCollisionQuery( const Triangle< N >& triangle )
          : triangle_( triangle )
        {
                Point< N > sideAB = triangle[0] - triangle[1];
                Point< N > sideAC = triangle[0] - triangle[2];
                Point< N > normal = cross( sideAB, sideAC ).normalized();

                separation_axis_[0]   = normal;
                intersection_axis_[0] = cross( normal, sideAB ).normalized();
                intersection_axis_[1] = cross( normal, triangle[2] - triangle[1] ).normalized();
                intersection_axis_[2] = cross( normal, sideAC ).normalized();
        }
        Range< PointIterator > getPoints() const override
        {
                return range( triangle_ );
        }
        Range< SeparationAxisIterator > getSeparationAxes() const override
        {
                return range( separation_axis_ );
        }
        Range< SeparationAxisIterator > getIntersectionAxes() const override
        {
                return range( intersection_axis_ );
        }
};

extern template class TriangleCollisionQuery< 3 >;

Point< 3 > getTriangleSphereCenter( const Triangle< 3 >& triangle );

class TriangleFastCollisionQuery : IFastCollisionItem< TriangleCollisionQuery< 3 >, 3 >
{
        Point< 3 >    sphere_center_;
        Radius        sphere_radius_;
        Triangle< 3 > triangle_;

      public:
        using ItemType = Triangle< 3 >;
        using Full     = TriangleCollisionQuery< 3 >;

        explicit TriangleFastCollisionQuery( const Triangle< 3 >& triangle )
          : sphere_radius_( 0 )
          , triangle_( triangle )
        {
                sphere_center_ = getTriangleSphereCenter( triangle );
                sphere_radius_ = Radius( sphere_center_.distance( triangle[0] ) );
        }

        Point< 3 > getMinSpherePosition() const override
        {
                return sphere_center_;
        }

        Radius getMinSphereRadius() const override
        {
                return sphere_radius_;
        }

        TriangleCollisionQuery< 3 > getFull() const override
        {
                return TriangleCollisionQuery< 3 >{triangle_};
        }

        bool isLeaf() const
        {
                return true;
        }

        std::vector< TriangleFastCollisionQuery > getChildren() const
        {
                return {};
        }
};

inline Triangle< 3 > transform( const Triangle< 3 >& t, const Pose& transformation )
{
        return Triangle< 3 >{
            transform( t[0], transformation ), transform( t[1], transformation ), transform( t[2], transformation )};
}

// TODO(squirrel): this is not semantically correct
template < unsigned N >
inline Triangle< N > scale( const Triangle< N >& t, const Point< N >& scale )
{
        return Triangle< N >{t[0] * scale, t[1] * scale, t[2] * scale};
}

}  // namespace schpin
