
/** Functions used to detect collision between 3D primitives.
 *
 * This is polymorphic code that uses templates. So for each primitive combination checked, separate code is generated.
 *
 * There are two primary functions:
 *   bool collides<A,B>(A,B) - full collision check between primitive of type A and primitive of type B, uses the
 * Separating Axis Test. bool fastCollides<A,B>(A,B) - fast collision check between primitive of type A and primitive
 * of type B, uses min-spheres to detect the collision
 *
 *
 * SAT solution inspired from:
 *   https://gamedevelopment.tutsplus.com/tutorials/collision-detection-using-the-separating-axis-theorem--gamedev-169
 *
 *
 */

#include "base/point.h"
#include "collision/comparators.h"
#include "collission_interfaces.h"
#include "util.h"

#pragma once

namespace schpin
{

/* Calculates the smallest and biggest distance of all points from 'T' projected on axis. These are all necessary
 * points for Separation Axis Test. The distances are between the projected points and [0,0,0] coordinate on that
 * axis. The axis is defined only by it's direction.
 *
 * O(n), where 'n' is number of points
 */
template < typename T, unsigned N >
MinMax< Distance > projectionRange( const T& t, const Point< N >& axisDirection )
{
        return minmax( t.getPoints(),
                       [&]( const Point< N >& p ) { return axisProjectionDistance( p, axisDirection ); } );
}

/* Checks if the provided axis detects an separation between 'LH' and 'RH' objects. The axis is defined only by it's
 * direction. All points of each object are projected on same axis and in case they do not overlap, the two objects
 * are not in collision - the separation was detected.
 *
 * O(n+k), where 'n' us number of points for 'LH' and 'k' is number of points for 'RH'.
 */
template < bool can_intersect, typename LH, typename RH, unsigned N >
bool axisSeparates( const LH& lh, const RH& rh, const Point< N >& axisDirection )
{
        MinMax< Distance > lh_range = projectionRange( lh, axisDirection );
        MinMax< Distance > rh_range = projectionRange( rh, axisDirection );
        bool               no_overlap;
        if ( can_intersect ) {
                no_overlap = lh_range.min >= rh_range.max || rh_range.min >= lh_range.max;
        } else {
                no_overlap = lh_range.min > rh_range.max || rh_range.min > lh_range.max;
        }
        // https://stackoverflow.com/questions/3269434/whats-the-most-efficient-way-to-test-two-integer-ranges-for-overlap

        return no_overlap;
}

/* Checks collision between 'LH' and 'RH' objects. The Separation Axis Test is used and only axes of the first
 * object are considered. For each axis, the axisSeparates function is used. The method returns true in case that not
 * a single axis separates the objects. In case at least one axis separates the objects, false is returned.
 *
 * O(a*n), where 'a' is number of axis of 'LH' and 'n' is the sum of points provided by 'LH' and 'RH'
 */
template < typename LH, typename RH, unsigned N >
bool collidesSingleAxes( const LH& lh, const RH& rh )
{
        return noneOf( lh.getSeparationAxes(),
                       [&]( const Point< N >& axisDirection ) {
                               return axisSeparates< false >( lh, rh, axisDirection );
                       } ) &&
               noneOf( lh.getIntersectionAxes(), [&]( const Point< N >& axisDirection ) {
                       return axisSeparates< true >( lh, rh, axisDirection );
               } );
}

/* Checks collision between provided objects. The Separation Axis Test is used. This simply calls the
 * collidesSingleAxes(A,B) on the axes of 'RH' object and than on the axes of 'RH' object.
 *
 * O( (x+z)*(k+n) ), where x,z are the number of axes for the LH and RH objects and k,n are the number of points for
 * LH and RH.
 */
template < unsigned N, typename LH, typename RH >
bool collides( const LH& lh, const RH& rh )
{
        static_assert(
            std::is_base_of_v< ICollisionItem< typename LH::PointIterator, typename LH::SeparationAxisIterator >,
                               LH > );
        static_assert(
            std::is_base_of_v< ICollisionItem< typename RH::PointIterator, typename RH::SeparationAxisIterator >,
                               RH > );
        return collidesSingleAxes< LH, RH, N >( lh, rh ) && collidesSingleAxes< RH, LH, N >( rh, lh );
}

/* Fast check of collision between provided objects. This uses min-spheres for each object. The min sphere is the
 * smallest sphere containing adequate object. The collision check between two spheres is easy - simply compare the
 * distance between their centers with their radiuses.
 *
 * O(1)
 */
template < unsigned N, typename LH, typename RH >
bool fastCollides( const LH& lh, const RH& rh )
{
        static_assert( std::is_base_of< IFastCollisionItem< typename LH::Full, N >, LH >::value );
        static_assert( std::is_base_of< IFastCollisionItem< typename RH::Full, N >, RH >::value );

        Distance dist{lh.getMinSpherePosition().distance( rh.getMinSpherePosition() )};

        return dist < ( lh.getMinSphereRadius() + rh.getMinSphereRadius() );
}

template < unsigned N, CollisionMode mode = CollisionMode::FAST, typename LH, typename RH, typename UnaryFunction >
bool treesNodesCollides( const LH& lh, const RH& rh, UnaryFunction f )
{
        if constexpr ( mode == CollisionMode::FULL ) {
                if ( !collides< N >( lh, rh ) ) {
                        return false;
                }
        } else if constexpr ( mode == CollisionMode::FAST ) {
                if ( !fastCollides< N >( lh, rh ) ) {
                        return false;
                }
        }

        if ( lh.isLeaf() && rh.isLeaf() ) {
                if constexpr ( mode == CollisionMode::FAST ) {
                        return collides< N >( lh.getFull(), rh.getFull() ) && f( lh, rh );
                } else {
                        return f( lh, rh );
                }
        }

        bool rather_right = false;

        if constexpr ( mode == CollisionMode::FAST ) {
                rather_right = lh.getMinSphereRadius() > rh.getMinSphereRadius();
        } else {
                rather_right = lh > rh;
        }

        if ( lh.isLeaf() || ( rather_right && !rh.isLeaf() ) ) {
                return anyOf( rh.getChildren(), [&]( const auto& child ) {  //
                        return treesNodesCollides< N, mode >( lh, child, f );
                } );
        }
        return anyOf( lh.getChildren(), [&]( const auto& child ) {  //
                return treesNodesCollides< N, mode >( child, rh, f );
        } );
}

template < unsigned      N,
           CollisionMode mode,
           typename LH,
           typename RH,
           typename CollisionFunction,
           typename CollectFunction >
void findCollisionNodes( const LH& lh, const RH& rh, CollisionFunction collision_f, CollectFunction collect_f )
{
        treesNodesCollides< N, mode >( lh, rh, [&]( const auto& lh_node, const auto& rh_node ) {
                if ( !collision_f( lh_node, rh_node ) ) {
                        return false;
                }

                collect_f( lh_node, rh_node );

                return false;
        } );
}

}  // namespace schpin
