#include "environment.h"
#include "octree/insert.h"

namespace schpin
{

class EnvironmentSurfaceVisitor : public IOcSimpleVisitor< EnvironmentOcTreeContent >
{
        const OcTree< EnvironmentOcTreeContent >& tree_;

      public:
        EnvironmentSurfaceVisitor( OcTree< EnvironmentOcTreeContent >& tree )
          : tree_( tree )
        {
        }

        OcNodeControl accessNode( const OcProxyNode< EnvironmentOcTreeContent >& node ) const
        {
                return OcNodeControl::FOLLOW_CHILDREN;
        }

        OcLockedLeafControl accessLockedLeaf( const OcProxyNode< EnvironmentOcTreeContent >& node ) const
        {
                return OcLockedLeafControl::EXPAND_CHILDREN;
        }

        void accessBottomLeaf( const OcProxyNode< EnvironmentOcTreeContent >& node,
                               EnvironmentOcTreeContent&                      content ) const
        {

                // verify that actual node holds 'air'
                if ( ( *node )->getOccupancy() >= ( ENV_MIN_OCCUPANCY + ENV_PRUNNING_DIFF ) ) {
                        return;
                }

                Point< 3 > desired_pos = node.getPosition() + Point< 3 >{0, 0, -node.getEdgeLength()};

                auto opt_below_node = findLeafNode( tree_, desired_pos );

                if ( !opt_below_node ) {
                        return;
                }

                const OcProxyNode< EnvironmentOcTreeContent >& below_node = *opt_below_node;

                // verify that node below holds ground
                if ( ( *below_node )->getOccupancy() < ENV_MAX_OCCUPANCY - ENV_PRUNNING_DIFF ) {
                        return;
                }

                content.setSurface( true );
        }
};

Either< Environment, EnvError > buildEnvironment( const EnvConfig& config )
{
        SCHPIN_INFO_LOG( "Building environment" );
        SCHPIN_INFO_LOG( "Default occ " << config.default_occ );
        SCHPIN_INFO_LOG( "Min edge length " << config.min_edge_length );
        forEach( config.paths, [&]( const std::filesystem::path& p ) {  //
                SCHPIN_INFO_LOG( "Path " << p );
        } );

        using Visitor =
            SingleCollisionInsert< FullCollisionChecker< TriangleCollisionQuery< 3 >, EnvironmentOcTreeContent > >;

        std::vector< ParseError >          errors;
        OcTree< EnvironmentOcTreeContent > tree{EnvironmentOcTreeContent( config.default_occ, false ),
                                                config.min_edge_length};

        std::size_t c = 0;

        auto handle_triangles_f = [&]( const std::vector< Triangle< 3 > >& triangles ) {
                forEach( triangles, [&]( const Triangle< 3 >& t ) {
                        ++c;
                        SCHPIN_INFO_LOG_DELAYED_THROTTLE(
                            5, "Triangle processed count: " << c << "/" << triangles.size() );
                        Visitor v{t, OccupancyToken::FULL};
                        tree.expand( t );
                        tree.modify( v );
                } );
        };

        forEach( config.paths, [&]( std::filesystem::path p ) {
                parseTriangles( p ).match( handle_triangles_f,
                                           [&]( ParseError error ) {  //
                                                   errors.push_back( error );
                                           } );
        } );

        if ( !errors.empty() ) {
                return {E( EnvError ).attach( errors ) << "Failed to build environment due to errors in parsing"};
        }

        EnvironmentSurfaceVisitor surface_visitor{tree};
        tree.modify( surface_visitor );

        SCHPIN_INFO_LOG( "Environment size: " << getNodeCount( tree.getProxyRoot() ) );

        return {Environment( std::move( tree ) )};
}

}  // namespace schpin
