#include "octree/interfaces.h"
#include "octree/occupancy.h"

#pragma once

namespace schpin
{
static constexpr uint16_t ENV_MAX_OCCUPANCY = 65535;
static constexpr uint16_t ENV_MIN_OCCUPANCY = 0;
static constexpr uint16_t ENV_PRUNNING_DIFF = 1024;  // #random

class EnvironmentOcTreeContent : public IOcContent< EnvironmentOcTreeContent, OccupancyToken >
{
        uint16_t occupancy_  = ENV_MAX_OCCUPANCY / 2;
        bool     is_surface_ = false;

      public:
        EnvironmentOcTreeContent() = default;

        EnvironmentOcTreeContent( uint16_t occupancy, bool is_surface )
          : occupancy_( occupancy )
          , is_surface_( is_surface )
        {
        }

        uint16_t getOccupancy() const
        {
                return occupancy_;
        }

        void setSurface( bool val )
        {
                is_surface_ = val;
        }

        bool isSurface() const
        {
                return is_surface_;
        }

        std::size_t getMemoryUsage( const OcNode< EnvironmentOcTreeContent >& node ) const override
        {
                return sizeof( node );
        }

        bool prunnable( const OcNode< EnvironmentOcTreeContent >& node ) const override
        {
                bool prunnable_occupancy = ( ENV_MAX_OCCUPANCY - occupancy_ ) < ENV_PRUNNING_DIFF ||
                                           ( occupancy_ - ENV_MIN_OCCUPANCY ) < ENV_PRUNNING_DIFF;

                bool prunnable_surface =
                    noneOf( node.getChildren(), [&]( const auto& child ) { return child->isSurface(); } );

                return prunnable_surface && prunnable_occupancy;
        }

        void childBasedUpdate( const OcNode< EnvironmentOcTreeContent >& node ) override
        {
                SCHPIN_ASSERT( node.size() != 0 );
                unsigned sum = accumulate(
                    node.getChildren(), unsigned( 0 ), [&]( const OcNode< EnvironmentOcTreeContent >& child ) {
                            return child->getOccupancy();
                    } );

                occupancy_ = uint16_t( sum / node.size() );
        }

        void setPrunned( const OcNode< EnvironmentOcTreeContent >& node ) override
        {
                if ( occupancy_ > ( ENV_MAX_OCCUPANCY + ENV_MIN_OCCUPANCY ) / 2 ) {
                        occupancy_ = ENV_MAX_OCCUPANCY;
                } else {
                        occupancy_ = ENV_MIN_OCCUPANCY;
                }
        }

        void update( const OccupancyToken& token ) override
        {
                switch ( token ) {
                        case OccupancyToken::FULL:
                                occupancy_ = ENV_MAX_OCCUPANCY;
                                return;
                }
        }
};
inline bool operator==( const EnvironmentOcTreeContent& lh, const EnvironmentOcTreeContent& rh )
{
        return lh.getOccupancy() == rh.getOccupancy() && lh.isSurface() == rh.isSurface();
}
inline bool operator!=( const EnvironmentOcTreeContent& lh, const EnvironmentOcTreeContent& rh )
{
        return !( lh == rh );
}
inline void from_json( const nlohmann::json& j, EnvironmentOcTreeContent& c )
{
        c = EnvironmentOcTreeContent{j["occ"].get< uint16_t >(), j["is_surface"].get< bool >()};
}
inline void to_json( nlohmann::json& j, const EnvironmentOcTreeContent& c )
{
        j["occ"]        = c.getOccupancy();
        j["is_surface"] = c.isSurface();
}
}  // namespace schpin
