#include "gait/chain.h"
#include "gait/edge.h"
#include "gait/vertex.h"
#include "graph/interface.h"
#include "leg/sidepath.h"
#include <schpin_lib/identificator_container.h>
#include <utility>

#pragma once

namespace schpin
{

struct ResetChain
{
        const RobotStateView&                          robot;
        const Environment&                             env;
        std::vector< std::pair< BodyVertexID, Pose > > body_path;
};

template < typename VertexContentT >
class GaitGraph : IGraph< GaitVertex< VertexContentT >, GaitEdge >
{
      public:
        using VertexContent          = VertexContentT;
        using Vertex                 = GaitVertex< VertexContent >;
        using Price                  = GaitPrice;
        using Edge                   = GaitEdge;
        using EdgeContent            = typename GaitEdge::Content;
        using EdgeID                 = GaitEdgeID;
        using VertexID               = GaitVertexID;
        using GaitChainIterator      = typename GaitChain< Vertex >::iterator;
        using GaitChainConstIterator = typename GaitChain< Vertex >::const_iterator;

      private:
        using ChainMap = std::unordered_map< BodyVertexID, GaitChainIterator >;

        GaitChain< Vertex > chain_;
        ChainMap            chain_mapping_;

      public:
        auto getChainIter( BodyVertexID bid )
        {
                auto iter = findIf( chain_, [&]( const auto& view ) { return view.source_pin.bid == bid; } );
                SCHPIN_ASSERT( iter != chain_.end() );
                return iter;
        }

        template < typename Visitor >
        void event( ResetChain e, Visitor )
        {
                // TODO visitor shall not be ignored
                chain_.clear();
                forEach( e.body_path, [&]( std::pair< BodyVertexID, Pose > p ) {
                        chain_.pushBack( GaitPin{p.second, p.first} );
                } );

                forEach( chain_, [&]( const auto& view ) {  //
                        view.link.body_segment = BodyViewSegment(
                            view.source_pin.body_pose, view.target_pin.body_pose, Angle{0.1f}, Distance{0.02f} );
                } );

                auto sidepaths = createSidepaths< 3 >( chain_, e.robot, e.env );
                forEach( sidepaths, [&]( const std::pair< BodyVertexID, LegContainer< LegSidepathSegment > >& item ) {
                        getChainIter( item.first )->link.sidepaths_segments = item.second;
                } );

        }

        const GaitChain< Vertex >& getChain() const
        {
                return chain_;
        }

        const GaitEdge& getEdge( const GaitEdgeID& eid ) const override
        {
                return getVertex( eid.vertex_id ).getEdge( eid.edge_offset );
        }

        const Vertex& getVertex( const GaitVertexID& id ) const override
        {
                auto chain_iter = chain_mapping_.find( id.bid );

                SCHPIN_ASSERT( chain_iter != chain_mapping_.end() );
                return chain_iter->second->link.vertexes[id.gid];
        }

        template < typename UnaryFunction >
        void modifyAll( UnaryFunction f )
        {
                forEach( chain_, [&]( GaitChainView< Vertex > view ) {
                        forEach( view.link.vertexes, [&]( Vertex& vertex ) {  //
                                f( *vertex );
                        } );
                } );
        }

        VertexContent& refContent( const GaitVertexID& id ) override
        {
                return *getChainIter( id.bid )->link.vertexes[id.gid];
        }

        template < typename UnaryFunction >
        void addEdge( const GaitVertexID source, const GaitVertexID target, UnaryFunction f )
        {
                GaitEdgeID eid =
                    getChainIter( source.bid )->link.vertexes[source.gid].addEdge( f( source, target ), target );
                getChainIter( target.bid )->link.vertexes[target.gid].addReverseEdge( eid );
        }

        EdgeContent& refEdgeContent( const GaitEdgeID& eid )
        {
                return getChainIter( eid.vertex_id.bid )
                    ->link.vertexes[eid.vertex_id.gid]
                    .refEdgeContent( eid.edge_offset );
        }

        const Vertex& emplaceVertex( BodyVertexID bid, VertexContent content )
        {
                GaitVertexID id{bid, getChainIter( bid )->link.vertexes.size()};
                getChainIter( bid )->link.vertexes.emplace_back( id, std::move( content ) );
                return getVertex( id );
        }
};

}  // namespace schpin
