#include "graph/astar/content.h"
#include "graph/base.h"
#include "graph/util.h"
#include <optional>
#include <schpin_lib/chain.h>

#pragma once

namespace schpin
{

template < typename VertexID, typename Price >
struct AStarQueueItem
{
        Price    p;
        VertexID id;

        AStarQueueItem( Price p, VertexID id )
          : p( p )
          , id( id )
        {
        }
};

template < typename VertexID, typename Price >
bool operator<( const AStarQueueItem< VertexID, Price >& lh, const AStarQueueItem< VertexID, Price >& rh )
{
        return lh.p < rh.p;
}
template < typename VertexID, typename Price >
bool operator>( const AStarQueueItem< VertexID, Price >& lh, const AStarQueueItem< VertexID, Price >& rh )
{
        return lh.p > rh.p;
}

template < typename AStarConfig >
class AStarOverlay
{
      public:
        using Graph             = typename AStarConfig::Graph;
        using Heap              = typename AStarConfig::Heap;
        using HeuristicFunction = typename AStarConfig::HeuristicFunction;
        using Vertex            = typename Graph::Vertex;
        using VertexID          = typename Vertex::ID;
        using VertexContent     = typename Vertex::Content;
        using Edge              = typename Graph::Edge;
        using EdgeID            = typename Edge::ID;
        using QueueItem         = typename AStarConfig::QueueItem;
        using Price             = typename Graph::Price;
        using HeapHandle        = typename AStarConfig::HeapHandle;

      private:
        Graph                                     graph_;
        Heap                                      heap_;
        std::optional< GraphSearchTask< Graph > > task_;
        GraphSearchState                          search_state_;
        HeuristicFunction                         heuristic_function_;

      public:
        AStarOverlay( Graph graph, HeuristicFunction hf )
          : graph_( std::move( graph ) )
          , heuristic_function_( hf )
        {
        }
        AStarOverlay( Graph graph )
          : graph_( std::move( graph ) )
        {
        }
        void removeVertex( VertexID vid )
        {
                graph_.removeVertex( vid );
                invalidateSearch();
        }
        auto& refContent( VertexID id )
        {
                return graph_.refContent( id );
        }
        const Graph& operator*() const
        {
                return graph_;
        }
        const Graph* operator->() const
        {
                return &graph_;
        }
        const std::optional< GraphSearchTask< Graph > >& getTask() const
        {
                return task_;
        }
        GraphSearchState getSearchState() const
        {
                return search_state_;
        }
        template < typename Tag >
        auto event( Tag&& tag )
        {
                auto visitor = makeAffectionVisitor< Graph >( [this]( EdgeID ) { this->invalidateSearch(); },
                                                              [this]( VertexID ) { this->invalidateSearch(); } );
        }
        void setTask( VertexID current, VertexID goal )
        {
                task_.reset();
                task_ = GraphSearchTask< Graph >{current, goal};
                invalidateSearch();
        }
        template < typename... Args >
        const Vertex& emplaceVertex( Args... args )
        {
                const Vertex& vertex = graph_.emplaceVertex( args... );
                invalidateSearch();
                return vertex;
        }
        template < typename Container, typename UnaryFunction >
        void addEdges( VertexID source, const Container& targets, UnaryFunction f )
        {
                forEach( targets, [&]( VertexID target_id ) {  //
                        graph_.addEdge( source, target_id, f );
                } );
                invalidateSearch();
        }
        template < typename UnaryFunction >
        void addEdge( VertexID source, VertexID target_id, UnaryFunction f )
        {
                graph_.addEdge( source, target_id, f );
                invalidateSearch();
        }
        const Graph& getGraph() const
        {
                return graph_;
        }
        const Chain< EdgeID, VertexID > getPath() const
        {

                Chain< EdgeID, VertexID > res;

                if ( !task_ ) {
                        return res;
                }

                if ( search_state_ != GraphSearchState::FOUND_PATH ) {
                        return res;
                }

                const Vertex* actual = &graph_.getVertex( task_->goal_id );

                res.pushBack( actual->getID() );

                while ( actual->getID() != task_->current_id ) {

                        const Edge& edge = graph_.getEdge( parent_edge( actual->getID() ) );

                        actual = &graph_.getVertex( edge.getSourceID() );

                        res.pushBack( actual->getID() );
                        res.end()->link = edge.getID();
                }

                res.reverse();

                return res;
                // TODO: method same as in dstar
        }
        std::size_t computeShortestPath( std::size_t step_limit )
        {
                return computeShortestPath(
                    step_limit, makeGraphProcessVisitor< Graph >( []( auto ) { return true; }, []( auto ) {} ) );
        }
        template < typename Visitor >
        std::size_t computeShortestPath( std::size_t step_limit, Visitor visitor )
        {
                if ( !task_ ) {
                        return 0;
                }

                return *findIf( range( step_limit ), [&]( std::size_t ) {
                        checkPathExistence();
                        if ( !hasCandidateQueueItem() ) {
                                return true;
                        }
                        processQueueItem( visitor );
                        return false;
                } );
        }

      private:
        void tryUpdateG( VertexID vertex, EdgeID eid, Price candidate_g )
        {
                if ( candidate_g >= g( vertex ) ) {
                        return;
                }

                // TODO: dirty trick to chekc validty of handle, fix...
                if ( g( vertex ) == std::numeric_limits< Price >::max() ) {
                        handle( vertex ) = heap_.emplace( candidate_g, vertex );
                } else {
                        heap_.update( handle( vertex ), QueueItem( candidate_g, vertex ) );
                }

                parent_edge( vertex ) = eid;
                g( vertex )           = candidate_g;
        }
        template < typename Visitor >
        void processQueueItem( Visitor visitor )
        {
                SCHPIN_ASSERT( task_ );
                VertexID      actual_id = heap_.top().id;
                const Vertex& actual    = graph_.getVertex( actual_id );

                if ( !visitor.vertexReady( actual ) ) {
                        visitor.initialize( actual_id );
                }
                heap_.pop();
                handle( actual ) = HeapHandle();

                forEach( actual.getEdges(), [&]( const Edge& e ) {
                        Price candidate_g = g( actual_id ) + e.getPrice();

                        tryUpdateG( e.getTargetID(), e.getID(), candidate_g );
                } );
        }

        void checkPathExistence()
        {
                SCHPIN_ASSERT( task_ );
                VertexID goal = task_->goal_id;
                if ( g( goal ) == std::numeric_limits< Price >::max() ) {
                        search_state_ = GraphSearchState::PATH_MISSING;
                } else {
                        search_state_ = GraphSearchState::FOUND_PATH;
                }
        }
        bool hasCandidateQueueItem()
        {
                SCHPIN_ASSERT( task_ );
                VertexID goal = task_->goal_id;
                if ( heap_.empty() ) {
                        return false;
                }

                if ( heap_.top().p <= g( goal ) ) {
                        return true;
                }
                return false;
        }
        Price& g( VertexID id )
        {
                return graph_.refContent( id ).search_content.g;
        }
        HeapHandle& handle( VertexID id )
        {
                return graph_.refContent( id ).search_content.handle;
        }
        HeapHandle& handle( const Vertex& v )
        {
                return graph_.refContent( v.getID() ).search_content.handle;
        }

        EdgeID& parent_edge( VertexID id )
        {
                return graph_.refContent( id ).search_content.parent_edge;
        }
        const EdgeID& parent_edge( VertexID id ) const
        {
                return graph_.getVertex( id )->search_content.parent_edge;
        }
        void invalidateSearch()
        {
                heap_.clear();
                graph_.modifyAll( [&]( VertexContent& content ) {  //
                        content.search_content.g = std::numeric_limits< Price >::max();
                } );
                if ( task_ ) {
                        g( task_->current_id ) = Price( 0. );
                        heap_.emplace( Price( 0. ), task_->current_id );
                }
                search_state_ = GraphSearchState::NOT_FINISHED;
        }
};

}  // namespace schpin
