/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 *
 * This file is part of Schpin.
 *
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */

#include "graph/base.h"
#include "graph/dstar/content.h"
#include "graph/interface.h"
#include "graph/util.h"
#include <algorithm>
#include <array>
#include <boost/heap/pairing_heap.hpp>
#include <schpin_lib/util.h>
#include <utility>
#include <vector>

#pragma once

namespace schpin
{

/*
 * QueueItem is used in priority queues
 */
template < typename VertexID, typename Price >
struct DstarQueueItem
{
        std::array< Price, 2 > keys;
        VertexID               vid;

        DstarQueueItem( std::array< Price, 2 > keys, VertexID vid )
          : keys( keys )
          , vid( vid )
        {
        }
};

template < typename VertexID, typename Price >
bool operator<( const DstarQueueItem< VertexID, Price >& lh, const DstarQueueItem< VertexID, Price >& rh )
{
        return lh.keys < rh.keys;
}

template < typename DStarConfig >
class AnytimeDStar
{
      public:
        using Graph             = typename DStarConfig::Graph;
        using Vertex            = typename Graph::Vertex;
        using Edge              = typename Graph::Edge;
        using Heap              = typename DStarConfig::Heap;
        using QueueItem         = typename DStarConfig::QueueItem;
        using HeuristicFunction = typename DStarConfig::HeuristicFunction;
        using VertexContent     = typename Vertex::Content;
        using VertexID          = typename Vertex::ID;
        using EdgeID            = typename Edge::ID;
        using Price             = typename Graph::Price;
        using HeapHandle        = typename DStarConfig::HeapHandle;

        static_assert( std::is_base_of< IGraph< Vertex, Edge >, Graph >::value );

        enum Flags
        {
                OPEN   = 1 << 0,
                INCONS = 1 << 1,
                CLOSED = 1 << 2
        };

      private:
        Graph                                     graph_;
        Heap                                      open_;
        std::vector< VertexID >                   incons_;
        std::vector< VertexID >                   closed_;
        GraphSearchState                          search_state_;
        HeuristicFunction                         heuristic_function_;
        std::optional< GraphSearchTask< Graph > > task_;

        static void initializeContent( VertexContent& content )
        {

                content.search_content.g   = std::numeric_limits< Price >::max();
                content.search_content.rhs = std::numeric_limits< Price >::max();
        }

      public:
        AnytimeDStar()
          : AnytimeDStar( Graph{}, HeuristicFunction{} )
        {
        }
        AnytimeDStar( Graph graph, HeuristicFunction hf )
          : graph_( std::move( graph ) )
          , open_()
          , search_state_( GraphSearchState::NOT_FINISHED )
          , heuristic_function_( hf )
          , task_()
        {
                graph_.modifyAll( AnytimeDStar::initializeContent );
        }

        const std::optional< GraphSearchTask< Graph > >& getTask() const
        {
                return task_;
        }

        template < typename Tag >
        auto event( Tag&& tag )
        {
                auto visitor = makeAffectionVisitor< Graph >(
                    [this]( EdgeID eid ) {  //
                            this->updateEdge( eid );
                    },
                    [this]( VertexID vid ) {  //
                            this->updateVertex( vid );
                    } );
                return graph_.event( std::forward< Tag >( tag ), visitor );
        }

        GraphSearchState getSearchState() const
        {
                return search_state_;
        }

        void setTask( VertexID current, VertexID goal )
        {
                if ( task_ ) {

                        g( task_->goal_id )   = std::numeric_limits< Price >::max();
                        rhs( task_->goal_id ) = calcualteRhs( task_->goal_id );

                        updateVertex( task_->goal_id );
                }
                g( goal )      = std::numeric_limits< Price >::max();
                rhs( goal )    = Price( 0 );
                g( current )   = std::numeric_limits< Price >::max();
                rhs( current ) = std::numeric_limits< Price >::max();
                updateVertex( goal );
                updateVertex( current );
                task_         = GraphSearchTask< Graph >{current, goal};
                search_state_ = GraphSearchState::NOT_FINISHED;
        }

        template < typename... Args >
        const Vertex& emplaceVertex( Args... args )
        {
                const Vertex& vertex = graph_.emplaceVertex( args... );

                AnytimeDStar::initializeContent( graph_.refContent( vertex.getID() ) );

                updateVertex( vertex.getID() );

                return vertex;
        }

        template < typename Container, typename UnaryFunction >
        void addEdges( VertexID source, const Container& targets, UnaryFunction f )
        {
                forEach( targets, [&]( VertexID target_id ) {  //
                        graph_.addEdge( source, target_id, f );
                } );

                updateVertex( source );
        }

        template < typename UnaryFunction >
        void addEdge( VertexID source, VertexID target_id, UnaryFunction f )
        {
                graph_.addEdge( source, target_id, f );

                updateVertex( source );
        }

        const Graph& getGraph() const
        {
                return graph_;
        }

        auto getVertexes() const
        {
                return graph_.getVertexes();
        }

        template < typename Visitor >
        const Visitor visitPath()
        {
                static_assert( std::is_base_of< IGraphPathVisitor< Graph >, Visitor >::value );
                Visitor visitor;

                if ( !task_ ) {
                        return visitor;
                }

                const Vertex* actual = &graph_.getVertex( task_->current_id );

                for ( ;; ) {
                        visitor.visitVertex( *actual, graph_.refContent( actual->getID() ) );

                        if ( actual->getID() == task_->goal_id ) {
                                return visitor;
                        }

                        auto edge_iter = findIf( actual->getEdges(), [&]( const Edge& edge ) {
                                const VertexID& target = edge.getTargetID();
                                return g( target ) + edge.getPrice() == g( actual->getID() );
                        } );

                        if ( edge_iter == actual->getEdges().end() ) {
                                return visitor;
                        }

                        visitor.visitEdge( *edge_iter );
                }

                return visitor;
        }

        std::size_t computeShortestPath( std::size_t step_limit )
        {
                return computeShortestPath(
                    step_limit, makeGraphProcessVisitor< Graph >( []( auto ) { return true; }, []( auto ) {} ) );
        }

        template < typename Visitor >
        std::size_t computeShortestPath( std::size_t step_limit, Visitor visitor )
        {
                if ( !task_ ) {
                        return 0;
                }
                static_assert( std::is_base_of_v< IGraphProcessVisitor< Graph >, Visitor > );

                return *findIf( range( step_limit ), [&]( std::size_t ) {  //
                        checkPathExistence();
                        if ( hasCandidateQueueItem() ) {
                                processQueueItem( visitor );
                                return false;
                        }
                        if ( heuristic_function_.minimal() ) {
                                return true;
                        }
                        heuristic_function_.decrease();

                        Heap tmp{};

                        forEach( open_, [&]( const QueueItem& qitem ) {  //
                                tmp.emplace( keys( qitem.vid ), qitem.vid );
                        } );
                        forEach( incons_, [&]( VertexID id ) {  //
                                flags( id )[INCONS] = 1;
                                tmp.emplace( keys( id ), id );
                        } );
                        forEach( closed_, [&]( VertexID id ) {  //
                                flags( id )[CLOSED] = 0;
                        } );
                        closed_.clear();
                        open_ = tmp;
                        incons_.clear();
                        return false;
                } );
        }

      public:
        HeapHandle& handle( VertexID id )
        {
                return graph_.refContent( id ).search_content.handle;
        }

        std::bitset< 8 >& flags( VertexID id )
        {
                return graph_.refContent( id ).search_content.flags;
        }

        Price& g( VertexID id )
        {
                return graph_.refContent( id ).search_content.g;
        }
        const Price& g( VertexID id ) const
        {
                return graph_.getVertex( id )->search_content.g;
        }
        Price& rhs( VertexID id )
        {
                return graph_.refContent( id ).search_content.rhs;
        }
        const Price& rhs( VertexID id ) const
        {
                return graph_.getVertex( id )->search_content.rhs;
        }

        Price hf( VertexID from, VertexID to )
        {
                return heuristic_function_( graph_.getVertex( from ), graph_.getVertex( to ) );
        }

        Price calcualteRhs( const VertexID& vid ) const
        {
                const auto& edges = graph_.getVertex( vid ).getEdges();

                if ( edges.empty() ) {
                        return std::numeric_limits< Price >::max();
                }

                return min( edges, [this]( const Edge& edge ) {  //
                        return edge.getPrice() + this->g( edge.getTargetID() );
                } );
        }

        std::array< Price, 2 > keys( VertexID vertex )
        {
                SCHPIN_ASSERT( task_ );
                if ( g( vertex ) > rhs( vertex ) ) {
                        return {rhs( vertex ) + hf( vertex, task_->goal_id ), rhs( vertex )};
                } else {
                        return {g( vertex ) + hf( vertex, task_->goal_id ), g( vertex )};
                }
        }

        void checkPathExistence()
        {
                SCHPIN_ASSERT( task_ );
                VertexID current = task_->current_id;
                if ( g( current ) == std::numeric_limits< Price >::max() ) {
                        search_state_ = GraphSearchState::PATH_MISSING;
                } else {
                        search_state_ = GraphSearchState::FOUND_PATH;
                }
        }

        bool hasCandidateQueueItem()
        {
                SCHPIN_ASSERT( task_ );
                VertexID current = task_->current_id;
                if ( open_.empty() ) {
                        return false;
                }

                if ( open_.top().keys < keys( current ) || rhs( current ) != g( current ) ) {
                        return true;
                }
                return false;
        }

        template < typename Visitor >
        bool isTopReadyForProcessing( Visitor visitor )
        {
                SCHPIN_ASSERT( task_ );
                VertexID      actual_vid = open_.top().vid;
                const Vertex& actual     = graph_.getVertex( actual_vid );

                if ( !visitor.vertexReady( actual ) ) {
                        visitor.initialize( actual_vid );
                        return false;
                }
                return true;
        }

        void updateEdge( const EdgeID& eid )
        {
                updateVertex( graph_.getEdge( eid ).getSourceID() );
                search_state_ = GraphSearchState::NOT_FINISHED;
        }

        void updateVertex( VertexID id )
        {
                if ( id != this->task_->goal_id ) {
                        rhs( id ) = calcualteRhs( id );
                }

                if ( inOpen( id ) ) {
                        eraseOpen( id );
                }

                // if rhs/g price are inconsistent, update the vertex
                if ( g( id ) != rhs( id ) ) {
                        if ( inClosed( id ) ) {
                                pushOpen( id );
                        } else {
                                pushIcons( id );
                        }
                }
        }

        void pushIcons( VertexID id )
        {
                incons_.push_back( id );
                flags( id )[INCONS] = 1;
        }

        void pushOpen( VertexID id )
        {
                flags( id )[OPEN] = 1;
                handle( id )      = open_.emplace( keys( id ), id );
        }

        void eraseOpen( VertexID id )
        {
                flags( id )[OPEN] = 0;
                open_.erase( handle( id ) );
        }

        bool inOpen( VertexID id )
        {
                return flags( id )[OPEN];
        }
        bool inClosed( VertexID id )
        {
                return flags( id )[CLOSED];
        }

        void popOpen( VertexID id )
        {
                flags( id )[OPEN] = 0;
                open_.pop();
        }

        void pushClosed( VertexID id )
        {
                flags( id )[CLOSED] = 1;
                closed_.emplace_back( id );
        }

        template < typename Visitor >
        void processQueueItem( Visitor visitor )
        {
                SCHPIN_ASSERT( task_ );

                if ( !isTopReadyForProcessing( visitor ) ) {
                        return;
                }

                VertexID vertex = open_.top().vid;

                popOpen( vertex );

                if ( g( vertex ) >= rhs( vertex ) ) {
                        g( vertex ) = rhs( vertex );
                        pushClosed( vertex );
                } else {
                        g( vertex ) = std::numeric_limits< Price >::max();
                        updateVertex( vertex );
                }
                forEach( graph_.getVertex( vertex ).getEdges(), [&]( const Edge& e ) {  //
                        updateVertex( e.getTargetID() );
                } );
        }
};

}  // namespace schpin
