#include "graph/interface.h"
#include <schpin_lib/identificator_container.h>

#pragma once

namespace schpin
{
template < typename VertexID, typename VertexContent, typename Edge >
class GenericVertex : IGraphVertex< VertexID, VertexContent, Edge >
{
      public:
        using ID          = VertexID;
        using Content     = VertexContent;
        using EdgeOffset  = typename Edge::Offset;
        using EdgeID      = typename Edge::ID;
        using EdgeContent = typename Edge::Content;

      private:
        using EdgeContainer            = Container< Edge, EdgeOffset, std::vector >;
        using ReverseEdgeContainer     = Container< EdgeID, EdgeOffset, std::vector >;
        using ConstEdgeIterator        = typename EdgeContainer::const_iterator;
        using ReverseConstEdgeIterator = typename ReverseEdgeContainer::const_iterator;

        EdgeContainer        edges_{};
        ReverseEdgeContainer reverse_edges_{};
        VertexID             id_;
        VertexContent        content_;

      public:
        GenericVertex( VertexID id, VertexContent content )
          : id_( std::move( id ) )
          , content_( std::move( content ) )
        {
        }

        GenericVertex( const GenericVertex& ) = default;
        GenericVertex( GenericVertex&& )      = default;
        GenericVertex& operator=( const GenericVertex& ) = default;
        GenericVertex& operator=( GenericVertex&& ) = default;

        ~GenericVertex() = default;

        const Edge& getEdge( EdgeOffset eid ) const override
        {
                return edges_[eid];
        }
        Range< ConstEdgeIterator > getEdges() const
        {
                return range( edges_ );
        }
        Range< ReverseConstEdgeIterator > getReverseEdges() const
        {
                return range( reverse_edges_ );
        }
        EdgeID addEdge( EdgeContent&& content, VertexID target ) override
        {
                EdgeID eid{id_, edges_.size()};
                edges_.emplace_back( eid, std::move( content ), target );
                return eid;
        }
        void addReverseEdge( EdgeID eid ) override
        {
                reverse_edges_.emplace_back( eid );
        }
        VertexID getID() const override
        {
                return id_;
        }

        const VertexContent& operator*() const override
        {
                return content_;
        }
        const VertexContent* operator->() const override
        {
                return &content_;
        }
        VertexContent& operator*() override
        {
                return content_;
        }
        VertexContent* operator->() override
        {
                return &content_;
        }
        EdgeContent& refEdgeContent( EdgeOffset eoffset ) override
        {
                return *edges_[eoffset];
        }
        void swapEdges( EdgeOffset lh, EdgeOffset rh )
        {
                using std::swap;
                swap( edges_[lh], edges_[rh] );
        }

        void removeLastEdge()
        {
                edges_.popBack();
        }

        void removeReverseEdge( EdgeID eid )
        {
                auto iter = find( reverse_edges_, eid );
                SCHPIN_ASSERT( iter != reverse_edges_.end() );
                reverse_edges_.erase( iter );
        }

      private:
        GenericVertex( VertexID               id,
                       VertexContent&&        content,
                       EdgeContainer&&        edges,
                       ReverseEdgeContainer&& reverse_edges )
          : edges_( edges )
          , reverse_edges_( reverse_edges )
          , id_( id )
          , content_( content )
        {
        }

        friend nlohmann::adl_serializer< GenericVertex >;
};
template < typename VertexID, typename VertexContent, typename Edge >
bool operator==( const GenericVertex< VertexID, VertexContent, Edge >& lh,
                 const GenericVertex< VertexID, VertexContent, Edge >& rh )
{
        return ( lh.getID() == rh.getID() );
}
template < typename VertexID, typename VertexContent, typename Edge >
bool operator!=( const GenericVertex< VertexID, VertexContent, Edge >& lh,
                 const GenericVertex< VertexID, VertexContent, Edge >& rh )
{
        return !( lh == rh );
}
}  // namespace schpin
namespace nlohmann
{
template < typename VertexID, typename VertexContent, typename Edge >
struct adl_serializer< schpin::GenericVertex< VertexID, VertexContent, Edge > >
{
        using Vertex = schpin::GenericVertex< VertexID, VertexContent, Edge >;
        static void to_json( nlohmann::json& json, const Vertex& vertex )
        {
                const auto& content = *vertex;

                json = nlohmann::json{{"id", vertex.getID()},
                                      {"content", content},
                                      {"edges", vertex.getEdges()},
                                      {"reverse_edges", vertex.getReverseEdges()}};
        }

        static Vertex from_json( const nlohmann::json& j )
        {
                using EdgeContainer        = typename Vertex::EdgeContainer;
                using ReverseEdgeContainer = typename Vertex::ReverseEdgeContainer;

                return Vertex( j["id"].get< VertexID >(),
                               j["content"].get< VertexContent >(),
                               j["edges"].get< EdgeContainer >(),
                               j["reverse_edges"].get< ReverseEdgeContainer >() );
        }
};
}  // namespace nlohmann
namespace schpin
{
template < typename EdgeID, typename ContentT, typename PriceT, typename VertexID, typename OffsetT >
class GenericEdge : IGraphEdge< ContentT, PriceT, VertexID >
{
        EdgeID   id_;
        ContentT content_;
        VertexID to_;

      public:
        using Price   = PriceT;
        using ID      = EdgeID;
        using Content = ContentT;
        using Offset  = OffsetT;

        GenericEdge() = default;
        GenericEdge( EdgeID id, Content content, VertexID vid )
          : id_( id )
          , content_( std::move( content ) )
          , to_( vid )
        {
        }

        Content& operator*() override
        {
                return content_;
        }
        Content* operator->() override
        {
                return &content_;
        }

        const Content& operator*() const override
        {
                return content_;
        }
        const Content* operator->() const override
        {
                return &content_;
        }
        Price getPrice() const override
        {
                return content_.getPrice();
        }

        EdgeID getID() const
        {
                return id_;
        }

        VertexID getSourceID() const override
        {
                return id_.vertex_id;
        }

        VertexID getTargetID() const override
        {
                return to_;
        }
};

template < typename Price >
class ZeroGraphHeuristics
{
      public:
        bool minimal()
        {
                return true;
        }

        void decrease()
        {
        }

        template < typename T >
        Price operator()( const T&, const T& )
        {
                return Price{};
        }
};
}  // namespace schpin
