#include "base/point.h"
#include "graph/interface.h"

#pragma once

namespace schpin
{

template < typename Graph, typename VertexToPointFunction >
auto graphFindClosestVertex( const Point< 3 >&               target_point,
                             const typename Graph::VertexID& start,
                             const Graph&                    graph,
                             VertexToPointFunction           v_to_p_f ) -> typename Graph::VertexID
{
        using Vertex = typename Graph::Vertex;
        using Edge   = typename Graph::Edge;

        const Vertex& vertex = graph.getVertex( start );

        auto iter = minIterator( vertex.getEdges(), [&]( const Edge& edge ) {
                return v_to_p_f( edge.getTargetID() ).distance( target_point );
        } );

        const Vertex& best_nb = graph.getVertex( iter->getTargetID() );

        Distance vertex_distance  = v_to_p_f( vertex.getID() ).distance( target_point );
        Distance best_nb_distance = v_to_p_f( best_nb.getID() ).distance( target_point );
        if ( vertex_distance > best_nb_distance ) {
                return graphFindClosestVertex( target_point, best_nb.getID(), graph, v_to_p_f );
        }

        return vertex.getID();
}

template < typename Graph, typename ReadyFunction, typename InitializeFunction >
class LambdaGraphProcessVisitor : IGraphProcessVisitor< Graph >
{
        ReadyFunction      ready_f_;
        InitializeFunction initialize_f_;

      public:
        LambdaGraphProcessVisitor( ReadyFunction ready_f, InitializeFunction initialize_f )
          : ready_f_( ready_f )
          , initialize_f_( initialize_f )
        {
        }

        using Vertex   = typename Graph::Vertex;
        using VertexID = typename Graph::Vertex::ID;

        bool vertexReady( const Vertex& vertex ) override
        {
                return ready_f_( vertex );
        }
        void initialize( VertexID vid ) override
        {
                initialize_f_( vid );
        };
};

template < typename Graph, typename ReadyFunction, typename InitializeFunction >
LambdaGraphProcessVisitor< Graph, ReadyFunction, InitializeFunction > makeGraphProcessVisitor( ReadyFunction      rf,
                                                                                               InitializeFunction inif )
{
        return {rf, inif};
}

template < typename Graph, typename EdgeFunction, typename VertexFunction >
class LambdaAffectionVisitor : IGraphAffectionVisitor< Graph >
{
        EdgeFunction   edge_f_;
        VertexFunction vertex_f_;

      public:
        using EdgeID   = typename Graph::Edge::ID;
        using VertexID = typename Graph::Vertex::ID;

        LambdaAffectionVisitor( EdgeFunction edge_f, VertexFunction vertex_f )
          : edge_f_( edge_f )
          , vertex_f_( vertex_f )
        {
        }

        void affectedEdge( EdgeID eid ) override
        {
                edge_f_( eid );
        }
        void affectedVertex( VertexID vid ) override
        {
                vertex_f_( vid );
        }
};

template < typename Graph, typename EdgeFunction, typename VertexFunction >
LambdaAffectionVisitor< Graph, EdgeFunction, VertexFunction > makeAffectionVisitor( EdgeFunction ef, VertexFunction vf )
{
        return {ef, vf};
}

}  // namespace schpin
