#include "base/physics.h"
#include "base/pose.h"
#include "base/xml.h"
#include "leg/graph.h"
#include "leg/graph/gen.h"
#include "leg/graph/mapping.h"
#include "leg/graph/space.h"
#include "leg/model.h"
#include "leg/storage.h"
#include "leg/workarea.h"
#include "octree/insert.h"
#include "octree/node.h"

#pragma once

namespace schpin
{

template < unsigned N >
class Leg
{
        LegData< N > data_;
        LegID        id_;

      public:
        explicit Leg( LegID id, LegData< N >&& data )
          : data_( std::move( data ) )
          , id_( id )
        {
        }

        Pose getOriginOffset() const
        {
                return data_.model.getOriginOffset();
        }

        LegID getID() const
        {
                return id_;
        }

        const SimplexStorage< LegVertexID, N >& getGraphSpace() const
        {
                return data_.graph_space;
        }

        const OcTree< OccupancyContent >& getWorkarea() const
        {
                return data_.workarea;
        }

        const LegDstar< N >& getDstar() const
        {
                return data_.dstar;
        }

        template < typename UnaryFunction >
        void accessDstar( UnaryFunction f )
        {
                f( data_.dstar );
        }

        const LegModel< N >& getModel() const
        {
                return data_.model;
        }

        std::string getName() const
        {
                return data_.name;
        }

        std::string getFrame() const
        {
                return data_.name;
        }

        const LegData< N >& getLegData() const
        {
                return data_;
        }

        template < typename T >
        Point< 3 > getTipPosition( const T& angles ) const
        {
                return transform( data_.model.tipPosition( angles ), getOriginOffset() );
        }
};

struct LegSRDFChain
{
        std::string base_link;
        std::string tip_link;
};

inline Either< std::vector< LegSRDFChain >, XMLStorageError > loadLegChains( const pt::ptree& tree )
{
        auto leg_group_iter = findIf( tree.get_child( "robot" ), []( const pt::ptree::value_type& v ) {
                return v.second.get< std::string >( "<xmlattr>.name", "" ) == "legs";
        } );

        if ( leg_group_iter == tree.get_child( "robot" ).end() ) {
                return {E( XMLStorageError ) << "SRDF XML file does not have group with name 'legs'"};
        }

        std::vector< LegSRDFChain > res;

        forEach( leg_group_iter->second, [&]( const pt::ptree::value_type& v ) {
                if ( v.first != "chain" ) {
                        return;
                }
                LegSRDFChain chain;
                chain.base_link = v.second.get< std::string >( "<xmlattr>.base_link" );
                chain.tip_link  = v.second.get< std::string >( "<xmlattr>.tip_link" );
                res.push_back( chain );
        } );

        return res;
}

inline std::vector< Either< Leg< 3 >, LegBootError > > mapSRDFChainsToLegs( const std::filesystem::path& cache_prefix,
                                                                            const urdf::Model&           model,
                                                                            const std::vector< LegSRDFChain >& chains )
{
        LegID id( 0 );
        return mapF< std::vector< Either< Leg< 3 >, LegBootError > > >(
            chains, [&]( LegSRDFChain chain ) -> Either< Leg< 3 >, LegBootError > {
                    auto base_ptr = model.getLink( chain.base_link );
                    auto tip_ptr  = model.getLink( chain.tip_link );
                    if ( !base_ptr ) {
                            return {E( LegBootError )
                                    << "Base link " << chain.base_link << " is missing in the model."};
                    }
                    if ( !tip_ptr ) {
                            return {E( LegBootError ) << "Tip link " << chain.base_link << " is missing in the model."};
                    }
                    return loadOrGenerateLegData< 3 >( cache_prefix, model, base_ptr, tip_ptr )
                        .convertLeft( [&]( LegData< 3 >&& data ) {
                                Leg< 3 > leg( id, std::move( data ) );
                                ++id;
                                return leg;
                        } );
            } );
}

inline Either< LegContainer< Leg< 3 > >, LegBootError > loadLegs( const std::filesystem::path& cache_prefix,
                                                                  const urdf::Model&           model,
                                                                  std::istream&                srdf_istream )
{
        return safeLoadXml< std::vector< LegSRDFChain > >( srdf_istream, loadLegChains )
            .convertRight( []( XMLStorageError e ) {
                    return E( LegBootError ).attach( e ) << "Failed to load legs due to xml storage error";
            } )
            .transformLeft(
                [&]( std::vector< LegSRDFChain >&& chain ) -> Either< LegContainer< Leg< 3 > >, LegBootError > {
                        auto legs = mapSRDFChainsToLegs( cache_prefix, model, chain );

                        PartitionedEithers< Leg< 3 >, LegBootError > partitioned =
                            partitionEithers( std::move( legs ) );

                        if ( !partitioned.right.empty() ) {
                                return {E( LegBootError ).attach( partitioned.right ) << "Failed to load legs:"};
                        }

                        return {LegContainer< Leg< 3 > >( std::move( partitioned.left ) )};
                } );
}

}  // namespace schpin

namespace nlohmann
{

template < unsigned N >
struct adl_serializer< schpin::Leg< N > >
{
        static void to_json( nlohmann::json& j, const schpin::Leg< N >& leg )
        {
                j["id"]   = leg.getID();
                j["data"] = leg.getLegData();
        }

        static schpin::Leg< N > from_json( const nlohmann::json& j )
        {
                return schpin::Leg< N >{j["id"].get< schpin::LegID >(), j["data"].get< schpin::LegData< N > >()};
        }
};
}  // namespace nlohmann
