#include <array>
#include <ostream>
#include <schpin_lib/identificator.h>
#include <utility>

#pragma once

namespace schpin
{

using LegID = Identificator< struct LegIDTag, 8 >;

template < typename T >
class LegContainer
{
      public:
        using ItemContainer  = std::vector< T >;
        using ConstIterator  = typename ItemContainer::const_iterator;
        using const_iterator = ConstIterator;

      private:
        ItemContainer items_;

      public:
        LegContainer()
          : items_()
        {
        }

        LegContainer( const LegContainer& other )     = default;
        LegContainer( LegContainer&& other ) noexcept = default;
        LegContainer& operator=( const LegContainer& other ) = default;
        LegContainer& operator=( LegContainer&& other ) noexcept = default;

        template < typename Container,
                   typename = std::enable_if_t<
                       std::is_same_v< std::remove_cv_t< std::remove_reference_t< Container > >, ItemContainer > > >
        explicit LegContainer( Container&& items )
          : items_( std::forward< Container >( items ) )
        {
        }

        bool empty() const
        {
                return items_.empty();
        }

        std::size_t size() const
        {
                return items_.size();
        }

        const T& operator[]( const LegID& id ) const
        {
                return items_[*id];
        }
        T& operator[]( const LegID& id )
        {
                return items_[*id];
        }

        ConstIterator begin() const
        {
                return items_.begin();
        }

        ConstIterator end() const
        {
                return items_.end();
        }

        auto getIDs() const
        {
                return range( items_.size() );
        }

        ~LegContainer() = default;
};

template < typename Container >
LegContainer( Container&& items )->LegContainer< typename Container::value_type >;

}  // namespace schpin

namespace nlohmann
{
template < typename T >
struct adl_serializer< schpin::LegContainer< T > >
{
        static void to_json( nlohmann::json& j, const schpin::LegContainer< T >& cont )
        {
                j = nlohmann::json( schpin::range( cont ) );
        }

        static schpin::LegContainer< T > from_json( const nlohmann::json& j )
        {
                return schpin::LegContainer< T >( j.get< std::vector< T > >() );
        }
};
}  // namespace nlohmann

namespace schpin
{

template < unsigned N >
class DiscreteAngles
{
      public:
        static constexpr unsigned n = N;
        using Container             = std::array< uint8_t, n >;
        using ConstIterator         = typename Container::const_iterator;

      private:
        Container angles_;

      public:
        explicit DiscreteAngles( const std::array< uint8_t, n >& cont )
          : angles_( cont )
        {
        }

        DiscreteAngles()
          : angles_()
        {
        }

        const Container& operator*() const
        {
                return angles_;
        }

        template < unsigned i >
        uint8_t get() const
        {
                static_assert( i < n, "Index for DiscreteAngles has to be within it's range" );
                return angles_[i];
        }

        template < unsigned i >
        void set( uint8_t val )
        {
                static_assert( i < n, "Index for DiscreteAngles has to be within it's range" );
                angles_[i] = val;
        }

        uint8_t operator[]( std::size_t i ) const
        {
                return angles_[i];
        }

        uint8_t& operator[]( std::size_t i )
        {
                return angles_[i];
        }

        ConstIterator begin() const
        {
                return angles_.begin();
        }

        ConstIterator end() const
        {
                return angles_.end();
        }

        std::size_t size() const
        {
                return angles_.size();
        }

        const std::array< uint8_t, n >& getData() const
        {
                return angles_;
        }
};

template < unsigned N >
bool operator==( const DiscreteAngles< N >& lh, const DiscreteAngles< N >& rh )
{
        return *lh == *rh;
}

template < unsigned N >
void to_json( nlohmann::json& j, const DiscreteAngles< N >& angles )
{
        j = nlohmann::json( angles.getData() );
}

template < unsigned N >
void from_json( const nlohmann::json& j, DiscreteAngles< N >& angles )
{
        angles = DiscreteAngles< N >( j.get< typename DiscreteAngles< N >::Container >() );
}

template < unsigned N >
inline std::ostream& operator<<( std::ostream& os, const DiscreteAngles< N >& angles )
{
        os << "[";
        forEach( angles, [&]( uint8_t item ) { os << item << ","; } );
        os << "]";
        return os;
}

template < unsigned N >
inline std::array< uint8_t, N > operator-( const DiscreteAngles< N >& lh, const DiscreteAngles< N >& rh )
{
        return mapFtoA< N >( lh, rh, []( uint8_t l, uint8_t r ) -> uint8_t {
                if ( l > r ) {
                        return uint8_t( l - r );
                }
                return uint8_t( r - l );
        } );
}
}  // namespace schpin
