#include "graph/interface.h"
#include "leg/graph/base.h"
#include "leg/graph/vertex.h"
#include "leg/model.h"
#include <schpin_lib/identificator_container.h>
#include <vector>

#pragma once

namespace schpin
{

template < typename VertexContentT >
class LegGraph : IGraph< LegVertex< VertexContentT >, LegEdge >
{
      public:
        using VertexContent       = VertexContentT;
        using Vertex              = LegVertex< VertexContent >;
        using VertexContainer     = Container< Vertex, LegVertexID, std::vector >;
        using VertexIterator      = typename VertexContainer::iterator;
        using ConstVertexIterator = typename VertexContainer::const_iterator;
        using const_iterator      = ConstVertexIterator;
        using Edge                = LegEdge;
        using EdgeContent         = typename LegEdge::Content;
        using Price               = LegPrice;
        using VertexID            = LegVertexID;

      private:
        VertexContainer vertexes_{};

      public:
        explicit LegGraph()
          : vertexes_()
        {
        }

        explicit LegGraph( const std::vector< Vertex >& vertexes )
          : vertexes_( vertexes )
        {
        }

        void emplaceVertex( VertexContent content )
        {
                vertexes_.emplaceBack( vertexes_.size(), std::move( content ) );
        }

        Range< ConstVertexIterator > getVertexes() const
        {
                return range( vertexes_ );
        }

        const Vertex& getVertex( const LegVertexID& id ) const override
        {
                const Vertex& res = vertexes_.at( id );
                assert( res.getID() == id );
                return res;
        }

        void addEdge( const LegVertexID& from, LegPrice content, const LegVertexID& to )
        {
                LegEdgeID eid = vertexes_[from].addEdge( std::move( content ), to );
                vertexes_[to].addReverseEdge( eid );
        }

        const LegEdge& getEdge( const LegEdgeID& id ) const override
        {
                return getVertex( id.vid ).getEdge( id.edge_offset );
        }

        void purgeAllEdges( LegVertexID vid )
        {
                const Vertex& vertex  = getVertex( vid );
                auto          out_ids = mapF< std::vector< LegEdgeID > >( vertex.getEdges(),
                                                                 [&]( const LegEdge& edge ) { return edge.getID(); } );

                forEach( reversed( out_ids ), [this]( LegEdgeID eid ) {
                        this->removeEdge( eid );
                        // ok, bucket removal would be more efficient :)
                } );

                std::vector< LegEdgeID > ingoing_ids{vertex.getReverseEdges().begin(), vertex.getReverseEdges().end()};

                forEach( ingoing_ids, [this]( LegEdgeID eid ) { this->removeEdge( eid ); } );
        }

        void removeEdge( LegEdgeID eid )
        {
                SCHPIN_ASSERT( vertexes_.hasID( eid.vid ) );
                Vertex&    source      = vertexes_[eid.vid];
                const Edge actual_edge = source.getEdges()[*eid.edge_offset];
                const Edge last_edge   = source.getEdges().back();

                SCHPIN_ASSERT( vertexes_.hasID( last_edge.getTargetID() ) );
                Vertex& last_edge_target = vertexes_[last_edge.getTargetID()];
                last_edge_target.removeReverseEdge( last_edge.getID() );
                last_edge_target.addReverseEdge( eid );

                source.swapEdges( eid.edge_offset, last_edge.getID().edge_offset );

                vertexes_[actual_edge.getTargetID()].removeReverseEdge( eid );
                source.removeLastEdge();
        }

        std::size_t size() const
        {
                return std::size_t{vertexes_.size()};
        }

        EdgeContent& refEdgeContent( const LegEdgeID& eid )
        {
                return vertexes_[eid.vid].refEdgeContent( eid.edge_offset );
        }

        VertexContent& refContent( const LegVertexID& id ) override
        {
                return *vertexes_[id];
        }

        template < typename UnaryFunction >
        void modifyAll( UnaryFunction f )
        {
                forEach( vertexes_, [&]( Vertex& vertex ) { f( *vertex ); } );
        }
};

}  // namespace schpin

namespace nlohmann
{
template < typename VertexContent >
struct adl_serializer< schpin::LegGraph< VertexContent > >
{
        static void to_json( nlohmann::json& json, const schpin::LegGraph< VertexContent >& graph )
        {
                json = nlohmann::json{{"vertexes", graph.getVertexes()}};
        }
        static auto from_json( const nlohmann::json& j )
        {
                using Vertex = typename schpin::LegGraph< VertexContent >::Vertex;
                return schpin::LegGraph< VertexContent >{j["vertexes"].get< std::vector< Vertex > >()};
        }
};
}  // namespace nlohmann
