#include "leg/graph/vertex.h"
#include "octree/insert.h"
#include "octree/interfaces.h"
#include "octree/load.h"
#include "octree/tree.h"
#include <vector>

#pragma once

namespace schpin
{

class LegGraphMappingTreeContent;
bool operator==( const LegGraphMappingTreeContent& lh, const LegGraphMappingTreeContent& rh );

class LegGraphMappingTreeContent : public IOcContent< LegGraphMappingTreeContent, LegVertexID >
{
        // TODO(squirrel): find all typedefs and replace with using?
        // TODO(squirrel): check consisenty of all typedefs names
        std::vector< LegVertexID > ids_{};

      public:
        LegGraphMappingTreeContent() = default;

        template < typename Container >
        explicit LegGraphMappingTreeContent( const Container& cont )
          : ids_( cont.begin(), cont.end() )
        {
        }

        const std::vector< LegVertexID >& get() const
        {
                return ids_;
        }

        std::size_t getMemoryUsage( const OcNode< LegGraphMappingTreeContent >& node ) const override
        {
                return sizeof( node ) + sizeof( LegVertexID ) * ids_.capacity();
        }

        bool prunnable( const OcNode< LegGraphMappingTreeContent >& node ) const override
        {
                SCHPIN_ASSERT( node.hasChildren() );
                // maybe write as 'equal(container)' in util.h
                const OcNode< LegGraphMappingTreeContent >& first = *node.getChildren().begin();
                return allOf( tail( node.getChildren() ),
                              [&]( const OcNode< LegGraphMappingTreeContent >& child ) { return *first == *child; } );
        }
        void childBasedUpdate( const OcNode< LegGraphMappingTreeContent >& /*node*/ ) override
        {
                /*    forEach( node.getChildren(), [&]( const OcNode< LegGraphMappingTreeContent >& child ) {
                            forEach( child.getContent().get(), [&]( LegVertexID nid ) {
                                    if ( !binarySearch( ids_, nid ) ) {
                                            ids_.insert( findIf( ids_, [=]( LegVertexID subnid ) { return subnid > nid;
                   } ), nid );
                                    }
                            } );
                    } );*/
        }
        void setPrunned( const OcNode< LegGraphMappingTreeContent >& /*node*/ ) override
        {
                // content of prunned is same as one that is regulary updated based on it's children.
        }
        void update( const LegVertexID& nid ) override
        {
                // TODO(squirrel): find script to regulary check source for begin()/end() and replace that with things
                // accepting container

                auto iter = std::lower_bound( ids_.begin(), ids_.end(), nid );
                if ( iter == ids_.end() ) {
                        ids_.push_back( nid );
                } else if ( *iter != nid ) {
                        ids_.insert( ++iter, nid );
                }
        }
};

inline void to_json( nlohmann::json& j, const LegGraphMappingTreeContent& cont )
{
        j = mapF< nlohmann::json >( cont.get(), std::mem_fn( &LegVertexID::operator*) );
}

inline void from_json( const nlohmann::json& j, LegGraphMappingTreeContent& cont )
{
        cont = LegGraphMappingTreeContent{mapF< std::vector< LegVertexID > >(
            j, [&]( const nlohmann::json& sub ) { return LegVertexID( sub.get< unsigned >() ); } )};
}

inline bool operator==( const LegGraphMappingTreeContent& lh, const LegGraphMappingTreeContent& rh )
{
        return equal( lh.get(), rh.get() );
}

class LegGraphMapping
{
        OcTree< LegGraphMappingTreeContent > tree_;

      public:
        explicit LegGraphMapping( OcTree< LegGraphMappingTreeContent >&& tree )
          : tree_( std::move( tree ) )
        {
        }

        const OcTree< LegGraphMappingTreeContent >& getTree() const
        {
                return tree_;
        }
};

template < typename Graph, unsigned N >
inline LegGraphMapping createLegGraphMapping( const LegModel< N >& model, const Graph& g )
{
        // TODO(squirrel): this needs tests with SingleCollisison Insert and check for equality between single/batch
        // TODO(squirrel): this should be burned by fire

        using Vertex = typename Graph::Vertex;

        OcTree< LegGraphMappingTreeContent > tree{LegGraphMappingTreeContent{}, Length{0.01f}};

        forEach( model.getChain(), [&]( const ModelChain& chain ) {
                SCHPIN_INFO_LOG( "Leg chain item size: " << chain.model.getSurface().size() );
        } );

        using TreeVisitor =
            SingleCollisionInsert< FullCollisionChecker< TriangleCollisionQuery< 3 >, LegGraphMappingTreeContent > >;

        forEach( g.getVertexes(), [&]( const Vertex& vertex ) {
                const typename LegModel< N >::ChainContainer& chain    = model.getChain();
                const DiscreteAngles< N >&                    d_angles = vertex->angles;

                std::vector< Pose > offsets = model.chainPoses( d_angles );

                SCHPIN_ASSERT( d_angles.size() == offsets.size() );

                forEach( range( d_angles.size() ), [&]( std::size_t i ) {
                        forEach( chain[i].model.getSurface(), [&]( const Triangle< 3 >& t ) {
                                Triangle< 3 > transed = transform( t, offsets[i] );

                                TreeVisitor vis{transed, vertex.getID()};
                                tree.expand( transed );

                                tree.modify< TreeVisitor >( vis );
                        } );
                } );
        } );

        return LegGraphMapping( std::move( tree ) );
}

}  // namespace schpin

namespace nlohmann
{
template <>
struct adl_serializer< schpin::LegGraphMapping >
{
        static void to_json( nlohmann::json& j, const schpin::LegGraphMapping& mapping )
        {
                j = nlohmann::json{{"tree", mapping.getTree()}};
        }

        static schpin::LegGraphMapping from_json( const nlohmann::json& j )
        {
                using namespace schpin;
                return LegGraphMapping( j.at( "tree" ).get< OcTree< LegGraphMappingTreeContent > >() );
        }
};
}  // namespace nlohmann
