
#include "base/simplex.h"
#include "collission_interfaces.h"
#include "leg/graph.h"
#include <schpin_lib/util.h>

#pragma once

namespace schpin
{

template < unsigned K, typename Vertex >
bool covers( const Simplex< LegVertexID, K >& simplex, const Vertex& vertex )
{
        return allOf( simplex, [&]( const LegVertexID& id ) {
                return anyOf( vertex.getEdges(), [&]( const LegEdge& edge ) {  //
                        return edge.getTargetID() == id;
                } );
        } );
}

template < unsigned K, unsigned N, typename Graph >
std::vector< LegVertexID > findPromotableVertexes( const Simplex< LegVertexID, K >&  actual,
                                                   const std::vector< LegVertexID >& neighbour_ids,
                                                   const Graph&                      g )
{
        std::vector< LegVertexID > res;

        forEach( neighbour_ids, [&]( LegVertexID id ) {
                if ( !covers( actual, g.getVertex( id ) ) ) {
                        return;
                }
                if ( noneOf( actual, [&]( LegVertexID pres_id ) { return pres_id == id; } ) ) {
                        res.push_back( id );
                }
        } );
        return res;
}

template < unsigned K, unsigned N, typename Graph >
struct SimplexSearch
{
        static std::vector< Simplex< LegVertexID, N > > findSimplexes( const Simplex< LegVertexID, K >&  actual,
                                                                       const std::vector< LegVertexID >& neighbour_ids,
                                                                       const Graph&                      g )
        {
                std::vector< Simplex< LegVertexID, N > > res;
                static_assert( K < N );
                std::vector< LegVertexID > promotables =
                    findPromotableVertexes< K, N, Graph >( actual, neighbour_ids, g );

                forEach( promotables, [&]( LegVertexID id ) {
                        Simplex< LegVertexID, K + 1 > simplex{actual, id};

                        std::vector< Simplex< LegVertexID, N > > tmp =
                            SimplexSearch< K + 1, N, Graph >::findSimplexes( simplex, neighbour_ids, g );
                        res.insert( res.end(), tmp.begin(), tmp.end() );
                } );
                return res;
        }
};

template < unsigned N, typename Graph >
struct SimplexSearch< N, N, Graph >
{
        static std::vector< Simplex< LegVertexID, N > >
        findSimplexes( const Simplex< LegVertexID, N >& actual,
                       const std::vector< LegVertexID >& /*neighbour_ids*/,
                       const Graph& /*g*/ )
        {
                std::vector< Simplex< LegVertexID, N > > res;
                res.push_back( actual );
                return res;
        }
};

inline std::array< Point< 3 >, 4 > getSeparationAxesSimplexcollision( std::array< Point< 3 >, 4 >& points )
{
        return std::array< Point< 3 >, 4 >{getNormal( Triangle< 3 >( points[0], points[1], points[2] ) ),
                                           getNormal( Triangle< 3 >( points[0], points[1], points[3] ) ),
                                           getNormal( Triangle< 3 >( points[0], points[2], points[3] ) ),
                                           getNormal( Triangle< 3 >( points[1], points[2], points[3] ) )};
}
inline std::array< Point< 2 >, 3 > getSeparationAxesSimplexcollision( std::array< Point< 2 >, 3 >& points )
{
        return std::array< Point< 2 >, 3 >{
            getNormal( points[0] - points[1] ), getNormal( points[1] - points[2] ), getNormal( points[2] - points[0] )};
}

template < unsigned N >
class SimplexGraphCollisionQuery : ICollisionItem< const Point< N >*, const Point< N >* >
{

        static constexpr unsigned T_COUNT = N + 1;

        std::array< Point< N >, N + 1 >   points_;
        std::array< Point< N >, T_COUNT > intersection_axes_;

        static_assert( N == 3 || N == 2,
                       " For this to work, we have to assert that separation axis theorem works for n-dimensional "
                       "simplexes "
                       "and we know how to get a relevant axes for the algo... that should be solved later and for now "
                       "deal "
                       "with 3 dimensions only " );

      public:
        using PointIterator          = const Point< N >*;
        using SeparationAxisIterator = const Point< N >*;

        template < typename Graph >
        SimplexGraphCollisionQuery( const Graph& g, const Simplex< LegVertexID, N >& simplex )
          : points_()
          , intersection_axes_()
        {
                forEach( range( N + 1 ), [&]( std::size_t i ) {
                        DiscreteAngles< N > angles = g.getVertex( simplex[i] )->angles;

                        points_[i] = Point< N >{mapF( *angles, []( uint8_t val ) { return float( val ); } )};
                } );

                intersection_axes_ = getSeparationAxesSimplexcollision( points_ );
        }

        Range< PointIterator > getPoints() const override
        {
                return range( points_ );
        }

        Range< SeparationAxisIterator > getIntersectionAxes() const override
        {
                return range( intersection_axes_ );
        }

        Range< SeparationAxisIterator > getSeparationAxes() const override
        {
                return {nullptr, nullptr};
        }
};

template < typename Graph, unsigned N >
SimplexStorage< LegVertexID, N > autogenerateSimplexes( const Graph& g )
{
        using Vertex = typename Graph::Vertex;

        std::vector< Simplex< LegVertexID, N > > res;

        forEach( g.getVertexes(), [&]( const Vertex& vertex ) {
                auto neighbours = mapF< std::vector< LegVertexID > >(
                    vertex.getEdges(), [&]( const LegEdge& edge ) { return edge.getTargetID(); } );

                Simplex< LegVertexID, 0 >                actual{vertex.getID()};
                std::vector< Simplex< LegVertexID, N > > tmp =
                    SimplexSearch< 0, N, Graph >::findSimplexes( actual, neighbours, g );

                forEach( tmp, [&]( const Simplex< LegVertexID, N >& simplex ) {
                        bool in_collision = anyOf( res, [&]( const Simplex< LegVertexID, N >& accepted ) {
                                SimplexGraphCollisionQuery< N > acc_q( g, accepted );
                                SimplexGraphCollisionQuery< N > sim_q( g, simplex );
                                return collides< N >( acc_q, sim_q );
                        } );

                        if ( in_collision ) {
                                return;
                        }

                        res.push_back( simplex );
                } );
        } );

        std::sort( res.begin(), res.end() );
        auto iter = std::unique( res.begin(), res.end() );

        res.erase( iter, res.end() );
        // forEach( res, []( const Simplex< LegVertexID, N >& s ) { SCHPIN_INFO_LOG( s ); } );

        return SimplexStorage< LegVertexID, N >{res};
}

}  // namespace schpin
