#include "leg/model.h"

namespace schpin
{

void to_json( nlohmann::json& j, const ModelChain& chain )
{
        j = nlohmann::json{{"axis", chain.axis},
                           {"offset", chain.offset},
                           {"min", chain.min},
                           {"max", chain.max},
                           {"steps", chain.steps},
                           {"model", chain.model}};
}

void from_json( const nlohmann::json& j, ModelChain& chain )
{
        chain.axis   = j["axis"].get< Point< 3 > >();
        chain.offset = j["offset"].get< Pose >();
        chain.min    = j["min"].get< Angle >();
        chain.max    = j["max"].get< Angle >();
        chain.steps  = j["steps"].get< uint8_t >();
        chain.model  = j["model"].get< StructureModel >();
}

std::size_t getModelChainHash( const ModelChain& chain )
{
        std::stringstream ss;
        ss << chain.axis << chain.offset << chain.min << chain.max << chain.steps;

        forEach( chain.model.getSurface(), [&]( const Triangle< 3 >& t ) { ss << t; } );

        return std::hash< std::string >{}( ss.str() );
}

Either< std::vector< ModelChain >, UrdfParseError > createModelChainRec( const urdf::Joint&              joint,
                                                                         const urdf::Model&              model,
                                                                         const urdf::LinkConstSharedPtr& end_baselink )
{

        urdf::LinkConstSharedPtr link = model.getLink( joint.child_link_name );

        if ( link->name == end_baselink->name ) {
                return std::vector< ModelChain >{};
        }

        if ( !link ) {
                return {E( UrdfParseError ) << "Link " << joint.child_link_name << " is missing"};
        };

        ModelChain                chain;
        std::vector< ModelChain > res;
        chain.axis = urdfPointToSchpinPoint( joint.axis );

        if ( link->child_joints.size() != 1 ) {
                return {E( UrdfParseError )
                        << "Link " << joint.child_link_name << " has more than one child jint, not supported"};
        }

        urdf::JointConstSharedPtr subjoint = link->child_joints.front();
        assert( subjoint );

        std::optional< UrdfParseError > sub_error;

        createModelChainRec( *subjoint, model, end_baselink )
            .match( [&]( std::vector< ModelChain > new_res ) { res = new_res; },
                    [&]( UrdfParseError error ) { sub_error = std::move( error ); } );

        if ( sub_error ) {
                return *sub_error;
        }

        chain.offset = urdfPoseToSchpinPose( subjoint->parent_to_joint_origin_transform );

        if ( !joint.limits ) {
                return {E( UrdfParseError ) << "Link " << joint.child_link_name << " is mising limit element"};
        }

        assert( joint.limits );

        chain.min   = Angle{float( joint.limits->lower )};
        chain.max   = Angle{float( joint.limits->upper )};
        chain.steps = 16;  // TODO(squirrel): ??? how to get steps?

        return extractModel( *link ).convertLeft( [&]( StructureModel model ) {
                chain.model = std::move( model );
                res.push_back( chain );
                return res;
        } );
}

Either< LegModel< 3 >, UrdfParseError > createModelChain( const urdf::Model&              model,
                                                          const urdf::LinkConstSharedPtr& start_baselink,
                                                          const urdf::LinkConstSharedPtr& end_baselink )
{
        if ( start_baselink->child_joints.size() != 1 ) {
                return {E( UrdfParseError ) << "Failed to parse model chain of leg, baselink has not 1 children, "
                                               "actually: "
                                            << start_baselink->child_joints.size()};
        }

        urdf::JointConstSharedPtr joint = start_baselink->child_joints.front();

        return createModelChainRec( *joint, model, end_baselink )
            .convertLeft( [&]( std::vector< ModelChain >&& chain ) -> Either< LegModel< 3 >, UrdfParseError > {
                    if ( chain.size() != 3 ) {
                            return {E( UrdfParseError ) << "Chain " << start_baselink->name << " - "
                                                        << end_baselink->name << " item, but only 3 are supported."};
                    }

                    Pose leg_offset = urdfPoseToSchpinPose( joint->parent_to_joint_origin_transform );
                    return {LegModel< 3 >{leg_offset, {chain[2], chain[1], chain[0]}}};
            } )
            .convertRight( []( UrdfParseError error ) -> Either< LegModel< 3 >, UrdfParseError > { return {error}; } )
            .join();
}

}  // namespace schpin
