
#include "base/physics.h"
#include "base/pose.h"
#include "base/triangle.h"
#include "leg/base.h"
#include "smodel.h"
#include <urdf/model.h>
#include <urdf_model/joint.h>
#include <utility>

#pragma once

namespace schpin
{

struct ModelChain
{
        Point< 3 >     axis;
        Pose           offset;  // offset of subchain, for last it tells the pos of tip
        Angle          min{0};
        Angle          max{0};
        uint8_t        steps{};
        StructureModel model;
};

void to_json( nlohmann::json& j, const ModelChain& chain );
void from_json( const nlohmann::json& j, ModelChain& chain );

std::size_t getModelChainHash( const ModelChain& chain );

inline std::size_t getMemoryUsage( const ModelChain& chain )
{
        return sizeof( chain ) + chain.model.getSurface().capacity() * sizeof( Triangle< 3 > );
}

constexpr unsigned clappedIndex( unsigned i, unsigned i_max )
{
        return i < i_max ? i : i_max;
}

Either< std::vector< ModelChain >, UrdfParseError > createModelChainRec( const urdf::Joint&              joint,
                                                                         const urdf::Model&              model,
                                                                         const urdf::LinkConstSharedPtr& end_baselink );

inline Angle getRealAngle( const ModelChain& chain_link, uint8_t discretized_angle )
{
        return chain_link.min +
               float( discretized_angle ) * ( chain_link.max - chain_link.min ) / float( chain_link.steps - 1 );
}

inline Point< 3 > transformByAngle( const Point< 3 >& point, const Pose& offset, const Point< 3 >& axis, Angle a )
{
        return rotate( transform( point, offset ), Quaternion( axis, a ) );
}

template < unsigned N >
class LegModel
{
        static constexpr unsigned n = N;

      public:
        using ChainContainer = std::array< ModelChain, N >;

      private:
        Pose           origin_offset_;
        ChainContainer chain_;
        Color          color_;

      public:
        explicit LegModel( Pose origin_offset, std::array< ModelChain, n > chain )
          : origin_offset_( std::move( origin_offset ) )
          , chain_( std::move( chain ) )
        {
                auto iter = findIf( chain_, [&]( const ModelChain& chain ) {
                        return chain.model.getColor() != Color{0.f, 0.f, 0.f};
                } );

                if ( iter != chain_.end() ) {
                        color_ = iter->model.getColor();
                }
        }

        const Color& getColor() const
        {
                return chain_.front().model.getColor();
        }

        const Pose& getOriginOffset() const
        {
                return origin_offset_;
        }

        const ChainContainer& getChain() const
        {
                return chain_;
        }

        Point< 3 > tipPosition( const DiscreteAngles< N >& angles ) const
        {
                return tipPositionRec< 0 >( angles );
        }

        Point< 3 > tipPosition( const std::array< Angle, N >& angles ) const
        {
                return tipPositionRec< 0 >( angles );
        }

        Point< 3 > neutralTipPosition() const
        {
                std::array< Angle, N > angles;
                return tipPosition( angles );
        }

        std::vector< Pose > chainPoses( const DiscreteAngles< N >& angles ) const
        {
                std::vector< Pose > res = chainPoses< 0 >( Pose{}, angles );
                return reversed( res ).as< std::vector< Pose > >();
        }

        std::size_t getHash() const
        {
                return accumulate( mapF< std::vector< std::size_t > >( chain_, &getModelChainHash ), std::size_t( 0 ) );
        }


      private:
        template < unsigned i >
        Point< 3 > tipPositionRec( const std::array< Angle, N >& angles ) const
        {
                if ( i == N ) {
                        return Point< 3 >( 0, 0, 0 );
                }

                return transformByAngle(
                    tipPositionRec< clappedIndex( i + 1, N ) >( angles ), chain_[i].offset, chain_[i].axis, angles[i] );
        }

        template < unsigned i >
        Point< 3 > tipPositionRec( const DiscreteAngles< N >& angles ) const
        {
                if ( i == N ) {
                        return Point< 3 >( 0, 0, 0 );
                }
                Angle a = getRealAngle( chain_[i], angles[clappedIndex( i, N - 1 )] );
                return transformByAngle(
                    tipPositionRec< clappedIndex( i + 1, N ) >( angles ), chain_[i].offset, chain_[i].axis, a );
        }

        template < unsigned i >
        std::vector< Pose > chainPoses( Pose pose, const DiscreteAngles< N >& angles ) const
        {
                if ( i == n ) {
                        return std::vector< Pose >{};
                }
                Angle a = getRealAngle( chain_[i], angles[clappedIndex( i, n - 1 )] );

                std::vector< Pose > res = chainPoses< clappedIndex( i + 1, n ) >(
                    pose + rotate( chain_[i].offset, Quaternion( chain_[i].axis, a ) ), angles );
                res.push_back( pose );
                return res;
        }
};
Either< LegModel< 3 >, UrdfParseError > createModelChain( const urdf::Model&              model,
                                                          const urdf::LinkConstSharedPtr& start_baselink,
                                                          const urdf::LinkConstSharedPtr& end_baselink );

template < unsigned N >
inline std::size_t getMemoryUsage( const LegModel< N >& model )
{
        return accumulate(
            model.getChain(), sizeof( model ), []( const ModelChain& chain ) { return getMemoryUsage( chain ); } );
}

}  // namespace schpin

namespace nlohmann
{
template < unsigned N >
struct adl_serializer< schpin::LegModel< N > >
{
        static void to_json( nlohmann::json& j, const schpin::LegModel< N >& model )
        {
                j = nlohmann::json{{"chain", model.getChain()}, {"origin_offset", model.getOriginOffset()}};
        }

        static schpin::LegModel< N > from_json( const nlohmann::json& j )
        {
                return schpin::LegModel< N >{j["origin_offset"].get< schpin::Pose >(),
                                             j["chain"].get< std::array< schpin::ModelChain, N > >()};
        }
};
}  // namespace nlohmann
