#include "base/point.h"
#include "environment.h"
#include "graph/astar.h"
#include "graph/util/dot.h"
#include "leg.h"
#include "leg/view.h"
#include "robot.h"
#include "robot/state_view.h"
#include <vector>

#pragma once

namespace schpin
{

class LegSidepathSegment
{
      public:
        using PointContainer = std::vector< Point< 3 > >;
        using ConstIterator  = typename PointContainer::const_iterator;
        using const_iterator = ConstIterator;

      private:
        PointContainer points_;

      public:
        Range< ConstIterator > getPoints() const
        {
                return range( points_ );
        }

      private:
        void pushPoint( Point< 3 > p )
        {
                points_.emplace_back( std::move( p ) );
        }

        template < unsigned N, typename GaitChain >
        friend std::unordered_map< BodyVertexID, LegContainer< LegSidepathSegment > >
        createSidepaths( const GaitChain& gchain, const RobotStateView& robot, const Environment& env );
};

using SidepathLayerVertexID   = Identificator< struct SidepathLayerVertexIDTag, 32 >;
using SidepathLayerEdgeOffset = Identificator< struct SidepathLayerEdgeOffsetTag, 32 >;
struct SidepathLayerEdgeID
{
        SidepathLayerVertexID   vertex_id;
        SidepathLayerEdgeOffset offset;
};

bool operator==( const SidepathLayerEdgeID& lh, const SidepathLayerEdgeID& rh )
{
        return lh.vertex_id == rh.vertex_id && lh.offset == rh.offset;
}

struct SidepathLayerEdgeContent
{
        Distance price;

      public:
        Distance getPrice() const
        {
                return price;
        }
};

using SidepathLayerEdge = GenericEdge< SidepathLayerEdgeID,
                                       SidepathLayerEdgeContent,
                                       Distance,
                                       SidepathLayerVertexID,
                                       SidepathLayerEdgeOffset >;

template < typename SearchContent >
struct SidepathLayerVertexContent
{
        BodyVertexID                    bid{};
        Pose                            body_pose{};
        SearchContent                   search_content;
        std::shared_ptr< EnvProxyNode > env_node{};
        Point< 3 >                      position{};
        Distance                        price{};
        bool                            collide_check = false;

        SidepathLayerVertexContent( std::shared_ptr< EnvProxyNode > env_node,
                                    Point< 3 >                      position,
                                    Distance                        price,
                                    BodyVertexID                    bid,
                                    Pose                            pose )
          : bid( bid )
          , body_pose( pose )
          , env_node( env_node )
          , position( position )
          , price( price )
        {
        }

        SidepathLayerVertexContent( Point< 3 > position = Point< 3 >{} )
          : position( position )
          , collide_check( true )
        {
        }
};

template < typename SearchContent >
using SidepathLayerVertex =
    GenericVertex< SidepathLayerVertexID, SidepathLayerVertexContent< SearchContent >, SidepathLayerEdge >;

// TODO: make generic graph :)
template < typename SearchContent >
class SidepathLayerGraph : IGraph< SidepathLayerVertex< SearchContent >, SidepathLayerEdge >
{
      public:
        using Price               = Distance;
        using Vertex              = SidepathLayerVertex< SearchContent >;
        using Edge                = SidepathLayerEdge;
        using EdgeContent         = SidepathLayerEdgeContent;
        using EdgeID              = SidepathLayerEdgeID;
        using VertexID            = SidepathLayerVertexID;
        using VertexContent       = SidepathLayerVertexContent< SearchContent >;
        using VertexContainer     = Container< Vertex, VertexID, std::vector >;
        using ConstVertexIterator = typename VertexContainer::const_iterator;

      private:
        VertexContainer vertexes_;

      public:
        SidepathLayerVertexID size() const
        {
                return vertexes_.size();
        }

        VertexID emplaceVertex( VertexContent vcontent )
        {
                vertexes_.emplace_back( VertexID{vertexes_.size()}, std::move( vcontent ) );
                return vertexes_.back().getID();
        }
        Range< ConstVertexIterator > getVertexes() const
        {
                return range( vertexes_ );
        }

        const Vertex& getVertex( const VertexID& id ) const override
        {
                return vertexes_[id];
        }

        const Edge& getEdge( const EdgeID& eid ) const override
        {
                return getVertex( eid.vertex_id ).getEdge( eid.offset );
        }

        VertexContent& refContent( const VertexID& id ) override
        {
                return *vertexes_[id];
        }

        auto& refEdgeContent( const EdgeID& eid )
        {
                return vertexes_[eid.vertex_id].refEdgeContent( eid.offset );
        }

        void removeVertex( VertexID vid )
        {
                auto& vertex = vertexes_[vid];
                forEach( reversed( vertex.getEdges() ), [&]( const Edge& e ) { removeEdge( e.getID() ); } );

                forEach( reversed( vertex.getReverseEdges() ), [&]( EdgeID eid ) { removeEdge( eid ); } );

                vertexes_.erase( vertexes_.begin() + *vid );
        }

        void removeEdge( EdgeID eid )
        {
                // TODO copypasta from leg graph :)
                Vertex&    source      = vertexes_[eid.vertex_id];
                const Edge actual_edge = source.getEdges()[*eid.offset];
                const Edge last_edge   = source.getEdges().back();

                Vertex& last_edge_target = vertexes_[last_edge.getTargetID()];
                last_edge_target.removeReverseEdge( last_edge.getID() );
                last_edge_target.addReverseEdge( eid );

                source.swapEdges( eid.offset, last_edge.getID().offset );

                vertexes_[actual_edge.getTargetID()].removeReverseEdge( eid );
                source.removeLastEdge();
        }

        void addEdge( VertexID from, VertexID to, EdgeContent content )
        {
                auto eid = vertexes_[from].addEdge( std::move( content ), to );
                vertexes_[to].addReverseEdge( eid );
        }

        template < typename UnaryFunction >
        void modifyAll( UnaryFunction f )
        {
                forEach( vertexes_, [&]( Vertex& v ) { f( *v ); } );
        }
};

template < unsigned N, typename CollectFunction >
void collectEligibleSurface( const Pose&        body_pose,
                             const Leg< N >&    leg,
                             const Environment& env,
                             CollectFunction    collect_f )
{
        OcNodeCollisionQuery< CollisionMode::FAST, EnvironmentOcTreeContent > env_query( env.getTree().getProxyRoot() );
        OcNodeCollisionQuery< CollisionMode::FAST, OccupancyContent >         leg_query(
            leg.getWorkarea().getProxyRoot(), transform( leg.getOriginOffset(), body_pose ) );
        // ^^ this query TODO should be made by leg_view, transposed....

        findCollisionNodes< 3, CollisionMode::FAST >(
            env_query,
            leg_query,
            [&]( const auto& env_query, const auto& leg_query ) -> bool {  // collision function
                    return ( **env_query )->isSurface() &&
                           ( **leg_query )->getOccupancy() > Occupancy( 0.5 );  // TODO: this is same in body planner :/

                    // TODO: I could write default collision function I think
            },
            [&]( const auto& env_query, const auto& ) {  // collect function of colliding leafs
                    collect_f( *env_query );
            } );
}

// detects that there exists veretx (filtered by 'filter' function) that does not collide with the environment
template < typename LView, typename UnaryFunction >
bool legCollisionlessConfigExists( const LView& leg_view, const Environment& env, UnaryFunction filter )
{
        return anyOf( leg_view->getDstar().getVertexes(), [&]( const auto& leg_vertex ) {
                if ( !filter( leg_vertex ) ) {
                        return false;
                }

                // TODO: maybe env class could return the collision object? so it does not have to be fucking created
                // each time? hint: grep this code in the codebase
                OcNodeCollisionQuery< CollisionMode::FAST, EnvironmentOcTreeContent > env_query(
                    env.getTree().getProxyRoot() );

                bool collission_free =
                    allOf( leg_view.getCollisionQueries( leg_vertex->angles ),
                           [&]( FastStructureModelCollisionQuery leg_query ) {
                                   // TODO: make the fucking sphere type
                                   return !treesNodesCollides< 3 >(
                                       env_query, leg_query, [&]( const auto& env_query, const auto& ) {
                                               return ( **env_query )->getOccupancy() > ENV_MAX_OCCUPANCY / 2;
                                       } );
                           } );

                return collission_free;
        } );
}

struct SidePathLayerAStarConfig
{
        using QueueItem     = AStarQueueItem< SidepathLayerVertexID, Distance >;
        using Heap          = boost::heap::pairing_heap< QueueItem, boost::heap::compare< std::greater< QueueItem > > >;
        using HeapHandle    = typename Heap::handle_type;
        using SearchContent = AStarContent< HeapHandle, Distance, SidepathLayerEdgeID >;
        using Graph         = SidepathLayerGraph< SearchContent >;
        using HeuristicFunction = ZeroGraphHeuristics< Distance >;
};

using SidepathLayerSearchOverlay = AStarOverlay< SidePathLayerAStarConfig >;

template < unsigned N, typename GaitChain >
std::unordered_map< BodyVertexID, LegContainer< LegSidepathSegment > > createSidepaths( const GaitChain&      gchain,
                                                                                        const RobotStateView& robot,
                                                                                        const Environment&    env )
{
        std::unordered_map< BodyVertexID, LegContainer< LegSidepathSegment > > res;

        forEach( gchain, [&]( const auto& segment_view ) {
                res[segment_view.source_pin.bid] =
                    LegContainer{std::vector< LegSidepathSegment >{robot.getLegs().size()}};
        } );

        forEach( robot.getLegs(), [&]( const LegStateView< N >& leg_state ) {
                using Graph         = typename SidePathLayerAStarConfig::Graph;
                using VertexContent = typename Graph::VertexContent;
                using VertexID      = typename Graph::VertexID;
                SCHPIN_INFO_LOG( "Leg: " << leg_state->getID() );
                SCHPIN_INFO_LOG( leg_state->getOriginOffset() );

                // #complexSetup
                auto relative_tip_to_absolute = [&]( const Point< 3 >& p, const Pose& body_pose ) {
                        return transform( transform( p, leg_state->getOriginOffset() ), body_pose );
                };

                Graph      layer_graph;
                Point< 3 > neutral_pos = leg_state->getModel().neutralTipPosition();
                VertexID   last_eid    = layer_graph.emplaceVertex( VertexContent{} );
                VertexID   first_eid =
                    layer_graph.emplaceVertex( VertexContent( robot.getTipPosition( leg_state->getID() ) ) );

                std::vector< VertexID > prev_layer = {first_eid};
                std::vector< VertexID > actual_layer;

                // function to connect all vertexes from layer to 'actual_id' vertex
                auto connect_layer_f =
                    [&]( const VertexID actual_id, const std::vector< VertexID >& layer, Distance edge ) {
                            const Point< 3 >& actual_pos = layer_graph.getVertex( actual_id )->position;
                            forEach( layer, [&]( VertexID prev_id ) {
                                    Distance dist = layer_graph.getVertex( prev_id )->position.distance( actual_pos );
                                    if ( dist >= edge * 1.5f ) {
                                            return;
                                    }
                                    Distance price = dist + layer_graph.getVertex( actual_id )->price;
                                    layer_graph.addEdge( prev_id, actual_id, SidepathLayerEdgeContent{price} );
                            } );
                    };

                // generate all layers and interconnect them
                forEach( gchain, [&]( const auto& segment_view ) {
                        forEach( segment_view.link.body_segment.getPoses(), [&]( const Pose& body_pose ) {
                                Point< 3 > ideal_tip_pos = relative_tip_to_absolute( neutral_pos, body_pose );
                                collectEligibleSurface( body_pose, *leg_state, env, [&]( const EnvProxyNode& node ) {
                                        Distance price_bias = node.getPosition().distance( ideal_tip_pos );
                                        VertexID actual_id  = layer_graph.emplaceVertex(
                                            VertexContent( std::make_shared< EnvProxyNode >( node ),
                                                           node.getPosition(),
                                                           price_bias,
                                                           segment_view.source_pin.bid,
                                                           body_pose ) );

                                        connect_layer_f( actual_id, prev_layer, node.getEdgeLength() );
                                        actual_layer.push_back( actual_id );
                                } );

                                prev_layer = actual_layer;
                                actual_layer.clear();
                        } );
                } );

                // connect last layer to final vertex
                forEach( prev_layer, [&]( VertexID vid ) {
                        layer_graph.addEdge( vid, last_eid, SidepathLayerEdgeContent{Distance( 0. )} );
                } );

                SCHPIN_INFO_LOG( "Sidepath graph generated" );
                // find shortes path
                SidepathLayerSearchOverlay search_overlay{std::move( layer_graph )};

                search_overlay.setTask( first_eid, last_eid );
                search_overlay.computeShortestPath( std::numeric_limits< std::size_t >::max() );

                auto sidepath = search_overlay.getPath();

                while ( !sidepath.empty() ) {

                        auto iter = findIf( sidepath.getPins(), [&]( const auto& vertex_id ) {
                                auto& layer_vertex = search_overlay->getVertex( vertex_id );
                                if ( layer_vertex->collide_check ) {

                                        return false;
                                }

                                LegView l_view{*leg_state, layer_vertex->body_pose};

                                bool has_collisionless_config =
                                    legCollisionlessConfigExists( l_view, env, [&]( const auto& vertex ) {
                                            return contains( *layer_vertex->env_node,
                                                             l_view.tipPosition( vertex->angles ) );
                                    } );

                                if ( has_collisionless_config ) {
                                        search_overlay.refContent( vertex_id ).collide_check = true;
                                        return false;
                                }

                                return true;
                        } );

                        if ( iter == sidepath.getPins().end() ) {
                                SCHPIN_INFO_LOG( "Sidepath is collision free" );
                                break;
                        } else {
                                SCHPIN_INFO_LOG(
                                    "Sidepath has to be recalculated, graph size: " << search_overlay->size() );
                                search_overlay.removeVertex( *iter );
                                sidepath = search_overlay.getPath();
                        }
                }

                // insert data of sidepath into result datastructure
                forEach( sidepath, [&]( const auto& sidepath_view ) {
                        const auto& layer_vertex = search_overlay->getVertex( sidepath_view.source_pin );

                        if ( layer_vertex.getID() == first_eid || layer_vertex.getID() == last_eid ) {
                                return;
                        }

                        res[layer_vertex->bid][leg_state->getID()].pushPoint( layer_vertex->position );
                } );
        } );
        return res;
}

template < typename Container >
inline visualization_msgs::MarkerArray
createSidepathVisualization( const Container& points, std::string frame, std::string ns, Color color )
{
        auto msg = createMarker( ns, 0, frame, 1. );

        msg.type = visualization_msgs::Marker::LINE_STRIP;

        msg.color.r = color.r;
        msg.color.g = color.g;
        msg.color.b = color.b;

        msg.scale.x = 0.004;

        msg.points = mapFtoV( points, pointToMsg );

        visualization_msgs::MarkerArray res;
        res.markers.push_back( msg );
        return res;
}
}  // namespace schpin
