#include "graph/dstar.h"
#include "graph/dstar/storage.h"
#include "leg/model.h"
#include "leg/workarea.h"
#include "octree/occupancy.h"
#include "octree/tree.h"
#include <boost/heap/pairing_heap.hpp>
#include <filesystem>
#include <utility>

#pragma once

namespace schpin
{

template < typename Vertex >
class LegHeuristics
{
      public:
        LegPrice operator()( const Vertex& /*from*/, const Vertex& /*to*/ )
        {
                return LegPrice( 0.f );
        }
};

template < unsigned N >
struct LegDStarConfig
{
        using QueueItem         = DstarQueueItem< LegVertexID, LegPrice >;
        using Heap              = boost::heap::pairing_heap< QueueItem >;
        using HeapHandle        = typename Heap::handle_type;
        using SearchContent     = DStarContent< HeapHandle, LegPrice >;
        using Graph             = LegGraph< LegVertexContent< SearchContent, N > >;
        using HeuristicFunction = LegHeuristics< typename Graph::Vertex >;
};

template < unsigned N >
using LegDstar = AnytimeDStar< LegDStarConfig< N > >;

template < unsigned N >
struct LegData
{

        // TODO(squirrel): I have bad feeling about consistency here...
        std::string                      name;
        LegModel< N >                    model;
        LegDstar< N >                    dstar;
        SimplexStorage< LegVertexID, 3 > graph_space;
        OcTree< OccupancyContent >       workarea;
        LegGraphMapping                  graph_mapping;
};

template < unsigned N >
LegData< N > generateLegData( const std::string& name, LegModel< N >&& model )
{
        SCHPIN_INFO_LOG( "Generating dataset for: " << name );
        using Graph             = typename LegDstar< N >::Graph;
        using HeuristicFunction = typename LegDstar< N >::HeuristicFunction;

        LegDstar< N > dstar{generateLegGraph< Graph >( model ), HeuristicFunction()};
        SCHPIN_INFO_LOG( "LegGraph created, size: " << dstar.getGraph().size() );

        SimplexStorage< LegVertexID, N > graph_space{};  // = autogenerateSimplexes< Graph, N >( dstar.getGraph() );
        SCHPIN_INFO_LOG( "Simplexes for graph created: " << graph_space.size() );

        auto angle_conversion = [&]( const DiscreteAngles< N >& dangles, std::size_t max_step ) -> Point< 3 > {
                auto angles = mapFtoA< N >( range( N ), [&]( std::size_t i ) {
                        float alpha = float( dangles[i] ) / float( max_step );

                        return model.getChain()[i].min + alpha * ( model.getChain()[i].max - model.getChain()[i].min );
                } );
                return model.tipPosition( angles );
        };

        /*
        OcTree< OccupancyContent > workarea =
            createWorkarea( dstar.getGraph(), graph_space, angle_conversion, Length( 0.01f ) );*/
        OcTree< OccupancyContent > workarea = createWorkarea< N >( angle_conversion, Length( 0.04f ) );
        SCHPIN_INFO_LOG( "Workarea created: " << getNodeCount( workarea.getBareRoot() ) );

        LegGraphMapping graph_mapping = createLegGraphMapping( model, dstar.getGraph() );
        SCHPIN_INFO_LOG( "Graph mapping created: " << getNodeCount( graph_mapping.getTree().getBareRoot() ) );

        return LegData< N >{name,
                            model,
                            std::move( dstar ),
                            std::move( graph_space ),
                            std::move( workarea ),
                            std::move( graph_mapping )};
}

using LegBootError = Error< struct LegBootErrorTag >;

template < unsigned N >
Either< LegData< N >, LegBootError > loadLegData( const std::filesystem::path& prefix,
                                                  const std::string&           name,
                                                  LegModel< N >&& /*model*/ )
{
        SCHPIN_INFO_LOG( "Loading generated data for: " << name );
        std::ifstream in_stream{prefix / "leg.json"};
        return safeLoadJson< LegData< N > >( in_stream ).convertRight( [&]( JsonStorageError e ) {
                return E( LegBootError ).attach( e ) << "Failed to load leg " << name << " data";
        } );
}

template < unsigned N >
OptError< LegBootError > saveLegData( const std::filesystem::path& prefix, const LegData< N >& data )
{
        SCHPIN_INFO_LOG( "Saving generated data for: " << data.name );
        std::ofstream out_stream{prefix / "leg.json"};
        return safeStoreJson( out_stream, data ).convert( [&]( JsonStorageError e ) {
                return E( LegBootError ).attach( e ) << "Failed to save leg data to json error";
        } );
}

template < unsigned N >
Either< LegData< N >, LegBootError > loadOrGenerateLegData( const std::filesystem::path&    cache_prefix,
                                                            const urdf::Model&              model,
                                                            const urdf::LinkConstSharedPtr& start_baselink,
                                                            const urdf::LinkConstSharedPtr& end_baselink )
{
        // TODO(squirrel): hash is not saved
        std::filesystem::path prefix{cache_prefix / start_baselink->name};

        std::error_code err_code;
        if ( !std::filesystem::exists( prefix ) && !create_directories( prefix, err_code ) ) {
                return {E( LegBootError ) << "Cache folder " << prefix << " does not exists, and failed to create."
                                          << err_code.message()};
        }

        return createModelChain( model, start_baselink, end_baselink )
            .convertRight( [&]( UrdfParseError&& error ) {
                    return E( LegBootError ).attach( error ) << "Failed to load leg data due to URDF errors";
            } )
            .transformLeft( [&]( LegModel< N >&& model ) -> Either< LegData< N >, LegBootError > {
                    std::ifstream hash_file{prefix / "hash"};
                    if ( hash_file ) {
                            std::string stored_hash;
                            hash_file >> stored_hash;

                            if ( stored_hash == std::to_string( model.getHash() ) ) {
                                    return {loadLegData( prefix, start_baselink->name, std::move( model ) )};
                            }
                    }

                    std::ofstream hash_out{prefix / "hash"};
                    hash_out << model.getHash();

                    LegData< N >             l_data = generateLegData( start_baselink->name, std::move( model ) );
                    OptError< LegBootError > err    = saveLegData( prefix, l_data );

                    return err || std::move( l_data );
            } );
}

}  // namespace schpin

namespace nlohmann
{
template < unsigned N >
struct adl_serializer< schpin::LegData< N > >
{
        static void to_json( nlohmann::json& j, const schpin::LegData< N >& data )
        {
                j = nlohmann::json{{"name", data.name},
                                   {"model", data.model},
                                   {"dstar", data.dstar},
                                   {"simplexes", data.graph_space},
                                   {"workarea", data.workarea},
                                   {"mapping", data.graph_mapping}};
        }

        static schpin::LegData< N > from_json( const nlohmann::json& j )
        {
                using namespace schpin;
                return LegData< N >{j["name"].get< std::string >(),
                                    j["model"].get< LegModel< N > >(),
                                    j["dstar"].get< LegDstar< N > >(),
                                    j["simplexes"].get< SimplexStorage< LegVertexID, N > >(),
                                    j["workarea"].get< OcTree< OccupancyContent > >(),
                                    j["mapping"].get< LegGraphMapping >()};
        }
};
}  // namespace nlohmann
