
#include "collission.h"
#include "octree/interfaces.h"
#include "octree/occupancy.h"
#include "octree/proxy_node.h"
#include "octree/tree.h"
#include "parser.h"

#pragma once

namespace schpin
{

template < typename T, typename Content >
class FullCollisionChecker : public IOcCollisionCheck< Content >
{
        using QueryPointIterator          = typename T::PointIterator;
        using QuerySeparationAxisIterator = typename T::SeparationAxisIterator;
        static_assert( std::is_base_of< ICollisionItem< QueryPointIterator, QuerySeparationAxisIterator >, T >::value );

        using Token = typename Content::Token;

        const T     item_query_;
        const Token token_;

      public:
        template < typename ItemType >
        FullCollisionChecker( const ItemType& item, const Token& token )
          : item_query_( item )
          , token_( token )
        {
        }
        bool innerCollission( const OcProxyNode< Content >& /*node*/,
                              const OcNodeFullCollisionQuery< Content >& node_query ) const override
        {
                bool res = collides< 3 >( item_query_, node_query );
                return res;
        }
        void bottomCollission( const OcProxyNode< Content >& /*node*/,
                               const OcNodeFullCollisionQuery< Content >& node_query,
                               Content&                               content ) const override
        {
                if ( collides< 3 >( item_query_, node_query ) ) {
                        content.update( token_ );
                }
        }
};

template < typename T, typename TFast, typename Content >
class FastCollisionChecker : public IOcCollisionCheck< Content >
{
        const T     item_query_;
        const TFast item_fast_query_;

        static_assert( std::is_same< typename T::ItemType, typename TFast::ItemType >::value );

        using Token = typename Content::Token;

        const Token token_;

      public:
        template < typename ItemType >
        FastCollisionChecker( const ItemType& item, const Token& token )
          : item_query_( item )
          , item_fast_query_( item )
          , token_( token )
        {
        }

        bool innerCollission( const OcProxyNode< Content >&          node,
                              const OcNodeFullCollisionQuery< Content >& node_query ) const override
        {
                OcNodeFastCollisionQuery fast_q{node};
                if ( !fastCollides< 3 >( item_fast_query_, fast_q ) ) {
                        return false;
                }
                return collides< 3 >( item_query_, node_query );
        }

        void bottomCollission( const OcProxyNode< Content >& /*node*/,
                               const OcNodeFullCollisionQuery< Content >& node_query,
                               Content&                               content ) const override
        {
                if ( collides< 3 >( item_query_, node_query ) ) {
                        content.update( token_ );
                }
        }
};

template < typename Checker >
class SingleCollisionInsert : public IOcSimpleVisitor< typename Checker::NodeContent >
{
      public:
        using NodeContent = typename Checker::NodeContent;
        using Token       = typename NodeContent::Token;

      private:
        static_assert( std::is_base_of< IOcCollisionCheck< NodeContent >, Checker >::value );
        Checker checker_;

      public:
        template < typename Item >
        SingleCollisionInsert( const Item& item, const Token& token )
          : checker_( item, token )
        {
        }
        OcNodeControl accessNode( const OcProxyNode< NodeContent >& node ) const override
        {
                OcNodeFullCollisionQuery< NodeContent > node_query( node );
                if ( checker_.innerCollission( node, node_query ) ) {
                        return OcNodeControl::FOLLOW_CHILDREN;
                }
                return OcNodeControl::END;
        }
        OcLockedLeafControl accessLockedLeaf( const OcProxyNode< NodeContent >& node ) const override
        {
                OcNodeFullCollisionQuery< NodeContent > node_query( node );
                if ( checker_.innerCollission( node, node_query ) ) {
                        return OcLockedLeafControl::EXPAND_CHILDREN;
                }
                return OcLockedLeafControl::END;
        }

        void accessBottomLeaf( const OcProxyNode< NodeContent >& node, NodeContent& content ) const override
        {
                OcNodeFullCollisionQuery< NodeContent > node_query( node );
                checker_.bottomCollission( node, node_query, content );
        }
};

using FullSingleTriangleInsert =
    SingleCollisionInsert< FullCollisionChecker< TriangleCollisionQuery< 3 >, OccupancyContent > >;
using FastSingleTriangleInsert = SingleCollisionInsert<
    FastCollisionChecker< TriangleCollisionQuery< 3 >, TriangleFastCollisionQuery, OccupancyContent > >;

/// Drop "Full" from any insert query that uses full collision, that is not 'full' that is standart. TODO

template < typename Checker, unsigned BATCH_SIZE >
class BatchCollisionInsert : public IOcVisitor< std::bitset< BATCH_SIZE >, typename Checker::NodeContent >
{
        using NodeContent = typename Checker::NodeContent;
        static_assert( std::is_base_of< IOcCollisionCheck< NodeContent >, Checker >::value );
        using InfoType = std::bitset< BATCH_SIZE >;
        using Token    = typename NodeContent::Token;

        std::vector< Checker > checkers_{};
        InfoType               basic_selection_;

      public:
        // TODO(squirrel): maybe convert to Range?
        template < typename Iterator >
        BatchCollisionInsert( Iterator begin, Iterator end, Token token )
          : checkers_()
          , basic_selection_( 0 )
        {
                using ItemType = typename Iterator::value_type;
                if ( std::distance( begin, end ) > BATCH_SIZE ) {
                        end = begin + BATCH_SIZE;
                }
                std::for_each( begin, end, [&]( const ItemType& item ) { checkers_.emplace_back( item, token ); } );
                forRange( checkers_.size(), [&]( std::size_t i ) { basic_selection_[i] = 1; } );
        }

        std::bitset< BATCH_SIZE > getBasicSelection() const
        {
                return basic_selection_;
        }

        std::pair< OcNodeControl, InfoType > accessNode( const OcProxyNode< NodeContent >& node,
                                                         const InfoType&                   parent_info ) const override
        {
                InfoType                            info = parent_info;
                OcNodeFullCollisionQuery< NodeContent > node_query( node );
                forRange( BATCH_SIZE, [&]( std::size_t i ) {
                        if ( !info[i] ) {
                                return;
                        }
                        if ( !checkers_[i].innerCollission( node, node_query ) ) {
                                info[i] = false;
                        }
                } );
                if ( info.any() ) {
                        return {OcNodeControl::FOLLOW_CHILDREN, info};
                }

                return {OcNodeControl::END, info};
        }

        std::pair< OcLockedLeafControl, InfoType > accessLockedLeaf( const OcProxyNode< NodeContent >& node,
                                                                     const InfoType& parent_info ) const override
        {
                std::pair< OcNodeControl, InfoType > res = accessNode( node, parent_info );
                switch ( res.first ) {
                        case OcNodeControl::FOLLOW_CHILDREN:
                                return {OcLockedLeafControl::EXPAND_CHILDREN, res.second};
                        case OcNodeControl::END:
                                return {OcLockedLeafControl::END, res.second};
                }
                return {OcLockedLeafControl::END, res.second};
        }

        void accessBottomLeaf( const OcProxyNode< NodeContent >& node,
                               const InfoType&                   info,
                               NodeContent&                      content ) const override
        {
                OcNodeFullCollisionQuery< NodeContent > node_query( node );

                forRange( BATCH_SIZE, [&]( std::size_t i ) {
                        if ( info[i] ) {
                                checkers_[i].bottomCollission( node, node_query, content );
                        }
                } );
        }
};
template < unsigned BATCH_SIZE >
using FullBatchTriangleInsert =
    BatchCollisionInsert< FullCollisionChecker< TriangleCollisionQuery< 3 >, OccupancyContent >, BATCH_SIZE >;
template < unsigned BATCH_SIZE >
using FastBatchTriangleInsert = BatchCollisionInsert<
    FastCollisionChecker< TriangleCollisionQuery< 3 >, TriangleFastCollisionQuery, OccupancyContent >,
    BATCH_SIZE >;
}  // namespace schpin
