#include "base/json.h"
#include "octree/tree.h"
#include "parser.h"
#include "schpin_lib/static_vector.h"
#include <nlohmann/json.hpp>

#pragma once

namespace schpin
{

template < typename InsertionVisitor >
OcTree< typename InsertionVisitor::NodeContent >
buildOcTreeIterative( const std::vector< Triangle< 3 > >&          triangles,
                      const typename InsertionVisitor::NodeContent content,
                      const Length                                 min_resolution,
                      const typename InsertionVisitor::Token       token )
{
        using NodeContent = typename InsertionVisitor::NodeContent;
        OcTree< NodeContent > res{content, min_resolution};

        forEach( triangles, [&]( const Triangle< 3 >& triangle ) {
                forEach( triangle, [&]( const Point< 3 >& p ) { res.expand( p ); } );

                InsertionVisitor visitor( triangle, token );

                res.modify( visitor );
        } );

        return res;
}

}  // namespace schpin

namespace nlohmann
{

template < typename Content >
struct adl_serializer< schpin::OcNode< Content > >
{
        static schpin::OcNode< Content > from_json( const nlohmann::json& j )
        {

                using Array = std::array< schpin::OcNode< Content >, 8 >;

                nlohmann::json j_children = j.at( "children" );

                std::unique_ptr< Array > ptr;

                if ( !j_children.empty() ) {
                        auto vec = j.at( "children" ).get< std::vector< schpin::OcNode< Content > > >();
                        // TODO(squirrel): exception should be raised in case vec.size() != 8
                        ptr = std::make_unique< Array >( Array{std::move( vec[0] ),
                                                               std::move( vec[1] ),
                                                               std::move( vec[2] ),
                                                               std::move( vec[3] ),
                                                               std::move( vec[4] ),
                                                               std::move( vec[5] ),
                                                               std::move( vec[6] ),
                                                               std::move( vec[7] )} );
                }

                return schpin::OcNode< Content >( std::move( ptr ), j.at( "content" ).get< Content >() );
        }

        static void to_json( nlohmann::json& j, const schpin::OcNode< Content >& node )
        {
                j = nlohmann::json{{"content", *node}, {"children", node.getChildren()}};
        }
};

template < typename Content >
struct adl_serializer< schpin::OcTree< Content > >
{
        static schpin::OcTree< Content > from_json( const nlohmann::json& j )
        {
                return schpin::OcTree< Content >{
                    j.at( "root" ).get< schpin::OcNode< Content > >(),
                    schpin::OcNodeContext( j.at( "depth" ).get< unsigned >(),
                                           j.at( "max_depth" ).get< unsigned >(),
                                           j.at( "length" ).get< schpin::Length >(),
                                           j.at( "position" ).get< schpin::Point< 3 > >() ),
                    j.at( "default_content" ).get< Content >(),
                };
        }

        static void to_json( nlohmann::json& j, const schpin::OcTree< Content >& tree )
        {
                const schpin::OcNodeContext& root_context = tree.getRootContext();

                j = nlohmann::json{{"root", tree.getBareRoot()},
                                   {"default_content", tree.getDefaultContent()},
                                   {"depth", root_context.getDepth()},
                                   {"max_depth", root_context.getMaxDepth()},
                                   {"length", root_context.getEdgeLength()},
                                   {"position", root_context.getPosition()}};
        }
};

}  // namespace nlohmann
