#include "base.h"
#include "util.h"
#include <array>
#include <memory>
#include <numeric>
#include <schpin_lib/util.h>
#include <utility>

#pragma once

namespace schpin
{

template < typename Content >
class OcNode
{
      public:
        using Container     = std::array< OcNode, 8 >;
        using Iterator      = OcNode*;
        using ConstIterator = const OcNode*;

        explicit OcNode( std::unique_ptr< Container >&& nodes, Content content )
          : children_ptr_( std::move( nodes ) )
          , content_( std::move( content ) )
        {
        }

        explicit OcNode( Content content )
          : children_ptr_( nullptr )
          , content_( std::move( content ) )
        {
        }

        template < typename UnaryFunction >
        void modifyContent( UnaryFunction f )
        {
                f( content_ );
        }

        template < typename Proxy, typename... Args >
        Proxy overlayChildren( std::size_t index, Args... args )
        {
                return Proxy( children_ptr_->at( index ), args... );
        }

        std::size_t size() const
        {
                if ( children_ptr_ ) {
                        return children_ptr_->size();
                }
                return 0;
        }

        bool hasChildren() const
        {
                return bool( children_ptr_ );
        }

        Range< ConstIterator > getChildren() const
        {
                if ( children_ptr_ ) {
                        return range( *children_ptr_ );
                }
                return range< ConstIterator >( nullptr, nullptr );
        }

        bool isLeaf() const
        {
                return !bool( children_ptr_ );
        }

        const Content& operator*() const
        {
                return content_;
        }

        const Content* operator->() const
        {
                return &content_;
        }

        void initChildren()
        {
                Container children{OcNode< Content >( content_ ),
                                   OcNode< Content >( content_ ),
                                   OcNode< Content >( content_ ),
                                   OcNode< Content >( content_ ),
                                   OcNode< Content >( content_ ),
                                   OcNode< Content >( content_ ),
                                   OcNode< Content >( content_ ),
                                   OcNode< Content >( content_ )};

                children_ptr_ = std::make_unique< Container >( std::move( children ) );
        }

        void deleteChildren()
        {
                children_ptr_.reset( nullptr );
        }

      private:
        std::unique_ptr< Container > children_ptr_{};
        Content                      content_;
};

template < typename Content >
bool operator!=( const OcNode< Content >& lh, const OcNode< Content >& rh );

template < typename Content >
bool operator==( const OcNode< Content >& lh, const OcNode< Content >& rh )
{
        if ( *lh != *rh ) {
                return false;
        }

        return allOf( lh.getChildren(),
                      rh.getChildren(),
                      []( const OcNode< Content >& lh, const OcNode< Content >& rh ) { return lh == rh; } );
}

template < typename Content >
bool operator!=( const OcNode< Content >& lh, const OcNode< Content >& rh )
{
        return !( lh == rh );
}

template < typename Content >
uint64_t getNodeCount( const OcNode< Content >& node )
{
        return accumulate( node.getChildren(), uint64_t( 1 ), []( const OcNode< Content >& n ) -> uint64_t {
                return getNodeCount( n );
        } );
}

template < typename Content >
uint64_t getTreeHeight( const OcNode< Content >& node )
{
        return max( node.getChildren(), getTreeHeight ) + 1;
}

template < typename Content >
uint64_t getMemoryUsage( const OcNode< Content >& node )
{
        return accumulate( node.getChildren(),
                           node.getContent().getMemoryUsage( node ),
                           []( const OcNode< Content >& n ) { return getMemoryUsage( n ); } );
}

}  // namespace schpin
