#include "base/physics.h"
#include "octree/base.h"
#include "octree/interfaces.h"
#include "octree/node.h"
#include "octree/proxy_node.h"
#include "octree/tree.h"
#include <nlohmann/json.hpp>

#pragma once

namespace schpin
{

static const Occupancy OC_DEFAULT_OCCUPANCY( 0.5f );
static const Occupancy OC_PRUNNING_DIFF_OCCUPANCY( 0.000001f );
static const Occupancy OC_MIN_OCCUPANCY( 0.1192f );
static const Occupancy OC_MAX_OCCUPANCY( 0.971f );

enum class OccupancyToken
{
        FULL
};

class OccupancyContent : public IOcContent< OccupancyContent, OccupancyToken >
{
        Occupancy oc_;

      public:
        OccupancyContent() = default;

        template < typename Item >
        explicit OccupancyContent( Item def )
          : oc_( def )
        {
        }

        const Occupancy& getOccupancy() const
        {
                return oc_;
        }

        std::size_t getMemoryUsage( const OcNode< OccupancyContent >& node ) const override
        {
                return sizeof( node );
        }

        bool prunnable( const OcNode< OccupancyContent >& node ) const override
        {
                // shouldn't we check for children count here? or something? like all children are leafs?
                return abs( OC_MAX_OCCUPANCY - oc_ ) < OC_PRUNNING_DIFF_OCCUPANCY ||
                       abs( OC_MIN_OCCUPANCY - oc_ ) < OC_PRUNNING_DIFF_OCCUPANCY;
        }
        void childBasedUpdate( const OcNode< OccupancyContent >& node ) override
        {
                oc_ = accumulate( node.getChildren(), Occupancy( 0 ), []( const OcNode< OccupancyContent >& child ) {
                        return child->getOccupancy();
                } );
                oc_ /= float( node.size() );
        }

        void setPrunned( const OcNode< OccupancyContent >& node ) override
        {
                if ( oc_ > ( OC_MAX_OCCUPANCY + OC_MIN_OCCUPANCY ) / 2.f ) {
                        oc_ = OC_MAX_OCCUPANCY;
                } else {
                        oc_ = OC_MIN_OCCUPANCY;
                }
        }

        void update( const OccupancyToken& token ) override
        {
                switch ( token ) {
                        case OccupancyToken::FULL:
                                oc_ = OC_MAX_OCCUPANCY;
                                return;
                }
        }
};

extern template class OcTree< OccupancyContent >;
extern template class OcNode< OccupancyContent >;
extern template class OcProxyNode< OccupancyContent >;

inline bool operator==( const OccupancyContent& lh, const OccupancyContent& rh )
{
        return lh.getOccupancy() == rh.getOccupancy();
}
inline bool operator!=( const OccupancyContent& lh, const OccupancyContent& rh )
{
        return !( lh == rh );
}

inline void from_json( const nlohmann::json& j, OccupancyContent& c )
{
        c = OccupancyContent{j.get< float >()};
}

inline void to_json( nlohmann::json& j, const OccupancyContent& c )
{
        j = *( c.getOccupancy() );
}

}  // namespace schpin
