#include "base/pose.h"
#include "collission_interfaces.h"
#include "octree/node.h"
#include "octree/node_context.h"
#include "util.h"
#include <utility>

#pragma once

namespace schpin
{

template < typename NodeContent >
class OcTree;

template < typename Content >
class OcProxyNode
{

      public:
        const OcNode< Content >& operator*() const
        {
                return bare_node_;
        }
        const OcNode< Content >* operator->() const
        {
                return &bare_node_;
        }

        template < typename UnaryFunction >
        void modifyContent( UnaryFunction f )
        {
                bare_node_.modifyContent( f );
        }

        std::size_t size() const
        {
                return bare_node_.size();
        }

        bool hasChildren() const
        {
                return bare_node_.hasChildren();
        }

        unsigned getMaxDepth() const
        {
                return context_.getMaxDepth();
        }

        unsigned getDepth() const
        {
                return context_.getDepth();
        }

        Length getEdgeLength() const
        {
                return context_.getEdgeLength();
        }

        Point< 3 > getPosition() const
        {
                return context_.getPosition();
        }

        bool isLeaf() const
        {
                return bare_node_.isLeaf();
        }

        // method possible to execute only some times? sucks
        void initChildren()
        {
                assert( context_.getDepth() < context_.getMaxDepth() );
                bare_node_.initChildren();
        }

        void deleteChildren()
        {
                bare_node_.deleteChildren();
        }

        OcProxyNode getChild( uint8_t index )
        {
                assert( index < bare_node_.size() );

                return bare_node_.template overlayChildren< OcProxyNode >( index, downscaleContext( context_, index ) );
        }

        const OcProxyNode getChild( uint8_t index ) const
        {
                assert( index < bare_node_.size() );

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
                return const_cast< OcProxyNode* >( this )->getChild( index );
        }

        std::array< OcProxyNode, 8 > generateChildren()
        {
                SCHPIN_ASSERT( bare_node_.hasChildren() );
                return {getChild( 0 ),
                        getChild( 1 ),
                        getChild( 2 ),
                        getChild( 3 ),
                        getChild( 4 ),
                        getChild( 5 ),
                        getChild( 6 ),
                        getChild( 7 )};
        }
        std::array< const OcProxyNode, 8 > generateChildren() const
        {
                SCHPIN_ASSERT( bare_node_.hasChildren() );
                return {getChild( 0 ),
                        getChild( 1 ),
                        getChild( 2 ),
                        getChild( 3 ),
                        getChild( 4 ),
                        getChild( 5 ),
                        getChild( 6 ),
                        getChild( 7 )};
        }

        void tryPrunning()
        {
                if ( !hasChildren() ) {
                        return;
                }
                if ( anyOf( bare_node_.getChildren(),
                            [&]( const OcNode< Content >& node ) { return node.hasChildren(); } ) ) {
                        return;
                }
                const OcNode< Content >& cref = bare_node_;
                if ( bare_node_->prunnable( cref ) ) {
                        bare_node_.modifyContent( [&]( Content& content ) { content.setPrunned( cref ); } );
                        bare_node_.deleteChildren();
                }
        }

      private:
        friend OcTree< Content >;
        friend OcNode< Content >;

        OcProxyNode( OcNode< Content >& node, OcNodeContext context )
          : bare_node_( node )
          , context_( std::move( context ) )
        {
        }

        OcNode< Content >&  bare_node_;
        const OcNodeContext context_;
};

template < typename NodeContent >
std::ostream& operator<<( std::ostream& os, const OcProxyNode< NodeContent >& node )
{
        os << " OcProxyNode pos: " << node.getPosition() << " size: " << node.size() << " depth: " << node.getDepth()
           << "/" << node.getMaxDepth() << " edge: " << node.getEdgeLength();
        return os;
}

template < typename NodeContent >
bool contains( const OcProxyNode< NodeContent >& node, const Point< 3 >& target )
{
        const Point< 3 >& pos = node.getPosition();
        Length            l   = node.getEdgeLength() / 2.f;

        return allOf( range( 3 ), [&]( std::size_t i ) {  //
                return pos[i] - *l <= target[i] && target[i] <= pos[i] + *l;
        } );
}

template < typename NodeContent >
uint64_t getNodeCount( const OcProxyNode< NodeContent >& node )
{
        return getNodeCount( *node );
}

template < typename NodeContent >
unsigned getTreeHeight( const OcProxyNode< NodeContent >& node )
{
        return getTreeHeight( node.getBareNode() );
}

template < typename NodeContent >
uint64_t getMemoryUsage( const OcProxyNode< NodeContent >& node )
{
        return getMemoryUsage( node.getBareNode() );
}
const static std::array< Point< 3 >, 8 > cube_offsets{{Point< 3 >{1, 1, 1},
                                                       Point< 3 >{-1, 1, 1},
                                                       Point< 3 >{1, -1, 1},
                                                       Point< 3 >{-1, -1, 1},
                                                       Point< 3 >{1, 1, -1},
                                                       Point< 3 >{-1, 1, -1},
                                                       Point< 3 >{1, -1, -1},
                                                       Point< 3 >{-1, -1, -1}}};

template < typename NodeContent >
std::array< Point< 3 >, 8 > getNodeCorners( const OcProxyNode< NodeContent >& node )
{
        return mapF( cube_offsets, [&]( const Point< 3 >& p ) {
                return p * float( *node.getEdgeLength() / 2 ) + node.getPosition();
        } );
}

const static std::array< Point< 3 >, 3 > cube_intersection_axis{{
    Point< 3 >( 0, 0, 1 ),
    Point< 3 >( 0, 1, 0 ),
    Point< 3 >( 1, 0, 0 ),
}};

template < typename NodeContent >
class OcNodeFullCollisionQuery : ICollisionItem< const Point< 3 >*, const Point< 3 >* >
{
        std::array< Point< 3 >, 8 > points_;
        std::array< Point< 3 >, 3 > intersection_axes_;

        const OcProxyNode< NodeContent > node_;
        const Pose                       offset_;

      public:
        using PointIterator          = const Point< 3 >*;
        using SeparationAxisIterator = const Point< 3 >*;

        explicit OcNodeFullCollisionQuery( const OcProxyNode< NodeContent >& node )
          : OcNodeFullCollisionQuery( node, Pose{} )
        {
        }

        OcNodeFullCollisionQuery( const OcProxyNode< NodeContent >& node, const Pose& offset )
          : node_( node )
          , offset_( offset )
        {

                points_ = mapF( getNodeCorners( node ), [&]( const Point< 3 >& p ) {  //
                        return transform( p, offset );
                } );

                intersection_axes_ = mapF( cube_intersection_axis,
                                           [&]( const Point< 3 >& p ) {  //
                                                   return rotate( p, offset.orientation );
                                           } );
        }

        Range< SeparationAxisIterator > getSeparationAxes() const override
        {
                return {nullptr, nullptr};
        }

        Range< SeparationAxisIterator > getIntersectionAxes() const override
        {
                return range( intersection_axes_ );
        }

        Range< PointIterator > getPoints() const override
        {
                return range( points_ );
        }

        bool isLeaf() const
        {
                return node_.isLeaf();
        }

        auto getChildren() const
        {
                return mapFtoV( node_.generateChildren(), [&]( const OcProxyNode< NodeContent >& child ) {
                        return OcNodeFullCollisionQuery( child, offset_ );
                } );
        }

        const OcProxyNode< NodeContent >& operator*() const
        {
                return node_;
        }

        const OcProxyNode< NodeContent >* operator->() const
        {
                return &node_;
        }
};

template < typename NodeContent >
class OcNodeFastCollisionQuery : IFastCollisionItem< OcNodeFullCollisionQuery< NodeContent >, 3 >
{
        const Point< 3 >                 pos_;
        const OcProxyNode< NodeContent > node_;
        const Radius                     rad_;
        const Pose                       pose_;

      public:
        using Full = OcNodeFullCollisionQuery< NodeContent >;
        explicit OcNodeFastCollisionQuery( const OcProxyNode< NodeContent >& node )
          : OcNodeFastCollisionQuery( node, Pose{} )
        {
        }
        explicit OcNodeFastCollisionQuery( const OcProxyNode< NodeContent >& node, const Pose& offset )
          : pos_( transform( node.getPosition(), offset ) )
          , node_( node )
          , rad_( node.getEdgeLength() * sqrt( 2.f ) / 2.f )
          , pose_( offset )
        {
        }

        Point< 3 > getMinSpherePosition() const override
        {
                return pos_;
        }

        Radius getMinSphereRadius() const override
        {
                return rad_;
        }

        OcNodeFullCollisionQuery< NodeContent > getFull() const override
        {
                return OcNodeFullCollisionQuery< NodeContent >{node_, pose_};
        }

        bool isLeaf() const
        {
                return node_.isLeaf();
        }

        auto getChildren() const
        {
                return mapFtoV( node_.generateChildren(), [&]( const OcProxyNode< NodeContent >& child ) {
                        return OcNodeFastCollisionQuery( child, pose_ );
                } );
        }

        const OcProxyNode< NodeContent >& operator*() const
        {
                return node_;
        }

        const OcProxyNode< NodeContent >* operator->() const
        {
                return &node_;
        }
};

template < typename LhContent, typename RhContent >
bool operator>( const OcNodeFullCollisionQuery< LhContent >& lh, const OcNodeFullCollisionQuery< RhContent >& rh )
{
        return lh->getEdgeLength() > rh->getEdgeLength();
}
template < typename LhContent, typename RhContent >
bool operator>( const OcNodeFastCollisionQuery< LhContent >& lh, const OcNodeFastCollisionQuery< RhContent >& rh )
{
        return lh->getEdgeLength() > rh->getEdgeLength();
}

template < CollisionMode mode, typename NodeContent >
using OcNodeCollisionQuery = std::conditional_t< mode == CollisionMode::FAST,
                                                 OcNodeFastCollisionQuery< NodeContent >,
                                                 OcNodeFullCollisionQuery< NodeContent > >;

template < typename NodeContent, typename UnaryFunction >
MinMax< Point< 3 > > getUsedSpace( const OcProxyNode< NodeContent >& node, UnaryFunction f )
{
        if ( f( *node ) ) {
                MinMax< Point< 3 > > res = minmax( getNodeCorners( node ) );
                return res;
        }
        if ( !node.hasChildren() ) {
                return startMinMax< Point< 3 > >();
        }

        std::array< Point< 3 >, 16 > p;
        forRange( node.size(), [&]( std::size_t i ) {
                MinMax< Point< 3 > > subres = getUsedSpace( node.getChild( uint8_t( i ) ), f );
                p.at( 2 * i )               = subres.min;
                p.at( 2 * i + 1 )           = subres.max;
        } );
        return minmax( p );
}

template < typename NodeContent, typename UnaryFunction >
Volume getVolume( const OcProxyNode< NodeContent >& node, UnaryFunction accept_f )
{
        if ( accept_f( node ) ) {
                return Volume( float( std::pow( *node.getEdgeLength(), 3 ) ) );
        }
        if ( !node.hasChildren() ) {
                return Volume( 0 );
        }
        return schpin::accumulate(
            node.generateChildren(), Volume( 0 ), [&]( const OcProxyNode< NodeContent >& child ) {
                    return getVolume( child, accept_f );
            } );
}

template < typename NodeContent >
bool validateOctree( const OcProxyNode< NodeContent >& node )
{
        if ( node.getDepth() == node.getMaxDepth() ) {
                return !node.hasChildren();
        }

        if ( !node.hasChildren() ) {
                return true;
        }

        return allOf( node.generateChildren(),
                      []( const OcProxyNode< NodeContent >& child ) { return validateOctree( child ); } );
}

template < typename NodeContent >
const OcProxyNode< NodeContent >& findLeafNode( const OcProxyNode< NodeContent >& node, const Point< 3 >& p )
{
        if ( !node.hasChildren() ) {
                return node;
        }
        auto children  = node.generateChildren();
        auto childIter = findIf( children,
                                 [&]( const OcProxyNode< NodeContent >& child ) {  //
                                         return contains( child, p );
                                 } );
        SCHPIN_ASSERT_MSG( childIter != children.end(), "point: " << p << " and node: " << node );

        return findLeafNode( *childIter, p );
}
}  // namespace schpin
