#include "base.h"
#include "base/physics.h"
#include "node.h"
#include "octree/interfaces.h"
#include "proxy_node.h"
#include <utility>

#pragma once

namespace schpin
{
template < typename NodeContent >
class OcTree
{
        OcNode< NodeContent > root_;
        OcNodeContext         root_context_;
        NodeContent           default_content_;

      public:
        OcTree( const NodeContent& default_content, Length root_length )
          : root_( default_content )
          , root_context_( 0, 0, root_length, Point< 3 >( 0, 0, 0 ) )
          , default_content_( default_content )
        {
        }

        OcTree( OcNode< NodeContent >&& root, const OcNodeContext& context, NodeContent default_content )
          : root_( std::move( root ) )
          , root_context_( context )
          , default_content_( std::move( default_content ) )
        {
                // data that we got may be completely wrong, wrong data-relative-to-childs in nodes, the depth of root
                // thing... I propose a static factory function, that is only one with access to this constructor It
                // shal return datatype with option for error and force user to handle that TODO
                SCHPIN_ASSERT( context.getDepth() == 0 );  // this bothers me and makes me feel insecure
                OcPassiveVisitor< NodeContent > visitor{};
                this->modify( visitor );
        }

        OcTree( const OcTree& other ) = delete;
        OcTree& operator=( const OcTree& other ) = delete;

        OcTree( OcTree&& other ) noexcept = default;
        OcTree& operator=( OcTree&& other ) noexcept = default;

        const OcProxyNode< NodeContent > getProxyRoot() const
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
                return {const_cast< OcNode< NodeContent >& >( root_ ), root_context_};
        }

        const OcNode< NodeContent >& getBareRoot() const
        {
                return root_;
        }

        template < typename Visitor >
        void modify( const Visitor& visitor )
        {
                static_assert( std::is_base_of< IOcSimpleVisitor< NodeContent >, Visitor >::value,
                               "This method accepts only datatype implementing IOcSimpleVisitor" );
                OcSimpleVisitorOverlay< Visitor, NodeContent > overlay( visitor );
                modify< OcEmptyInfoType, OcSimpleVisitorOverlay< Visitor, NodeContent > >( OcEmptyInfoType{}, overlay );
        }

        template < typename InfoType, typename Visitor >
        void modify( const InfoType& info, const Visitor& visitor )
        {
                static_assert( std::is_base_of< IOcVisitor< InfoType, NodeContent >, Visitor >::value,
                               "This method accepts only datatype implementing IOcVisitor" );
                OcProxyNode< NodeContent > n{root_, root_context_};
                SCHPIN_ASSERT( validateOctree( n ) );
                modifyNode( n, info, visitor );
                SCHPIN_ASSERT( validateOctree( n ) );
        }

        template < typename Container >
        void expand( const Container& cont )
        {
                forEach( cont, [&]( const Point< 3 >& p ) { this->expand( p ); } );
        }

        void expand( Point< 3 > target )
        {
                if ( contains( getProxyRoot(), target ) ) {
                        return;
                }
                // The actual root cube has to be expanded in direction of that target.
                // The center of new root cube has to exists in direction of the tagert point.
                //
                // To solve this, the relative coordinates of actual root cube, in frame of the new root, have to be
                // known.

                Point< 3 > root_pos  = root_context_.getPosition();
                OcChildPos child_pos = getOcChildPosBasedOnCenter( target, root_pos );

                // New root for this tree is created and actual root is set as it's child at appropiate position

                std::array< OcNode< NodeContent >, 8 > new_nodes( {
                    OcNode< NodeContent >( *root_ ),
                    OcNode< NodeContent >( *root_ ),
                    OcNode< NodeContent >( *root_ ),
                    OcNode< NodeContent >( *root_ ),
                    OcNode< NodeContent >( *root_ ),
                    OcNode< NodeContent >( *root_ ),
                    OcNode< NodeContent >( *root_ ),
                    OcNode< NodeContent >( *root_ ),
                } );
                std::swap( new_nodes.at( child_pos.getIndex() ), root_ );
                root_ = OcNode< NodeContent >(
                    std::make_unique< std::array< OcNode< NodeContent >, 8 > >( std::move( new_nodes ) ),
                    default_content_ );

                root_.modifyContent( [&]( NodeContent& content ) { content.childBasedUpdate( root_ ); } );

                // Scale the new properties of the root cube.
                root_context_ = upscaleContext( root_context_, child_pos );

                expand( target );
        }

        const OcNodeContext& getRootContext() const
        {
                return root_context_;
        }

        const NodeContent& getDefaultContent() const
        {
                return default_content_;
        }

        ~OcTree() = default;

      private:
        template < typename InfoType, typename Visitor >
        void modifyNode( OcProxyNode< NodeContent >& node, const InfoType& parent_info, const Visitor& visitor )
        {
                const OcProxyNode< NodeContent >& c_node = node;
                InfoType                          info;
                if ( node.getDepth() == node.getMaxDepth() ) {
                        node.modifyContent(
                            [&]( NodeContent& content ) { visitor.accessBottomLeaf( c_node, parent_info, content ); } );
                        return;
                }
                if ( node.hasChildren() ) {
                        std::pair< OcNodeControl, InfoType > res = visitor.accessNode( c_node, parent_info );
                        if ( res.first == OcNodeControl::END ) {
                                return;
                        }
                        info = res.second;
                } else {
                        std::pair< OcLockedLeafControl, InfoType > res =
                            visitor.accessLockedLeaf( c_node, parent_info );
                        if ( res.first == OcLockedLeafControl::UNLOCK ) {
                                node.modifyContent( [&]( NodeContent& content ) {
                                        visitor.accessBottomLeaf( c_node, parent_info, content );
                                } );
                                return;
                        } else if ( res.first == OcLockedLeafControl::END ) {
                                return;
                        }
                        node.initChildren();
                        info = res.second;
                }
                if ( node.hasChildren() ) {
                        forEach( node.generateChildren(),
                                 [&]( OcProxyNode< NodeContent > child ) { modifyNode( child, info, visitor ); } );
                }

                node.modifyContent( [&]( NodeContent& content ) { content.childBasedUpdate( *node ); } );
                node.tryPrunning();
        }
};

template < typename NodeContent >
bool operator==( const OcTree< NodeContent >& lh, const OcTree< NodeContent >& rh )
{
        return lh.getBareRoot() == rh.getBareRoot() && lh.getRootContext() == rh.getRootContext();
}

template < typename NodeContent, typename UnaryFunction >
void logOctreeSpace( const OcTree< NodeContent >& tree, UnaryFunction f )
{
        MinMax< Point< 3 > > res = getUsedSpace( tree.getProxyRoot(), f );
        SCHPIN_DEBUG_LOG( "Octree min: " << res.min << " max: " << res.max );
}

template < typename NodeContent >
std::optional< OcProxyNode< NodeContent > > findLeafNode( const OcTree< NodeContent >& tree, const Point< 3 >& p )
{
        if ( !contains( tree.getProxyRoot(), p ) ) {
                return {};
        }
        return {findLeafNode( tree.getProxyRoot(), p )};
}

}  // namespace schpin
