#include "parser.h"
#include "schpin_lib/util.h"
#include <assimp/Importer.hpp>
#include <assimp/postprocess.h>
#include <assimp/scene.h>
#include <fstream>

namespace schpin
{

inline Point< 3 > aiVecToPoint( aiVector3D vec )
{
        return Point< 3 >( vec.x, vec.y, vec.z );
}

Either< std::vector< Triangle< 3 > >, ParseError > parseTriangles( std::filesystem::path filename )
{
        filename = getResolvedPath( filename );
        Assimp::Importer importer;

        SCHPIN_DEBUG_LOG( "Loading file: " << filename );

        std::ifstream file{filename, std::ios::binary | std::ios::ate};

        if ( !file ) {
                return {E( ParseError ) << "Failed to open file: " << filename};
        }

        std::streamsize size = file.tellg();
        file.seekg( 0, std::ios::beg );

        std::vector< char > buffer( static_cast< uint64_t >( size ) );  // this conversion is ugly TODO: find a proper
                                                                        // way to do this and do this in all places!
        if ( !file.read( buffer.data(), size ) ) {
                return {E( ParseError ) << "Failed to load a file: " << filename};
        }

        std::filesystem::path file_t = filename.extension();
        if ( file_t == "stlb" ) {
                file_t = "stl";
        }

        const aiScene* scene = importer.ReadFileFromMemory( static_cast< void* >( buffer.data() ),
                                                            buffer.size(),
                                                            aiProcess_CalcTangentSpace | aiProcess_Triangulate |
                                                                aiProcess_JoinIdenticalVertices |
                                                                aiProcess_SortByPType | aiProcess_ImproveCacheLocality,
                                                            file_t.c_str() );

        if ( scene == nullptr ) {
                return {E( ParseError ) << "Assimp failed to parse the file: " << importer.GetErrorString()};
        }

        std::vector< Triangle< 3 > > res;

        forEach( range( scene->mMeshes, scene->mNumMeshes ), [&]( const aiMesh* mesh ) {
                forEach( range( mesh->mFaces, mesh->mNumFaces ), [&]( const aiFace& face ) {
                        SCHPIN_ASSERT( face.mNumIndices == 3 );

                        std::array< Point< 3 >, 3 > points = mapFtoA< 3 >(
                            range( face.mIndices, face.mNumIndices ), [&]( unsigned int indice ) -> Point< 3 > {
                                    // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-pointer-arithmetic)
                                    return aiVecToPoint( mesh->mVertices[indice] );
                            } );

                        res.emplace_back( points[0], points[1], points[2] );
                } );
        } );

        return {res};
}

}  // namespace schpin
