#include "body/graph.h"
#include "body/rand.h"
#include "environment.h"
#include "graph/astar.h"
#include "graph/astar/content.h"
#include "planner/base.h"
#include "robot.h"

#pragma once

namespace schpin
{

struct BodyPlannerErrorTag
{
};

using BodyPlannerError = Error< BodyPlannerErrorTag >;
template < CollisionMode mode >
bool bodyCollidesWithEnvironment( const Body& body, const Pose& offset, const Environment& env )
{
        // TODO: UGLYYYYY

        OcNodeCollisionQuery< mode, OccupancyContent > body_query( body.getCollisionTree().getProxyRoot(), offset );
        OcNodeCollisionQuery< mode, EnvironmentOcTreeContent > env_query( env.getTree().getProxyRoot() );

        return treesNodesCollides< 3, mode >(
            body_query, env_query, [&]( const auto& body_query, const auto& env_query ) {
                    return ( **body_query )->getOccupancy() > Occupancy( 0.5 ) &&
                           ( **env_query )->getOccupancy() > ENV_MAX_OCCUPANCY / 2;
            } );
}

template < CollisionMode mode >
bool allLegsTouchesWithEnvironment( const LegContainer< Leg< 3 > >& legs,
                                    const Pose&                     body_offset,
                                    const Environment&              env )
{

        OcNodeCollisionQuery< mode, EnvironmentOcTreeContent > env_query( env.getTree().getProxyRoot() );

        return allOf( legs, [&]( const Leg< 3 >& leg ) -> bool {
                Pose                                           offset = transform( leg.getOriginOffset(), body_offset );
                OcNodeCollisionQuery< mode, OccupancyContent > leg_query( leg.getWorkarea().getProxyRoot(), offset );
                return treesNodesCollides< 3, mode >(
                    leg_query, env_query, [&]( const auto& leg_node, const auto& env_node ) {
                            return ( **env_node )->isSurface() && ( **leg_node )->getOccupancy() > Occupancy( 0.5 );
                    } );
        } );
}

// usefull for tests
struct ValidPathRes
{
        const bool is_valid;
        const Pose failed_pose;

        operator bool() const
        {
                return is_valid;
        }
};

template < CollisionMode mode >
ValidPathRes isValidPath( const Pose& source_pose, const Pose& target_pose, const Robot& robot, const Environment& env )
{

        Distance dist_step = min( robot.getBody().getCollisionTree().getRootContext().getMinEdgeLength(),
                                  robot.getLegs()[0].getWorkarea().getRootContext().getMinEdgeLength() );

        // TODO(squirrel):
        // the angle could be calculated by getting radius of smalles sphere for entire robot strechted at maximum
        // the angle is than angle between two points, which are 'radius' from [0,0,0] and 'dist_step' from each other
        auto angle_step = Angle( 0.1f );

        PoseDistance actual_dist = source_pose.distance( target_pose );

        std::size_t steps_c = steps( actual_dist, dist_step, angle_step );

        steps_c = std::max( steps_c, std::size_t( 2 ) );

        // find problematic step for body
        // instead of steps: 1,2,3,4,5,6,7,8,9,10
        // make it: 1,6,11,..2,7,12...,3,8,13...
        std::size_t maxstep     = steps_c + 1;
        std::size_t step_offset = 5;
        auto        offset      = *findIf( range( maxstep ), [&]( std::size_t i ) {
                i = i * step_offset % maxstep + i * step_offset / maxstep;

                float factor        = float( i ) / float( steps_c );
                Pose  actual_offset = interp( source_pose, target_pose, factor );

                return bodyCollidesWithEnvironment< mode >( robot.getBody(), actual_offset, env );
        } );

        if ( offset != steps_c + 1 ) {
                return ValidPathRes{false, interp( source_pose, target_pose, float( offset ) / float( steps_c ) )};
        }

        // find problematic step for legs
        offset = *findIf( range( maxstep ), [&]( std::size_t i ) {
                i                   = i * step_offset % maxstep + i * step_offset / maxstep;
                float factor        = float( i ) / float( steps_c );
                Pose  actual_offset = interp( source_pose, target_pose, factor );

                return !allLegsTouchesWithEnvironment< mode >( robot.getLegs(), actual_offset, env );
        } );

        return ValidPathRes{offset == steps_c + 1,
                            interp( source_pose, target_pose, float( offset ) / float( steps_c ) )};
}

template < typename HeapHandle >
std::vector< std::pair< Distance, BodyVertexID > > getNeighboursDistances( const Pose&                    new_pose,
                                                                           const Robot&                   robot,
                                                                           const BodyGraph< HeapHandle >& graph )
{
        auto distances = mapF< std::vector< std::pair< Distance, BodyVertexID > > >(
            graph.getVertexes(), [&]( const auto& vertex ) {  //
                    PoseDistance actual_dist = vertex->pose.distance( new_pose );
                    return std::make_pair( actual_dist.dist + ( *actual_dist.angle ) * robot.getPriceRadius(),
                                           vertex.getID() );
            } );

        std::sort( distances.begin(), distances.end(), [&]( const auto& lh, const auto& rh ) {  //
                return lh.first < rh.first;
        } );

        return distances;
}

template < CollisionMode mode, typename HeapHandle >
std::vector< BodyVertexID >
filterCandidateNeighbours( const Pose&                                               new_pose,
                           const BodyGraph< HeapHandle >&                            graph,
                           const std::vector< std::pair< Distance, BodyVertexID > >& distances,
                           const Robot&                                              robot,
                           const Environment&                                        env,
                           std::size_t                                               count )
{

        std::vector< BodyVertexID > res{};

        // ugly way to do while :)
        findIf( distances, [&]( const std::pair< Distance, BodyVertexID >& pair ) -> bool {
                const auto& vertex = graph.getVertex( pair.second );
                if ( isValidPath< mode >( new_pose, vertex->pose, robot, env ) ) {
                        res.emplace_back( vertex.getID() );
                }

                return res.size() == count;
        } );

        return res;
}

struct BodyAStarConfig
{
        using QueueItem     = AStarQueueItem< BodyVertexID, BodyPrice >;
        using Heap          = boost::heap::pairing_heap< QueueItem, boost::heap::compare< std::greater< QueueItem > > >;
        using HeapHandle    = typename Heap::handle_type;
        using SearchContent = AStarContent< HeapHandle, BodyPrice, BodyEdgeID >;
        using Graph         = BodyGraph< BodyVertexContent< SearchContent > >;
        using HeuristicFunction = ZeroGraphHeuristics< BodyPrice >;
};

using BodySearchOverlay = AStarOverlay< BodyAStarConfig >;

struct MakeVertexesEvent
{
        unsigned count;
};

struct FindPathEvent
{
};

template < typename OnPathFunction >
class BodyPlannerLambdaVisitor
{
        OnPathFunction on_path_f_;

      public:
        BodyPlannerLambdaVisitor( OnPathFunction on_path_f )
          : on_path_f_( on_path_f )
        {
        }

        void hasPlan( Chain< BodyEdgeID, BodyVertexID > chain )
        {
                on_path_f_( std::move( chain ) );
        }
};

template < CollisionMode mode >
class BodyPlanner
{
      public:
        using Graph         = typename BodySearchOverlay::Graph;
        using Heuristics    = typename BodySearchOverlay::HeuristicFunction;
        using VertexContent = typename Graph::VertexContent;

      private:
        PlannerContext    plan_context_;
        BodySearchOverlay graph_search_;
        BodyGenerator     gen_;

      public:
        BodyPlanner( PlannerContext plan_context )
          : plan_context_( std::move( plan_context ) )
          , graph_search_( Graph(), Heuristics() )
          , gen_()
        {

                // this does not check collision, fix!
                BodyVertexID current = graph_search_.emplaceVertex( VertexContent( plan_context_.current ) ).getID();
                if ( plan_context_.goal ) {
                        BodyVertexID goal = graph_search_.emplaceVertex( VertexContent( *plan_context_.goal ) ).getID();
                        graph_search_.setTask( current, goal );
                }
        }

        const Robot& getRobot() const
        {
                return *plan_context_.robot_ptr;
        }

        const Environment& getEnvironment() const
        {
                return *plan_context_.env_ptr;
        }

        bool hasPlan() const
        {
                return graph_search_.getSearchState() == GraphSearchState::FOUND_PATH;
        }

        OptError< BodyPlannerError > setGoal( Pose goal_pose )
        {
                plan_context_.goal = goal_pose;
                // Search first, check for collsion not present...
                std::optional< BodyVertexID > gid = createVertex< false >( goal_pose );

                if ( !gid ) {
                        return E( BodyPlannerError ) << "Failed to create vertex for goal pose";
                }

                graph_search_.setTask( graph_search_.getTask()->current_id, *gid );
                return {};
        }

        OptError< BodyPlannerError > setCurrent( Pose pose )
        {
                plan_context_.current = pose;

                // Search first!, check for collision...
                std::optional< BodyVertexID > cid = createVertex< false >( pose );

                if ( !cid ) {
                        return E( BodyPlannerError ) << "Failed to create vertex for current pose";
                }

                graph_search_.setTask( *cid, graph_search_.getTask()->goal_id );

                return {};
        }

        template < typename Visitor >
        void event( MakeVertexesEvent e, Visitor )
        {
                if ( !plan_context_.goal ) {
                        return;
                }

                while ( e.count > 0 ) {
                        Pose offset = gen_.getRandomBodyCoord( plan_context_.current, *plan_context_.goal );

                        std::optional< BodyVertexID > vid = createVertex< false >( offset );
                        if ( vid ) {
                                e.count -= 1;
                        }
                }
        }

        template < typename Visitor >
        void event( FindPathEvent, Visitor visitor )
        {

                // replan path
                auto limit = std::numeric_limits< std::size_t >::max();
                graph_search_.computeShortestPath( limit );

                if ( hasPlan() ) {
                        visitor.hasPlan( getPath() );
                }
        }

        Chain< BodyEdgeID, BodyVertexID > getPath() const
        {
                return graph_search_.getPath();
        }

        const Graph& getGraph() const
        {
                return graph_search_.getGraph();
        }

        const Pose& getBodyPose( BodyVertexID vid ) const
        {
                return getGraph().getVertex( vid )->pose;
        }

      private:
        template < bool requires_neighbours >
        std::optional< BodyVertexID > createVertex( Pose offset )
        {
                const Environment& env   = *plan_context_.env_ptr;
                const Robot&       robot = *plan_context_.robot_ptr;

                // check collision with environment
                if ( bodyCollidesWithEnvironment< mode >( robot.getBody(), offset, env ) ||
                     !allLegsTouchesWithEnvironment< mode >( robot.getLegs(), offset, env ) ) {
                        return {};
                }

                std::vector< std::pair< Distance, BodyVertexID > > distances =
                    getNeighboursDistances( offset, robot, graph_search_.getGraph() );

                // check for closest neighbour
                if ( distances.front().first < Distance( 0.001f ) ) {
                        return {};
                }

                // try to attach to existing
                std::vector< BodyVertexID > nbrs = filterCandidateNeighbours< mode >(
                    offset, graph_search_.getGraph(), distances, robot, env, 8 );  // TODO: what the number means? :)
                if ( requires_neighbours && nbrs.empty() ) {
                        return {};
                }

                // create graph node, attach it
                const auto& new_vertex = graph_search_.emplaceVertex( VertexContent{offset} );

                auto id_to_pose_f = [&]( BodyVertexID id ) -> const Pose& {
                        return graph_search_.getGraph().getVertex( id )->pose;
                };

                auto calc_edge_price_f = [&]( BodyVertexID from, BodyVertexID to ) {
                        PoseDistance dist = id_to_pose_f( from ).distance( id_to_pose_f( to ) );
                        BodyPrice    p    = *dist.dist + ( *dist.angle ) * ( *robot.getPriceRadius() );
                        return BodyEdgeContent{p};
                };

                graph_search_.addEdges( new_vertex.getID(), nbrs, calc_edge_price_f );
                forEach( nbrs, [&]( BodyVertexID nb ) {
                        graph_search_.addEdge( nb, new_vertex.getID(), calc_edge_price_f );
                } );

                return {new_vertex.getID()};
        }
};

extern template class BodyPlanner< CollisionMode::FULL >;
extern template class BodyPlanner< CollisionMode::FAST >;

}  // namespace schpin
