#include "body/view.h"
#include "gait/graph.h"
#include "gait/vertex.h"
#include "graph/util.h"
#include "leg/graph.h"
#include "robot/view.h"

#pragma once

namespace schpin
{
template < typename GaitVertex >
class GaitHeuristics
{
      public:
        bool minimal()
        {
                return true;
        }
        void decrease()
        {
        }
        GaitPrice operator()( const GaitVertex& /*from*/, const GaitVertex& /*to*/ )
        {
                return GaitPrice( 0.f );
        }
};
struct GaitDStarConfig
{
        using QueueItem         = DstarQueueItem< GaitVertexID, GaitPrice >;
        using Heap              = boost::heap::pairing_heap< QueueItem >;
        using HeapHandle        = typename Heap::handle_type;
        using SearchContent     = DStarContent< HeapHandle, GaitPrice >;
        using Graph             = GaitGraph< GaitVertexContent< SearchContent > >;
        using HeuristicFunction = GaitHeuristics< typename Graph::Vertex >;
};

struct MoveLegOption
{
        LegVertexID actual_vertex_id;
        Point< 3 >  target;
};

struct StandingLegOption
{
        LegVertexID actual_vertex_id;
        Point< 3 >  actual_pos;
};

struct BodyOption
{
        Pose goal_pose;
};

using LegChangeOption = std::variant< MoveLegOption, StandingLegOption >;

template < typename LegView >
std::vector< LegChangeOption > generateGaitLegOptions( const GaitStandingLeg& actual, const LegView& leg_view )
{
        LegVertexID closest_vertex_id = leg_view.findClosestVertex( actual.position, actual.lid );
        auto        res               = mapF< std::vector< LegChangeOption > >(
            leg_view.getSidepathView().getCandidatePositions( actual.position ), [&]( Point< 3 > target ) {
                    return MoveLegOption{closest_vertex_id, target};
            } );
        res.push_back( StandingLegOption{actual.lid, actual.position} );
        return res;
}

template < typename LegView >
std::vector< LegChangeOption > generateGaitLegOptions( const GaitMovingLeg& actual, const LegView& leg_view )
{
        LegVertexID closest_vertex_id = leg_view.findClosestVertex( actual.target, actual.actual_lid );
        std::vector< LegChangeOption > res;

        if ( closest_vertex_id == actual.actual_lid ) {
                res.emplace_back( StandingLegOption{actual.actual_lid, actual.target} );
        } else {
                res.emplace_back( MoveLegOption{actual.actual_lid, actual.target} );
        }
        return res;
}

template < typename BodyView >
inline std::vector< BodyOption > generateBodyGaitOptions( const GaitState& center, const BodyView& body_view )
{

        return mapF< std::vector< BodyOption > >( body_view.getCandidateOffsets( center.body_pose ),
                                                  [&]( const Pose& offset ) { return BodyOption{offset}; } );
}

template < typename GaitGraph, typename ProcessFunction >
void generateNeighbourGaitOptions( const GaitState&              center,
                                   const RobotView< GaitGraph >& view,
                                   ProcessFunction               process_f )
{
        const GaitLegContainer& leg_states = center.legs;

        auto legs_options = mapF< std::vector< std::vector< LegChangeOption > > >(
            view.getLegViews(), [&]( const auto& leg_view ) -> std::vector< LegChangeOption > {
                    LegID id = leg_view.getLeg().getID();
                    return std::visit( [&]( const auto& actual ) { return generateGaitLegOptions( actual, leg_view ); },
                                       leg_states[*id] );
            } );

        std::vector< BodyOption > body_options = generateBodyGaitOptions( center, view.getBodyView() );

        cartesianProduct( legs_options, [&]( const std::vector< LegChangeOption >& variants ) {
                LegContainer< LegChangeOption > container( variants );

                forEach( body_options,
                         [&]( const BodyOption& body_option ) { process_f( body_option, std::move( container ) ); } );
        } );
}

}  // namespace schpin
