#include "body.h"
#include "config.h"
#include "leg.h"
#include <utility>

#pragma once

namespace schpin
{

struct RobotErrorTag
{
};

using RobotError = Error< RobotErrorTag >;

class Robot
{

        Body                     body_;
        LegContainer< Leg< 3 > > legs_;

      public:
        Robot( Body&& body, LegContainer< Leg< 3 > >&& legs )
          : body_( std::move( body ) )
          , legs_( std::move( legs ) )
        {
        }

        const Body& getBody() const
        {
                return body_;
        }

        const LegContainer< Leg< 3 > >& getLegs() const
        {
                return legs_;
        }

        // radius used when calculationg price of rotation in pose distance
        Radius getPriceRadius() const
        {
            //TODO: THIS SHALL NOT BE HARDCODED
            return Radius{0.1f};
        }

        template < typename UnaryFunction >
        void accessDstar( LegID id, UnaryFunction f )
        {
                legs_[id].accessDstar( f );
        }
};

inline Either< Robot, RobotError > loadRobot( std::filesystem::path cache_prefix,
                                              const urdf::Model&    model,
                                              std::istream&         srdf_istream )
{
        return assembleLeftCollectRight(  //
                   loadLegs( cache_prefix, model, srdf_istream ).convertRight( [&]( LegBootError&& e ) -> RobotError {
                           return E( RobotError ).attach( e ) << "Errors occured during the load of leg";
                   } ),
                   loadBody( model, "body" ).convertRight( [&]( BodyBootError&& e ) -> RobotError {
                           return E( RobotError ).attach( e ) << "Errors occured during the load of body";
                   } ) )
            .convertLeft( [&]( std::tuple< LegContainer< Leg< 3 > >, Body >&& data ) {
                    return Robot( std::get< Body >( std::move( data ) ),
                                  std::get< LegContainer< Leg< 3 > > >( std::move( data ) ) );
            } )
            .convertRight( [&]( std::vector< RobotError >&& right ) -> RobotError {
                    return E( RobotError ).attach( right ) << "Failed to load the robot";
            } );
}

inline Either< Robot, RobotError > loadRobotROS( ros::NodeHandle handle, ROSConfig cfg )
{
        urdf::Model model;
        model.initParam( cfg.robot_description_topic );

        std::string srdf;
        handle.getParam( cfg.robot_semantics_topic, srdf );

        std::stringstream srdf_istream{srdf};

        return loadRobot( cfg.cache_prefix, model, srdf_istream );
}

}  // namespace schpin

namespace nlohmann
{
template <>
struct adl_serializer< schpin::Robot >
{
        static void to_json( nlohmann::json& j, const schpin::Robot& robot )
        {
                j = nlohmann::json{{"legs", robot.getLegs()}, {"body", robot.getBody()}};
        }

        static schpin::Robot from_json( const nlohmann::json& j )
        {
                return schpin::Robot( j["body"].get< schpin::Body >(),
                                      j["legs"].get< schpin::LegContainer< schpin::Leg< 3 > > >() );
        }
};
}  // namespace nlohmann
