#include "smodel.h"

namespace schpin
{

Either< std::vector< Triangle< 3 > >, UrdfParseError > extractTrinaglesFromUrdf( const urdf::Pose&     offset,
                                                                                 const urdf::Geometry& gem )
{
        switch ( gem.type ) {
                case urdf::Geometry::SPHERE:
                case urdf::Geometry::BOX:
                case urdf::Geometry::CYLINDER:
                        return {E( UrdfParseError ) << "Unsupported geometry type"};
                        break;
                case urdf::Geometry::MESH:
                        const auto* mesh_ptr = dynamic_cast< const urdf::Mesh* >( &gem );

                        return parseTriangles( mesh_ptr->filename )
                            .convertLeft( [&]( std::vector< Triangle< 3 > >&& triangles ) {
                                    forEach( triangles, [&]( Triangle< 3 >& t ) {
                                            t = transform( scale( t, urdfPointToSchpinPoint( mesh_ptr->scale ) ),
                                                           urdfPoseToSchpinPose( offset ) );
                                    } );

                                    return std::move( triangles );
                            } )
                            .convertRight( [&]( ParseError e ) {
                                    return E( UrdfParseError ).attach( e ) << "Failed to parse URDF file";
                            } );
        }
        return {E( UrdfParseError ) << "Unknown geometry type"};
}

Either< StructureModel, UrdfParseError > extractModel( const urdf::Link& link )
{
        std::vector< std::pair< const urdf::Pose*, const urdf::Geometry* > > data;
        std::vector< UrdfParseError >                                        errors;
        Color                                                                color{0.0f, 0.0f, 0.0f};

        if ( !link.collision_array.empty() ) {
                forEach( link.collision_array, [&]( const urdf::CollisionSharedPtr col_ptr ) {
                        if ( !col_ptr->geometry ) {
                                errors.emplace_back( "The collision element is missing geometry" );
                        }
                        data.emplace_back( &col_ptr->origin, &*col_ptr->geometry );
                } );
        } else if ( link.collision ) {
                if ( !link.collision->geometry ) {
                        errors.emplace_back( "The collision element is missing geometry" );
                }
                data.emplace_back( &link.collision->origin, &*link.collision->geometry );
        } else {  // link->visual_array
                forEach( link.visual_array, [&]( const urdf::VisualSharedPtr vis_ptr ) {
                        if ( !vis_ptr->geometry ) {
                                errors.emplace_back( "The visual element is missing geometry" );
                        }
                        data.emplace_back( &vis_ptr->origin, &*vis_ptr->geometry );

                        // yeah, just take color of the last visual...
                        if ( vis_ptr->material ) {
                                color = urdfColorToSchpinColor( vis_ptr->material->color );
                        }
                } );
        }

        if ( data.empty() ) {
                return {E( UrdfParseError ) << "Link " << link.name << " does not have any 3D model."};
        }

        std::vector< Triangle< 3 > > surface;

        forEach( data, [&]( std::pair< const urdf::Pose*, const urdf::Geometry* >& p ) {
                extractTrinaglesFromUrdf( *p.first, *p.second )
                    .match(
                        [&]( const std::vector< Triangle< 3 > >& triangles ) {
                                surface.insert( surface.end(), triangles.begin(), triangles.end() );
                        },
                        [&]( const UrdfParseError& error ) { errors.push_back( error ); } );
        } );

        if ( !errors.empty() ) {
                return {E( UrdfParseError ).attach( errors )
                        << "There were errors during creation of model from link " << link.name};
        }

        return {StructureModel( surface, color )};
}
}  // namespace schpin
