#include "base/color.h"
#include "base/triangle.h"
#include "parser.h"
#include "schpin_lib/either.h"
#include "schpin_lib/error.h"
#include <string>
#include <urdf/model.h>
#include <urdf_model/joint.h>
#include <urdf_model/link.h>
#include <utility>
#include <vector>

#pragma once

namespace schpin
{

class StructureModel
{
        // this is pointlessly class... could be const struct instead :/
        std::vector< Triangle< 3 > > surface_;
        Color                        color_;
        Point< 3 >                   center_;
        Radius                       mini_ball_radius_;

      public:
        explicit StructureModel( std::vector< Triangle< 3 > > surface, Color c )
          : surface_( std::move( surface ) )
          , color_( c )
        {
                center_ = accumulate( surface_, Point< 3 >{}, [&]( const Triangle< 3 >& t ) -> Point< 3 > {  //
                        return center( t ) / surface_.size();
                } );

                mini_ball_radius_ = max( surface_, [&]( const Triangle< 3 >& t ) {  //
                        return max( t, [&]( const Point< 3 >& p ) {                 //
                                return p.distance( center_ );
                        } );
                } );
        }

        StructureModel() = default;

        const Point< 3 >& getCenter() const
        {
                return center_;
        }

        const Radius getMinballRadius() const
        {
                return mini_ball_radius_;
        }

        const Color& getColor() const
        {
                return color_;
        }

        const std::vector< Triangle< 3 > >& getSurface() const
        {
                return surface_;
        }

        auto getConvexHull() const
        {
                return range( surface_ );
        }
};

inline void to_json( nlohmann::json& j, const StructureModel& model )
{
        j["surface"] = model.getSurface();
        j["color"]   = model.getColor();
}

inline void from_json( const nlohmann::json& j, StructureModel& model )
{
        model = StructureModel{j["surface"].get< std::vector< Triangle< 3 > > >(), j["color"].get< Color >()};
}

inline Color urdfColorToSchpinColor( urdf::Color c )
{
        return Color{c.r, c.g, c.b};
}

inline Point< 3 > urdfPointToSchpinPoint( urdf::Vector3 v )
{
        return Point< 3 >{float( v.x ), float( v.y ), float( v.z )};
}

inline Quaternion urdfRotationToSchpinQuaternion( urdf::Rotation r )
{
        return Quaternion{r.x, r.y, r.z, r.w};
}

inline Pose urdfPoseToSchpinPose( urdf::Pose p )
{
        return Pose{urdfPointToSchpinPoint( p.position ), urdfRotationToSchpinQuaternion( p.rotation )};
}
using UrdfParseError = Error< struct UrdfParseTag >;

Either< StructureModel, UrdfParseError > extractModel( const urdf::Link& link );

using PointVectorIterator = std::vector< Point< 3 > >::const_iterator;
using PointSetIterator    = std::set< Point< 3 > >::const_iterator;

class StructureModelCollisionQuery : ICollisionItem< PointSetIterator, PointVectorIterator >
{
        std::set< Point< 3 > >    points_;
        std::vector< Point< 3 > > normals_;

      public:
        using PointIterator          = PointSetIterator;
        using SeparationAxisIterator = PointVectorIterator;

        explicit StructureModelCollisionQuery( const StructureModel& model, Pose offset )
        {
                forEach( model.getConvexHull(), [&]( const Triangle< 3 >& tri ) {
                        Triangle< 3 > t = transform( tri, offset );

                        forEach( t, [&]( const Point< 3 >& p ) {  //
                                points_.insert( p );
                        } );

                        normals_.push_back( getNormal( t ) );
                } );
        }

        Range< PointSetIterator > getPoints() const override
        {
                return range( points_ );
        }

        Range< PointVectorIterator > getSeparationAxes() const override
        {
                return range( normals_ );
        }

        Range< PointVectorIterator > getIntersectionAxes() const override
        {
                return {normals_.end(), normals_.end()};
        }
};

class FastStructureModelCollisionQuery : IFastCollisionItem< StructureModelCollisionQuery, 3 >
{
        const StructureModel& model_;
        Pose                  offset_;

      public:
        using Full = StructureModelCollisionQuery;
        FastStructureModelCollisionQuery( const StructureModel& model, Pose offset )
          : model_( model )
          , offset_( offset )
        {
        }
        Point< 3 > getMinSpherePosition() const
        {
                return transform( model_.getCenter(), offset_ );
        }
        Radius getMinSphereRadius() const
        {
                return model_.getMinballRadius();
        }
        StructureModelCollisionQuery getFull() const
        {
                return StructureModelCollisionQuery{model_, offset_};
        }

        bool isLeaf() const
        {
                return false;
        }

        auto getChildren() const
        {
                // TODO: this rather be lazy?
                return mapFtoV( model_.getConvexHull(), [&]( const Triangle< 3 >& t ) {
                        return TriangleFastCollisionQuery( transform( t, offset_ ) );
                } );
        }
};

}  // namespace schpin
