#include <interactive_markers/interactive_marker_server.h>
#include <tf2_ros/transform_broadcaster.h>

#pragma once

namespace schpin
{

visualization_msgs::InteractiveMarker makePoseMarker( std::string name, std::string frame )
{
        visualization_msgs::InteractiveMarker int_marker;
        int_marker.header.frame_id = frame;
        int_marker.name            = name + "_marker";
        int_marker.description     = name + " marker";
        int_marker.scale           = 0.1f;

        visualization_msgs::Marker box_marker;
        box_marker.frame_locked = true;
        box_marker.type         = visualization_msgs::Marker::SPHERE;
        box_marker.scale.x      = 0.05;
        box_marker.scale.y      = 0.05;
        box_marker.scale.z      = 0.05;
        box_marker.color.r      = 0.5;
        box_marker.color.g      = 0.5;
        box_marker.color.b      = 0.5;
        box_marker.color.a      = 0.5;

        visualization_msgs::InteractiveMarkerControl box_control;
        box_control.always_visible = true;
        box_control.markers.push_back( box_marker );

        int_marker.controls.push_back( box_control );

        visualization_msgs::InteractiveMarkerControl control;
        control.always_visible = true;

        control.orientation.w    = 1;
        control.orientation.x    = 1;
        control.orientation.y    = 0;
        control.orientation.z    = 0;
        control.name             = name + "_move_x";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::MOVE_AXIS;
        int_marker.controls.push_back( control );
        control.name             = name + "_rotate_x";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::ROTATE_AXIS;
        int_marker.controls.push_back( control );

        control.orientation.w    = 1;
        control.orientation.x    = 0;
        control.orientation.y    = 1;
        control.orientation.z    = 0;
        control.name             = name + "_move_y";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::MOVE_AXIS;
        int_marker.controls.push_back( control );
        control.name             = name + "_rotate_y";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::ROTATE_AXIS;
        int_marker.controls.push_back( control );

        control.orientation.w    = 1;
        control.orientation.x    = 0;
        control.orientation.y    = 0;
        control.orientation.z    = 1;
        control.name             = name + "_move_z";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::MOVE_AXIS;
        int_marker.controls.push_back( control );
        control.name             = name + "_rotate_z";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::ROTATE_AXIS;
        int_marker.controls.push_back( control );

        return int_marker;
}

class InteractivePose
{
        interactive_markers::InteractiveMarkerServer server_;
        std::string                                  name_;
        Pose                                         pose_;
        std::string                                  frame_;
        tf2_ros::TransformBroadcaster                br_;

      public:
        InteractivePose( std::string name, std::string frame )
          : server_( name )
          , name_( name )
          , pose_()
          , frame_( frame )
          , br_()
        {
                show();
                setTF();
        }

        const Pose& getPose() const
        {
                return pose_;
        }

        void setPose( const Pose& p )
        {
                pose_ = p;
                setTF();
        }

      private:
        void show()
        {
                auto marker = makePoseMarker( name_, "world" );
                server_.insert( marker,
                                [this]( const visualization_msgs::InteractiveMarkerFeedbackConstPtr& feedback ) {
                                        this->onFeedback( feedback );
                                } );
                server_.applyChanges();
        }

        void onFeedback( const visualization_msgs::InteractiveMarkerFeedbackConstPtr& feedback )
        {

                pose_ = msgToPose( feedback->pose );
                setTF();
        }
        void setTF()
        {
                geometry_msgs::TransformStamped transformStamped;

                transformStamped.header.stamp            = ros::Time::now();
                transformStamped.header.frame_id         = "world";
                transformStamped.child_frame_id          = frame_;
                transformStamped.transform.translation.x = double( pose_.position[0] );
                transformStamped.transform.translation.y = double( pose_.position[1] );
                transformStamped.transform.translation.z = double( pose_.position[2] );
                transformStamped.transform.rotation      = quaternionToMsg( pose_.orientation );
                br_.sendTransform( transformStamped );
        }
};

}  // namespace schpin
