#include "base/triangle.h"
#include "collission.h"
#include "octree/occupancy.h"
#include "octree/tree.h"
#include "schpin_test/test_environment.h"
#include <gtest/gtest.h>

using namespace schpin;

TEST( Collision, SimplexTriangles )
{
        Triangle< 3 > t1{Point< 3 >( 0, 1, 0 ), Point< 3 >( 0, 0, 0 ), Point< 3 >( 1, 0, 0 )};
        Triangle< 3 > t2{Point< 3 >( 0, 0, 0 ), Point< 3 >( 1, 1, 0 ), Point< 3 >( 1, 0, 0 )};
        Triangle< 3 > t3{Point< 3 >( 0, 1, 0 ), Point< 3 >( 1, 1, 0 ), Point< 3 >( 1, 0, 0 )};
        Triangle< 3 > t4{Point< 3 >( 10, 1, 0 ), Point< 3 >( 11, 1, 0 ), Point< 3 >( 11, 0, 0 )};

        TriangleCollisionQuery< 3 > q1( t1 );
        TriangleCollisionQuery< 3 > q2( t2 );
        TriangleCollisionQuery< 3 > q3( t3 );
        TriangleCollisionQuery< 3 > q4( t4 );

        EXPECT_TRUE( collides< 3 >( q1, q2 ) );
        EXPECT_TRUE( collides< 3 >( q2, q3 ) );
        EXPECT_FALSE( collides< 3 >( q1, q3 ) );
        EXPECT_FALSE( collides< 3 >( q3, q4 ) );
}

TEST( Collision, SimplexTriangles2 )
{
        Triangle< 3 > t1{Point< 3 >( 1, 0, 0 ), Point< 3 >( 0, 1, 0 ), Point< 3 >( 1, 1, 0 )};
        Triangle< 3 > t2{Point< 3 >( 1, 0, 0 ), Point< 3 >( 2, 0, 0 ), Point< 3 >( 1, 1, 0 )};

        TriangleCollisionQuery< 3 > q1( t1 );
        TriangleCollisionQuery< 3 > q2( t2 );

        EXPECT_FALSE( collides< 3 >( q1, q2 ) );
}

using Tree = OcTree< EnvironmentOcTreeContent >;
// TODO: ugly lambdas! fix this
auto nodes_collides_f = []( const auto& lh_node_query, const auto& rh_node_query ) {
        return ( **lh_node_query )->getOccupancy() > ENV_MAX_OCCUPANCY / 2 &&
               ( **rh_node_query )->getOccupancy() > ENV_MAX_OCCUPANCY / 2;
};
const EnvironmentOcTreeContent TREE_CONTENT{ENV_MAX_OCCUPANCY, false};

TEST( TreeCollision, selfCollision )
{
        Tree tree{TREE_CONTENT, Length( 1 )};

        bool has_collided =
            treesNodesCollides< 3, CollisionMode::FULL >( OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                                                          OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                                                          nodes_collides_f );
        EXPECT_TRUE( has_collided ) << "same tree at same place should collide";

        Pose offset{Point< 3 >( 10, 0, 0 ), NEUTRAL_QUAT};

        has_collided =
            treesNodesCollides< 3, CollisionMode::FULL >( OcNodeFullCollisionQuery( tree.getProxyRoot(), offset ),
                                                          OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                                                          nodes_collides_f );
        EXPECT_FALSE( has_collided ) << "same tree moved by size bigger than itself should not collide";
}

std::vector< Point< 3 > > TREE_POINTS{Point< 3 >{0, 0, 0},
                                      Point< 3 >{1, 0, 0},
                                      Point< 3 >{0, 1, 0},
                                      Point< 3 >{0, 0, 1}};

TEST( TreeCollision, selfCollisionComplex )
{
        Tree tree{TREE_CONTENT, Length( 1 )};

        forEach( TREE_POINTS, [&]( const Point< 3 >& p ) { tree.expand( p ); } );

        bool has_collided =
            treesNodesCollides< 3, CollisionMode::FULL >( OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                                                          OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                                                          nodes_collides_f );
        EXPECT_TRUE( has_collided ) << "same tree at same place should collide";

        Pose offset_big{Point< 3 >( 10, 0, 0 ), NEUTRAL_QUAT};

        has_collided =
            treesNodesCollides< 3, CollisionMode::FULL >( OcNodeFullCollisionQuery( tree.getProxyRoot(), offset_big ),
                                                          OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                                                          nodes_collides_f );
        EXPECT_FALSE( has_collided ) << "tree moved by size bigger than his should not collide";

        Pose offset_small{Point< 3 >( 2.f - 0.001f, 0, 0 ), NEUTRAL_QUAT};

        has_collided =
            treesNodesCollides< 3, CollisionMode::FULL >( OcNodeFullCollisionQuery( tree.getProxyRoot(), offset_small ),
                                                          OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                                                          nodes_collides_f );
        EXPECT_TRUE( has_collided ) << "tree moved by size smaller than his should collide";
}

TEST( Environment, selfCollision )
{
        Tree tree{TREE_CONTENT, Length( 1 )};

        runIn( "environment/column_octree.json", [&]( std::tuple< Environment, Robot >&& envrob ) {
                Environment& env          = std::get< Environment >( envrob );
                bool         has_collided = treesNodesCollides< 3, CollisionMode::FULL >(
                    OcNodeFullCollisionQuery( tree.getProxyRoot() ),
                    OcNodeFullCollisionQuery( env.getTree().getProxyRoot() ),
                    nodes_collides_f );
                EXPECT_TRUE( has_collided );
        } );
}
