#include "planner/body.h"
#include "schpin_test/test_environment.h"
#include <gtest/gtest.h>
#include <tuple>

using namespace schpin;

template < CollisionMode mode >
void checkPath( bool should_be_valid, const Pose& from, const Pose& to, const Robot& robot, const Environment& env )
{
        ValidPathRes is_valid = isValidPath< mode >( from, to, robot, env );
        EXPECT_EQ( should_be_valid, bool( is_valid ) ) << "from          : " << from << "\n"
                                                       << "to            : " << to << "\n"
                                                       << "failed at pose: " << is_valid.failed_pose << "\n";
}

std::shared_ptr< Environment > ENV;
std::shared_ptr< Robot >       ROBOT;

template < typename T >
class BodyPlannerTest : public ::testing::Test
{
      public:
        float h = 0.05f;
        Pose  NEUTRAL_POSE{Point< 3 >{0, 0, h}, NEUTRAL_QUAT};
        Pose  SHIFTED_POSE{Point< 3 >{0.5, 0, h}, NEUTRAL_QUAT};
        Pose  CORNER_POSE{Point< 3 >{0.5, 0.5, h}, NEUTRAL_QUAT};

        static void SetUpTestCase()
        {
                if ( ENV && ROBOT ) {
                        return;
                }
                loadTestEnvRobot( "environment/column_octree.json", "robot/koke.json" )
                    .match(
                        [&]( std::tuple< Environment, Robot >&& data ) {
                                ENV   = std::make_unique< Environment >( std::get< Environment >( std::move( data ) ) );
                                ROBOT = std::make_unique< Robot >( std::get< Robot >( std::move( data ) ) );
                        },
                        [&]( TestError e ) { FAIL() << e; } );
        }
};

template < CollisionMode mode_arg >
struct PlannerConf
{
        static constexpr CollisionMode mode = mode_arg;
};

using BodyPlannerTestTypes = ::testing::Types< PlannerConf< CollisionMode::FAST >, PlannerConf< CollisionMode::FULL > >;

TYPED_TEST_CASE( BodyPlannerTest, BodyPlannerTestTypes );

TYPED_TEST( BodyPlannerTest, bodyCollidesWithEnvironment )
{
        static constexpr CollisionMode mode  = TypeParam::mode;
        Environment&                   env   = *ENV;
        Robot&                         robot = *ROBOT;

        EXPECT_TRUE( bodyCollidesWithEnvironment< mode >( robot.getBody(), this->NEUTRAL_POSE, env ) );
        EXPECT_FALSE( bodyCollidesWithEnvironment< mode >( robot.getBody(), this->SHIFTED_POSE, env ) );
        EXPECT_FALSE( bodyCollidesWithEnvironment< mode >( robot.getBody(), this->CORNER_POSE, env ) );
}

TYPED_TEST( BodyPlannerTest, allLegsTouchesWithEnvironment )
{
        static constexpr CollisionMode mode  = TypeParam::mode;
        Environment&                   env   = *ENV;
        Robot&                         robot = *ROBOT;

        EXPECT_TRUE( allLegsTouchesWithEnvironment< mode >( robot.getLegs(), this->NEUTRAL_POSE, env ) );
        EXPECT_TRUE( allLegsTouchesWithEnvironment< mode >( robot.getLegs(), this->SHIFTED_POSE, env ) );
        EXPECT_TRUE( allLegsTouchesWithEnvironment< mode >( robot.getLegs(), this->CORNER_POSE, env ) );
}

TYPED_TEST( BodyPlannerTest, isValidPath )
{
        static constexpr CollisionMode mode  = TypeParam::mode;
        Environment&                   env   = *ENV;
        Robot&                         robot = *ROBOT;

        std::vector< std::tuple< bool, Pose, Pose > > tests{
            {false, this->NEUTRAL_POSE, this->SHIFTED_POSE},
            {false, this->NEUTRAL_POSE, Pose{Point< 3 >{1, 1, this->h}, NEUTRAL_QUAT}},
            {false, this->NEUTRAL_POSE, Pose{Point< 3 >{-5, 0, this->h}, NEUTRAL_QUAT}},
            {true, this->SHIFTED_POSE, this->CORNER_POSE},
            {true, this->CORNER_POSE, Pose{Point< 3 >{0.3, 0.3, this->h}, NEUTRAL_QUAT}}};

        forEach( tests, [&]( const std::tuple< bool, Pose, Pose >& tup ) {
                checkPath< mode >( std::get< 0 >( tup ), std::get< 1 >( tup ), std::get< 2 >( tup ), robot, env );
                checkPath< mode >( std::get< 0 >( tup ), std::get< 2 >( tup ), std::get< 1 >( tup ), robot, env );
        } );
}

TYPED_TEST( BodyPlannerTest, simple )
{
        static constexpr CollisionMode mode = TypeParam::mode;

        PlannerContext context{this->SHIFTED_POSE, std::make_optional( this->CORNER_POSE ), ENV, ROBOT};

        BodyPlanner< mode > planner{std::move( context )};

        std::size_t max_steps = 50;
        std::size_t step      = *findIf( range( max_steps ), [&]( std::size_t ) -> bool {
                planner.tick();
                return planner.hasPlan();
        } );

        EXPECT_NE( max_steps, step );
}
