#include "config.h"
#include "graph/util/rviz.h"
#include "octree/rviz.h"
#include "planner.h"
#include "ros/ros.h"
#include "util/interactive_pose.h"
#include <filesystem>
#include <schpin_lib/console.h>
#include <tf2_ros/transform_broadcaster.h>

using namespace schpin;

struct MakeUntilPathEvent
{
};
struct LoadEnvironmentEvent
{
        std::string filename;
};
template < typename T >
struct BodyLevel
{
        T sub_cmd;
};
template < typename T >
struct RobotLevel
{
        T sub_cmd;
};
struct ShowLegGraphEvent
{
        std::optional< unsigned > index;
};
struct ShowWorkareaEvent
{
        std::optional< unsigned > index;
};
struct ShowBodyCollision
{
};
struct PrintModelEvent
{
};
struct ShowEnvironmentEvent
{
};
struct ClearEvent
{
};
struct SetCurrentPoseEvent
{
        std::optional< float > x;
        std::optional< float > y;
        std::optional< float > z;
};
struct SetGoalPoseEvent
{
        std::optional< float > x;
        std::optional< float > y;
        std::optional< float > z;
};

Console CONSOLE{
    //
    command< LoadEnvironmentEvent >( "load_env",
                                     "Loads enivornment",
                                     Arg< std::string >{"filename", "File which should be loaded"} ),
    command< ShowEnvironmentEvent >( "show_environment", "Show the environment" ),
    command< SetCurrentPoseEvent >( "set_current_pose",
                                    "Sets current pose of the robot",
                                    Arg< std::optional< float > >{"x", ""},
                                    Arg< std::optional< float > >{"y", ""},
                                    Arg< std::optional< float > >{"z", ""} ),
    command< SetGoalPoseEvent >( "set_goal_pose",
                                 "Sets the goal pose of motion planner",
                                 Arg< std::optional< float > >{"x", ""},
                                 Arg< std::optional< float > >{"y", ""},
                                 Arg< std::optional< float > >{"z", ""} ),
    command< ClearEvent >( "clear", "Clears all data" ),
    level< RobotLevel >(
        "robot",
        "Access the dataset around robot",
        command< ShowBodyCollision >( "show_body_collision", "Shows collision strucutre used for body" ),
        command< ShowLegGraphEvent >( "show_leg_graph",
                                      "Shows all graphs of legs",
                                      Arg< std::optional< unsigned > >{"index", "Optional index of the leg"} ),
        command< ShowWorkareaEvent >( "show_workarea",
                                      "Shows workarea of all legs",
                                      Arg< std::optional< unsigned > >{"index", "Optional index of the leg"} ),
        command< PrintModelEvent >( "print_model", "Prints data about the model of all legs" ),
        command< ClearEvent >( "clear", "Clears all data" ) ),
    level< BodyLevel >(
        "body_planning",
        "Access planner of the body movement",  //
        command< MakeVertexesEvent >( "make_vertexes",
                                      "Create specified number of vertexes ",
                                      Arg< unsigned >{"count", "Count of vertexes"} ),
        command< FindPathEvent >( "find_path", "Finds path in existing graph" ),
        command< MakeUntilPathEvent >( "make_until_path", "Creates vertexes and finds path until patch exists" ) )};

struct ManualPlanner
{
        Planner       planner;
        MarkerManager marker_manager;

        InteractivePose goal_marker{"goal", "goal"};
        InteractivePose current_marker{"current", "body"};

        bool onEvent( LoadEnvironmentEvent e )
        {
                std::ifstream env_data{e.filename};
                safeLoadJson< Environment >( env_data )
                    .match( [&]( Environment&& env ) { planner.setEnvironment( env ); },
                            [&]( JsonStorageError e ) { std::cerr << "Failed to load env due: " << e << std::endl; } );

                return bool( env_ptr );
        }

        bool onEvent( LevelEvent )
        {
                return false;
        }

        bool onEvent( SetGoalPoseEvent e )
        {
                Pose p;
                if ( e.x || e.y || e.z ) {
                        if ( !e.x && e.y && e.z ) {
                                std::cout << "you did not speficied all coordinates for current pose" << std::endl;
                                return false;
                        }
                        p = Pose{Point< 3 >{*e.x, *e.y, *e.z}, NEUTRAL_QUAT};
                } else {
                        p = goal_marker.getPose();
                }
                goal_marker.setPose( p );
                planner.setGoal( p );
                std::cout << "Setting goal pose to: " << p << std::endl;
                return false;
        }

        bool onEvent( SetCurrentPoseEvent e )
        {
                Pose p;
                if ( e.x || e.y || e.z ) {
                        if ( !e.x && e.y && e.z ) {
                                std::cout << "you did not speficied all coordinates for current pose" << std::endl;
                                return false;
                        }
                        p = Pose{Point< 3 >{*e.x, *e.y, *e.z}, NEUTRAL_QUAT};
                } else {
                        p = current_marker.getPose();
                }
                current_marker.setPose( p );

                RobotState rstate = planner.getRobotState();
                rstate.body.pose  = p;
                planner.setRobotState( rstate );
                std::cout << "Setting current pose to: " << p << std::endl;
                return false;
        }

        bool onEvent( ShowEnvironmentEvent )
        {
                std::cout << "Loading environment into rviz " << std::endl;
                auto msg = createOcVisualization(
                    "environment", planner.getEnvironment().getTree(), "world", 1, [&]( const auto& node ) -> bool {
                            return node->getOccupancy() > ENV_MAX_OCCUPANCY / 4;
                    } );
                marker_manager.publish( msg );

                msg = createOcVisualization(
                    "surface", planner.getEnvironment().getTree(), "world", 0.5, [&]( const auto& node ) -> bool {
                            return node->isSurface();
                    } );
                marker_manager.publish( msg );
                return false;  // ignored
        }

        bool onEvent( ExitEvent )
        {
                return true;
        }

        template < typename Variant >
        bool onEvent( BodyLevel< Variant > lvl )
        {
                std::visit( [&]( auto item ) { onBodyEvent( item ); }, lvl.sub_cmd );
                return false;
        }

        void showBodyGraph()
        {
                auto msg = createGraphVisualization(
                    planner.getBodyPlanner().getGraph(), "world", [&]( const auto& vertex ) -> Point< 3 > {
                            return vertex->pose.position;
                    } );
                marker_manager.publish( msg );
        }

        void onBodyEvent( FindPathEvent e )
        {
                planner.bodyEvent( e );
                Chain< BodyEdgeID, BodyVertexID > g_path = planner.getBodyPlanner().getPath();
                auto msg = createChainVisualization( g_path, "world", [&]( BodyVertexID vid ) -> Point< 3 > {
                        return planner.getBodyPlanner().getGraph().getVertex( vid )->pose.position;
                } );
                marker_manager.publish( msg );

                forEach( planner.getRobot().getLegs(), [&]( const Leg< 3 >& leg ) {
                        // TODO: factory needed

                        auto leg_view = planner.getLegSidepathView( leg.getID() );

                        auto msg = createSidepathVisualization(
                            leg_view.getPoints(), "world", leg.getFrame(), leg.getModel().getColor() );

                        marker_manager.publish( msg );
                } );
        }

        void onBodyEvent( MakeVertexesEvent e )
        {
                planner.bodyEvent( e );
                showBodyGraph();
        }

        void onBodyEvent( MakeUntilPathEvent )
        {
                while ( !planner.getBodyPlanner().hasPlan() ) {
                        planner.bodyEvent( MakeVertexesEvent{5} );
                        planner.bodyEvent( FindPathEvent{} );
                }
                showBodyGraph();
        }

        void onBodyEvent( LevelEvent )
        {
                showBodyGraph();
        }

        void onBodyEvent( ExitEvent )
        {
        }

        template < typename Variant >
        bool onEvent( RobotLevel< Variant > lvl )
        {
                std::visit( [&]( auto& item ) { robotEvent( item ); }, lvl.sub_cmd );
                return false;
        }

        void robotEvent( PrintModelEvent )
        {
                forEach( planner.getRobot().getLegs(), [&]( const Leg< 3 >& leg ) {
                        std::cout << leg.getName() << std::endl;
                        forEach( leg.getModel().getChain(), [&]( const ModelChain& chain ) {
                                std::cout << "axis: " << chain.axis << " offset: " << chain.offset
                                          << " angles: " << chain.min << "/" << chain.max << std::endl;
                        } );
                } );
        }

        void robotEvent( LevelEvent )
        {
                // TODO: make this focus rviz on the robot ?
        }
        void robotEvent( ExitEvent )
        {
        }

        void robotEvent( ClearEvent )
        {
                marker_manager.clear();
        }

        void robotEvent( ShowBodyCollision )
        {
                const auto& body = planner.getRobot().getBody();
                auto        msg  = createOcVisualization(
                    "body_collision", body.getCollisionTree(), body.getFrame(), 0.5, [&]( const auto& node ) {
                            return node->getOccupancy() > ( OC_MAX_OCCUPANCY / 2.f + OC_MIN_OCCUPANCY / 2.f );
                    } );
                marker_manager.publish( msg );
        }

        void robotEvent( ShowLegGraphEvent e )
        {
                forEach( range( planner.getRobot().getLegs().size() ), [&]( std::size_t i ) {
                        const Leg< 3 >& leg = planner.getRobot().getLegs()[i];
                        if ( e.index && i != *e.index ) {
                                return;
                        }
                        auto msg = createGraphVisualization(
                            leg.getDstar().getGraph(), leg.getFrame(), [&]( const auto& vertex ) -> Point< 3 > {
                                    return leg.getModel().tipPosition( vertex->angles );
                            } );
                        marker_manager.publish( msg );
                } );
        }

        void robotEvent( ShowWorkareaEvent e )
        {
                forEach( range( planner.getRobot().getLegs().size() ), [&]( std::size_t i ) {
                        const Leg< 3 >& leg = planner.getRobot().getLegs()[i];
                        if ( e.index && i != *e.index ) {
                                return;
                        }
                        auto msg = createOcVisualization(
                            leg.getFrame() + "-workarea",
                            leg.getWorkarea(),
                            leg.getFrame(),
                            0.5,
                            [&]( const auto& node ) {
                                    return node->getOccupancy() > ( OC_MAX_OCCUPANCY / 2.f + OC_MIN_OCCUPANCY / 2.f );
                            } );
                        marker_manager.publish( msg );
                } );
        }
};

int main( int argc, char* argv[] )
{
        ros::init( argc, argv, "manual" );

        ros::NodeHandle   handle{};
        ros::AsyncSpinner spinner( 4 );  // Use 4 threads
        spinner.start();

        bool should_exit = false;

        PlannerContext context;
        context.goal = std::optional< Pose >();

        ROSConfig cfg;
        loadRobotROS( handle, cfg )
            .match( [&]( Robot&& robot ) { context.robot_ptr = std::make_shared< Robot >( std::move( robot ) ); },
                    [&]( RobotError e ) { std::cerr << "Failed to load robot due: " << e << std::endl; } );

        if ( !context.robot_ptr ) {
                return 0;
        }

        ManualPlanner planner{Planner{std::move( context )}, MarkerManager{handle}};

        ros::Duration dur( 0.5 );
        dur.sleep();

        should_exit = false;

        while ( !should_exit ) {
                auto console_res = CONSOLE.promptMatch( console_res );

                if ( console_res.item ) {
                        should_exit =
                            std::visit( [&]( auto item ) { return planner.onEvent( item ); }, *console_res.item );
                }
        }
        return 0;
}
