/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 *
 * This file is part of Schpin.
 *
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "include_barrier/tf/tf.h"
#include "util/physical_quantity.h"
#include <tf/transform_datatypes.h>

#pragma once

namespace schpin
{
typedef ros::Duration Duration;
typedef ros::Time     Time;

typedef tf::Vector3    Vector3;
typedef tf::Quaternion Quaternion;

namespace anchor
{
  struct World;
  struct Leg;
};

template < typename RelativeAnchor >
class Point : public Vector3
{
public:
  Point()
    : Vector3()
  {
  }
  Point( double x, double y, double z )
    : Vector3( x, y, z )
  {
  }

  Point( const Vector3& other )
    : Vector3( other )
  {
  }

  Point& operator=( Point other )
  {
    using std::swap;
    swap( other, *this );
    return *this;
  }
};

//* pose in space
/**
 * Contains position and orientation for space, thus defining proper 6DOF
 *
 */
struct Pose
{
  Quaternion             orientation{1, 0, 0, 0}; /**< orientation in space */
  Point< anchor::World > position{0, 0, 0};       /**< position in space */

  void stackUp( const Pose& other )
  {
    position    = quatRotate( orientation, other.position ) + position;
    orientation = orientation + other.orientation;
  }
};

class Length : public util::PhysicalQuantity< Length >
{
  using util::PhysicalQuantity< Length >::PhysicalQuantity;
  static constexpr char unit[] = "m";
};

class Velocity : public util::PhysicalQuantity< Velocity >
{
  using util::PhysicalQuantity< Velocity >::PhysicalQuantity;
  static constexpr char unit[] = "m/s";
};

class Angle : public util::PhysicalQuantity< Angle >
{
  using PhysicalQuantity< Angle >::PhysicalQuantity;
  static constexpr char unit[] = "rad";
};

Duration operator*( Velocity velocity, Length length )
{
  return Duration( double( velocity ) * double( length ) );
}
}
