/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "../basic.h"
#include "../leg.h"
#include "./graph/edges.h"
#include "./graph/vertex.h"
#include <queue>
#include <set>
#include <tf/tf.h>

#pragma once

namespace schpin
{
namespace leg
{

  /*
   * Container for BFS style exploration of graph.
   */
  class ExploreContext
  {
    std::queue< Vertex > queue_;
    std::set< VertexID > open_;

  public:
    void enqueue( Vertex vertex )
    {
      queue_.push( vertex );
      open_.insert( vertex.getID() );
    }

    bool queueEmpty()
    {
      return queue_.empty();
    }

    bool isOpen( VertexID vid )
    {
      return open_.find( vid ) != open_.end();
    };

    Vertex popQueue()
    {
      Vertex tmp = queue_.front();
      queue_.pop();
      return tmp;
    };
  };

  /*
   * approx_* - approximated line segment
   * edge_* - edge for which weight is calculated
   */
  double calcApproxWeight( const Point< anchor::Leg >& approx_start,
                           const Point< anchor::Leg >& approx_goal,
                           const Point< anchor::Leg >& edge_start,
                           const Point< anchor::Leg >& edge_goal )
  {

    ROS_ASSERT( !( approx_start == edge_start && approx_goal == edge_goal ) );  // if this case passes, it results in
                                                                                // 'NaN' in calculation

    Vector3 edge_direction   = edge_start - edge_goal;
    Vector3 approx_direction = approx_start - approx_goal;

    Angle angle( edge_direction.angle( approx_direction ) );

    double weight = 1 / ( approx_start.distance( edge_start ) + approx_goal.distance( edge_goal ) );

    weight *= std::max< double >( 0, cos( angle ) );

    return weight;
  }

  /*
   * Approximate duration density in actual 'leg'. Makes exploration of graph from 'start' vertex, explores every edge
   * that has target within 'range' of 'goal_point'. Makes weighted average of velocity on those edges.
   */
  Velocity approxVelocity(
      const Leg& leg, Vertex start, Point< anchor::Leg > start_point, Point< anchor::Leg > goal_point, Length range )
  {

    const Graph& graph = leg.getGraph();

    ExploreContext context;

    context.enqueue( start );

    // used for weighted average
    double sum        = 0;
    double weight_sum = 0;

    while ( !context.queueEmpty() )
    {
      Vertex actual = context.popQueue();

      Graph::ConstOutgoingEdgeIterator begin = graph.outgoingEdgeBegin( actual );
      Graph::ConstOutgoingEdgeIterator end   = graph.outgoingEdgeEnd( actual );
      for ( ; begin != end; ++begin )
      {
        const Edge&          edge   = *begin;
        Point< anchor::Leg > target = leg.getPoint( edge.getTargetID() );
        Point< anchor::Leg > source = leg.getPoint( edge.getSourceID() );

        if ( Length( target.distance( goal_point ) ) > range )
        {
          continue;
        }
        else if ( source == start_point && target == goal_point )  // bingo, same edge!
        {
          double val = edge.getData().getDuration().toSec() / target.distance( source );
          return Velocity( val );
        }

        if ( !context.isOpen( edge.getTargetID() ) )
        {
          context.enqueue( graph.getVertex( edge.getTargetID() ) );
        }

        double length = target.distance( source );
        double weight = calcApproxWeight( start_point, goal_point, source, target );

        sum += weight * ( edge.getData().getDuration().toSec() / length );
        weight_sum += weight;
      }
    }

    ROS_ASSERT( sum != 0 );
    return Velocity( sum / weight_sum );
  }

  /*
   * Uses greedy approach to find closest vertex to point 'center' in 'leg', starting from vertex 'start'
   */
  Vertex greedyGotoClosest( const Leg& leg, const Point< anchor::Leg >& center, const Vertex& start )
  {
    VertexID                         actual_id       = start.getID();
    Length                           actual_distance = Length( 0 );
    Graph::ConstOutgoingEdgeIterator edge_iter =
        leg.getGraph().outgoingEdgeEnd( leg.getGraph().getVertex( actual_id ) );

    // each iteration tries to find edge with smaller 'distance' to center point
    // iteration ends when no such edge is found and actual vertex is returned
    do
    {
      if ( edge_iter != leg.getGraph().outgoingEdgeEnd( leg.getGraph().getVertex( actual_id ) ) )
      {
        actual_id = edge_iter->getTargetID();
      }

      const Vertex actual = leg.getGraph().getVertex( actual_id );
      actual_distance     = Length( leg.getPoint( actual_id ).distance( center ) );

      edge_iter = std::find_if(
          leg.getGraph().outgoingEdgeBegin( actual ), leg.getGraph().outgoingEdgeEnd( actual ), [&]( Edge edge ) {
            Length new_distance( leg.getPoint( edge.getTargetID() ).distance( center ) );
            return new_distance < actual_distance;
          } );

    } while ( edge_iter != leg.getGraph().outgoingEdgeEnd( leg.getGraph().getVertex( actual_id ) ) );

    return leg.getGraph().getVertex( actual_id );
  };

  Length edgeLength( const Leg& leg, const Edge& edge )
  {
    Point< anchor::Leg > source = leg.getPoint( edge.getSourceID() );
    Point< anchor::Leg > target = leg.getPoint( edge.getTargetID() );

    return Length( source.distance( target ) );
  }

  /*
   * Returns maximal edge length around given vertex
   */
  Length localEdgeLength( const Leg& leg, const Vertex& vertex )
  {

    auto iter = std::max_element(
        leg.getGraph().outgoingEdgeBegin( vertex ),
        leg.getGraph().outgoingEdgeEnd( vertex ),
        [&]( const Edge lh, const Edge rh ) { return edgeLength( leg, lh ) < edgeLength( leg, rh ); } );

    return edgeLength( leg, *iter );
  }

  Point< anchor::Leg > advancePoint( const Point< anchor::Leg >& start, const Vector3& direction, Length distance )
  {
    return start + direction * ( double( distance ) / direction.length() );
  }

  Duration approxPointMovementDuration( const Leg&                  leg,
                                        const Point< anchor::Leg >& start,
                                        const Vertex&               start_hint,
                                        const Point< anchor::Leg >& goal )
  {
    Duration             result{0};
    Point< anchor::Leg > actual_point      = start;
    Vertex               actual_vertex     = greedyGotoClosest( leg, actual_point, start_hint );
    Vector3              direction         = goal - start;
    Length               local_edge_length = localEdgeLength( leg, actual_vertex );
    Length section_length = Length( std::min< double >( double( local_edge_length ), actual_point.distance( goal ) ) );
    Point< anchor::Leg > next_point = advancePoint( actual_point, direction, section_length );

    // TODO: get rid of Length destructors
    // epsilon math should be used for this maybe?
    while ( actual_point.distance( goal ) > 0 )
    {
      result +=
          approxVelocity( leg, actual_vertex, actual_point, next_point, local_edge_length * 1.5 ) * section_length;

      actual_point      = next_point;
      actual_vertex     = greedyGotoClosest( leg, actual_point, actual_vertex );
      local_edge_length = localEdgeLength( leg, actual_vertex );
      section_length    = Length( std::min< double >( double( local_edge_length ), actual_point.distance( goal ) ) );
      next_point        = advancePoint( actual_point, direction, section_length );
    }
    return result;
  }
}
}
