/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "../util.h"
#include "../util/lazy_iterator.h"
#include "graph/base.h"
#include "graph/edges.h"
#include "graph/vertex.h"
#include "ros/assert.h"
#include <boost/signals2.hpp>

#pragma once

namespace schpin
{
namespace leg
{


  class Graph
  {
  public:
    class VertexIterator : public LazyIterator< VertexIterator, Vertex, VertexID >
    {
      std::reference_wrapper< Graph > graph_;

    public:
      VertexIterator( Graph& graph, VertexID vid )
        : LazyIterator( vid )
        , graph_( graph )
      {
      }

      Vertex operator()( const VertexID& actual )
      {
        return graph_.get().getVertex( actual );
      }
    };

    class ConstVertexIterator : public LazyIterator< ConstVertexIterator, const Vertex, VertexID >
    {
      std::reference_wrapper< const Graph > graph_;

    public:
      ConstVertexIterator( const Graph& graph, VertexID vid )
        : LazyIterator( vid )
        , graph_( graph )
      {
      }

      const Vertex operator()( const VertexID& actual )
      {
        return graph_.get().getVertex( actual );
      }
    };

    class OutgoingEdgeIterator : public LazyIterator< OutgoingEdgeIterator, Edge, EdgeID >
    {
      std::reference_wrapper< Graph > graph_;

    public:
      OutgoingEdgeIterator( Graph& graph, EdgeID eid )
        : LazyIterator( eid )
        , graph_( graph )
      {
      }

      Edge operator()( const EdgeID& actual )
      {
        return graph_.get().getOutgoingEdge( actual );
      }
    };

    class ConstOutgoingEdgeIterator : public LazyIterator< ConstOutgoingEdgeIterator, const Edge, EdgeID >
    {
      std::reference_wrapper< const Graph > graph_;

    public:
      ConstOutgoingEdgeIterator( const Graph& graph, EdgeID eid )
        : LazyIterator( eid )
        , graph_( graph )
      {
      }

      const Edge operator()( const EdgeID& actual )
      {
        return graph_.get().getOutgoingEdge( actual );
      }
    };

    class IngoingEdgeIterator : public LazyIterator< IngoingEdgeIterator, Edge, EdgeID >
    {
      std::reference_wrapper< Graph > graph_;

    public:
      IngoingEdgeIterator( Graph& graph, EdgeID eid )
        : LazyIterator( eid )
        , graph_( graph )
      {
      }

      Edge operator()( const EdgeID& actual )
      {
        return graph_.get().getIngoingEdge( actual );
      }
    };

  private:
    typedef util::IdentificatorVector< OutgoingEdge, EdgeID >    OutgoingEdgeContainer;
    typedef util::IdentificatorVector< IngoingEdge, EdgeID >     IngoingEdgeContainer;
    typedef util::IdentificatorVector< VertexStorage, VertexID > VertexContainer;

  public:
    typedef Vertex   Vertex;
    typedef VertexID VertexID;
    typedef EdgeData EdgeData;
    typedef Edge     Edge;
    typedef Price    Price;
    /*
     * When graph is constructed, we pernamently set maximal number of edges per
     * vertex => we have to know it
     * beforehand
     */
    Graph( EdgeCount max_edges )
      : max_edge_count_( max_edges )
      , vertexes_()
      , outgoing_edges_()
      , ingoing_edges_()
      , size_signal_(){};

  private:
    EdgeCount max_edge_count_;

    VertexContainer       vertexes_;
    OutgoingEdgeContainer outgoing_edges_;
    IngoingEdgeContainer  ingoing_edges_;

    boost::signals2::signal< void( void ) > size_signal_;

    Edge getOutgoingEdge( EdgeID eid )
    {
      OutgoingEdge& oedge = outgoing_edges_[eid];
      IngoingEdge&  iedge = ingoing_edges_[oedge.getIngoingEdgeID()];

      return Edge( iedge, oedge.getIngoingEdgeID(), oedge, eid );
    }

    const Edge getOutgoingEdge( const EdgeID eid ) const
    {
      OutgoingEdge& oedge = const_cast< OutgoingEdge& >( outgoing_edges_[eid] );
      IngoingEdge&  iedge = const_cast< IngoingEdge& >( ingoing_edges_[oedge.getIngoingEdgeID()] );

      return Edge( iedge, oedge.getIngoingEdgeID(), oedge, eid );
    }

    Edge getIngoingEdge( EdgeID eid )
    {
      IngoingEdge&  iedge = ingoing_edges_[eid];
      OutgoingEdge& oedge = outgoing_edges_[iedge.getOutgoingEdgeID()];

      return Edge( iedge, eid, oedge, iedge.getOutgoingEdgeID() );
    }

  public:
    EdgeCount getMaxEdgeCount()
    {
      return max_edge_count_;
    }

    EdgeCount getEdgeCount() const
    {
      EdgeCount res;
      std::for_each( vertexes_.cbegin(), vertexes_.cend(), [&]( const VertexStorage& storage ) {
        res += storage.getOutgoingEdgeCount();
      } );

      return res;
    };

    /*
     * Prints memory usage of graph and number of vertexes into log.
     */
    void logStatistics() const
    {
      DEBUG_LOG( "Graph statistics:" );
      DEBUG_LOG( "Maximal edge count: " << max_edge_count_ );

      DEBUG_LOG( "vertexes :" );
      vertexes_.logStatistics();

      DEBUG_LOG( "Outgoing edges:" );
      outgoing_edges_.logStatistics();

      DEBUG_LOG( "Ingoing edges:" );
      ingoing_edges_.logStatistics();
    }

    Vertex getVertex( const VertexID& vid )
    {
      return Vertex( vid, vertexes_[vid] );
    }

    const Vertex getVertex( const VertexID& vid ) const
    {
      // There is no other way to construct Vertex, or more KISS way
      return Vertex( vid, const_cast< VertexStorage& >( vertexes_[vid] ) );
    }

    template < typename Func >
    boost::signals2::connection connectSizeSignal( Func f )
    {
      return size_signal_.connect( f );
    }

    VertexID getVertexCount() const
    {
      return vertexes_.size();
    }

    VertexIterator vertexBegin()
    {
      return VertexIterator( *this, VertexID( 0 ) );
    }

    ConstVertexIterator vertexBegin() const
    {
      return ConstVertexIterator( *this, VertexID( 0 ) );
    };

    VertexIterator vertexEnd()
    {
      return VertexIterator( *this, getVertexCount() );
    }

    ConstVertexIterator vertexEnd() const
    {
      return ConstVertexIterator( *this, getVertexCount() );
    }

    /*
     * Reserves memory for 'vertex_count' vertexes and appropiate memory for edges
     */
    void reserve( const VertexID& vertex_count )
    {
      EdgeID edge_count( max_edge_count_ * vertex_count.get() );
      vertexes_.reserve( vertex_count );
      outgoing_edges_.reserve( edge_count );
      ingoing_edges_.reserve( edge_count );
    }

    /*
     * Shrinks all internal containers to fit data
     */
    void shrinkToFit()
    {
      vertexes_.shrinkToFit();
      outgoing_edges_.shrinkToFit();
      ingoing_edges_.shrinkToFit();
    }

    void insertEdge( const Vertex& source, const Vertex& target, const EdgeData& edata )
    {
      ROS_ASSERT( vertexes_[source.getID()].getOutgoingEdgeCount() < max_edge_count_ );
      ROS_ASSERT( vertexes_[target.getID()].getIngoingEdgeCount() < max_edge_count_ );

      EdgeID ingoing_edge_id  = getEdgeIndex( target, vertexes_[target.getID()].getIngoingEdgeCount() );
      EdgeID outgoing_edge_id = getEdgeIndex( source, vertexes_[source.getID()].getOutgoingEdgeCount() );

      vertexes_[source.getID()].incrementOutgoingEdgeCount();
      vertexes_[target.getID()].incrementIngoingEdgeCount();

      outgoing_edges_[outgoing_edge_id] = OutgoingEdge( ingoing_edge_id, edata, target.getID() );
      ingoing_edges_[ingoing_edge_id]   = IngoingEdge( outgoing_edge_id, source.getID() );
    }

    /*
     * Creates new vertex and allocates max_edge_count edges for it
     */
    Vertex emplaceVertex()
    {
      VertexID vid = vertexes_.emplaceBack();

      std::fill_n( outgoing_edges_.makeBackInserter(), std::size_t( max_edge_count_ ), OutgoingEdge() );
      std::fill_n( ingoing_edges_.makeBackInserter(), std::size_t( max_edge_count_ ), IngoingEdge() );

      size_signal_();

      return getVertex( vid );
    }

    void removeEdge( Edge& edge )
    {
      IngoingEdge&  iedge = edge.refIngoingEdge();
      OutgoingEdge& oedge = edge.refOutgoingEdge();

      Vertex target = getVertex( oedge.getTargetID() );
      Vertex source = getVertex( iedge.getSourceID() );

      // check for ingoingEdgeID validity
      ROS_ASSERT( iedge.getOutgoingEdgeID() == edge.getOutgoingEdgeID() );
      // check for outgoingEdgeID validity
      ROS_ASSERT( oedge.getIngoingEdgeID() == edge.getIngoingEdgeID() );

      // swap ingoing edge in it's storage
      swapIngoingEdges( edge.getIngoingEdgeID(), getLastIngoingEdgeID( target ) );
      // swap outgoing edge in it's storage
      swapOutgoingEdges( edge.getOutgoingEdgeID(), getLastOutgoingEdgeID( source ) );

      // check for ingoingEdgeID validity
      ROS_ASSERT( iedge.getOutgoingEdgeID() == edge.getOutgoingEdgeID() );
      // check for outgoingEdgeID validity
      ROS_ASSERT( oedge.getIngoingEdgeID() == edge.getIngoingEdgeID() );

      // deincrement vectors
      vertexes_[target.getID()].decrementIngoingEdgeCount();
      vertexes_[source.getID()].decrementOutgoingEdgeCount();
    }

    // maybe we want iterator ranges for edgeid's, not only the structures itselfs
    void purgeIngoingEdges( const Vertex& vertex )
    {
      std::for_each(
          ingoingEdgeBegin( vertex ), ingoingEdgeEnd( vertex ), [this]( Edge& edge ) { this->removeEdge( edge ); } );

      ROS_ASSERT( vertexes_[vertex.getID()].getIngoingEdgeCount() == 0 );
    };

    IngoingEdgeIterator ingoingEdgeBegin( const Vertex& vertex )
    {

      EdgeID eid = getEdgeIndex( vertex, EdgeCount( 0 ) );
      return IngoingEdgeIterator( *this, eid );
    }
    IngoingEdgeIterator ingoingEdgeEnd( const Vertex& vertex )
    {
      EdgeID eid = getEdgeIndex( vertex, vertexes_[vertex.getID()].getIngoingEdgeCount() );
      return IngoingEdgeIterator( *this, eid );
    }

    OutgoingEdgeIterator outgoingEdgeBegin( const Vertex& vertex )
    {
      EdgeID eid = getEdgeIndex( vertex, EdgeCount( 0 ) );
      return OutgoingEdgeIterator( *this, eid );
    }

    ConstOutgoingEdgeIterator outgoingEdgeBegin( const Vertex& vertex ) const
    {
      EdgeID eid = getEdgeIndex( vertex, EdgeCount( 0 ) );
      return ConstOutgoingEdgeIterator( *this, eid );
    }

    OutgoingEdgeIterator outgoingEdgeEnd( const Vertex& vertex )
    {
      EdgeID eid = getEdgeIndex( vertex, vertexes_[vertex.getID()].getOutgoingEdgeCount() );
      return OutgoingEdgeIterator( *this, eid );
    }

    ConstOutgoingEdgeIterator outgoingEdgeEnd( const Vertex& vertex ) const
    {
      EdgeID eid = getEdgeIndex( vertex, vertexes_[vertex.getID()].getOutgoingEdgeCount() );
      return ConstOutgoingEdgeIterator( *this, eid );
    }

    Vertex copyVertex( const Vertex& vertex )
    {
      Vertex copy = emplaceVertex();

      copy.refData() = vertex.getData();

      size_signal_();

      return copy;
    }

    std::vector< Vertex > copyVertex( const Vertex& vertex, std::size_t count )
    {
      std::vector< Vertex > res;

      for ( std::size_t i = 0; i < count; ++i )
      {
        res.push_back( copyVertex( vertex ) );
      }

      return res;
    }

  private:
    void swapIngoingEdges( EdgeID lhid, EdgeID rhid )
    {
      IngoingEdge& lh = ingoing_edges_[lhid];
      IngoingEdge& rh = ingoing_edges_[rhid];

      using std::swap;
      swap( outgoing_edges_[lh.getOutgoingEdgeID()].refIngoingEdgeID(),
            outgoing_edges_[rh.getOutgoingEdgeID()].refIngoingEdgeID() );
      swap( lh, rh );
    }

    EdgeID getLastIngoingEdgeID( const Vertex& vertex ) const
    {
      return getEdgeIndex( vertex, vertexes_[vertex.getID()].getIngoingEdgeCount() - 1 );
    }

    void swapOutgoingEdges( EdgeID lhid, EdgeID rhid )
    {
      OutgoingEdge& lh = outgoing_edges_[lhid];
      OutgoingEdge& rh = outgoing_edges_[rhid];

      using std::swap;
      swap( ingoing_edges_[lh.getIngoingEdgeID()].refOutgoingEdgeID(),
            ingoing_edges_[rh.getIngoingEdgeID()].refOutgoingEdgeID() );
      swap( lh, rh );
    }

    EdgeID getLastOutgoingEdgeID( const Vertex& vertex ) const
    {
      return getEdgeIndex( vertex, vertexes_[vertex.getID()].getOutgoingEdgeCount() - 1 );
    }

    EdgeID getEdgeIndex( const Vertex& vertex, const EdgeCount& edge_number ) const
    {
      return vertex.getID().get() * max_edge_count_ + edge_number;
    }
  };

}  // ns leg
}  // ns schpin
