/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "../../basic.h"
#include "base.h"

#pragma once

namespace schpin
{
namespace leg
{

  class EdgeData
  {
    Duration duration_{0};

  public:
    EdgeData() = default;
    EdgeData( Duration duration )
      : duration_( duration )
    {
    }

    void setDuration( Duration duration )
    {
      duration_ = duration;
    }

    const Duration& getDuration() const
    {
      return duration_;
    }

    Price getPrice() const
    {
      return Price( getDuration().toSec() );
    }
  };

  class OutgoingEdge
  {
    EdgeData data_{};
    VertexID target_{0};
    EdgeID   ingoing_edge_id_;

  public:
    OutgoingEdge()
      : OutgoingEdge( EdgeID{0}, EdgeData{}, VertexID{0} )
    {
    }

    OutgoingEdge( EdgeID ieid, EdgeData data, VertexID target )
      : data_( data )
      , target_( target )
      , ingoing_edge_id_( ieid )
    {
    }

    EdgeID& refIngoingEdgeID()
    {
      return ingoing_edge_id_;
    }

    const EdgeID& getIngoingEdgeID() const
    {
      return ingoing_edge_id_;
    }

    EdgeData& refData()
    {
      return data_;
    }

    const EdgeData& getData() const
    {
      return data_;
    }

    const VertexID& getTargetID() const
    {
      return target_;
    }
  };

  std::ostream& operator<<( std::ostream& os, const OutgoingEdge& edge )
  {
    os << "Outgoing Edge: target: " << edge.getTargetID();
    return os;
  }

  class IngoingEdge
  {
    EdgeID   outgoing_edge_id_;
    VertexID source_;

  public:
    IngoingEdge()
      : IngoingEdge( EdgeID{0}, VertexID{0} )
    {
    }

    IngoingEdge( EdgeID outgoing_edge_id, VertexID source )
      : outgoing_edge_id_( outgoing_edge_id )
      , source_( source )
    {
    }

    const EdgeID& getOutgoingEdgeID() const
    {
      return outgoing_edge_id_;
    }

    EdgeID& refOutgoingEdgeID()
    {
      return outgoing_edge_id_;
    }

    const VertexID& getSourceID() const
    {
      return source_;
    }
  };

  std::ostream& operator<<( std::ostream& os, const IngoingEdge& edge )
  {
    os << "Ingoing Edge: source: " << edge.getSourceID();
    return os;
  }

  class Edge
  {
    std::reference_wrapper< OutgoingEdge > oedge_;
    EdgeID                                 oedge_id_;
    std::reference_wrapper< IngoingEdge >  iedge_;
    EdgeID                                 iedge_id_;

    friend class Graph;

    Edge( IngoingEdge& iedge, EdgeID iedge_id, OutgoingEdge& oedge, EdgeID oedge_id )
      : oedge_( oedge )
      , oedge_id_( oedge_id )
      , iedge_( iedge )
      , iedge_id_( iedge_id )
    {
    }

    OutgoingEdge& refOutgoingEdge()
    {
      return oedge_.get();
    }

    IngoingEdge& refIngoingEdge()
    {
      return iedge_.get();
    }

    EdgeID getOutgoingEdgeID()
    {
      return oedge_id_;
    }

    EdgeID getIngoingEdgeID()
    {
      return iedge_id_;
    }

  public:
    const VertexID& getSourceID() const
    {
      return iedge_.get().getSourceID();
    }

    const VertexID& getTargetID() const
    {
      return oedge_.get().getTargetID();
    }

    // TODO: remove this
    EdgeData& refData()
    {
      return oedge_.get().refData();
    }

    const EdgeData& getData() const
    {
      return oedge_.get().getData();
    }

    Price getPrice() const
    {
      return getData().getPrice();
    };
  };
}
}
