/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "base.h"

#pragma once

namespace schpin
{
namespace leg
{

  class VertexData
  {
    Price       g_{0};
    Price       rhs_{0};
    JointStates joint_states_;
    bool        is_open_;

    friend bool operator==( const VertexData&, const VertexData& );

  public:
    void setClosed()
    {
      is_open_ = false;
    }

    void setOpen()
    {
      is_open_ = true;
    }

    bool isOpen() const
    {
      return is_open_;
    }

    void setG( Price price )
    {
      g_ = price;
    }

    void setRhs( Price price )
    {
      rhs_ = price;
    }

    Price getG() const
    {
      return g_;
    }

    Price getRhs() const
    {
      return rhs_;
    }

    const JointStates& getJointStates() const
    {
      return joint_states_;
    }
    void setJointStates( JointStates& states )
    {
      joint_states_ = states;
    }
  };

  bool operator==( const VertexData& lh, const VertexData& rh )
  {
    return lh.g_ == rh.g_ && lh.g_ == rh.g_ && lh.joint_states_ == rh.joint_states_;
  }

  class VertexStorage
  {
    VertexData data_;
    EdgeCount  outgoing_edge_count_{0};
    EdgeCount  ingoing_edge_count_{0};

    friend bool operator==( const VertexStorage&, const VertexStorage& );

  public:
    VertexData& refData()
    {
      return data_;
    }
    const VertexData& getData() const
    {
      return data_;
    }

    const EdgeCount& getOutgoingEdgeCount() const
    {
      return outgoing_edge_count_;
    }

    void incrementOutgoingEdgeCount()
    {
      ++outgoing_edge_count_;
    }

    void decrementOutgoingEdgeCount()
    {
      --outgoing_edge_count_;
    }

    void setOutgoingEdgeCount( EdgeCount ecount )
    {
      outgoing_edge_count_ = ecount;
    }

    const EdgeCount& getIngoingEdgeCount() const
    {
      return ingoing_edge_count_;
    }

    void incrementIngoingEdgeCount()
    {
      ++ingoing_edge_count_;
    }

    void decrementIngoingEdgeCount()
    {
      --ingoing_edge_count_;
    };

    void setIngoingEdgeCount( EdgeCount ecount )
    {
      ingoing_edge_count_ = ecount;
    }
  };

  bool operator==( const VertexStorage& lh, const VertexStorage& rh )
  {
    return lh.data_ == rh.data_ && lh.outgoing_edge_count_ == rh.outgoing_edge_count_ &&
           lh.ingoing_edge_count_ == rh.ingoing_edge_count_;
  }

  class Vertex
  {

    std::reference_wrapper< VertexStorage > storage_;
    VertexID                                id_;

    friend class Graph;

    Vertex( VertexID vid, VertexStorage& storage )
      : storage_( storage )
      , id_( vid )
    {
    }

    VertexData& refData()
    {
      return storage_.get().refData();
    }

    const VertexData& getData() const
    {
      return storage_.get().getData();
    }

    friend bool operator==( const Vertex&, const Vertex& );

  public:
    const JointStates& getJointStates() const
    {
      return getData().getJointStates();
    }

    void setJointStates( JointStates states )
    {
      refData().setJointStates( states );
    }

    void setClosed()
    {
      refData().setClosed();
    }

    void setOpen()
    {
      refData().setOpen();
    }

    void maximalizeG()
    {
      refData().setG( std::numeric_limits< Price >::max() );
    }
    void setG( Price price )
    {
      refData().setG( price );
    }

    Price getG() const
    {
      return getData().getG();
    }

    Price getRhs() const
    {
      return getData().getRhs();
    }

    void maximalizeRhs()
    {
      refData().setRhs( std::numeric_limits< Price >::max() );
    }
    void setRhs( Price price )
    {
      refData().setRhs( price );
    }

    void improveRhs( Price price )
    {
      setRhs( std::min( getRhs(), price ) );
    }

    const VertexID& getID() const
    {
      return id_;
    }

    bool isOpen() const
    {
      return getData().isOpen();
    }
  };

  bool operator==( const Vertex& lh, const Vertex& rh )
  {
    return lh.getID() == rh.getID();
  }

  bool operator<( const Vertex& lh, const Vertex& rh )
  {
    return lh.getID() < rh.getID();
  }
}
}
