/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "../leg.h"
#include "./axis.h"
#include "waypoint.h"
#include <tuple>

#pragma once

namespace schpin
{
namespace solver
{

  class BodyState
  {
    typedef std::tuple< XAxis::Point,      //
                        YAxis::Point,      //
                        ZAxis::Point,      //
                        YawAxis::Point,    //
                        PitchAxis::Point,  //
                        RollAxis::Point >
              Container;
    Container container_;

  public:
    template < typename Axis >
    const typename Axis::Point& getAxisPoint() const
    {
      return std::get< Axis::Point >( container_ );
    }

    template < typename Axis >
    void setAxisPoint( typename Axis::Point point )
    {
      std::get< Axis::Point >( container_ ) = point;
    }
  };

  Pose projectStateToPose( const BodyState& state, const Config& config )
  {
    double x = config.getValue< XAxis >( state.getAxisPoint< XAxis >() );
    double y = config.getValue< YAxis >( state.getAxisPoint< YAxis >() );
    double z = config.getValue< ZAxis >( state.getAxisPoint< ZAxis >() );

    double yaw   = config.getValue< YawAxis >( state.getAxisPoint< YawAxis >() );
    double pitch = config.getValue< PitchAxis >( state.getAxisPoint< PitchAxis >() );
    double roll  = config.getValue< RollAxis >( state.getAxisPoint< RollAxis >() );

    Pose pose;

    pose.orientation.setEuler( yaw, pitch, roll );
    pose.position = Point< anchor::World >( x, y, z );

    return pose;
  }

  class StaticLeg
  {
    LegID                  lid_;
    leg::SidepathDistance  sidepath_distance_;
    leg::VertexID          vertex_hint_;
    Point< anchor::World > tip_position_;

  public:
    LegID getID() const
    {
      return lid_;
    }

    leg::SidepathDistance getSidepathDistance() const
    {
      return sidepath_distance_;
    }

    leg::VertexID getVertexHintID() const
    {
      return vertex_hint_;
    }

    Point< anchor::World > getTipPosition() const
    {
      return tip_position_;
    }
  };

  class MovingLeg
  {
    LegID                 lid_;
    leg::VertexID         actual_;
    leg::SidepathDistance target_distance_;
  };

  class State
  {
    WaypointID               wpid_;
    BodyState                body_;
    std::vector< MovingLeg > moving_legs_;
    std::vector< StaticLeg > static_legs_;

  public:
    const BodyState& getBodyState() const
    {
      return body_;
    }

    template < typename Axis >
    const Axis& getBodyAxisPoint() const
    {
      return body_.getAxisPoint< Axis >();
    }

    template < typename Axis >
    void setBodyAxisPoint( typename Axis::point point )
    {
      body_.setAxisPoint( point );
    }

    void replaceStaticLeg( const StaticLeg& static_leg, MovingLeg moving_leg )
    {
      auto iter = std::find( static_legs_.begin(), static_legs_.end(), static_leg );

      ROS_ASSERT( iter != static_legs_.end() );

      static_legs_.erase( iter );

      moving_legs_.push_back( std::move( moving_leg ) );
    }

    const std::vector< MovingLeg >& getMovingLegs() const
    {
      return moving_legs_;
    }
    const std::vector< StaticLeg >& getStaticLegs() const
    {
      return static_legs_;
    }
  };

  Point< anchor::Leg > transformPoint( const Point< anchor::World >& world_point,
                                       const Pose&                   body_pose,
                                       const Leg&                    leg )
  {
    Pose base = leg.getActualPose( body_pose );

    Point< anchor::Leg > res = quatRotate( base.orientation.inverse(), world_point );

    res -= base.position;

    return res;
  }
}
}
