/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "config.h"
#include "leg/config.h"
#include "solver/basic.h"
#include "solver/evaluation.h"

namespace schpin
{
namespace solver
{

  template < typename Axis, typename NextMethod >
  void bodyAxisDuplication( const State& start, State state, std::vector< State >& output, const Config& config )
  {
    const Axis&          axis         = config.getAxis< Axis >();
    typename Axis::Point actual_point = state.getBodyAxisPoint< Axis >();

    if ( actual_point > axis.getMinPoint() )
    {
      State prev_state;
      prev_state.setBodyAxisPoint< Axis >( axis.getPrevPoint( actual_point ) );
      NextMethod( start, prev_state, output, config );
    }

    if ( actual_point < axis.getMaxPoint() )
    {
      State next_state;
      next_state.setBodyAxisPoint< Axis >( axis.getNextPoint( actual_point ) );
      NextMethod( start, next_state, output, config );
    }

    NextMethod( start, state, output, config );
  }

  void finish_generation( State state, std::vector< State >& output )
  {
    output.push_back( state );
  }

  // TODO: this finishes advances and generation
  void moveLegs( const State& start, State state, std::vector< State >& output, const Robot& robot )
  {
    Duration body_shift_duration = evaluateBodyShiftDuration( start, state, robot );

    // TODO: move moving legs
    // TODO: check if leg arrived in destination
    // TODO: each movement proression, picks farthest reachable vertexid within time window
    finish_generation( state, output );
  }

  void legsDuplication( const State& start, State state, std::vector< State >& output, const Robot& robot )
  {
    moveLegs( start, state, output, robot );  // don't change anything about legs, just move

    if ( state.getMovingLegs().size() >= state.getStaticLegs().size() )
    {
      return;
    }

    std::for_each( state.getStaticLegs().begin(), state.getStaticLegs().end(), [&]( const StaticLeg& static_leg ) {
      auto& move_distances = robot.getLeg( static_leg.getID() ).getConfig().getMoveDistances();

      std::for_each( move_distances.begin(), move_distances.end(), [&]( const leg::SidepathDistance& offset ) {

        MovingLeg moving_leg{
            static_leg.getID(), static_leg.getVertexHintID(), static_leg.getSidepathDistance() - offset};

        State new_state = state;

        new_state.replaceStaticLeg( static_leg, moving_leg );

        moveLegs( start, new_state, output, robot );

      } );

    } );
  }

  using PitchDuplication = bodyAxisDuplication< PitchAxis, outputInsert >;
  using RollDuplication  = bodyAxisDuplication< RollAxis, PitchDuplication >;
  using YawDuplication   = bodyAxisDuplication< YawAxis, RollDuplication >;
  using ZDuplication     = bodyAxisDuplication< ZAxis, YawDuplication >;
  using YDuplication     = bodyAxisDuplication< YAxis, ZDuplication >;
  using XDuplication     = bodyAxisDuplication< XAxis, YawDuplication >;

  std::vector< State > generateNeighboursStates( State state, const Config& config )
  {
    std::vector< State > result;
    XDuplication( state, state, result, config );

    return result;
  }
}
}
