/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "solver/graph/types.h"

#pragma once

namespace schpin
{
namespace solver
{

  class Graph
  {
    class VertexGenerator
    {
      VertexID                        actual_;
      std::reference_wrapper< Graph > graph_;

    public:
      typedef Vertex result_type;

      VertexGenerator( Graph& graph, VertexID vid )
        : actual_( vid )
        , graph_( graph_ )
      {
      }

      result_type operator()()
      {
        ++actual_;
        ROS_ASSERT( actual_ < graph_.get().getVertexCount() );
        return graph_.get().getVertex( actual_ );
      }
    };

  public:
    typedef ArrayIndex< IndexLeaf, RollAxis::PointType >  RollIndex;
    typedef ArrayIndex< RollIndex, PitchAxis::PointType > PitchIndex;
    typedef ArrayIndex< PitchIndex, YawAxis::PointType >  YawIndex;
    typedef ArrayIndex< YawIndex, ZAxis::PointType >      ZIndex;
    typedef ArrayIndex< ZIndex, YAxis::PointType >        YIndex;
    typedef VectorIndex< YIndex, XAxis::PointType >       XIndex;

    typedef IdentificatorVector< Vertex > VertexContainer;
    typedef boost::function_input_iterator< VertexGenerator, VertexID > VertexIterator;

  private:
    XIndex index_;

    VertexContainer vertexes_;

  public:
    Vertex getVertex( const VertexID& vid );
    {
      return Vertex( vid, vertexes_[vid] );
    }

    template < typename Func >
    boost::signals2::connection connectSizeSignal( Func f )
    {
      return size_signal_.connect( f );
    }

    boost::iterator_range< VertexIterator > getVertexesRange()
    {
      VertexGenerator start_generator( *this, VertexID( 0 ) );
      VertexGenerator end_generator( *this, getVertexCount() );

      return boost::make_iterator_range( VertexIterator( start_generator, VertexID( 0 ) ),
                                         VertexIterator( end_generator, getVertexCount() ) );
    }

    void reserve( const VertexID& vertex_count );
    {
      vertexes_.reserve( vertex_count );
    }
    void shrinkToFit()
    {
      vertexes_.shrinkToFit();
      for ( VertexStorage& storage : vertexes_ )
      {
        // TODO: this may cause reference invalidation, warning!
        storage.refOutgoingEdges().shrink_to_fit();
        storage.refIngoingEdges().shrink_to_fit();
      }
    }

    std::vector< Edge > makeOutgoingEdgesRange( const Vertex& vertex )
    {
      std::vector< Edge > result;

      for ( IngoingEdge& oedge : vertexes_[vertex.getID()].refOutgoingEdges() )
      {
        VertexStorage& target_storage = vertexes_[oedge.getTargetID()];

        IngoingEdge& counteredge = target_storage.refIngoingEdge( oedge.getIngoingEdgeIndex() );

        result.emplace_back( oedge, counteredge );
      }

      return result;
    }

    std::vector< Edge > makeIngoingEdgesRange( const Vertex& vertex )
    {
      std::vector< Edge > result;

      for ( IngoingEdge& iedge : vertexes_[vertex.getID()].refIngoingEdges() )
      {
        // TODO: maybe rename strucutres, Vertex - VertexStorage is intuitive, Edge - IngoingEdge/OutgoingEdge is not
        VertexStorage& source_storage = vertexes_[iedge.getSourceID()];

        OutgoingEdge& counteredge = source_storage.refOutgoingEdge( iedge.getOutgoingEdgeIndex() );

        result.emplace_back( counteredge, iedge );
      }

      return result;
    }

    VertexID getVertexCount() const
    {
      return container_.getVertexCount();
    }
  };
}
}
