/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "basic.h"
#include "leg/graph.h"
#include "util/identificator.h"
#include "util/physical_quantity.h"

#pragma once

namespace schpin
{
namespace leg_graph_generator
{

  using util::Identificator;
  using util::PhysicalQuantity;
  using namespace schpin::leg;

  class Mass : public PhysicalQuantity< Mass >
  {
    using PhysicalQuantity< Mass >::PhysicalQuantity;
    static constexpr char unit[] = "kg";
  };

  class Torque : public PhysicalQuantity< Torque >
  {
    using PhysicalQuantity< Torque >::PhysicalQuantity;
    static constexpr char unit[] = "N*m";
  };

  class LengthPow2 : public PhysicalQuantity< LengthPow2 >
  {
    using PhysicalQuantity< LengthPow2 >::PhysicalQuantity;
    static constexpr char unit[] = "m^2";
  };

  class Force : public PhysicalQuantity< Force >
  {
    using PhysicalQuantity< Force >::PhysicalQuantity;
    static constexpr char unit[] = "N";
  };

  // TODO: maybe make it electric current?
  class Current : public PhysicalQuantity< Current >
  {
    using PhysicalQuantity< Current >::PhysicalQuantity;
    static constexpr char unit[] = "A";
  };

  class AngularVelocity : public PhysicalQuantity< AngularVelocity >
  {
    using PhysicalQuantity< AngularVelocity >::PhysicalQuantity;
    static constexpr char unit[] = "rad/s";
  };

  class AngularVelocityPow2 : public PhysicalQuantity< AngularVelocityPow2 >
  {
    using PhysicalQuantity< AngularVelocityPow2 >::PhysicalQuantity;
    static constexpr char unit[] = "rad^2/s^2";
  };

  class AngularAcceleration : public PhysicalQuantity< AngularAcceleration >
  {
    using PhysicalQuantity< AngularAcceleration >::PhysicalQuantity;
    static constexpr char unit[] = "rad/s^2";
  };

  class MomentOfInertia : public PhysicalQuantity< MomentOfInertia >
  {
    using PhysicalQuantity< MomentOfInertia >::PhysicalQuantity;
    static constexpr char unit[] = "kg*m^2";
  };

  Force operator*( Mass mass, AngularAcceleration accel )
  {
    return Force( mass.get() * accel.get() );
  }

  Angle operator*( AngularVelocity velocity, Duration duration )
  {
    return Angle( velocity.get() * duration.get() );
  }

  Torque operator*( Force f, Length l )
  {
    return Torque( f.get() * l.get() );
  }

  AngularAcceleration operator/( Torque t, MomentOfInertia moi )
  {
    ROS_ASSERT( moi.get() != 0 );
    return AngularAcceleration( t.get() / moi.get() );
  }

  AngularVelocity operator*( AngularAcceleration dist, Duration dur )
  {
    return AngularVelocity( dist.get() * dur.get() );
  }

  Duration operator/( Angle a, AngularVelocity vel )
  {
    ROS_ASSERT( vel.get() != 0 );
    return Duration( a.get() / vel.get() );
  }

  Duration operator/( AngularVelocity v, AngularAcceleration accel )
  {
    ROS_ASSERT( accel.get() != 0 );
    return Duration( v.get() / accel.get() );
  }

  LengthPow2 pow2( Length l )
  {
    return LengthPow2( std::pow( l.get(), 2 ) );
  }

  AngularVelocityPow2 operator*( AngularAcceleration accel, Angle a )
  {
    return AngularVelocityPow2( accel.get() * a.get() );
  }

  AngularVelocityPow2 pow2( AngularVelocity v )
  {
    return AngularVelocityPow2( std::pow( v.get(), 2 ) );
  }

  MomentOfInertia operator*( Mass m, LengthPow2 l_pow2 )
  {
    return MomentOfInertia( m.get() * l_pow2.get() );
  }

  AngularVelocity operator/( Angle a, Duration d )
  {
    ROS_ASSERT( d.get() != 0 );
    return AngularVelocity( a.get() / d.get() );
  }

  AngularVelocity sqrt( AngularVelocityPow2 apow2 )
  {
    return AngularVelocity( std::sqrt( apow2.get() ) );
  }

  class SegmentID : public Identificator< SegmentID, 16 >
  {
    using Identificator< SegmentID, 16 >::Identificator;
  };

  class Segment
  {
  public:
    Pose        pose;
    unsigned    steps;
    std::string next_segment;
    Point       mass_center;
    Mass        mass{0};
    Length      mass_r{0};
    bool        has_joint{false};
    struct
    {
      Angle           min_angle{0};
      Angle           max_angle{0};
      AngularVelocity max_velocity{0};
      Torque          max_torque{0};
      Current         stall_current{0};
      Current         running_current{0};
      Point           axis;
    } joint;

    Angle stateToAngle( JointState joint_state ) const
    {
      ROS_ASSERT( has_joint );
      Angle range = joint.max_angle - joint.min_angle;
      // we need to get double as result of division
      return range * ( double( joint_state ) / double( steps ) );
    }
  };

  typedef std::vector< Segment > Segments;

  struct Config
  {

    Segments segments;
    Pose     pose;
  };

  struct SegmentEval
  {
    Duration duration{0};
    Current  current{0};
    Torque   available_torque{0};
    Torque   gravity_torque{0};
  };

  struct EvalRes
  {
    Mass  mass{0};
    Point mass_center;
  };
}
}
