/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "base.h"
#include "physics.h"
#include <numeric>

#pragma once

namespace schpin
{
namespace leg_graph_generator
{
  namespace generation
  {
    using namespace physics;

    class CubePattern
    {
      std::vector< JointStates > patterns_;

    public:
      CubePattern( unsigned size )
        : patterns_()
      {
        ROS_ASSERT( size != 0 );
        JointStates start{};
        for ( unsigned i = 0; i < size; ++i )
        {
          start.push_back( 0 );
        }
        generateGeneralPattern( start, start.size() - 1 );

        std::sort( patterns_.begin(), patterns_.end() );
        /*
         *
         * After generation of pattenrs they should be sorted like this:
         *
         *	0, 0, 0
         *	0, 0, 1
         *	0, 0, 2
         *	0, 1, 0
         *	0, 1, 1
         *	.
         *	.
         *	.
         *
         */
      }

      std::size_t size() const
      {
        return patterns_.size();
      }

      /* returns all JointStates for this pattern around <center>
       */
      std::vector< JointStates > generatePattern( const JointStates& center ) const
      {
        // we just have to make -1 offset, because we want to move from 0,1,2 to
        // -1,0,1
        std::vector< JointStates > res;
        JointStates                offset = center;
        for ( auto& js : offset )
        {
          js -= 1;
        }

        std::transform( patterns_.begin(),
                        patterns_.end(),
                        std::back_inserter( res ),
                        [this, &offset]( const JointStates& pattern ) { return this->sum( pattern, offset ); } );

        return res;
      }

    private:
      JointStates sum( const JointStates& lhs, const JointStates& rhs ) const
      {
        ROS_ASSERT( lhs.size() == rhs.size() );
        JointStates res = lhs;
        for ( unsigned i = 0; i < lhs.size(); ++i )
        {
          res[i] += rhs[i];
        }
        return res;
      }

      JointStates deduct( const JointStates& lhs, const JointStates& rhs ) const
      {
        ROS_ASSERT( lhs.size() == rhs.size() );
        JointStates res = lhs;
        for ( unsigned i = 0; i < lhs.size(); ++i )
        {
          res[i] -= rhs[i];
        }
        return res;
      }
      void generateGeneralPattern( JointStates pattern, std::size_t i )
      {
        for ( const int j : {0, 1, 2} )
        {
          // TODO: is this conversion OK?
          // TODO: are other conversions OK?
          pattern[i] = decltype( pattern[i] )( j );
          if ( i > 0 )
          {
            generateGeneralPattern( pattern, i - 1 );
          }
          else
          {
            patterns_.push_back( pattern );
          }
        }
      }
    };

    /* make it clear that this drops edge data */
    void multiplyVertexByDirectionChange( Graph& graph, Vertex vertex )
    {
      std::vector< Vertex > precedessors;
      std::map< JointStates, Vertex > map_states_to_precedessors;

      std::for_each( graph.ingoingEdgeBegin( vertex ), graph.ingoingEdgeEnd( vertex ), [&]( const Edge& edge ) {

        Vertex source = graph.getVertex( edge.getSourceID() );
        precedessors.push_back( source );

        // the vertex itself is not relevant now

        map_states_to_precedessors.insert( std::make_pair( source.getJointStates(), vertex ) );
      } );

      std::size_t ingoing_keys_count = map_states_to_precedessors.size();

      std::vector< Vertex > group = graph.copyVertex( vertex, ingoing_keys_count - 1 );
      group.push_back( vertex );

      graph.purgeIngoingEdges( vertex );

      // fill in real data
      for ( auto& map_item : map_states_to_precedessors )
      {
        map_states_to_precedessors.at( map_item.first ) = group.back();
        group.pop_back();
      }

      ROS_ASSERT( group.size() == 0 );

      std::for_each( precedessors.begin(), precedessors.end(), [&]( const Vertex& predecessor ) {

        const Vertex& group_target = map_states_to_precedessors.at( predecessor.getJointStates() );
        graph.insertEdge( predecessor, group_target, EdgeData() );
      } );
    }

    void connectVertexToNeighbours( Graph& graph,
                                    const std::map< JointStates, Vertex >& index,
                                    const Vertex&      vertex,
                                    const CubePattern& pattern )
    {

      std::vector< Vertex > candidates;
      candidates.reserve( pattern.size() );

      std::vector< JointStates > neighbour_keys = pattern.generatePattern( vertex.getJointStates() );

      for ( JointStates& nb_key : neighbour_keys )
      {
        auto iter = index.find( nb_key );
        if ( iter != index.end() )
        {
          candidates.push_back( iter->second );
        }
      }

      std::for_each( candidates.begin(), candidates.end(), [&]( const Vertex& target ) {
        graph.insertEdge( vertex, target, EdgeData{} );
      } );
    }
    /** Recursively generates vertexes inside graph based on segments.
     *
     * Idea is to recursively traverse the joint_state space in DFS way.
     *
     * Each call evaluates segments[sid]. For jointless segments, it just continues
     * onto next segment with call.
     * For segments with joint, it calls recursion multiple times, for each joint
     * configuration.
     *
     * Last recursive call creates the Vertex and inserts it into graph with
     * 'joint_state'
     *
     */
    void recurGenerateVertexes( const Segments& segments,
                                SegmentID       sid,
                                JointStates     joint_states,
                                Graph&          graph,
                                std::map< JointStates, Vertex >& index )
    {
      if ( std::size_t( sid ) == segments.size() )
      {
        ROS_ASSERT( joint_states.size() != 0 );

        Vertex vertex = graph.emplaceVertex();

        vertex.setJointStates( joint_states );

        index.insert( std::make_pair( joint_states, vertex ) );

        return;
      }

      const Segment& segment = segments[std::size_t( sid )];

      ++sid;

      if ( !segment.has_joint )
      {
        recurGenerateVertexes( segments, sid, joint_states, graph, index );
        return;
      }

      for ( unsigned j = 0; j < segment.steps; ++j )
      {
        JointStates new_states = joint_states;
        new_states.emplace_back( j );
        recurGenerateVertexes( segments, sid, new_states, graph, index );
      }
    }

    /** Calculates duration of 'from' -> 'to' move
     *
     * Calculates duration of moving between 'from' and 'to' angles, given
     * 'acceleration', 'max_velocity' and
     * 'initial_velocity'
     *
     * It asserts that move will try to accelerate until 'max_velocity' is hit.
     *
     */

    Duration getMovementDuration( Angle               from,
                                  Angle               to,
                                  AngularAcceleration acceleration,
                                  AngularVelocity     max_velocity,
                                  AngularVelocity     initial_velocity )
    {

      Duration velocity_maximization_duration = accelerationDuration( max_velocity - initial_velocity, acceleration );

      Angle angle_diff = abs( to - from );

      Duration move_duration = travelDuration( angle_diff, initial_velocity, acceleration );

      if ( move_duration < velocity_maximization_duration )
      {
        return move_duration;
      }

      Angle angle_traveled = travelAngle( initial_velocity, velocity_maximization_duration, acceleration );

      ROS_ASSERT( angle_traveled < angle_diff );

      return velocity_maximization_duration + travelDuration( abs( angle_diff - angle_traveled ), max_velocity );
    }

    void evaluateTorque( EvalRes&       eval_res,
                         SegmentEval&   segment_eval,
                         const Segment& segment,
                         Quaternion     rotational_position )
    {

      ROS_ASSERT( segment.has_joint );

      Torque g_torque = gravityTorque( segment.mass, rotational_position, segment.mass_r ) +
                        gravityTorque( eval_res.mass, rotational_position, zDistance( eval_res.mass_center ) );

      Torque available_torque = segment.joint.max_torque - g_torque;

      segment_eval.available_torque = available_torque;
      segment_eval.gravity_torque   = g_torque;
    }

    Duration evaluateDuration( EvalRes&       eval_res,
                               SegmentEval&   segment_eval,
                               const Segment& segment,
                               Angle          prev_angle,
                               Angle          start_angle,
                               Angle          goal_angle )
    {

      MomentOfInertia I = momentOfInertia( segment.mass, segment.mass_r ) +
                          momentOfInertia( eval_res.mass, zDistance( eval_res.mass_center ) );

      AngularAcceleration acceleration = accel( segment_eval.available_torque, I );  // maximal rotational acceleration

      Duration prev_movement_duration = getMovementDuration(
          prev_angle, start_angle, acceleration, segment.joint.max_velocity, AngularVelocity( 0 ) );  // L

      AngularVelocity initial_velocity = velocity( start_angle - prev_angle, prev_movement_duration );  // actual speed

      return getMovementDuration( start_angle, goal_angle, acceleration, segment.joint.max_velocity, initial_velocity );
    }

    /**
     *
     *
     *
     */

    SegmentEval evalSegment( EvalRes&           eval_res,
                             const Segment&     segment,
                             const Pose         base,
                             unsigned&          joint_index,
                             const JointStates& prev_states,
                             const JointStates& start_states,
                             const JointStates& goal_states )
    {

      SegmentEval segment_eval;
      Point       eval_res_pos;

      if ( segment.has_joint )
      {
        Angle prev_angle  = segment.stateToAngle( prev_states[joint_index] );
        Angle start_angle = segment.stateToAngle( start_states[joint_index] );
        Angle goal_angle  = segment.stateToAngle( goal_states[joint_index] );

        joint_index += 1;

        Quaternion rotational_position =
            Quaternion( segment.joint.axis, double( avg( start_angle, goal_angle ) ) ) + base.orientation;

        evaluateTorque( eval_res, segment_eval, segment, rotational_position );
        segment_eval.duration =
            evaluateDuration( eval_res, segment_eval, segment, prev_angle, start_angle, goal_angle );

        eval_res_pos = quatRotate( rotational_position, eval_res.mass_center );
      }
      else
      {
        eval_res_pos = eval_res.mass_center;
      }

      Point merged_pos =
          weightedAverage( eval_res_pos, double( eval_res.mass ), segment.mass_center, double( segment.mass ) );

      eval_res.mass        = segment.mass + eval_res.mass;
      eval_res.mass_center = merged_pos;

      return segment_eval;
    }

    JointStates getIngoingJointStates( Graph& graph, Vertex vertex )
    {

      auto begin = graph.ingoingEdgeBegin( vertex );
      auto end   = graph.ingoingEdgeEnd( vertex );

      ROS_ASSERT( begin != end );

      const Vertex& source = graph.getVertex( begin->getSourceID() );

      JointStates res = source.getJointStates();

      ROS_ASSERT( std::all_of( begin, end, [&]( const Edge& edge ) {
        return graph.getVertex( edge.getSourceID() ).getJointStates() == res;
      } ) );

      return res;
    }

    Current evalCurrent( const Segment& segment, const SegmentEval& segment_eval, Duration real_duration )
    {

      if ( !segment.has_joint )
      {
        return Current( 0 );
      }

      Torque real_torque = scaleTorque( segment_eval.available_torque, segment_eval.duration, real_duration ) +
                           segment_eval.gravity_torque;

      return scaleCurrent(
          segment.joint.running_current, segment.joint.stall_current, real_torque, segment.joint.max_torque );
    };

    void evaluateEdge( Edge& edge, Graph& graph, const Config& config, JointStates& prev_states )
    {
      const Vertex& target = graph.getVertex( edge.getTargetID() );
      const Vertex& source = graph.getVertex( edge.getSourceID() );

      unsigned                   joint_index = 0;
      EvalRes                    eval_res;
      std::vector< SegmentEval > segment_evaluations;
      Pose                       base;

      base.stackUp( config.pose );

      for ( const Segment& segment : config.segments )
      {
        base.stackUp( segment.pose );
        SegmentEval seg_eval = evalSegment(
            eval_res, segment, base, joint_index, prev_states, source.getJointStates(), target.getJointStates() );
        segment_evaluations.push_back( seg_eval );
      }

      auto iter =
          std::max_element( segment_evaluations.begin(),
                            segment_evaluations.end(),
                            []( const SegmentEval& lh, const SegmentEval& rh ) { return lh.duration < rh.duration; } );

      Duration final_duration = iter->duration;

      std::vector< Current > currents;

      std::transform( config.segments.begin(),
                      config.segments.end(),
                      segment_evaluations.begin(),
                      std::back_inserter( currents ),
                      [final_duration]( const Segment& segment, const SegmentEval& eval ) {
                        return evalCurrent( segment, eval, final_duration );
                      } );

      Current final_current = std::accumulate( currents.begin(), currents.end(), Current( 0 ) );

      edge.refData().setDuration( double( final_duration ) );
      edge.refData().setCurrent( double( final_current ) );
    }

    void evaluateVertex( Graph& graph, Vertex vertex, const Config& config )
    {
      JointStates prev_states = getIngoingJointStates( graph, vertex );

      std::for_each( graph.ingoingEdgeBegin( vertex ), graph.ingoingEdgeEnd( vertex ), [&]( Edge& edge ) {

        evaluateEdge( edge, graph, config, prev_states );
      } );
    }

  }  // ns gene
}
}
