/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "include_barrier/ros/ros.h"

#pragma once

namespace schpin
{
namespace leg_graph_generator
{
  namespace parse
  {

    std::vector< double > parse_param_data( ros::NodeHandle& local_node_handle, std::string name, unsigned c )
    {
      std::vector< double > data;
      bool                  has_param = local_node_handle.getParam( name, data );
      ROS_ASSERT_MSG( has_param, "Parameter %s is missing", name.c_str() );
      ROS_ASSERT_MSG(
          data.size() == c, "Parameter %s should be array of three values, got %lu values", name.c_str(), data.size() );

      return data;
    }

    void processPoint( ros::NodeHandle& local_node_handle, std::string name, Point& target )
    {
      std::vector< double > data = parse_param_data( local_node_handle, name, 3 );
      target                     = Point( data.at( 0 ), data.at( 1 ), data.at( 2 ) );
    }

    // TODO : A. hidden todo! B. figure out a way how to print config for user into
    // stdout or something

    template < typename T >
    void processValue( ros::NodeHandle& local_node_handle, std::string name, T& target )
    {
      bool has_param = local_node_handle.getParam( name, target );
      ROS_ASSERT_MSG( has_param, "Parameter %s is missing", name.c_str() );
    }

    template < typename T >
    void processPhysicalQuantity( ros::NodeHandle& local_node_handle, std::string name, PhysicalQuantity< T >& target )
    {
      double value = 0;
      processValue( local_node_handle, name, value );
      target = T( value );
    }

    void processRotation( ros::NodeHandle& local_node_handle, std::string name, Quaternion& target )
    {
      std::vector< double > data = parse_param_data( local_node_handle, name, 3 );
      target.setEuler( data.at( 1 ), data.at( 0 ), data.at( 2 ) );
    }

    unsigned loadSegments( std::string segment_name, Segments& segments )
    {

      unsigned        n_joints = 0;
      ros::NodeHandle local_node_handle{"segments/" + segment_name};

      segments.emplace_back();
      Segment& segment = segments.back();

      DEBUG_LOG( "segment name " << segment_name );

      processPhysicalQuantity( local_node_handle, "mass", segment.mass );

      processPoint( local_node_handle, "pos", segment.pose.position );
      processPoint( local_node_handle, "mass_center", segment.mass_center );

      processRotation( local_node_handle, "rot", segment.pose.orientation );

      segment.has_joint = local_node_handle.hasParam( "joint" );
      if ( segment.has_joint )
      {

        double steps = 0;
        processValue( local_node_handle, "joint/graph/steps", steps );
        segment.steps = unsigned( steps );
        processPhysicalQuantity( local_node_handle, "joint/angle/min", segment.joint.min_angle );
        processPhysicalQuantity( local_node_handle, "joint/angle/max", segment.joint.max_angle );
        processPhysicalQuantity( local_node_handle, "joint/max_velocity", segment.joint.max_velocity );
        processPhysicalQuantity( local_node_handle, "joint/max_torque", segment.joint.max_torque );
        processPhysicalQuantity( local_node_handle, "joint/stall_current", segment.joint.stall_current );
        processPhysicalQuantity( local_node_handle, "joint/running_current", segment.joint.running_current );

        processPoint( local_node_handle, "joint/axis", segment.joint.axis );

        n_joints += 1;
      }

      bool has_next_segment = local_node_handle.getParam( "next_segment", segment.next_segment );
      if ( has_next_segment )
      {
        DEBUG_LOG( "next_joint: " << segment.next_segment );
        n_joints += loadSegments( segment.next_segment, segments );
      }

      return n_joints;
    }

    unsigned loadConfig( Config& config )
    {
      ros::NodeHandle local_node_handle{};

      std::string chain_root;
      local_node_handle.getParam( "chain_root", chain_root );

      processPoint( local_node_handle, "pos", config.pose.position );
      processRotation( local_node_handle, "rot", config.pose.orientation );

      return loadSegments( chain_root, config.segments );
    }
  }
}
}
