/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "basic.h"
#include "ros/assert.h"

#pragma once
namespace schpin
{
namespace leg_graph_generator
{
  namespace physics
  {

    AngularAcceleration gravity( 9.18 );

    template < typename T >
    inline T avg( const T& a, const T& b )
    {
      return ( a + b ) / 2;
    }

    // torque on joint made by 'mass' at 'r' distance from the joint
    inline Torque gravityTorque( Mass mass, Quaternion joint, Length r )
    {

      Force force( mass * gravity );
      Point g( 0, 0, -1 );

      Angle g_axis_angle( g.angle( joint.getAxis() ) );
      Angle joint_angle( joint.getAngle() );

      Force f_perp( sin( g_axis_angle ) * force * cos( joint_angle ) );

      return f_perp * r;
    }

    // scale torque based on time needed for move with maximal torque and actual
    // time
    inline Torque scaleTorque( Torque max_torque, Duration actual_duration, Duration max_torque_duration )
    {
      ROS_ASSERT( actual_duration <= max_torque_duration );
      return max_torque * ( actual_duration / max_torque_duration );
    }

    // scale current between min and max, based on actual torque and maximal torque
    inline Current scaleCurrent( Current min_cur, Current max_cur, Torque torq, Torque max_torq )
    {
      return ( max_cur - min_cur ) * ( torq / max_torq ) + min_cur;
    }

    // distance of point from zero coordinate
    inline Length zDistance( Point p )
    {
      return Length( p.distance( Point{0, 0, 0} ) );
    }

    // moment of intertia for 'mass' at distance 'r' TODO: rethink if this shouldn't
    // be a constructor
    inline MomentOfInertia momentOfInertia( Mass mass, Length r )
    {
      return mass * pow2( r );
    }

    // TODO: rethink this as constructor
    // acceleration with 'torque' and 'momentOfInertia' for angular movement
    inline AngularAcceleration accel( Torque torque, MomentOfInertia moment_of_inertia )
    {
      return torque / moment_of_inertia;
    }

    // TODO: rethink this as constructor
    inline AngularVelocity velocity( Angle angle, Duration duration )
    {
      return angle / duration;
    }

    // calcualtes weighetd average of two points, used for mass_center
    inline Point weightedAverage( Point a, double aw, Point b, double bw )
    {
      if ( aw + bw == 0 )
      {
        return ( a + b ) / 2;
      }
      return ( a * aw + b * bw ) / ( aw + bw );
    }

    // calculates time to reach velocity 'v' with 'accel'eration
    inline Duration accelerationDuration( AngularVelocity v, AngularAcceleration accel )
    {
      return v / accel;
    }

    // calculates time needed to travel 'angle' distance with velocity 'velocity'
    inline Duration travelDuration( Angle angle, AngularVelocity velocity )
    {
      return angle / velocity;
    }

    // calculates time needed to travel 'angle' distance with base velocity
    // 'velocity' and acceleration 'accel',
    // asserts  linear acceleration
    inline Duration travelDuration( Angle angle, AngularVelocity velocity, AngularAcceleration accel )
    {
      angle = abs( angle );
      /*
       * solved 't' for: angle = v * t + 1/2 * accel * t^2
       * which is standart way to calculate traveled angle for accelerated movement
       *	t =  sqrt( 2*angle*accel + v^2 ) / accel - v/accel;
       */

      // with doubling everything, we enter zone where type checks won't check
      // sanity of aclculation
      return sqrt( 2 * accel * angle + pow2( velocity ) ) / accel - velocity / accel;
    }

    // calculates angle traveled with base velocity 'velocity', after traveling for
    // 'time' seconds with 'accel',
    // asserts
    // linear behavior
    inline Angle travelAngle( AngularVelocity velocity, Duration duration, AngularAcceleration accel )
    {
      return velocity * duration + ( 1 / 2 ) * accel * duration * duration;
    }
  }
}
}
