/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#include "../util.h"
#include <boost/integer.hpp>
#include <iterator>
#include <ros/assert.h>

#pragma once

namespace schpin
{
namespace util
{
  // CRTP
  template < typename T, unsigned bits >
  class Identificator
  {

  public:
    typedef typename boost::uint_t< bits >::least    ValueType;
    typedef typename boost::int_t< bits + 1 >::least DifferenceType;

  private:
    ValueType value_;

  public:
    Identificator()
      : value_( 0 )
    {
    }

    Identificator( uintmax_t val )
      : value_( ValueType( val ) )
    {
      // TODO: as long as ValueType is unsigned, this is enough
      ROS_ASSERT( val < std::numeric_limits< ValueType >::max() );
    }

    explicit operator std::size_t() const
    {
      return std::size_t( value_ );
    }

    const ValueType& get() const
    {
      return value_;
    }

    T& operator++()
    {
      ROS_ASSERT( value_ != std::numeric_limits< ValueType >::max() );
      value_ += 1;
      return *( static_cast< T* >( this ) );
    }
    T operator++( int )
    {
      T tmp( *( static_cast< T* >( this ) ) );
      operator++();
      return tmp;
    }
    T& operator--()
    {
      ROS_ASSERT( value_ != std::numeric_limits< ValueType >::min() );
      value_ -= 1;
      return *( static_cast< T* >( this ) );
    }
    T operator--( int )
    {
      T tmp( *( static_cast< T* >( this ) ) );
      operator--();
      return tmp;
    }
  };

  template < typename T, unsigned                   bits >
  typename Identificator< T, bits >::DifferenceType operator-( const Identificator< T, bits >& lhs,
                                                               const Identificator< T, bits >& rhs )
  {
    typedef typename Identificator< T, bits >::DifferenceType DifferenceType;
    return DifferenceType( lhs.get() ) - DifferenceType( rhs.get() );
  }

  template < typename T, unsigned bits >
  inline bool operator==( const Identificator< T, bits >& lhs, const Identificator< T, bits >& rhs )
  {
    return lhs.get() == rhs.get();
  }
  template < typename T, unsigned bits >
  inline bool operator!=( const Identificator< T, bits >& lhs, const Identificator< T, bits >& rhs )
  {
    return !operator==( lhs, rhs );
  }
  template < typename T, unsigned bits >
  inline bool operator<( const Identificator< T, bits >& lhs, const Identificator< T, bits >& rhs )
  {
    return lhs.get() < rhs.get();
  }
  template < typename T, unsigned bits >
  inline bool operator>( const Identificator< T, bits >& lhs, const Identificator< T, bits >& rhs )
  {
    return operator<( rhs, lhs );
  }
  template < typename T, unsigned bits >
  inline bool operator<=( const Identificator< T, bits >& lhs, const Identificator< T, bits >& rhs )
  {
    return !operator>( lhs, rhs );
  }
  template < typename T, unsigned bits >
  inline bool operator>=( const Identificator< T, bits >& lhs, const Identificator< T, bits >& rhs )
  {
    return !operator<( lhs, rhs );
  }
  template < typename T, unsigned bits >
  std::ostream& operator<<( std::ostream& os, const Identificator< T, bits >& id )
  {
    os << id.get();
    return os;
  }

  template < typename Item, typename Identificator >
  class IdentificatorVector
  {
  public:
    typedef std::vector< Item >                    ContainerType;
    typedef typename ContainerType::iterator       Iterator;
    typedef typename ContainerType::const_iterator ConstIterator;
    typedef Item                                   ValueType;
    typedef ValueType&                             Reference;
    typedef ValueType const&                       ConstReference;
    typedef Identificator                          SizeType;

  private:
    ContainerType container_;

  public:
    void logStatistics() const
    {
      DEBUG_LOG( size() << "/" << container_.capacity() << " items | " << sizeof( typename ContainerType::value_type )
                        << " bytes per item | "
                        << ( sizeof( typename ContainerType::value_type ) * container_.capacity() ) / ( 1024 * 1024 )
                        << " MB of memory usage" );
    }

    ConstReference operator[]( const Identificator& id ) const
    {
      return container_[id.get()];
    }
    Reference operator[]( const Identificator& id )
    {
      return container_[id.get()];
    }

    ConstIterator cbegin() const
    {
      return container_.cbegin();
    }

    ConstIterator cend() const
    {
      return container_.cend();
    }

    SizeType size() const
    {
      return container_.size();
    }

    SizeType capacity() const
    {
      return container_.capacity();
    }

    bool hasID( const Identificator& id )
    {
      return id.get() > 0 && id.get() < container_.size();
    }

    void resize( const Identificator& id )
    {
      container_.resize( id.get() );
    }

    void reserve( const Identificator& id )
    {
      container_.reserve( id.get() );
    }

    void shrinkToFit()
    {
      container_.shrink_to_fit();
    }

    std::back_insert_iterator< ContainerType > makeBackInserter()
    {
      return std::back_inserter( container_ );
    }

    // this gets on my nerve
    Identificator emplaceBack()
    {
      container_.emplace_back();
      Identificator res( size() );
      --res;
      return res;
    }

    ValueType& back()
    {
      return container_.back();
    }
  };
}
}
