/**
 * Copyright 2017 by Jan Koniarik <jan.koniarik@gmail.com>
 * 
 * This file is part of Schpin.
 * 
 * Schpin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Schpin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Schpin.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * @license LPGL-3.0+ <https://www.gnu.org/licenses/lgpl.html>
 */
#pragma once

namespace schpin
{
namespace util
{
  template < typename T >
  class PhysicalQuantity
  {
    double value_;

  public:
    typedef T Derived;

    explicit PhysicalQuantity( double val )
      : value_( val )
    {
    }

    const double& get() const
    {
      return value_;
    }

    explicit operator double() const
    {
      return double( this->get() );
    }

    T& operator+=( const PhysicalQuantity< T >& other )
    {
      value_ += other.get();
      return *( static_cast< T* >( this ) );
    }
    T& operator-=( const PhysicalQuantity< T >& other )
    {
      value_ -= other.get();
      return *( static_cast< T* >( this ) );
    }
    T& operator/=( const double val )
    {
      value_ /= val;
      return *( static_cast< T* >( this ) );
    }
    T& operator*=( const double val )
    {
      value_ *= val;
      return *( static_cast< T* >( this ) );
    }
  };

  template < typename T >
  double operator/( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    return lhs.get() / rhs.get();
  }

  template < typename T >
  inline T operator+( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    T res( lhs.get() );
    res += rhs;
    return res;
  }
  template < typename T >
  inline T operator-( PhysicalQuantity< T > lhs, const PhysicalQuantity< T >& rhs )
  {
    T res( lhs.get() );
    res -= rhs;
    return res;
  }

  template < typename T >
  inline bool operator==( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    return lhs.get() == rhs.get();
  }
  template < typename T >
  inline bool operator<( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    return lhs.get() < rhs.get();
  }

  template < typename T >
  T operator*( const PhysicalQuantity< T > pq, const double val )
  {
    T res( pq.get() );
    res *= val;
    return res;
  }

  template < typename T >
  inline T operator/( const PhysicalQuantity< T >& pq, const double val )
  {
    T res( pq.get() );
    res /= val;
    return res;
  }

  template < typename T >
  T abs( PhysicalQuantity< T > a )
  {
    return T( std::abs( a.get() ) );
  }

  template < typename T >
  double cos( const PhysicalQuantity< T > u )
  {
    return std::cos( u.get() );
  }

  template < typename T >
  double sin( const PhysicalQuantity< T > u )
  {
    return std::sin( u.get() );
  }

  //---------------------------------------------------------------------------

  template < typename T >
  inline bool operator!=( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    return !operator==( lhs, rhs );
  }
  template < typename T >
  inline bool operator>( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    return operator<( rhs, lhs );
  }
  template < typename T >
  inline bool operator<=( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    return !operator>( lhs, rhs );
  }
  template < typename T >
  inline bool operator>=( const PhysicalQuantity< T >& lhs, const PhysicalQuantity< T >& rhs )
  {
    return !operator<( lhs, rhs );
  }
  //---------------------------------------------------------------------------

  template < typename T >
  T operator*( const double val, const PhysicalQuantity< T > pq )
  {
    return pq * val;
  }
  template < typename T >
  inline T operator/( const double val, const PhysicalQuantity< T >& pq )
  {
    return pq / val;
  }
  template < typename T >
  std::ostream& operator<<( std::ostream& os, const PhysicalQuantity< T >& id )
  {
    os << id.get() << " " << T::unit;
    return os;
  }
}
}
