#include "leg/graph.h"
#include <gtest/gtest.h>
#include <random>
#include <vector>

namespace schpin
{
std::random_device                       rd;
std::mt19937                             gen{rd()};
std::uniform_real_distribution< double > dist{0, 10};
// DATASOURCE: http://yaroslavvb.com/upload/graphs2.txt

struct GraphData
{
  leg::Graph                                      graph{1}; 
  boost::optional< leg::Vertex >                  start; // vertex that should be used for algorithms as 'start' point
  boost::optional< std::vector< leg::VertexID > > vertex_order; // order of vertex discovery, for shortest path discovery from 'start' vertex
};

GraphData getSimple()
{
  GraphData res;

  auto v1 = res.graph.emplaceVertex();
  auto v2 = res.graph.emplaceVertex();

  res.graph.insertEdge( v1, v2, leg::EdgeData( Duration( 1 ) ) );

  res.vertex_order = std::vector< leg::VertexID >{v1.getID(), v2.getID()};

  res.start = v1;

  return res;
};

GraphData getLine( std::size_t size )
{

  GraphData res;

  res.graph = leg::Graph( leg::EdgeCount( size ) );
  res.start = res.graph.emplaceVertex();

  std::vector< leg::VertexID > vertex_order;
  vertex_order.push_back( res.start->getID() );

  leg::Vertex actual = *res.start;

  while ( res.graph.getVertexCount() < leg::VertexID( size ) )
  {
    leg::Vertex target = res.graph.emplaceVertex();

    leg::EdgeData data( Duration( dist( gen ) ) );

    res.graph.insertEdge( actual, target, data );

    vertex_order.push_back( target.getID() );

    actual = target;
  }

  res.vertex_order = vertex_order;

  return res;
}

GraphData getStar( std::size_t size )
{
  GraphData res;

  res.graph = leg::Graph( leg::EdgeCount( size ) );
  res.start = res.graph.emplaceVertex();

  std::map< leg::Price, leg::VertexID > mapping;

  while ( res.graph.getVertexCount() < leg::VertexID( size + 1 ) )
  {
    leg::Vertex target = res.graph.emplaceVertex();

    leg::EdgeData data( Duration( dist( gen ) ) );

    auto iter = mapping.find( data.getPrice() );

    if ( iter != mapping.end() )
    {
      continue;
    }

    mapping.insert( std::make_pair( data.getPrice(), target.getID() ) );

    EXPECT_NE( res.start->getID(), target.getID() );
    res.graph.insertEdge( *res.start, target, data );
  }

  std::vector< leg::VertexID > vertex_order;

  vertex_order.push_back( res.start->getID() );

  std::transform( mapping.begin(),
                  mapping.end(),
                  std::back_inserter( vertex_order ),
                  []( std::pair< leg::Price, leg::VertexID > data ) { return std::get< 1 >( data ); } );

  res.vertex_order = vertex_order;
  return res;
}

GraphData getPetersen()
{

  GraphData res;

  res.graph = leg::Graph( 3 );

  std::vector< std::pair< unsigned, unsigned > > matching{
      {1, 3}, {1, 5}, {1, 7},  {2, 4},  {2, 6},  {2, 8},   {3, 1},  {3, 5},   {3, 9},   {4, 2},  {4, 6},  {4, 10},
      {5, 1}, {5, 3}, {5, 11}, {6, 2},  {6, 4},  {6, 12},  {7, 1},  {7, 8},   {7, 12},  {8, 2},  {8, 7},  {8, 9},
      {9, 3}, {9, 8}, {9, 10}, {10, 4}, {10, 9}, {10, 11}, {11, 5}, {11, 10}, {11, 12}, {12, 6}, {12, 7}, {12, 11}};

  std::vector< std::size_t > order{1, 3, 5, 7, 9, 11, 8, 12, 10, 2, 4, 6};

  std::vector< leg::Vertex > vertexes;

  unsigned counter = 0;

  for ( auto pair : matching )
  {
    while ( std::max( pair.first, pair.second ) >= vertexes.size() )
    {
      vertexes.push_back( res.graph.emplaceVertex() );
    }

    // TODO: make random edge prices
    res.graph.insertEdge( vertexes[pair.first], vertexes[pair.second], leg::EdgeData( Duration( counter ) ) );

    counter += 1;
  }

  res.start        = vertexes[1];
  res.vertex_order = std::vector< leg::VertexID >{};

  std::transform( order.begin(), order.end(), std::back_inserter( *res.vertex_order ), [&]( std::size_t s ) {
    return vertexes[s].getID();
  } );
  return res;
};
}
