#include "leg/approx.h"
#include "basic.h"
#include "graph/util.h"
#include <algorithm>
#include <gtest/gtest.h>

using namespace schpin;

// TODO:
//  - increase lineary edge price as it goes from [0,0,0] point away, test approximation of segments startin from
//  [0,0,0]
//      note: with linear increase, average should work just fine
//  - test with 0-cost edges

struct GridData
{
  leg::Graph graph{1};
  std::map< leg::JointStates, Point< anchor::Leg > > states_to_point{};
};

GridData getGridGraph( unsigned x_max, unsigned y_max )
{
  leg::Graph      graph{4};
  leg::JointState j_state{0};
  std::map< leg::JointStates, Point< anchor::Leg > > states_to_point;

  for ( unsigned x = 0; x < x_max; ++x )
  {
    for ( unsigned y = 0; y < y_max; ++y )
    {
      leg::JointStates states = {j_state};
      leg::Vertex      vertex = graph.emplaceVertex();

      vertex.setJointStates( states );

      states_to_point[states] = Point< anchor::Leg >( x, y, 0 );

      ++j_state;
    }
  }

  std::for_each( graph.vertexBegin(), graph.vertexEnd(), [&]( leg::Vertex& source ) {
    std::for_each( graph.vertexBegin(), graph.vertexEnd(), [&]( leg::Vertex& target ) {
      if ( source.getID() == target.getID() )
      {
        return;
      }

      if ( states_to_point[target.getJointStates()].distance( states_to_point[source.getJointStates()] ) > 1.1 )
      {
        return;
      }
      graph.insertEdge( source, target, leg::EdgeData( Duration( 1 ) ) );
    } );
  } );

  GridData res;

  res.graph           = std::move( graph );
  res.states_to_point = states_to_point;

  return res;
}

TEST( ApproxDurationDensityTest, all_one )
{
  GridData data = getGridGraph( 6, 6 );

  graph::for_each_edge( data.graph, [&]( leg::Edge edge ) { edge.refData().setDuration( Duration( 1 ) ); } );

  leg::ConfigData cdata;

  leg::Config config( cdata );

  Leg leg( data.graph, data.states_to_point, config );

  std::vector< Vector3 > directions{
      {1, 0, 0}, {1, 1, 0}, {0, 1, 0}, {-1, 1, 0}, {-1, 0, 0}, {-1, -1, 0}, {0, -1, 0}, {1, -1, 0}};

  std::for_each( leg.getGraph().vertexBegin(), leg.getGraph().vertexEnd(), [&]( const leg::Vertex vertex ) {

    Point< anchor::Leg > vertex_coord = leg.getPoint( vertex.getID() );

    std::for_each( directions.begin(), directions.end(), [&]( Vector3 direction ) {
      Velocity velocity = approxVelocity( leg, vertex, vertex_coord, vertex_coord + direction, Length( 100 ) );

      ASSERT_NEAR( double( velocity ), 1, 0.00000001 );
    } );
  } );
}

TEST( ApproxDurationDensityTest, direction )
{
  GridData data = getGridGraph( 6, 6 );

  graph::for_each_edge( data.graph, [&]( leg::Edge edge ) {
    Point< anchor::Leg > target = data.states_to_point[data.graph.getVertex( edge.getTargetID() ).getJointStates()];
    Point< anchor::Leg > source = data.states_to_point[data.graph.getVertex( edge.getSourceID() ).getJointStates()];

    if ( target.x() > source.x() )  // [1,0,0]
    {
      edge.refData().setDuration( Duration( 1 ) );
    }
    else if ( target.x() < source.x() )  // [-1,0,0]
    {
      edge.refData().setDuration( Duration( 100 ) );
    }
    else if ( target.y() < source.y() )  // [0,-1,0]
    {
      edge.refData().setDuration( Duration( 10000 ) );
    }
    else  // [0,1,0]
    {
      edge.refData().setDuration( Duration( 1000000 ) );
    }
  } );

  leg::ConfigData cdata;

  leg::Config config( cdata );

  Leg leg( data.graph, data.states_to_point, config );

  struct TestApproximation
  {
    Vector3 direction;
    double  value;
    double  tolerance;
  };

  std::vector< TestApproximation > test_approximations{
      // These 4 cases mirror value distribution of edges
      {{1, 0, 0}, 1, 0.00000001},
      {{0, -1, 0}, 10000, 0.00000001},
      {{-1, 0, 0}, 100, 0.00000001},
      {{0, 1, 0}, 1000000, 0.00000001},
      // these are 45 angle between the value distribution above
      {{1, -1, 0}, 5000, 550},
      {{-1, -1, 0}, 5000, 550},
      {{-1, 1, 0}, 500000, 55000},
      {{1, 1, 0}, 500000, 55000},
      //
  };

  std::for_each( leg.getGraph().vertexBegin(), leg.getGraph().vertexEnd(), [&]( const leg::Vertex vertex ) {

    Point< anchor::Leg > vertex_coord = leg.getPoint( vertex.getID() );

    std::for_each( test_approximations.begin(), test_approximations.end(), [&]( const TestApproximation& test_data ) {
      Velocity velocity =
          approxVelocity( leg, vertex, vertex_coord, vertex_coord + test_data.direction, Length( 100 ) );
      ASSERT_NEAR( double( velocity ), test_data.value, test_data.tolerance );
    } );
  } );
}

TEST( ApproxPointMovementDurationTest, allones )
{
  GridData data = getGridGraph( 6, 6 );

  Point< anchor::Leg > start{0, 0, 0};

  graph::for_each_edge( data.graph, [&]( leg::Edge edge ) { edge.refData().setDuration( Duration( 1 ) ); } );

  leg::ConfigData cdata;
  leg::Config     config( cdata );

  Leg leg( data.graph, data.states_to_point, config );

  const leg::Vertex    start_hint = leg.getGraph().getVertex( leg::VertexID( 0 ) );
  Point< anchor::Leg > goal{5, 5, 0};
  Duration             res = approxPointMovementDuration( leg, start, start_hint, goal );

  ASSERT_NEAR( res.toSec(), 7.07, 0.1 );
}
TEST( ApproxPointMovementDurationTest, linear )
{
  // this test case should generate grid network like:
  //
  //  O--2--O--2--O
  //  |     |     |
  //  2     2     2
  //  |     |     |
  //  O--1--O--2--O
  //  |     |     |
  //  1     1     2
  //  |     |     |
  //  O--1--O--2--O
  //
  //  ie. value of edge is max(x,y) of [x,y,z] coordinate of it's target
  //
  //  goal is to executed approximation from bottom left corner, to bottom right coner
  //
  unsigned size = 20;
  GridData data = getGridGraph( size, size );

  Point< anchor::Leg > start{0, 0, 0};

  graph::for_each_edge( data.graph, [&]( leg::Edge edge ) {
    Point< anchor::Leg > target = data.states_to_point[data.graph.getVertex( edge.getTargetID() ).getJointStates()];

    edge.refData().setDuration( Duration( std::max( target.x(), target.y() ) ) );
  } );

  leg::ConfigData cdata;
  leg::Config     config( cdata );

  Leg leg( data.graph, data.states_to_point, config );

  const leg::Vertex    start_hint = leg.getGraph().getVertex( leg::VertexID( 0 ) );
  Point< anchor::Leg > goal{size - 1., size - 1., 0};
  Duration             res = approxPointMovementDuration( leg, start, start_hint, goal );

  Duration expected{0};
  for ( unsigned i = 1; i < size; ++i )
  {
    expected += Duration( i * sqrt( 2 ) );
  };

  ASSERT_NEAR( res.toSec(), expected.toSec(), 0.03 * expected.toSec() );
}

int main( int argc, char** argv )
{
  testing::InitGoogleTest( &argc, argv );
  return RUN_ALL_TESTS();
}
