#!/usr/bin/python2

from math import sqrt

from interactive_markers.interactive_marker_server import (
    InteractiveMarkerServer)
from visualization_msgs.msg import InteractiveMarker, InteractiveMarkerControl
from urdf_parser_py.urdf import URDF

from tf.transformations import (quaternion_multiply, quaternion_conjugate,
                                quaternion_from_euler, quaternion_about_axis)

from numpy import dot, clip, cross

import std_msgs.msg as std_msg
import rospy
import geometry_msgs


class JointMarker:
    def __init__(self, joint, interactive_marker_server):
        """ construction takes joint from URDF file and InteractiveMarkerServer
            initializes class, creates JointMove publisher and
            adequate interactive marker """

        self.base_orientation = shortest_arc_quat([1, 0, 0], joint.axis)
        self.joint = joint
        # TODO: topic name would 'enjoy' some parametrization
        self.publisher = rospy.Publisher( 
            "/control/" + joint.name +"_joint_position_controller/command", std_msg.Float64, queue_size=1 )
        self.interactive_marker_server =interactive_marker_server

        self.set_marker(0)

    def set_marker(self, angle):

        inter_marker = InteractiveMarker()

        inter_marker.header.frame_id = self.joint.parent
        inter_marker.name = "joint_marker/" + self.joint.name
        inter_marker.scale = rospy.get_param("scale", 0.03)
        inter_marker.pose = get_base_pose(self.joint)

        rotate_control = InteractiveMarkerControl()
        rotate_control.name = "joint_marker/" + self.joint.name + "/move"
        rotate_control.interaction_mode = InteractiveMarkerControl.ROTATE_AXIS

        angle_quat = quaternion_about_axis(angle, self.joint.axis)
        rotate_control.orientation = quat_list_to_msg(quaternion_multiply(self.base_orientation, angle_quat))

        inter_marker.controls.append(rotate_control)

        self.interactive_marker_server.insert(inter_marker, self.feedback_cb)
        self.interactive_marker_server.applyChanges()

    def feedback_cb(self, feedback_msg):
        """ callback to be executed on interactive marker feedback,
            sends new angle to joint """

        actual_quat = quat_msg_to_list(feedback_msg.pose.orientation)

        angle = quat_angle(actual_quat, self.base_orientation)
        cliped_angle = clip(angle, self.joint.limit.lower, self.joint.limit.upper)

        move_msg = std_msg.Float64()
        move_msg.data = cliped_angle

        if cliped_angle != angle:
            self.set_marker( cliped_angle )

        self.publisher.publish(move_msg)


def get_base_pose(joint):
    """ generates pose from origin of given joint """

    pose = geometry_msgs.msg.Pose()

    pose.position.x = joint.origin.position[0]
    pose.position.y = joint.origin.position[1]
    pose.position.z = joint.origin.position[2]

    base_quaternion = quaternion_from_euler(*joint.origin.rotation)

    pose.orientation.x = base_quaternion[0]
    pose.orientation.y = base_quaternion[1]
    pose.orientation.z = base_quaternion[2]
    pose.orientation.w = base_quaternion[3]

    return pose


def quat_msg_to_list(msg):
    return [msg.x, msg.y, msg.z, msg.w]

def quat_list_to_msg(l):
    msg = geometry_msgs.msg.Quaternion()
    msg.x = l[0]
    msg.y = l[1]
    msg.z = l[2]
    msg.w = l[3]
    return msg

def quat_multiply_vec(quat, vector):
    """ multiplies quaternion with vector  """

    return [
        quat[3] * vector[0] + quat[1] * vector[2] - quat[2] * vector[1],
        quat[3] * vector[1] + quat[2] * vector[0] - quat[0] * vector[2],
        quat[3] * vector[2] + quat[0] * vector[1] - quat[1] * vector[0],
        -quat[0] * vector[0] - quat[1] * vector[1] - quat[2] * vector[2]
    ]


def quat_rotate(quat, vector):
    """ rotates vector by quaternion """

    # just standart formula: Q * v * -Q

    tmp = quat_multiply_vec(quat, vector)

    res = quaternion_multiply(tmp, quaternion_conjugate(quat))

    return res[:3]


def get_quaternion_axis(quat):
    """ returns axis represented by this quat """

    s = sqrt(1. - pow(quat[3], 2))
    return [quat[0] / s, quat[1] / s, quat[2] / s]


def quat_angle(lh_quat, rh_quat):
    """ calculates angle between two quaternions """
    lh_lengh = float(dot(lh_quat, lh_quat))
    rh_lengh = float(dot(rh_quat, rh_quat))

    s = sqrt(lh_lengh * rh_lengh)
    return dot(lh_quat, rh_quat) / s * 2.


def shortest_arc_quat(lh_vector, rh_vector):
    """ calculates quaternion represeting change from lh_vector to rh_vector"""
    # inspired by C++ tf::shortestArcQuat

    vec_cross = cross(lh_vector, rh_vector)
    vec_dot = dot(lh_vector, rh_vector)

    if vec_dot < -1:
        raise Exception("unhandled part of calculations, fix in case needed")

    s = sqrt((1.0 + vec_dot) * 2)
    rs = 1. / s

    return [vec_cross[0] * rs,
    vec_cross[1] * rs,
    vec_cross[2] * rs,
    s / 2]


if __name__ == "__main__":
    rospy.init_node("joint_marker", log_level=rospy.INFO)

    robot = URDF.from_parameter_server()

    server = InteractiveMarkerServer("joint_marker")

    joints = list()

    for joint in robot.joints:

        if joint.type == "fixed":
            continue
        print(joint)
        joints.append(JointMarker(joint, server))
    rospy.spin()
